<?php
/**
 * Handle logic for rendering a Content Widget based on external taxonomy.
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20150121.2343
 * @copyright Copyright (C) 2009-2021  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Widget;


/**
 * A short teaser of what ContentWidget does.
 *
 * More lengthy description of what ContentWidget does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for ContentWidget
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class ContentTaxonomyWidget extends Widget {
	public $is_simple = true;
	
	public $title = 'Content Taxonomy Widget';

	public $settings = [
		[
			'type'        => 'select',
			'name'        => 'template',
			'title'       => 'Taxonomy Template',
			'description' => 'Select the taxonomy template to use for retrieving results',
			'required' => true,
		],
		[
			'type' => 'number',
			'name' => 'limit',
			'title' => 'Number of Results',
			'description' => 'Enter the number of results to retrieve at maximum',
			'min' => 1,
			'max' => 100,
			'step' => 1,
			'value' => 10,
			'required' => true,
		],
		[
			'type' => 'select',
			'name' => 'sort',
			'title' => 'Sort By',
			'required' => true,
			'options' => [
				'random' => 'Random',
				'popular' => 'Popular',
				'newest' => 'Newest First',
				'oldest' => 'Oldest First',
			]
		],
		[
			'type' => 'select',
			'name' => 'pagination',
			'title' => 'Pagination',
			'value' => 'none',
			'options' => [
				'none' => 'No Pagination',
				'standard' => 'Standard Pagination',
			]
		],
	];

	public function getFormSettings() {
		// Retrieve what templates are created to provide the administrator a list of options to select

		$tmpname = 'pages/content/view/';
		$matches = array();

		foreach(\Core\Templates\Template::GetPaths() as $d){
			if(is_dir($d . $tmpname)){
				// Yay, sift through that and get the files!
				$dir = \Core\Filestore\Factory::Directory($d . $tmpname);
				foreach($dir->ls(['tpl', 'phtml']) as $file){
					// Skip directories
					if($file instanceof \Core\Filestore\Directory) continue;

					/** @var $file \Core\Filestore\File */
					//$fullpath = $tmpname . $file->getBaseFilename();
					$fullpath = $file->getBaseFilename(true);
					// Do some template updates and make it a little more friendlier to read.
					$name = ucwords(str_replace(['-', '_'], ' ', $fullpath));
					$matches[ $fullpath ] = $name;
				}
			}
		}

		// Sort!
		asort($matches);

		$this->settings[0]['options'] = $matches;


		return $this->settings;
	}

	/**
	 * Get the path for the preview image for this widget.
	 *
	 * Should be an image of size 210x70, 210x140, or 210x210.
	 *
	 * @return string
	 */
	public function getPreviewImage() {
		// Extend this method in your class and return the path you need.
		// Optional.
		return 'assets/images/previews/templates/widgets/content/custom-content-area.png';
	}

	/**
	 * Widget to display a simple site search box
	 */
	public function execute() {
		$view = $this->getView();
		$request = \Core\Router::$DefaultRequest;

		// Retrieve the data for this widget based on admin settings
		$component = 'content';
		$template = $this->getSetting('template');
		$limit = (int)$this->getSetting('limit');
		$sort = $this->getSetting('sort');
		$now = \Core\Date\DateTime::NowGMT();
		$pagination = $this->getSetting('pagination');

		$ds = \Core\Datamodel\Dataset::Init()
			->select('site', 'baseurl')
			->table('page')
			->whereEquals('component', $component)
			->whereEquals('page_template', $template)
			->whereEquals('published_status', 'published')
			->where('published <= ' . $now);

		// Expiration requires a group where because it supports NULL
		$expires = new \Core\Datamodel\DatasetWhereClause();
		$expires->setSeparator('OR');
		$expires->addWhere('published_expires = NULL');
		$expires->addWhere('published_expires = ');
		$expires->addWhere('published_expires >= ' . $now);
		$ds->where($expires);


		// Support pagination
		if($pagination === 'standard') {
			// Duplicate the original dataset to retrieve how many total results are available
			// This will provide how many total results, and thus the total number of pages
			$cloneDS = clone $ds;
			$cloneDS->select(null)->count();
			$results = $cloneDS->executeAndGet();

			if($request && $request->getParameter('taxonomy') === $template) {
				$page = $request->getParameter('page') ?? 1;
			}
			else {
				$page = 1;
			}

			// Calculate the number of pages total
			$pagesTotal = ceil($results / $limit);

			// Ensure the requested page falls within the acceptable area
			$page = max(1, $page);
			$page = min($pagesTotal, $page);

			if($page > 1) {
				$previousLink = '?taxonomy=' . $template . '&page=' . ($page - 1);
			}
			else {
				$previousLink = null;
			}

			if($page < $pagesTotal) {
				$nextLink = '?taxonomy=' . $template . '&page=' . ($page + 1);
			}
			else {
				$nextLink = null;
			}

			$ds->limit($limit, $limit * ($page - 1));
		}
		else {
			$page = null;
			$pagesTotal = null;
			$previousLink = null;
			$nextLink = null;
			$ds->limit($limit);
		}

		switch($sort) {
			case 'random':
				$ds->order('RAND()');
				break;
			case 'popular':
				$ds->order('popularity DESC');
				break;
			case 'newest':
				$ds->order('published DESC');
				break;
			case 'oldest':
				$ds->order('published ASC');
				break;
		}

		$results = [];
		$pages = $ds->executeAndGet();

		if($limit === 1) {
			// Limit of 1 will trigger executeAndGet to return only a single record.
			// To keep the logic cleaner, convert those back to multi-selects
			$pages = [ $pages ];
		}

		foreach($pages as $data) {
			$page = PageModel::Construct($data['site'], $data['baseurl']);
			$insertables = $page->getLink('InsertableModel');
			$meta = $page->getMetasArray();
			$fields = [];

			foreach($insertables as $insertable) {
				$fields[ $insertable->name ] = $insertable->value;
			}

			$results[] = [
				'page' => $page,
				'meta' => $meta,
				'fields' => $fields,
			];
		}

		// If no entries, just return a blank string.
		if(!count($results)) {
			return '';
		}

		$view->assign('results', $results);
		$view->assign('page', $page);
		$view->assign('pages_total', $pagesTotal);
		$view->assign('prev_link', $previousLink);
		$view->assign('next_link', $nextLink);
	}
}