<?php

/*
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\CLI\CLI;
use Core\Date\DateTime;
use Core\Model;
use mikehaertl\shellcommand\Command;


/**
 * Description of PersistentJobModel
 *
 * @author charlie
 */
class CronPersistentJobModel extends Model {
	public static $Schema = [
		'id' => [
			'type' => Model::ATT_TYPE_UUID,
		],
		'key' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 64,
		],
		'call' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
		],
		'parameters' => [
			'type' => Model::ATT_TYPE_DATA,
			'encoding' => Model::ATT_ENCODING_JSON,
		],
		'state' => [
			'type' => Model::ATT_TYPE_ENUM,
			'options' => ['pending', 'starting', 'running', 'pausing', 'paused', 'stopping', 'stopped', 'zombie'],
			'default' => 'pending',
		],
		'server' => [
			'type' => Model::ATT_TYPE_STRING,
			'comment' => 'Server this job is running on',
		],
		'pid' => [
			'type' => Model::ATT_TYPE_INT,
			'comment' => 'Process ID of the job',
		],
		'memory' => [
			'type' => Model::ATT_TYPE_INT,
			'comment' => 'Memory usage at the last check',
		],
		'timestamp_checked' => [
			'type' => Model::ATT_TYPE_INT,
			'comment' => 'Timestamp of the last check',
		],
	];
	
	public static $Indexes = [
		'primary' => 'id',
	];

	public function getLabel() {
		$title = $this->call;

		$params = $this->parameters;
		if($params) {
			if(is_array($params)) {
				$title .= '(' . implode(', ', $params) . ')';
			}
			elseif(is_scalar($params)) {
				$title .= '(' . $params . ')';
			}
			else {
				$title .= '(Object)';
			}
		}
		else {
			$title .= '()';
		}

		return $title;
	}

	/**
	 * Get the actual process information for this job.
	 *
	 * Will return an array containing the following keys:
	 *
	 * * state {string}
	 * * memory {int}
	 * * pid {int}
	 *
	 * @return array
	 */
	public function getProcessInformation() : array {
		$cmd = new Command();
		$cmd->setCommand('ps');
		$cmd->addArg('-p', $this->pid);
		$cmd->addArg('-o', 'state,size');
		$cmd->execute();

		$out = explode("\n", $cmd->getOutput(true));


		if(count($out) === 1) {
			// One line shows header only; no processes found.
			return [
				'state' => 'stopped',
				'memory' => 0,
				'pid' => 0,
			];
		}
		else {
			list($s, $m) = explode(' ', $out[1]);
			if($s === 'Z') {
				return [
					'state' => 'zombie',
					'memory' => $m,
					'pid' => $this->pid,
				];
			}
			else {
				return [
					'state' => 'running',
					'memory' => $m,
					'pid' => $this->pid,
				];
			}
		}
	}

	/**
	 * Scan the server for this process and get some of its details.
	 *
	 * Nothing is returned, but the relevant details are saved to the Model
	 */
	public function scanProcessInformation() {
		if(!defined('SERVER_ID')) {
			// No server ID defined, not able to manage any tasks!
			return;
		}

		if($this->server !== SERVER_ID) {
			// Only the server that started the process has its process information.
			return;
		}

		if($this->state === 'starting') {
			// These are states that are in flux; wait until they complete before checking the status.
			return;
		}

		if(!$this->pid) {
			// No process ID set for this task.
			if($this->state === 'paused') {
				// Paused or starting, either way acceptable for the task not running.
				return;
			}
			else {
				// It's not running, but should be.
				$this->state = 'stopped';
				$this->memory = 0;
				$this->timestsamp_checked = DateTime::NowGMT();
				$this->save();
			}
		}
		else {
			$status = $this->getProcessInformation();

			if($status['state'] === 'stopped' && $this->state === 'pausing') {
				// Pausing tasks that have stopped are good.
				$this->state = 'paused';
			}
			elseif($status['state'] === 'stopped' && $this->state === 'stopping') {
				// Stopping tasks that have stopped are good.
				$this->state = 'stopped';
			}
			else {
				$this->state = $status['state'];
			}

			$this->memory = $status['memory'];
			$this->pid = $status['pid'];
			$this->timestsamp_checked = DateTime::NowGMT();
			$this->save();
		}
	}
	
	/**
	 * Stop a process that is currently running and mark is as stopped.
	 */
	public function stopProcess() {
		if($this->server !== SERVER_ID) {
			return;
		}

		$this->state = 'stopping';
		$this->save();

		$this->_killProcess();

		$status = $this->getProcessInformation();
		if($status['state'] === 'stopped') {
			// Process was able to successfully stop
			$this->state = 'stopped';
			$this->pid = 0;
			$this->memory = 0;
			$this->save();
		}
	}
	
	/**
	 * Queue a process to start, will do nothing if already running.
	 *
	 * @throws Exception
	 */
	public function startProcess() {
		if($this->state === 'running') {
			return;
		}

		$this->state = 'starting';
		$this->server = SERVER_ID;
		$this->save();

		$command = new Command();
		$command->setCommand('php');
		$command->procCwd = ROOT_PDIR;
		$command->addArgValue('./cli');
		$command->addArgValue('/cron/persistent/' . $this->id);
		$command->executeBackground();
	}

	public function pauseProcess() {
		if($this->server !== SERVER_ID) {
			return;
		}

		$this->state = 'pausing';
		$this->save();

		$this->_killProcess();

		$status = $this->getProcessInformation();
		if($status['state'] === 'stopped') {
			// Process was able to successfully stop
			$this->state = 'paused';
			$this->pid = 0;
			$this->memory = 0;
			$this->save();
		}
	}

	public function restartProcess() {
		$this->stopProcess();
		$this->startProcess();
	}

	/**
	 * Kill the process and wait a short time until it's shutdown.
	 *
	 */
	private function _killProcess() {
		if($this->pid) {
			// Send the instruction to kill this process.
			exec('kill ' . $this->pid);

			// Wait a short time for the process to exit gracefully.
			$counter = 0;
			$status = $this->getProcessInformation();
			while($status['state'] === 'running' && $counter < 10) {
				usleep(10000);
				$counter++;
				$status = $this->getProcessInformation();
			}
		}
	}

	/**
	 * Find all jobs that match the requested parameters
	 *
	 * @param string $call
	 *
	 * @return CronPersistentJobModel[]
	 */
	public static function FindJobs(string $call) : array {
		return self::Find(['call = ' . $call]);
	}

	/**
	 * Find all jobs that match the requested parameters
	 *
	 * @param string $call
	 *
	 * @return CronPersistentJobModel|null
	 */
	public static function FindJob(string $call, array $params = null) : ?CronPersistentJobModel {
		return CronPersistentJobModel::Find(['key = ' . self::CalculateHash($call, $params)], 1);
	}

	public static function CalculateHash(string $call, array $params = null) : string {
		$hash = $call;

		if(is_array($params)) {
			foreach($params as $k => $v){
				if(is_scalar($v)){
					$hash .= ';' . $k . ':' . $v;
				}
				elseif(is_object($v)){
					$hash .= ';' . $k . ':' . get_class($v);
				}
			}
		}

		return md5($hash);
	}
	
	/**
	 * Register a persistent job to be background executed on this server, (or another server in the cluster)
	 *
	 * If this job is already registered, it will simply be marked as pending.
	 * 
	 * @param string $call
	 * @param array|mixed $params
	 *
	 * @return bool
	 */
	public static function RegisterJob(string $call, array $params = null) : bool {
		$hash = self::CalculateHash($call, $params);

		$model = CronPersistentJobModel::Find(['key = ' . $hash], 1);
		if($model !== null) {
			if($model->state === 'paused' || $model->state === 'stopped') {
				$model->startProcess();
				return true;
			}
			else {
				return false;
			}
		}
		
		$model = new CronPersistentJobModel();
		$model->setFromArray([
			'key'        => $hash,
			'call'       => $call,
			'parameters' => $params,
			'state'      => 'pending',
		]);
		$model->save();
		return true;
	}
	
	/**
	 * Method to check all active/pending jobs and start them if necessary.
	 * 
	 */
	public static function CheckJobs() {
		
		if(!defined('SERVER_ID')) {
			CLI::PrintError('Unable to support persistent jobs without a valid SERVER ID set!');
			return false;
		}

		$models = CronPersistentJobModel::Find();

		if(!count($models)) {
			CLI::PrintLine('No persistent jobs are queued.');
			return true;
		}

		foreach($models as $m) {
			/** @var $m CronPersistentJobModel */
			CLI::PrintActionStart('Job ' . $m->getLabel());
			$m->scanProcessInformation();

			if($m->state === 'pending' || $m->state === 'stopped') {
				// If the process is marked as pending, then queue to get it started.
				try {
					$m->startProcess();
					CLI::PrintActionStatus('starting');
				}
				catch(Exception $ex) {
					CLI::PrintActionStatus(false);
					CLI::PrintError($ex->getMessage());
					\Core\log_exception($ex);
				}
			}
			elseif($m->state === 'starting') {
				// The process is starting
				// This is generally not a good thing, as the job has failed to start since the last time this check ran.
				CLI::PrintActionStatus(false);
				CLI::PrintError('Job has not started yet, retrying');
				\Core\log_warning('Persistent job ' . $m->getLabel() . ' failed to start between checks!');

				$m->stopProcess();
			}
			elseif($m->state === 'paused') {
				CLI::PrintActionStatus('skip');
			}
			elseif($m->state === 'running') {
				CLI::PrintActionStatus(true);
			}
			else {
				CLI::PrintActionStatus($m->state);
			}


			// Print some details about this job.
			CLI::PrintLine('Job State: ' . $m->state);
			if($m->state === 'running') {
				CLI::PrintLine('PID: ' . $m->pid);
				CLI::PrintLine('Memory: ' . \Core\Formatter\GeneralFormatter::Filesize($m->memory));
			}
		}
		
		return true;
	}
}
