<?php
/**
 * Class file for FormBuilderHelper
 *
 * @package FormBuilder
 * @author Nicholas Hinsch <nicholas@evalagency.com>
 */

use Core\ConfigHandler;
use Core\Core;
use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\EmailAPI\Email;
use Core\Forms\Form;
use Core\Model;
use Core\ModelValidationException;
use SecuritySuite\SpamCan\SpamCheck;


abstract class FormBuilderHelper {
	/**
	 * Helper function to save custom form pages, both new and existing.
	 *
	 * @static
	 *
	 * @param Form $form
	 * @return string Redirect URL
	 */
	public static function FormBuilderFormHandler(Form $form) {
		try{
			/** @var FormBuilderModel $model */
			$model = $form->getModel();
			/** @var WidgetModel $widget */
			$widget = $model->getLink('Widget');
			$isnew = $model->checkIsNew();

			if(!isset($_POST['fields']) || !is_array($_POST['fields'])){
				Core::SetMessage('Creating a form with no fields on it seems a little silly, don\'t you think?', 'error');
				return false;
			}

			$model->save();

			// Set the necessary widget criteria
			$widget->set('title', $model->get('title'));
			$widget->set('component', 'formbuilder');
			$widget->set('editurl', '/formbuilder/update/' . $model->id);
			// This widget will NOT have a delete option!
			// Otherwise it may seem that the admin is deleting the widget without knowing that it'll also delete the corresponding form.
			//$widget->set('deleteurl', '/formbuilder/delete/' . $model->get('id'));
			$widget->set('editpermissions', $model->get('editpermissions'));
			$widget->save();


			// Set the admin page for this view so it's registered in the admin menu for convenience.
			// We do Find here just in case the site ID was changed on the parent page.
			$adminPage = PageModel::Find(['baseurl = ' . '/formbuilder/listing/' . $model->id], 1, null);
			if(!$adminPage) {
				// New!
				$adminPage = new PageModel();
			}
			$adminPage->setFromArray([
				'site'        => $model->site,
				'baseurl'     => '/formbuilder/listing/' . $model->id,
				'admin'       => 1,
				'title'       => $model->title,
				'parenturl'   => '/formbuilder/admin',
				'component'   => 'formbuilder',
				'access'      => $model->listpermissions,
				'admin_group' => 't:STRING_CONTENT',
				'selectable'  => 0,
				'indexable'   => 0,
				'editurl'     => '/formbuilder/update/' . $model->id,
				'deleteurl'   => '/formbuilder/delete/' . $model->id,
			]);
			$adminPage->setMeta('image', 'assets/images/icons/formbuilder-entries.svg');
			$adminPage->save();

			// Save the individual form fields now too!
			// These have to be done manually since they are generated with Javascript on the frontend.
			// The parent model has already been saved and is done, so we will be managing the children fields directly here.
			$fields = $model->getLink('FormBuilderField');
			$currentmap = [];

			// Run through currently linked fields and either map them for future use or delete them if they're not included in the POST.
			foreach($fields as $f){
				/** @var FormBuilderFieldModel $f */
				$id = $f->get('id');
				if(!isset($_POST['fields'][$id])){
					$f->delete();
				}
				else{
					$currentmap[$id] = $f;
				}
			}

			// Now that I have a flat map of current fields, I can run through the incoming data
			// and easily update the existing fields and create the necessary new ones.
			$weight = 0;
			foreach($_POST['fields'] as $id => $dat){
				if(strpos($id, 'new-') === 0 || !isset($currentmap[$id])){
					$field = new FormBuilderFieldModel();
					$field->set('formid', $model->get('id'));
				}
				else{
					$field = $currentmap[$id];
				}

				$field->set('type', $dat['type']);
				$field->set('title', $dat['title']);
				$field->set('conditional', isset($dat['conditional']) ? $dat['conditional'] : 0);
				//$field->set('conditional_element', $dat['conditional_element']);
				//$field->set('conditional_value', $dat['conditional_value']);
				$field->set('weight', ++$weight);
				$field->setOptions(isset($dat['options']) ? $dat['options'] : null);

				$field->save();
			}

			if($isnew){
				Core::SetMessage('Created custom form successfully!', 'success');
			}
			else{
				Core::SetMessage('Updated custom form successfully!', 'success');
			}
			return 'back';
		}
		catch(ModelValidationException $e){
			Core::SetMessage($e->getMessage(), 'error');
			return false;
		}
		catch(Exception $e){
			Core::SetMessage($e->getMessage(), 'error');
			\Core\ErrorManagement\exception_handler($e);
			return false;
		}
	}

	/**
	 * Form handler for the public submission form.
	 *
	 * @param Form $form
	 * @return bool|string
	 */
	public static function SaveEntry(Form $form) {
		try {
			$secureForm  = false;
			$entry       = new FormBuilderEntryModel();
			$contactform = FormBuilderModel::Construct($form->getElementValue('formid'));
			$spamCheck   = new SpamCheck();

			if(!$contactform->checkExists()) {
				Core::SetMessage('Submission page not found', 'error');
				return false;
			}

			// It'll have some information not in the form, ie: IP address and user agent.
			$entry->setFromArray(
				[
					'formid'    => $contactform->get('id'),
					'ip'        => REMOTE_IP,
					'useragent' => $_SERVER['HTTP_USER_AGENT'] ?? '',
					// Remember, saved from the INITIAL source.
					'referrer'  => $_SESSION['contactformreferrer'] ?? '',
				    'source_page' => $form->originalurl,
				]
			);
			$spamCheck->setUserAgent($_SERVER['HTTP_USER_AGENT'] ?? '');


			if(\Core\Core::IsComponentAvailable('security-suite')) {
				$spamChecker = new SpamCheck();
			}
			else {
				$spamChecker = null;
			}


			// Set the user's ID
			if(\Core\user()) {
				$entry->set('user_id', \Core\user()->get('id'));
				if($spamChecker) {
					$spamChecker->setEmail(\Core\user()->get('email'));
				}
			}

			$fields     = $contactform->getLink('FormBuilderField');

			foreach($fields as $f){
				/** @var FormBuilderFieldModel $f */
				$title    = $f->get('title');
				$key      = \Core\str_to_url($title);
				$name     = 'fields[' . $key . ']';
				$type     = $f->get('type');
				$multiple = false;
				$secure   = false;

				if(!$title){
					// Skip inputs that do not have names.
					continue;
				}

				if($type == 'checkboxes'){
					$name .= '[]';
					$multiple = true;
				}
				$value = $form->getElementValue($name);

				if($multiple && is_array($value)){
					$value = implode("\n", $value);
				}

				$ef = new FormBuilderEntryFieldModel();
				$ef->setFromArray(
					[
						'formid'  => $contactform->get('id'),
					    'fieldid' => $f->get('id'),
					    'type'    => $type,
					    'title'   => $f->get('title'),
					    'value'   => $value,
					]
				);

				foreach($f->getOptions() as $k => $v){
					if($k == 'secure' && $v == 'private'){
						$secureForm = true;
						$ef->set('decrypted_value', $value);
						$ef->set('value', '--OMITTED--');
					}
					elseif($k == 'secure' && $v == 'encrypted'){
						$secureForm = true;
						$ef->set('decrypted_value', $value);
						$ef->set('value', Model::EncryptValue($value));
						$ef->set('encrypted', true);
					}
				}

				$entry->setLink('FormBuilderEntryField', $ef);

				if($spamChecker) {
					switch($type) {
						case 'email':
							$spamChecker->setEmail($value);
							break;
						case 'text':
						case 'textarea':
							$spamChecker->addContent($value);
					}
				}
			}

			if($spamChecker) {
				if($spamChecker->checkIsBlockable()) {
					// Really horrible spam, just gets blocked outright
					\SecuritySuite\IpBlacklistHelper::BlockIP(
						REMOTE_IP,
						SECONDS_TWO_HOUR,
						'Blocked for submitting spam content you have been.',
						'Spam submission score of ' . $spamChecker->getScore()
					);
					return false;
				}
				elseif($spamChecker->checkIsSpam()) {
					$entry->junk = 1;
					$results = $spamChecker->getResults();
					$entry->spam_checks = implode(
						"\n",
						array_map(
							function($a) {
								/** @var $a \SecuritySuite\SpamCan\SpamCheckResult */
								return 'Score +' . $a->score . ' - ' . $a->comment;
							},
							$results
						)
					);
				}
			}

			if($contactform->geo_restrict && defined('REMOTE_COUNTRY')) {
				// This form has geo-restrict set.
				// This will not block a submission, but will automatically flag it as junk.
				$geoList = array_map('trim', explode(',', $contactform->geo_restrict));
				$found = false;
				foreach($geoList as $geo) {
					if(strpos($geo, ':') !== false) {
						list($country, $province) = explode(':', $geo);
					}
					else {
						$country = $geo;
						$province = null;
					}

					if(
						REMOTE_COUNTRY === $country &&
						($province === null || $province === REMOTE_PROVINCE)
					) {
						$found = true;
						break;
					}
				}

				if(!$found) {
					$entry->junk = 1;
					$entry->spam_checks .= "\nGeo check failed";
				}
			}

			$entry->save();
		} catch (ModelValidationException $e) {
			Core::SetMessage($e->getMessage(), 'error');
			return false;
		}
		catch (Exception $e) {
			Core::SetMessage('Something went wrong with the form submission, please try again later', 'error');
			\Core\log_exception($e);
			return false;
		}

		$replyto = $entry->getEmailRecipient();

		// Also, send the notification email in this case, since it's a public submission.
		if ($contactform->get('notify_email') && !$entry->get('junk')) {
			try {
				$email = Email::Factory('/pages/formbuilder/entry_view.tpl');
				$email->setEncryption($secureForm);
				$notify = $contactform->get('notify_email');
				// Replace windows newlines with unix newlines
				$notify = str_replace(["\r\n", "\r"], "\n", $notify);
				// And explode on them.
				$notify = explode("\n", $notify);
				foreach ($notify as $eaddress) {
					if(!trim($eaddress)){
						// There may be blank lines here.
						continue;
					}
					$email->addTo($eaddress);
				}
				$formname = $contactform->get('title');
				$email->setSubject("New Contact Submission on $formname");
				$email->assign('formurl', $form->originalurl);
				$email->assign('formname', $formname);
				$email->assign('entry', $entry);
				$email->assign('form', $contactform);
				if($replyto){
					$email->setReplyTo($replyto);
					//$email->setOnBehalfOf($replyto, $entry->getRecipient());
				}
				$email->send();
			} catch (Exception $e) {
				\Core\ErrorManagement\exception_handler($e);
				//Core::SetMessage('Submission was saved, but unable to notify an administrator of the submission, please call us about the submission', 'info');
			}
		}

		// The user should get a confirmation email automatically, if enabled.
		if($replyto && !$entry->get('junk')) {
			// I need to check the contents of the email WYSIWYG entry; if it's empty.... don't do anything.
			$text = trim(strip_tags($contactform->get('thankyou_email')));
			if($text != '' && $text != '&nbsp;'){
				try {
					$email = Email::Factory();
					$email->addTo($replyto);
					$email->setSubject('Contact Form Confirmation');
					$email->setBody($contactform->get('thankyou_email'));
					$email->send();
				} catch (Exception $e) {
					\Core\ErrorManagement\exception_handler($e);
					//Core::SetMessage('Submission was saved, but unable to notify an administrator of the submission, please call us about the submission', 'info');
				}
			}
		}

		Core::SetMessage('Submission Successfully Submitted', 'success');
		return '/formbuilder/submitted/' . $contactform->get('id');
	}

	/**
	 * Used to track inbound user sessions and sources.
	 *
	 * @static
	 */
	public static function RegisterUserSession() {
		if (!isset($_SESSION['contactformreferrer'])) {
			// Yup, this is it. ^_^
			$_SESSION['contactformreferrer'] = (isset($_SERVER['HTTP_REFERER'])) ? $_SERVER['HTTP_REFERER'] : 'anonymous-source';
		}
	}
}
