<?php
/**
 * Class file for the controller FormBuilderController
 *
 * @package FormBuilder
 * @author Nicholas Hinsch <nicholas@evalagency.com>
 */

use Core\BadRequestException;
use Core\Controller;
use Core\Date\DateTime;
use Core\Filestore\Factory;
use Core\Forms\Form;
use Core\ListingTable\Table;
use Core\View;
use Core\NotFoundException;
use Core\AccessDeniedException;


class FormBuilderController extends Controller {

	/**
	 * @core-access-string p:/formbuilder/manage_all
	 */
	public function admin() {
		$view = $this->getView();
		$request = $this->getRequest();

		$forms = FormBuilderModel::Find();

		$view->addControlTLI('Create New Form', '/formbuilder/create', 'add');

		$view->mastertemplate = 'admin';
		$view->title = 'Custom Forms';
		$view->assign('forms', $forms);
	}

	/**
	 * Admin page to create a new form.
	 *
	 * @core-access-string p:/formbuilder/manage_all
	 *
	 * @return int
	 */
	public function create() {
		$view = $this->getView();
		$request = $this->getRequest();

		$formbuilder = new FormBuilderModel();
		$form = new Form();
		$form->set('callsmethod', 'FormBuilderHelper::FormBuilderFormHandler');

		$form->addModel($formbuilder, 'model');

		$form->addElement('submit', array('value' => 'Create'));

		$view->addBreadcrumb('Custom Forms', '/formbuilder/admin');
		$view->mastertemplate = 'admin';
		$view->templatename = '/pages/formbuilder/update.tpl';
		$view->title = 'Create New Form';
		$view->assignVariable('form', $form);
		$view->assign('formbuilder', $formbuilder);
	}

	/**
	 * Admin page to update an existing form.
	 *
	 * @return int
	 */
	public function update() {
		$view     = $this->getView();
		$request  = $this->getRequest();

		$formbuilder    = new FormBuilderModel($request->getParameter(0));
		$this->_checkEditAccess($formbuilder);

		$form = new Form();
		$form->set('callsmethod', 'FormBuilderHelper::FormBuilderFormHandler');

		$form->addModel($formbuilder, 'model');

		$form->addElement('submit', array('value' => 'Update'));

		$view->addBreadcrumb('Custom Forms', '/formbuilder/admin');
		$view->mastertemplate = 'admin';
		$view->title = 'Update Form';
		$view->assignVariable('form', $form);
		$view->assign('formbuilder', $formbuilder);
	}

	/**
	 * View a contact form, this is the controller method that is mapped to a frontend page.
	 */
	public function view(){
		$page    = $this->getPageModel();
		$view    = $this->getView();
		$request = $this->getRequest();

		if(!$this->setAccess($page->get('access'))){
			return View::STATUS_ACCESSDENIED;
		}

		$m = new FormBuilderModel($request->getParameter(0));
		$formtitle = Core\str_to_url($page->get('title'));

		if(!$m->exists()){
			return View::STATUS_NOTFOUND;
		}
		
		$form = $m->getForm();
		
		if($m->isSecure()){
			$view->ssl = true;
		}

		$view->addControls($m);
		$view->assign('model', $m);
		$view->assign('form', $form);
		$view->assign('title', $formtitle);
		$view->assign('page', $page);
	}

	/**
	 * Delete an entire form and all submissions
	 *
	 * @return int
	 */
	public function delete() {
		$view    = $this->getView();
		$request = $this->getRequest();

		$form = new FormBuilderModel($request->getParameter(0));
		if (!$form->checkExists()) {
			throw new NotFoundException();
		}
		$manager = \Core\user()->checkAccess('p:/formbuilder/manage_all');

		if (!$manager) {
			throw new AccessDeniedException();
		}

		if (!$request->isPost()) {
			throw new BadRequestException();
		}

		// Delete the admin page first while we still know the ID
		$adminPage = PageModel::Find(['baseurl = ' . '/formbuilder/listing/' . $form->id], 1, null);
		if($adminPage) {
			$adminPage->delete();
		}

		$form->delete();
		\core\go_back();
	}


	/**
	 * Page after "view".
	 * @return int
	 */
	public function submitted(){
		$view    = $this->getView();
		$request = $this->getRequest();

		/** @var FormBuilderModel $m */
		$m = FormBuilderModel::Construct($request->getParameter(0));

		if(!$m->exists()){
			return View::STATUS_NOTFOUND;
		}

		$view->title = 'Thank You';
		$view->assign('body', $m->get('confirmation_text'));
		$view->templatename = '/pages/formbuilder/submitted.tpl';
	}

	/**
	 * List all the submissions for this given form.
	 */
	public function listing() {
		$view     = $this->getView();
		$request  = $this->getRequest();

		/** @var FormBuilderModel $form */
		$form = FormBuilderModel::Construct($request->getParameter(0));
		$this->_checkListAccess($form);

		$table = new \FormBuilder\FormBuilderEntryTable($form);
		$table->setCSVDownloadable();
		$table->loadFromRequest($request);
		
		$view->mastertemplate = 'admin';
		$view->addControls($form);
		$view->title = $form->get('title') . ' Submissions';
		$view->assign('listings', $table);
	}

	/**
	 * Show the details for a given entry.
	 */
	public function entry_view(){
		$view     = $this->getView();
		$request  = $this->getRequest();
		$entry = new FormBuilderEntryModel($request->getParameter(0));

		if(!$entry->checkExists()) {
			throw new NotFoundException();
		}

		$form = $entry->getLink('FormBuilder');
		$this->_checkListAccess($form);

		if(!$entry->read) {
			$entry->read = DateTime::NowGMT();
			$entry->save();
		}

		$view->addBreadcrumb('Custom Forms', '/formbuilder/admin');
		$view->addBreadcrumb($form->get('title') . ' Submissions', '/formbuilder/listing/' . $form->get('id'));
		$view->mastertemplate = 'admin';
		$view->title = 'Submission Details';
		$view->assign('entry', $entry);
		$view->assign('formurl', $form->get('baseurl'));
		$view->assign('formname', $form->get('title'));
	}

	/**
	 * Show the details for a given entry.
	 */
	public function entry_delete(){
		$view     = $this->getView();
		$request  = $this->getRequest();
		$entry = new FormBuilderEntryModel($request->getParameter(0));

		if(!$entry->exists()){
			return View::STATUS_NOTFOUND;
		}

		$form = $entry->getLink('FormBuilder');
		$access  = $this->_checkListAccess($form);
		if($access != View::STATUS_OK){
			return $access;
		}
		if (!$request->isPost()) {
			return View::STATUS_BADREQUEST;
		}

		$entry->delete();
		\core\go_back();
	}

	/**
	 * Just a helper function to return the error on a given form for the current user.
	 *
	 * This is only used internally and is just to reduce code duplication.
	 *
	 * @throws NotFoundException
	 * @throws AccessDeniedException
	 */
	private function _checkListAccess(FormBuilderModel $form) {
		if(!$form->checkExists()) {
			throw new NotFoundException();
		}

		// This requires at least listing access for the form.
		$lister  = \Core\user()->checkAccess($form->get('listpermissions'));
		$editor  = \Core\user()->checkAccess($form->get('editpermissions'));
		$manager = \Core\user()->checkAccess('p:/formbuilder/manage_all');

		if(!($lister || $editor || $manager)) {
			throw new AccessDeniedException();
		}
	}

	/**
	 * Just a helper function to return the error on a given form for the current user.
	 *
	 * This is only used internally and is just to reduce code duplication.
	 *
	 * @throws NotFoundException
	 * @throws AccessDeniedException
	 */
	private function _checkEditAccess(FormBuilderModel $form) {
		if(!$form->checkExists()){
			throw new NotFoundException();
		}

		// This requires at least listing access for the form.
		//$lister  = \Core\user()->checkAccess($form->get('listpermissions'));
		$editor  = \Core\user()->checkAccess($form->get('editpermissions'));
		$manager = \Core\user()->checkAccess('p:/formbuilder/manage_all');

		if(!($editor || $manager)){
			throw new AccessDeniedException();
		}
	}
}
