<?php
/**
 * Class file for the model FormBuilderEntryFieldModel
 *
 * @package FormBuilder
 * @author Nicholas Hinsch <nicholas@evalagency.com>
 */

use Core\Model;
use Core\View;
use Core\Filestore\Factory as FileFactory;


class FormBuilderEntryFieldModel extends Model {

	public static $ExportGroup = 'Custom Form Submissions';

	/**
	 * Schema definition for FormBuilderEntryFieldModel
	 *
	 * @static
	 * @var array
	 */
	public static $Schema = array(
		'id' => [
			'type' => Model::ATT_TYPE_UUID,
		],
		'formid' => [
			'type'     => Model::ATT_TYPE_UUID_FK,
			'required' => true,
			'null'     => false,
			'formtype' => 'system',
		],
		'fieldid' => [
			'type'     => Model::ATT_TYPE_UUID_FK,
		],
		'entryid' => [
			'type' => Model::ATT_TYPE_UUID_FK,
		],
	    'type' => [
		    'type' => Model::ATT_TYPE_STRING,
	    ],
	    'title' => [
		    'type' => Model::ATT_TYPE_STRING,
	    ],
	    'value' => [
		    'type' => Model::ATT_TYPE_DATA,
			'formatter' => ['this', 'getValue'],
	    ],
		'encrypted' => [
			'type' => Model::ATT_TYPE_BOOL,
		    'default' => false,
		    'comment' => 'Requried for on-the-fly decryption',
		],
	);

	/**
	 * Index definition for FormEntryModel
	 *
	 * @static
	 * @var array
	 */
	public static $Indexes = array(
		'primary' => array('id'),
	);
	
	public function getValue(string $format = null) : string {
		if($this->get('decrypted_value')){
			// This supports on-the-fly encryption and actual private fields.
			$val = $this->get('decrypted_value');
		}
		elseif($this->get('encrypted')){
			// This field is requested to be encrypted.
			// I need to decrypt it first.
			$val = Model::DecryptValue($this->get('value'));
			$this->set('decrypted_value', $val);
		}
		else{
			$val = $this->get('value');
		}
		
		if($format === View::CTYPE_HTML) {
			switch($this->get('type')) {
				case 'file':
					return $this->_translateFileValue($val);
				case 'checkboxes':
					$vals = explode("\n", $val);
					return '<ul><li>' . implode('</li><li>', $vals) . '</li></ul>';
					break;
				default:
					return $this->_translateGenericValue($val);
			}
		}
		else{
			// No translations needed.
			return $val;
		}
	}
	
	private function _translateGenericValue($val) : string {
		return nl2br(htmlentities($val));
	}
	
	private function _translateFileValue($val) : string {
		
		if($val === '' || $val === null) {
			return '';
		}
		
		$file = FileFactory::File($val);
		
		$base = $file->getBaseFilename();
		
		
		if(!$file->exists()) {
			return $base . ' (File Not Located)';
		}
		
		$url  = $file->getURL();
		$size = $file->getFilesize(true);
		
		if($file->isImage()) {
			return '<a href="' . $url . '" target="_blank" title="' . $base . ' (' . $size . ')">' .
				'<img src="' . $file->getPreviewURL('100x80') . '"/>' .
				'</a>';
		}
		else{
			return '<a href="' . $url . '" target="_blank">' .
				$base . ' (' . $size . ')' .
				'</a>';
		}
	}
}
