<?php
/**
 * Class file for the model FormBuilderEntryModel
 *
 * @package FormBuilder
 * @author Nicholas Hinsch <nicholas@evalagency.com>
 */

use Core\Filestore\Factory;
use Core\Formatter\GeneralFormatter;
use Core\Model;


class FormBuilderEntryModel extends Model {

	public static $ExportGroup = 'Custom Form Submissions';

	/**
	 * Schema definition for FormBuilderEntryModel
	 *
	 * @static
	 * @var array
	 */
	public static $Schema = array(
		'id' => [
			'type' => Model::ATT_TYPE_UUID,
		],
		'formid' => array(
			'type'     => Model::ATT_TYPE_UUID_FK,
			'required' => true,
			'null'     => false,
			'formtype' => 'system',
		),
		'ip' => array(
			'type'      => Model::ATT_TYPE_STRING,
			'required'  => false,
			'formtype'  => 'hidden',
			'formatter' => 'Core\Formatter\GeneralFormatter::IPAddress',
		),
		'user_id'    => array(
			'type'      => Model::ATT_TYPE_UUID_FK,
			'default'   => 0,
			'formtype'  => 'disabled',
			'formatter' => 'Core\Formatter\GeneralFormatter::User',
		),
		'useragent' => array(
			'type'      => Model::ATT_TYPE_TEXT,
			'required'  => false,
			'formtype'  => 'hidden',
			'formatter' => 'Core\Formatter\GeneralFormatter::UserAgent',
		),
		'referrer'  => array(
			'type'     => Model::ATT_TYPE_STRING,
			'required' => false,
			'formtype' => 'hidden',
		),
		'source_page' => [
			'type' => Model::ATT_TYPE_STRING,
			'required' => false,
			'formtype' => 'hidden',
		    'comment' => 'Source page, useful for form widgets!',
		],
		'junk' => [
			'type' => Model::ATT_TYPE_BOOL,
			'default' => false,
			'formtype' => 'disabled',
			'formatter' => ['this', 'getJunkStatus'],
			'comment' => 'Set to true when flagged as junk.',
		],
		'spam_checks' => [
			'type'      => Model::ATT_TYPE_TEXT,
			'required'  => false,
			'formtype'  => 'disabled',
			'comment'   => 'Contents of spam checks',
		],
		'read'      => [
			'type' => Model::ATT_TYPE_INT,
			'default' => 0,
			'comment' => 'Timestamp when this entry is read',
		],
		'created'   => [
			'type'      => Model::ATT_TYPE_CREATED,
			'null'      => false,
			'formatter' => 'Core\Formatter\GeneralFormatter::DateStringSDT',
		],
		'updated'   => [
			'type'      => Model::ATT_TYPE_UPDATED,
			'null'      => false,
			'formatter' => 'Core\Formatter\GeneralFormatter::DateStringSDT',
		],
	);

	/**
	 * Index definition for FormEntryModel
	 *
	 * @static
	 * @var array
	 */
	public static $Indexes = array(
		'primary' => array('id'),
	);

	public function __construct(...$keys) {
		$this->_linked = array(
			'FormBuilder' => [
				'link' => Model::LINK_BELONGSTOONE,
			    'on' => ['id' => 'formid'],
			],
			'FormBuilderEntryField' => [
				'link' => Model::LINK_HASMANY,
				'on' => array('entryid' => 'id'),
			],
		);

		parent::__construct(...$keys);
	}
	
	public function get($key, $format = null) {
		if(strpos($key, 'field_') === 0){
			// This is a request for a field value; use the helper method to retrieve that.
			return $this->getValueByFieldId(substr($key, 6), $format);
		}
		else{
			return parent::get($key, $format);
		}
	}

	/**
	 * Method to retrieve a field's submitted value based on its ID.
	 *
	 * @param $id
	 *
	 * @return string|null
	 */
	public function getValueByFieldId($id, $format = null){
		$fields = $this->getLink('FormBuilderEntryField');
		$val    = null;
		$type   = $this->getTypeByFieldId($id);

		foreach($fields as $f){
			/** @var FormBuilderEntryFieldModel $f */
			if($f->get('fieldid') == $id){
				return $f->getValue($format);
			}
		}
		
		return null;
	}

	/**
	 * Method to retrieve a field's submitted value based on its ID.
	 *
	 * @param $id
	 *
	 * @return string|null
	 */
	public function getTypeByFieldId($id){
		$fields = $this->getLink('FormBuilderEntryField');

		foreach($fields as $f){
			/** @var FormBuilderEntryFieldModel $f */
			if($f->get('fieldid') == $id){
				return $f->get('type');
			}
		}

		return null;
	}

	public function getJunkStatus($format = null) : string {
		if($this->junk) {
			if($format === 'text/html') {
				return '<i class="icon icon-poop" title="Spam"></i> Spam';
			}
			else {
				return 'Junk';
			}
		}
		else {
			if($format === 'text/html') {
				return '<i class="icon icon-thumbs-up" title="Not Spam"></i> Not Spam';
			}
			else {
				return 'Not Junk';
			}
		}
	}

	/**
	 * Check to see if this form has an email address.
	 *
	 * @return bool
	 */
	public function hasEmailRecipient(){
		$fields = $this->getLink('FormBuilderEntryField');

		foreach($fields as $f){
			/** @var FormBuilderEntryFieldModel $f */
			if($f->get('type') == 'email' && $f->get('value')){
				return true;
			}
		}

		return false;
	}

	/**
	 * Get the email recipient, (useful for reply-to)
	 *
	 * @return string|null
	 */
	public function getEmailRecipient(){
		$fields = $this->getLink('FormBuilderEntryField');

		foreach($fields as $f){
			/** @var FormBuilderEntryFieldModel $f */
			if($f->get('type') == 'email' && $f->get('value')){
				return $f->get('value');
			}
		}

		return null;
	}

	/**
	 * Get the recipient's NAME, (useful for reply-to)
	 *
	 * @return string|null
	 */
	public function getRecipient(){
		$fields = $this->getLink('FormBuilderEntryField');

		foreach($fields as $f){
			/** @var FormBuilderEntryFieldModel $f */
			if($f->get('type') == 'name' && $f->get('value')){
				return $f->get('value');
			}
		}

		return null;
	}
	
	public function getControlLinks() : array {
		$ret = [];
		
		$manager = \Core\user()->checkAccess('p:/formbuilder/manage_all');
		$editor  = (\Core\user()->checkAccess($this->get('editpermissions')) || $manager);
		$lister  = (\Core\user()->checkAccess($this->get('listpermissions')) || $editor);
		$id      = $this->get('id');
		
		if($lister){
			$ret[] = [
				'title' => 'Details',
				'link' => '/formbuilder/entry/view/' . $id,
				'icon' => 'view',
			];
			
			$ret[] = [
				'title' => 'Delete',
				'link' => '/formbuilder/entry/delete/' . $id,
				'icon' => 'delete',
				'confirm' => 'Are you sure you want to delete this submission?',
			];
		}
		
		return $ret;
	}
}
