<?php
/**
 * Class file for the model FormBuilderModel
 *
 * @package FormBuilder
 * @author Nicholas Hinsch <nicholas@evalagency.com>
 */

use Core\Forms\Form;
use Core\Model;


class FormBuilderModel extends Model {

	public static $ExportGroup = 'Custom Forms';

	/**
	 * Schema definition for FormModel
	 *
	 * @static
	 * @var array
	 */
	public static $Schema = [
		'id' => [
			'type' => Model::ATT_TYPE_UUID,
		],
		'site' => [
			'type' => Model::ATT_TYPE_SITE,
			'formtype' => 'system',
		],
		'title' => [
			'type' => Model::ATT_TYPE_STRING,
			'form' => [
				'group' => 't:STRING_BASIC',
				'grouptype' => 'tabs',
			],
		],
		'geo_restrict' => [
			'type' => Model::ATT_TYPE_STRING,
			'form' => [
				'group' => 't:STRING_BASIC',
				'grouptype' => 'tabs',
			],
		],
		'notify_email' => [
			'type' => Model::ATT_TYPE_TEXT,
			'required' => false,
			'form' => [
				'title' => 'Notify Email Addresses',
				'description' => 'List of email(s) to send submissions to, list multiple emails by a new line',
				'group' => 'Confirmation / Thank You',
				'grouptype' => 'tabs',
			],
		],
		'editpermissions' => [
			'type' => Model::ATT_TYPE_STRING,
			'default' => '!*',
			'form' => [
				'type' => 'access',
				'title' => 'Edit Permissions',
				'description' => 'Permissions for who is allowed to edit this page and list the entries',
				'group' => 'Access & Advanced',
				'grouptype' => 'tabs',
			]
		],
		'listpermissions' => [
			'type' => Model::ATT_TYPE_STRING,
			'default' => '!*',
			'form' => [
				'type' => 'access',
				'title' => 'Listing Permissions',
				'description' => 'Permissions for who is allowed to list the submissions',
				'group' => 'Access & Advanced',
				'grouptype' => 'tabs',
			]
		],
		'confirmation_text' => [
			'type' => Model::ATT_TYPE_TEXT,
			'form' => [
				'type' => 'wysiwyg',
				'title' => 'Successful confirmation content',
				'description' => 'The text that is displayed on the "thank you" page',
				'group' => 'Confirmation / Thank You',
				'grouptype' => 'tabs',
			]
		],
		'thankyou_email' => [
			'type' => Model::ATT_TYPE_TEXT,
			'form' => [
				'type' => 'wysiwyg',
				'title' => 'Thank You Email Text',
				'description' => 'The text that is displayed on the confirmation email.  If left blank, no email is sent to the user.',
				'group' => 'Confirmation / Thank You',
				'grouptype' => 'tabs',
			]
		],
		'created' => [
			'type' => Model::ATT_TYPE_CREATED,
			'null' => false,
		],
		'updated' => [
			'type' => Model::ATT_TYPE_UPDATED,
			'null' => false,
		],
	];

	/**
	 * Index definition for FormModel
	 *
	 * @static
	 * @var array
	 */
	public static $Indexes = array(
		'primary' => array('id'),
	);

	public function __construct(...$keys) {
		$this->_linked = array(
			'FormBuilderEntry' => array(
				'link' => Model::LINK_HASMANY,
				'on' => array('formid' => 'id'),
			),
			'FormBuilderField' => array(
				'link' => Model::LINK_HASMANY,
				'on' => array('formid' => 'id'),
			    'order' => 'weight',
			),
			'Widget' => array(
				'link' => Model::LINK_HASONE,
				'on' => 'baseurl',
			),
		);

		parent::__construct(...$keys);
	}

	public function get($k, $format = null) {
		switch ($k) {
			case 'baseurl':
			case 'rewriteurl':
				return '/formbuilder/view/' . $this->get('id', $format);
				break;
			default:
				return parent::get($k, $format);
		}
	}

	/**
	 * Get the directory to upload images to, excluding the public/private component.
	 *
	 * @return string
	 */
	public function getUploadDirectory() : string {
		// Determine the directory to upload to.  This is just a nit-picky backend thing.
		// This will keep the files organized into their own individual directories (for each album)
		$dir = $this->title;
		// Trim off any invalid characters
		$dir = \Core\str_to_url($dir);
		// And the directory character.
		$dir = str_replace('/', '', $dir);

		return $dir . '/';
	}

	public function getFieldCount() : int {
		$links = $this->getLink('FormBuilderField');
		return count($links);
	}

	public function getSubmissionCount() : int {
		return \Core\Datamodel\Dataset::Init()
			->count()
			->table('form_builder_entry')
			->whereEquals('formid', $this->id)
			->executeAndGet();
	}

	public function getUnreadSubmissionCount() : int {
		return \Core\Datamodel\Dataset::Init()
			->count()
			->table('form_builder_entry')
			->whereEquals('formid', $this->id)
			->whereEquals('read', 0)
			->executeAndGet();
	}

	/**
	 * Get the directory to upload images to, including the public/private component.
	 *
	 * @return mixed
	 */
	public function getFullUploadDirectory(){
		return 'public/formbuilder/' . $this->getUploadDirectory();
	}

	/**
	 * Scan through all the form elements in this form and check if one of them is marked as 'secure' or 'private'.
	 * 
	 * This is meant to toggle the page's ssl boolean, but could be used for other purposes.
	 * 
	 * @return bool
	 */
	public function isSecure(){
		$fields = $this->getLink('FormBuilderField');
		foreach($fields as $f){
			/** @var FormBuilderFieldModel $f */
			foreach($f->getOptions() as $k => $v){
				if(
					$k == 'secure' && 
					( $v == 'private' || $v == 'secure' )
				){
					// SECURE FIELD!
					return true;
				}
			}
		}
		
		// No secure elements hit!
		return false;
	}

	/**
	 * Get a resolved form from this builder that links to the necessary helper script.
	 * 
	 * @return Form
	 */
	public function getForm(){
		$form = new Form();
		$form->addElement('system', ['name' => 'formid', 'value' => $this->get('id')]);
		$form->set('callsmethod', 'FormBuilderHelper::SaveEntry');

		$fields = $this->getLink('FormBuilderField');
		foreach($fields as $f){
			/** @var FormBuilderFieldModel $f */
			$title  = $f->get('title');
			$key    = Core\str_to_url($title);
			$type   = $f->get('type');
			$name   = 'fields[' . $key . ']';
			$secure = false;
			//$options = explode("\n", str_replace(array("\r", "\r\n"), "\n", $data['options']));
			//$required = $data['required'];

			if($type === 'checkboxes'){
				// Checkboxes submit multiple values, so it requires the [] suffix.
				$name .= '[]';
			}

			$atts = [
				'title' => $title,
				'name' => $name,
				'class' => '',
				'description' => '',
			];

			if($type === 'state') {
				// Try to auto-resolve this to the user's location, if available
				if(defined('REMOTE_PROVINCE')) {
					$atts['value'] = REMOTE_PROVINCE;
				}
			}

			if($f->get('conditional_element')){
				$atts['data-conditional'] = Core\str_to_url($f->get('conditional_element'));
				$atts['data-conditional-value'] = $f->get('conditional_value');
				$atts['class'] = 'has-conditional';
			}

			foreach($f->getOptions() as $k => $v){
				if($k == 'options'){
					// Options are always exploded out
					$atts[$k] = explode("\n", str_replace(["\r\n", "\r"], "\n", $v));
				}
				elseif($k == 'secure'){
					$secure = $v;
				}
				elseif($k == 'readonly'){
					if($v == 'Yes') $atts[$k] = true;
				}
				elseif($k == 'required'){
					$atts[$k] = ($v == 'Yes');
				}
				else{
					$atts[$k] = $v;
				}
			}

			if($f->get('type') == 'file'){
				$atts['basedir'] = $this->getFullUploadDirectory();
			}

			if($secure == 'private'){
				//$atts['description'] .= ($atts['description'] ? '<br/><br/>' : '') . 'This field will be sent to us ONE TIME via a GPG encrypted email and <strong>will not be stored in our database</strong>.';
				$atts['description'] .= ($atts['description'] ? '<br/><br/>' : '') . t('MESSAGE_FORMBUILDER_ELEMENT_DESCRIPTION_AMENDMENT_FOR_PRIVATE');
			}
			elseif($secure == 'secure'){
				//$atts['description'] .= ($atts['description'] ? '<br/><br/>' : '') . 'This field will be sent to us via a GPG encrypted email and will be stored in our database using encryption.';
				$atts['description'] .= ($atts['description'] ? '<br/><br/>' : '') . t('MESSAGE_FORMBUILDER_ELEMENT_DESCRIPTION_AMENDMENT_FOR_SECURE');
			}

			$form->addElement( $f->get('type'), $atts );
		}
		
		return $form;
	}

	public function getControlLinks() : array {
		$ret = [];

		$manager = \Core\user()->checkAccess('p:/formbuilder/manage_all');
		$editor  = (\Core\user()->checkAccess($this->get('editpermissions')) || $manager);
		$lister  = (\Core\user()->checkAccess($this->get('listpermissions')) || $editor);

		if($lister){
			$ret[] = [
				'title' => 'Submissions',
				'link' => '/formbuilder/listing/' . $this->get('id'),
				'icon' => 'list',
				'data-formbuilder-type' => 'listing',
			];
		}
		if($editor){
			$ret[] = [
				'title' => 'Edit Form',
			    'link' => '/formbuilder/update/' . $this->get('id'),
			    'icon' => 'edit',
				'data-formbuilder-type' => 'edit',
			];
		}
		if($manager){
			$ret[] = [
				'title'   => 'Delete Form',
				'link'    => '/formbuilder/delete/' . $this->get('id'),
				'icon'    => 'delete',
				'confirm' => 'Are you sure you want to completely delete this form AND all submissions?',
			];
		}

		return $ret;
	}
}
