<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace GeoLookup;

use Core\Datamodel\Dataset;
use Core\Cache;
use Exception;

/**
 * Description of Lookup
 *
 * @author charlie
 */
class Lookup {
	/**
	 * Search for a result based on its IP
	 * 
	 * @param string $ip
	 * 
	 * @return \GeoLookup\Result
	 * @throws Exception
	 */
	public static function ByIP(string $ip) : Result {
		$ipLong = ip2long($ip);
		
		if(($cache = Cache::Get('geoip-' . $ipLong))) {
			return new Result($cache);
		}
		
		// Check if this IP is a local/internl IP first!
		if(\Core\is_ip_private($ip)) {
			// Use the server's timezone to at least provide the country of this request.
			$result = Lookup::ByTimezone(date_default_timezone_get());
			
			Cache::Set('geoip-' . $ipLong, $result->raw);
			return $result;
		}
		else{
			$db = DBHelper::GetConnection();
		
			$dataset = Dataset::Init()
				->select('c1.continent', 'c1.code country', 'p1.code province', 'c2.name city', 'p2.code postal', 'p2.lat', 'p2.lng', 't.name timezone')
				->table('ip')
				->joinLeft('country', 'ip.country = c1._rowid_', 'c1')
				->joinLeft('province', 'ip.province = p1._rowid_', 'p1')
				->joinLeft('city', 'ip.city = c2._rowid_', 'c2')
				->joinLeft('postal', 'ip.postal = p2._rowid_', 'p2')
				->joinLeft('timezone', 'c2.timezone = t._rowid_', 't')
				->where('ip.start < ' . $ipLong)
				->where('ip.end > ' . $ipLong)
				->limit(1);

			if(!($row = $dataset->executeAndGet($db))) {
				throw new Exception($ip . ' not located in database');
			}
			
			Cache::Set('geoip-' . $ipLong, $row);
			return new Result($row);
		}
	}
	
	/**
	 * Search for a result based on geo lat/lng coordinates
	 * 
	 * @param float $lat
	 * @param float $lng
	 * 
	 * @return \GeoLookup\Result
	 * 
	 * @throws Exception
	 */
	public static function ByLatLng(float $lat, float $lng) : Result {
		$checkLat = round($lat, 1);
		$checkLng = round($lng, 1);
		
		$db = DBHelper::GetConnection();
		
		$dataset = Dataset::Init()
			->select('c1.continent', 'c1.code country', 'p1.code province', 'c2.name city', 'p2.code postal', 'p2.lat', 'p2.lng', 't.name timezone')
			->table('postal', 'p2')
			->joinLeft('city', 'p2.city = c2._rowid_', 'c2')
			->joinLeft('timezone', 'c2.timezone = t._rowid_', 't')
			->joinLeft('province', 'c2.province = p1._rowid_', 'p1')
			->joinLeft('country', 'p1.country = c1._rowid_', 'c1')
			->where('p2.lat >= ' . ($checkLat - 0.2))
			->where('p2.lat <= ' . ($checkLat + 0.2))
			->where('p2.lng >= ' . ($checkLng - 0.2))
			->where('p2.lng <= ' . ($checkLng + 0.2));
		
		$rows = $dataset->executeAndGet($db);
		
		$closest = null;
		$distance = 99;
		
		foreach($rows as $k => $dat) {
			$checkDist = abs(abs($dat['lat']) - abs($lat)) + abs(abs($dat['lng']) - abs($lng));
			
			if($checkDist < $distance) {
				$distance = $checkDist;
				$closest = $k;
			}
		}
		
		if($closest === null) {
			throw new Exception('No nearby IP addresses located in database');
		}
		
		return new Result($rows[ $closest ]);
	}
	
	/**
	 * Search for a result based on a timezone, this will only be able to provide country-level accuracy.
	 * 
	 * @param string $timezone
	 * 
	 * @return \GeoLookup\Result
	 * 
	 * @throws Exception
	 */
	public static function ByTimezone(string $timezone) : Result {
		
		$db = DBHelper::GetConnection();
		
		$dataset = Dataset::Init()
			->select('c1.continent', 'c1.code country', 't.name timezone')
			->table('timezone', 't')
			->joinLeft('city', 't._rowid_ = c2.timezone', 'c2')
			->joinLeft('province', 'c2.province = p1._rowid_', 'p1')
			->joinLeft('country', 'p1.country = c1._rowid_', 'c1')
			->where('t.name = ' . $timezone)
			->unique()
			->limit(1);
		
		$row = $dataset->executeAndGet($db);
		
		if(!$row) {
			throw new Exception('No matching timezone located in database');
		}
		
		return new Result($row);
	}
	
	/**
	 * Get all countries in the database
	 * 
	 * @return Result[]
	 */
	public static function GetCountries() : array {
		$db = DBHelper::GetConnection();
		
		$dataset = Dataset::Init()
			->select('continent', 'code country')
			->table('country')
			->order('continent, code')
			->unique();
		
		$rows = $dataset->executeAndGet($db);
		$results = [];
		
		foreach($rows as $row) {
			$results[] = new Result($row);
		}
		
		return $results;
	}
	
	/**
	 * Get all provinces in the database based on a given ISO2 country code
	 * 
	 * @return Result[]
	 */
	public static function GetProvinces(string $countryCode) : array {

		if(strlen($countryCode) !== 2) {
			throw new Exception('Please provide a 2-digit country code to lookup provinces within');
		}
		
		$db = DBHelper::GetConnection();
		
		$dataset = Dataset::Init()
			->select('c1.continent', 'c1.code country', 'p1.code province')
			->table('country', 'c1')
			->joinLeft('province', 'p1.country = c1._rowid_', 'p1')
			->where('c1.code = ' . $countryCode)
			->order('province');
		
		$rows = $dataset->executeAndGet($db);
		$results = [];
		
		foreach($rows as $row) {
			$results[] = new Result($row);
		}
		
		return $results;
	}
}
