/**
 * Inline style editor for Core Plus,
 * allows live editing of CSS variables in real time
 */

// a = window.getMatchedCSSRules(document.querySelector('.page-header'))
// Object.values(a[0].style).forEach(e => {console.log(e, a[0].style[e]); })

(() => {

	let styleEditorUI;

	class VariableStyle {
		constructor() {
			/** @var {Element} */
			this.target = null;
			/** @var {string} */
			this.style = '';
			/** @var {string} */
			this.variable = '';
		}

		/**
		 * Get the current set value for this style
		 *
		 * @returns {string}
		 */
		getValue() {
			let computedStyle = window.getComputedStyle(this.target),
				value = computedStyle.getPropertyValue('--' + this.variable);

			// An empty string here means empty, "inherit", or some other weird value.
			// .... probably, this is a javascript library for parsing CSS styles after all
			// anyway, skim up through the DOM until an element can be located with a valid value for this given style
			if(value === '') {
				let parent = this.target;
				while(value === '' && parent !== null) {
					value = window.getComputedStyle(parent).getPropertyValue(this.style);

					parent = parent.parentElement;
				}
			}

			//console.log(this.style, value);

			return value
		}

		/**
		 * Get the state modifier for this variable, dependent on the variable suffix
		 *
		 * @returns {string} one of 'active', 'hover', 'disabled', 'selected', 'default'
		 */
		getStateModifier() {
			if(this.variable.endsWith('-active')) {
				return 'active';
			}
			else if(this.variable.endsWith('-hover')) {
				return 'hover';
			}
			else if(this.variable.endsWith('-disabled')) {
				return 'disabled';
			}
			else if(this.variable.endsWith('-selected')) {
				return 'selected';
			}
			else {
				return 'default';
			}
		}

		/**
		 * Get the title / label for this variable style
		 *
		 * @returns {string}
		 */
		getTitle() {
			let title = this.style.replace(/-/g, ' ').ucwords(),
				state = this.getStateModifier();

			if(state !== 'default') {
				title += ' (' + state + ')';
			}

			return title;
		}

		getInput() {
			let input = document.createElement('input');

			input.type = 'text';
			input.classList.add('value');
			input.value = this.getValue();

			// Assign the change event to keep track of the values
			input.addEventListener('change', evt => {
				document.documentElement.style.setProperty('--' + this.variable, evt.target.value);
			});

			return input;
		}


		/**
		 *
		 * @param {Element} target
		 * @param {string} style
		 * @param {string} variable
		 *
		 * @return {VariableStyle}
		 */
		static Factory(target, style, variable) {
			let vs;

			// Different keys will use different classes for custom rendering.
			// Colours for example should render as a color input.
			if(style === 'color' || style === 'background-color' || style === 'border-color') {
				// Style selections that should render as a color picker
				vs = new VariableStyleColor();
			}
			else if(style === 'border-style') {
				vs = new VariableStyleBorderStyle();
			}
			else if(
				[
					'height', 'min-height', 'max-height',
					'width', 'min-width', 'max-width'
				].indexOf(style) !== -1
			) {
				vs = new VariableStyleSize();
			}
			else if(
				[
					'font-size', 'line-height'
				].indexOf(style) !== -1
			) {
				vs = new VariableStyleFontSize();
			}
			else if(style === 'font-family') {
				vs = new VariableStyleFontFamily();
			}
			else if(['border-width', 'padding', 'margin'].indexOf(style) !== -1) {
				vs = new VariableStyleSizes();
			}
			else {
				// Everything else, just gets a text field by default.
				vs = new VariableStyle();
			}

			vs.target = target;
			vs.style = style;
			vs.variable = variable;

			return vs;
		}
	}

	/**
	 * Color selector variable style, will render a color input
	 */
	class VariableStyleColor extends VariableStyle {

		/**
		 * Get the current set value for this style
		 *
		 * @returns {string}
		 */
		getValue() {
			let computedStyle = window.getComputedStyle(this.target),
				value = computedStyle.getPropertyValue('--' + this.variable);

			// An empty string here means empty, "inherit", or some other weird value.
			// .... probably, this is a javascript library for parsing CSS styles after all
			// anyway, skim up through the DOM until an element can be located with a valid value for this given style
			if(value === '') {
				let parent = this.target;
				while(value === '' && parent !== null) {
					value = window.getComputedStyle(parent).getPropertyValue(this.style);

					if(value === 'rgba(0, 0, 0, 0)') {
						// Transparent doesn't count, FF will use this for elements that do not have a colour defined.
						value = '';
					}

					parent = parent.parentElement;
				}
			}

			//console.log(this.style, value);

			return value
		}

		getProcessedValue() {
			let value = this.getValue().trim();

			if(value.match(/^#[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]$/) !== null) {
				// 3-digit hex codes need converted to full 6-digit codes
				return value.replace(/^#([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])$/, '#$1$1$2$2$3$3');
			}
			else if(typeof(VariableStyleColor.ColorNames[value.toLowerCase()]) !== 'undefined') {
				// Convert CSS colour names to their hex counterpart
				return VariableStyleColor.ColorNames[value.toLowerCase()];
			}
			else if(value.match(/^rgb\([0-9]{1,3}, [0-9]{1,3}, [0-9]{1,3}\)/) !== null) {
				// FF doesn't like RGB values for colours
				let r = value.replace(/^rgb\(([0-9]{1,3}), ([0-9]{1,3}), ([0-9]{1,3})\)/, '$1'),
					g = value.replace(/^rgb\(([0-9]{1,3}), ([0-9]{1,3}), ([0-9]{1,3})\)/, '$2'),
					b = value.replace(/^rgb\(([0-9]{1,3}), ([0-9]{1,3}), ([0-9]{1,3})\)/, '$3');
				return '#' +
					(r < 16 ? '0' : '') +
					parseInt(r).toString(16) +
					(g < 16 ? '0' : '') +
					parseInt(g).toString(16) +
					(b < 16 ? '0' : '') +
					parseInt(b).toString(16);
			}
			else {
				// Default
				return value;
			}
		}

		getInput() {
			let input = document.createElement('input');

			input.type = 'color';
			input.classList.add('value');
			input.value = this.getProcessedValue();

			// Assign the change event to keep track of the values
			input.addEventListener('change', evt => {
				document.documentElement.style.setProperty('--' + this.variable, evt.target.value);
			});

			return input;
		}

		/**
		 * Calculate the luminance of a given hex value for most humans
		 *
		 * @param {string} hexValue Full hex number with '#' prefix
		 * @returns {number} Luminance value of colour
		 */
		static Luminance(hexValue) {
			let rgb = hexValue.substr(1).match(/.{2}/g),
				a = rgb
					.map(v => {
						// Convert hex to int
						return parseInt(v, 16);
					})
					.map(v => {
						// Calculate base luminance of individual colour
						v /= 255;
						return v <= 0.03928
							? v / 12.92
							: Math.pow( (v + 0.055) / 1.055, 2.4 );
					});

			// Each colour has a different brightness correspondence to human eyes.
			return a[0] * 0.2126 + a[1] * 0.7152 + a[2] * 0.0722;
		}

		/**
		 * Calculate the colour contrast between two hex values
		 *
		 * For WCAG AA:
		 * 4.5 minimum for normal text
		 * 3.0 minimum for large text
		 *
		 * For WCAG AAA:
		 * 7.0 minimum for normal text
		 * 4.5 minimum for large text
		 *
		 * @param {string} hexValue1
		 * @param {string} hexValue2
		 * @returns {number} Larger the number the more contrast between colours
		 */
		static Contrast(hexValue1, hexValue2) {
			let lum1 = VariableStyleColor.Luminance(hexValue1),
				lum2 = VariableStyleColor.Luminance(hexValue2),
				brightest = Math.max(lum1, lum2),
				darkest = Math.min(lum1, lum2);

			return (brightest + 0.05) / (darkest + 0.05);
		}
	}

	/** @var {string[]} List of colour names for CSS-to-HTML translation */
	VariableStyleColor.ColorNames = {
		'aliceblue': '#F0F8FF',
		'antiquewhite': '#FAEBD7',
		'aqua': '#00FFFF',
		'aquamarine': '#7FFFD4',
		'azure': '#F0FFFF',
		'beige': '#F5F5DC',
		'bisque': '#FFE4C4',
		'black': '#000000',
		'blanchedalmond': '#FFEBCD',
		'blue': '#0000FF',
		'blueviolet': '#8A2BE2',
		'brown': '#A52A2A',
		'burlywood': '#DEB887',
		'cadetblue': '#5F9EA0',
		'chartreuse': '#7FFF00',
		'chocolate': '#D2691E',
		'coral': '#FF7F50',
		'cornflowerblue': '#6495ED',
		'cornsilk': '#FFF8DC',
		'crimson': '#DC143C',
		'cyan': '#00FFFF',
		'darkblue': '#00008B',
		'darkcyan': '#008B8B',
		'darkgoldenrod': '#B8860B',
		'darkgray': '#A9A9A9',
		'darkgrey': '#A9A9A9',
		'darkgreen': '#006400',
		'darkkhaki': '#BDB76B',
		'darkmagenta': '#8B008B',
		'darkolivegreen': '#556B2F',
		'darkorange': '#FF8C00',
		'darkorchid': '#9932CC',
		'darkred': '#8B0000',
		'darksalmon': '#E9967A',
		'darkseagreen': '#8FBC8F',
		'darkslateblue': '#483D8B',
		'darkslategray': '#2F4F4F',
		'darkslategrey': '#2F4F4F',
		'darkturquoise': '#00CED1',
		'darkviolet': '#9400D3',
		'deeppink': '#FF1493',
		'deepskyblue': '#00BFFF',
		'dimgray': '#696969',
		'dimgrey': '#696969',
		'dodgerblue': '#1E90FF',
		'firebrick': '#B22222',
		'floralwhite': '#FFFAF0',
		'forestgreen': '#228B22',
		'fuchsia': '#FF00FF',
		'gainsboro': '#DCDCDC',
		'ghostwhite': '#F8F8FF',
		'gold': '#FFD700',
		'goldenrod': '#DAA520',
		'gray': '#808080',
		'grey': '#808080',
		'green': '#008000',
		'greenyellow': '#ADFF2F',
		'honeydew': '#F0FFF0',
		'hotpink': '#FF69B4',
		'indianred': '#CD5C5C',
		'indigo': '#4B0082',
		'ivory': '#FFFFF0',
		'khaki': '#F0E68C',
		'lavender': '#E6E6FA',
		'lavenderblush': '#FFF0F5',
		'lawngreen': '#7CFC00',
		'lemonchiffon': '#FFFACD',
		'lightblue': '#ADD8E6',
		'lightcoral': '#F08080',
		'lightcyan': '#E0FFFF',
		'lightgoldenrodyellow': '#FAFAD2',
		'lightgray': '#D3D3D3',
		'lightgrey': '#D3D3D3',
		'lightgreen': '#90EE90',
		'lightpink': '#FFB6C1',
		'lightsalmon': '#FFA07A',
		'lightseagreen': '#20B2AA',
		'lightskyblue': '#87CEFA',
		'lightslategray': '#778899',
		'lightslategrey': '#778899',
		'lightsteelblue': '#B0C4DE',
		'lightyellow': '#FFFFE0',
		'lime': '#00FF00',
		'limegreen': '#32CD32',
		'linen': '#FAF0E6',
		'magenta': '#FF00FF',
		'maroon': '#800000',
		'mediumaquamarine': '#66CDAA',
		'mediumblue': '#0000CD',
		'mediumorchid': '#BA55D3',
		'mediumpurple': '#9370DB',
		'mediumseagreen': '#3CB371',
		'mediumslateblue': '#7B68EE',
		'mediumspringgreen': '#00FA9A',
		'mediumturquoise': '#48D1CC',
		'mediumvioletred': '#C71585',
		'midnightblue': '#191970',
		'mintcream': '#F5FFFA',
		'mistyrose': '#FFE4E1',
		'moccasin': '#FFE4B5',
		'navajowhite': '#FFDEAD',
		'navy': '#000080',
		'oldlace': '#FDF5E6',
		'olive': '#808000',
		'olivedrab': '#6B8E23',
		'orange': '#FFA500',
		'orangered': '#FF4500',
		'orchid': '#DA70D6',
		'palegoldenrod': '#EEE8AA',
		'palegreen': '#98FB98',
		'paleturquoise': '#AFEEEE',
		'palevioletred': '#DB7093',
		'papayawhip': '#FFEFD5',
		'peachpuff': '#FFDAB9',
		'peru': '#CD853F',
		'pink': '#FFC0CB',
		'plum': '#DDA0DD',
		'powderblue': '#B0E0E6',
		'purple': '#800080',
		'rebeccapurple': '#663399',
		'red': '#FF0000',
		'rosybrown': '#BC8F8F',
		'royalblue': '#4169E1',
		'saddlebrown': '#8B4513',
		'salmon': '#FA8072',
		'sandybrown': '#F4A460',
		'seagreen': '#2E8B57',
		'seashell': '#FFF5EE',
		'sienna': '#A0522D',
		'silver': '#C0C0C0',
		'skyblue': '#87CEEB',
		'slateblue': '#6A5ACD',
		'slategray': '#708090',
		'slategrey': '#708090',
		'snow': '#FFFAFA',
		'springgreen': '#00FF7F',
		'steelblue': '#4682B4',
		'tan': '#D2B48C',
		'teal': '#008080',
		'thistle': '#D8BFD8',
		'tomato': '#FF6347',
		'turquoise': '#40E0D0',
		'violet': '#EE82EE',
		'wheat': '#F5DEB3',
		'white': '#FFFFFF',
		'whitesmoke': '#F5F5F5',
		'yellow': '#FFFF00',
		'yellowgreen': '#9ACD32'
	};

	/**
	 * Width/Size variable style, will render a number input with unit selection
	 */
	class VariableStyleSize extends VariableStyle {
		getSuffixes() {
			return [
				// Common suffixes first
				'em', 'pt', 'px', 'rem', '%',
				// Obscure suffixes next
				'cm', 'in', 'mm', 'pc', 'vh', 'vmax', 'vmin', 'vw'
			];
		}
		getInput() {
			let inputSize = document.createElement('input'),
				inputUnit = document.createElement('select'),
				container = document.createElement('span'),
				value = this.getValue(),
				suffixes = this.getSuffixes(),
				suffix = '';

			inputSize.type = 'number';
			inputSize.step = '0.05';
			container.classList.add('value', 'value-size');

			// Build the list of options
			for(let i = 0; i < suffixes.length; i++) {
				let option = document.createElement('option');
				option.value = suffixes[i];
				option.innerHTML = suffixes[i];
				inputUnit.appendChild(option);
			}

			{
				// Blank option for "0/none"
				let option = document.createElement('option');
				option.value = '';
				option.innerHTML = '';
				inputUnit.appendChild(option);
			}

			// Check what suffix this unit currently is set to
			if(value !== '0' && value !== '') {
				for(let i = 0; i < suffixes.length; i++) {
					if(value.match(new RegExp('^[0-9.]*' + suffixes[i] + '$')) !== null) {
						// Regex needs to be used because "em" and "rem" would conflict with a simple endsWith
						suffix = suffixes[i];
						value = value.substr(0, (value.length - suffix.length));
					}
				}
			}

			// Set initial values
			inputSize.value = value;
			inputUnit.value = suffix;

			// Assign the change event to keep track of the values
			inputSize.addEventListener('change', evt => {
				let newValue = evt.target.value;

				if(newValue === '0' || newValue === '') {
					// Value is 0, units should disappear too
					inputUnit.value = '';
				}
				else if(inputUnit.value === '') {
					// Unit value was unset, (probably because value was set to 0), restore it
					inputUnit.value = suffix;
				}

				value = newValue;

				document.documentElement.style.setProperty('--' + this.variable, newValue + inputUnit.value);
			});
			inputUnit.addEventListener('change', evt => {
				let newValue = evt.target.value;

				if(newValue === '') {
					// The empty value has been set.
					inputSize.value = '0';
				}
				else {
					if(inputSize.value === '0') {
						// Restore the previously-set size
						inputSize.value = value;
					}

					// Changed, for convenience update the value to try to match the new unit
					inputSize.value = VariableStyleSize.TranslateSize(parseFloat(inputSize.value), suffix, newValue).toString();
					value = inputSize.value;
					suffix = newValue;
				}

				document.documentElement.style.setProperty('--' + this.variable, inputSize.value + newValue);
			});

			container.appendChild(inputSize);
			container.appendChild(inputUnit);


			return container;
		}

		/**
		 * Translate a value between an old unit and new unit
		 *
		 * @param {number} value
		 * @param {string} oldSuffix
		 * @param {number} newSuffix
		 *
		 * @return {number} The new raw number given the new prefix
		 */
		static TranslateSize(value, oldSuffix, newSuffix) {
			if(
				oldSuffix === '' || newSuffix === '' ||
				oldSuffix === '%' || newSuffix === '%' ||
				oldSuffix === newSuffix
			) {
				// No calculation available/required, just return as-is
				return value;
			}

			// Use a common translation table for units, won't be perfect, but it'll be something
			let translations = {
				'em': 16,
				'pt': 1.33,
				'px': 1,
				'rem': 16,
				'cm': 37.8,
				'in': 96,
				'mm': 3.8,
				'pc': 16,
				'vh': window.innerHeight / 100,
				'vmax': Math.max(window.innerHeight, window.innerWidth) / 100,
				'vmin': Math.min(window.innerHeight, window.innerWidth) / 100,
				'vw': window.innerWidth / 100
			};

			if(
				typeof(translations[oldSuffix]) !== 'undefined' &&
				typeof(translations[newSuffix]) !== 'undefined'
			) {
				// Both metrics are available in the translation table,
				// this means we can convert from the old unit to px,
				// then from px to the new unit.
				value = (value * translations[oldSuffix]) / translations[newSuffix];
			}

			if(newSuffix === 'px') {
				// px really should have no precision.
				value = Math.round(value);
			}
			else {
				// Round this number to the closest 0.05
				value = Math.round(value * 20) / 20;
			}

			return value;
		}

		/**
		 * Split the merged value to its numeric value and unit suffix
		 *
		 * @param {string} value
		 * @param {string[]} suffixes
		 *
		 * @return {Array.<{number, string}>}
		 */
		static SplitValueUnit(value, suffixes) {
			let suffix = '';

			if(value !== '0' && value !== '') {
				for(let i = 0; i < suffixes.length; i++) {
					if(value.match(new RegExp('^[0-9.]*' + suffixes[i] + '$')) !== null) {
						// Regex needs to be used because "em" and "rem" would conflict with a simple endsWith
						suffix = suffixes[i];
						value = value.substr(0, (value.length - suffix.length));
					}
				}
			}

			return [value, suffix];
		}
	}

	/**
	 * Sizes input, suitable for multi-position styles such as border-width, padding, or margin.
	 */
	class VariableStyleSizes extends VariableStyle {
		getSuffixes() {
			return [
				// Common suffixes first
				'em', 'pt', 'px', 'rem', '%',
				// Obscure suffixes next
				'cm', 'in', 'mm', 'pc', 'vh', 'vmax', 'vmin', 'vw'
			];
		}
		/**
		 * Get the current set value for this style
		 *
		 * @returns {string}
		 */
		getValue() {
			let value = super.getValue();

			if(value === '') {
				// If the default lookup fails, try the specific parameters instead.
				// This is after all broken up into 4 segments.
				let parent = this.target,
					valueTop, valueRight, valueBottom, valueLeft,
					computedStyle;

				while(value === '' && parent !== null) {
					computedStyle = window.getComputedStyle(parent);

					valueTop = computedStyle.getPropertyValue(this.style + '-top');
					valueRight = computedStyle.getPropertyValue(this.style + '-right');
					valueBottom = computedStyle.getPropertyValue(this.style + '-bottom');
					valueLeft = computedStyle.getPropertyValue(this.style + '-left');

					value = [valueTop, valueRight, valueBottom, valueLeft].join(' ').trim();
					parent = parent.parentElement;
				}
			}

			return value
		}
		getProcessedValue() {
			let value = this.getValue().trim().split(' '),
				cookedValue;

			if(value.length === 0) {
				cookedValue = {top: ['',''], right: ['',''], bottom: ['',''], left: ['','']};
			}
			else if(value.length === 1) {
				let v = VariableStyleSize.SplitValueUnit(value[0], this.getSuffixes());

				cookedValue = {
					top: v, right: v, bottom: v, left: v
				};
			}
			else if(value.length === 2) {
				let tb = VariableStyleSize.SplitValueUnit(value[0], this.getSuffixes()),
					lr = VariableStyleSize.SplitValueUnit(value[1], this.getSuffixes());

				cookedValue = {
					top: tb, right: lr, bottom: tb, left: lr
				};
			}
			else if(value.length === 3) {
				let t = VariableStyleSize.SplitValueUnit(value[0], this.getSuffixes()),
					lr = VariableStyleSize.SplitValueUnit(value[1], this.getSuffixes()),
					b = VariableStyleSize.SplitValueUnit(value[2], this.getSuffixes());

				cookedValue = {
					top: t, right: lr, bottom: b, left: lr
				};
			}
			else {
				let t = VariableStyleSize.SplitValueUnit(value[0], this.getSuffixes()),
					r = VariableStyleSize.SplitValueUnit(value[1], this.getSuffixes()),
					b = VariableStyleSize.SplitValueUnit(value[2], this.getSuffixes()),
					l = VariableStyleSize.SplitValueUnit(value[3], this.getSuffixes());

				cookedValue = {
					top: t, right: r, bottom: b, left: l
				};
			}

			return cookedValue;
		}
		getInput() {
			let inputTopSize = document.createElement('input'),
				inputRightSize = document.createElement('input'),
				inputBottomSize = document.createElement('input'),
				inputLeftSize = document.createElement('input'),
				inputTopUnit = document.createElement('select'),
				inputRightUnit = document.createElement('select'),
				inputBottomUnit = document.createElement('select'),
				inputLeftUnit = document.createElement('select'),
				containerTop = document.createElement('span'),
				containerRight = document.createElement('span'),
				containerBottom = document.createElement('span'),
				containerLeft = document.createElement('span'),
				container = document.createElement('span'),
				sizes = [inputTopSize, inputRightSize, inputBottomSize, inputLeftSize],
				units = [inputTopUnit, inputRightUnit, inputBottomUnit, inputLeftUnit],
				value = this.getProcessedValue(),
				suffixes = this.getSuffixes(),
				suffix = '',
				icon;

			for(let i = 0; i < sizes.length; i++) {
				sizes[i].type = 'number';
				sizes[i].step = '0.05';
				sizes[i].dataset.index = i.toString();

				// Assign the change event to keep track of the values
				sizes[i].addEventListener('change', evt => {
					let newValue = evt.target.value,
						index = parseInt(evt.target.dataset.index);

					if(newValue === '0' || newValue === '') {
						// Value is 0, units should disappear too
						units[index].value = '';
						newValue = '0';
					}
					else if(units[index].value === '') {
						// Unit value was unset, (probably because value was set to 0), restore it
						if(units[index].dataset.value === '') {
							units[index].dataset.value = 'px';
							units[index].value = 'px';
						}
						else {
							units[index].value = units[index].dataset.value;
						}
					}

					// Store for reverting units back
					evt.target.dataset.value = newValue;

					// Calculate the overall string to save
					newValue = VariableStyleSizes.MergeValues(
						sizes[0].value + units[0].value,
						sizes[1].value + units[1].value,
						sizes[2].value + units[2].value,
						sizes[3].value + units[3].value
					);

					document.documentElement.style.setProperty('--' + this.variable, newValue);
				});
			}

			for(let i = 0; i < units.length; i++) {
				units[i].dataset.index = i.toString();

				// Build the list of options
				for(let s = 0; s < suffixes.length; s++) {
					let option = document.createElement('option');
					option.value = suffixes[s];
					option.innerHTML = suffixes[s];
					units[i].appendChild(option);
				}

				{
					// Blank option for "0/none"
					let option = document.createElement('option');
					option.value = '';
					option.innerHTML = '';
					units[i].appendChild(option);
				}

				// Add the event listener to update the size when changed
				units[i].addEventListener('change', evt => {
					let newValue = evt.target.value,
						index = parseInt(evt.target.dataset.index);

					if(newValue === '') {
						// The empty value has been set.
						sizes[index].value = '0';
					}
					else {
						if(sizes[index].value === '0') {
							// Restore the previously-set size
							sizes[index].value = sizes[index].dataset.value;
						}

						// Changed, for convenience update the value to try to match the new unit
						sizes[index].value = VariableStyleSize.TranslateSize(parseFloat(sizes[index].value), evt.target.dataset.value, newValue).toString();
						sizes[index].dataset.value = sizes[index].value;
						evt.target.dataset.value = newValue;
					}

					// Calculate the overall string to save
					newValue = VariableStyleSizes.MergeValues(
						sizes[0].value + units[0].value,
						sizes[1].value + units[1].value,
						sizes[2].value + units[2].value,
						sizes[3].value + units[3].value
					);

					document.documentElement.style.setProperty('--' + this.variable, newValue);
				});
			}


			// Set initial values
			inputTopSize.value = value.top[0];
			inputRightSize.value = value.right[0];
			inputBottomSize.value = value.bottom[0];
			inputLeftSize.value = value.left[0];

			inputTopUnit.value = inputTopUnit.dataset.value = value.top[1];
			inputRightUnit.value = inputRightUnit.dataset.value = value.right[1];
			inputBottomUnit.value = inputBottomUnit.dataset.value = value.bottom[1];
			inputLeftUnit.value = inputLeftUnit.dataset.value = value.left[1];
			
			// Add everything to the appropriate containers
			icon = document.createElement('i');
			icon.classList.add('icon', 'icon-arrow-up');
			containerTop.append(icon, inputTopSize, inputTopUnit);

			icon = document.createElement('i');
			icon.classList.add('icon', 'icon-arrow-right');
			containerRight.append(icon, inputRightSize, inputRightUnit);

			icon = document.createElement('i');
			icon.classList.add('icon', 'icon-arrow-down');
			containerBottom.append(icon, inputBottomSize, inputBottomUnit);

			icon = document.createElement('i');
			icon.classList.add('icon', 'icon-arrow-left');
			containerLeft.append(icon, inputLeftSize, inputLeftUnit);
			
			container.classList.add('value', 'value-sizes');
			container.append(containerTop, containerRight, containerBottom, containerLeft);

			return container;
		}

		/**
		 * Translate a group of values back to a shorthand version for storing
		 *
		 * @param {string} top
		 * @param {string} right
		 * @param {string} bottom
		 * @param {string} left
		 *
		 * @return {string} The merged value
		 */
		static MergeValues(top, right, bottom, left) {

			// Data sanity
			if(top === '') {
				top = '0';
			}
			if(right === '') {
				right = '0';
			}
			if(bottom === '') {
				bottom = '0';
			}
			if(left === '') {
				left = '0';
			}

			if(top === right && top === bottom && top === left) {
				// Single value
				return top;
			}
			else if(top === bottom && left === right) {
				// 2 values
				return top + ' ' + left;
			}
			else if(left === right) {
				// 3 values
				return top + ' ' + left + ' ' + bottom;
			}
			else {
				// Full 4 values
				return top + ' ' + right + ' ' + bottom + ' ' + left;
			}
		}
	}

	/**
	 * Width/Size variable style, will render a number input with unit selection
	 *
	 * This is nearly identical to StyleSize, with the exception that the suffixes are a bit different
	 */
	class VariableStyleFontSize extends VariableStyleSize {
		getSuffixes() {
			return [
				// Common suffixes first
				'em', 'pt', 'px', 'rem',
				// Obscure suffixes next
				'cm', 'in', 'mm', 'pc', 'vh', 'vmax', 'vmin', 'vw'
			];
		}
	}

	/**
	 * Border style, will render a select box with valid options
	 */
	class VariableStyleBorderStyle extends VariableStyle {
		getInput() {
			let input = document.createElement('select'),
				options = [
					'none',
					'hidden',
					'dotted',
					'dashed',
					'solid',
					'double',
					'groove',
					'ridge',
					'inset',
					'outset',
					'initial',
					'inherit'
				];

			input.classList.add('value-style');

			// Build the list of options
			for(let i = 0; i < options.length; i++) {
				let option = document.createElement('option');
				option.value = options[i];
				option.innerHTML = options[i];
				input.appendChild(option);
			}

			// Set initial values
			input.value = this.getValue();

			// Assign the change event to keep track of the values
			input.addEventListener('change', evt => {
				let newValue = evt.target.value;
				document.documentElement.style.setProperty('--' + this.variable, newValue);
			});

			return input;
		}
	}

	/**
	 * Font family selection, will render a preferred and fallback font select box
	 */
	class VariableStyleFontFamily extends VariableStyle {

		getPreferredValue() {
			let value = this.getValue(),
				fallbackFonts = VariableStyleFontFamily.GetFallbackFonts(),
				installedFonts = VariableStyleFontFamily.GetInstalledFonts();

			if(value === 'inherit') {
				// Inherit is a special value; this switches over to the fallback font, so preferred is empty.
				return '';
			}

			// Resolve "Lato, Another Font, default"
			// to ['Lato', 'Another Font', 'default']
			// so we can iterate through and check default vs fallback fonts
			// since we only support one primary + one fallback set, this will not
			// catch all cases, notable ones with multiple primary fonts, but
			// should do a decent job at catching common settings.
			value = value
				.split(',')
				.map(v => {
					return v.trim();
				});

			for(let i = 0; i < installedFonts.length; i++) {
				if(installedFonts[i].toLowerCase() === value[0].toLowerCase()) {
					// Installed font was located!
					return installedFonts[i];
				}
			}

			// Preferred font was not detected in the installed fonts,
			// this doesn't mean it's not a preferred, just means it's not installed.
			// Parse the fallbacks to see if it's one of them, if it's not, then it's
			// just not installed.
			let preferredIsFallback = false;
			for(let i = 0; i < fallbackFonts.length; i++) {
				let f = fallbackFonts[i]
					.toLowerCase()
					.split(',');
				if(f.indexOf(value[0].toLowerCase()) !== -1) {
					// Preferred font is one of the fallback, set it to none.
					preferredIsFallback = true;
				}
			}

			if(!preferredIsFallback) {
				// Through the default scan, the first font in the value is NOT one of the fallbacks.
				// This means it's a custom font, but just not installed.
				return value[0];
			}
			else {
				// Preferred font is just a default web-safe font, that's cool too.
				return '';
			}
		}

		getFallbackValue() {
			let value = this.getValue(),
				fallbackFonts = VariableStyleFontFamily.GetFallbackFonts(),
				fallbackFont = '';

			if(value === 'inherit') {
				// Inherit is a special value; all the logic can be skipped
				return 'inherit';
			}

			// Resolve "Lato, Another Font, default"
			// to ['Lato', 'Another Font', 'default']
			// so we can iterate through and check default vs fallback fonts
			// since we only support one primary + one fallback set, this will not
			// catch all cases, notable ones with multiple primary fonts, but
			// should do a decent job at catching common settings.
			value = value
				.toLowerCase()
				.split(',')
				.map(v => {
					return v.trim();
				});


			// Process the fallback fonts now, these don't matter about the order,
			// but the priority does matter.
			// a value of "Font Name, Arial" should know that "Arial" means "Arial, sans-serif",
			// despite not being in the original value.
			// Since the more specific declaration falls later in the list, we can just be lazy about this.
			for(let i = 0; i < fallbackFonts.length; i++) {
				let f = fallbackFonts[i]
					.toLowerCase()
					.split(',');
				if(value.indexOf(f[0]) !== -1) {
					// Remember, don't break.
					// [sans-serif], [arial, sans-serif] will probably match both,
					// but we want to use the more specific one later in the list.
					fallbackFont = fallbackFonts[i];
				}
			}

			return fallbackFont;
		}

		getInput() {
			let inputFont = document.createElement('select'),
				inputFallback = document.createElement('select'),
				container = document.createElement('span'),
				fallbackFonts = VariableStyleFontFamily.GetFallbackFonts(),
				installedFonts = VariableStyleFontFamily.GetInstalledFonts(),
				preferredFont = this.getPreferredValue(),
				fallbackFont = this.getFallbackValue(),
				preferredFontInstalled = false;

			
			container.classList.add('value-font-family');

			// Build the list of options for preferred
			for(let i = 0; i < installedFonts.length; i++) {
				let option = document.createElement('option');
				option.value = installedFonts[i];
				option.innerHTML = installedFonts[i];
				inputFont.appendChild(option);

				if(installedFonts[i].toLowerCase() === preferredFont.toLowerCase()) {
					preferredFontInstalled = true;
				}
			}

			// Not installed, but still custom font
			if(!preferredFontInstalled && preferredFont !== '') {
				let option = document.createElement('option');
				option.value = preferredFont;
				option.innerHTML = preferredFont + ' (NOT INSTALLED)';
				inputFont.appendChild(option);
			}

			{
				// Blank option for "0/none"
				let option = document.createElement('option');
				option.value = '';
				option.innerHTML = '- No preferred font -';
				inputFont.appendChild(option);
			}

			// Build the list of options for fallback
			for(let i = 0; i < fallbackFonts.length; i++) {
				let option = document.createElement('option');
				option.value = fallbackFonts[i];
				option.innerHTML = fallbackFonts[i];
				inputFallback.appendChild(option);
			}

			{
				// Blank option for "0/none"
				let option = document.createElement('option');
				option.value = 'inherit';
				option.innerHTML = '- Inherit from Parent -';
				inputFallback.appendChild(option);
			}

			{
				// Blank option, not recommended.
				let option = document.createElement('option');
				option.value = '';
				option.innerHTML = '- No fallback font - (NOT RECOMMENDED)';
				inputFallback.appendChild(option);
			}

			// Set initial values
			inputFont.value = preferredFont;
			inputFallback.value = fallbackFont;

			// Assign the change event to keep track of the values
			inputFont.addEventListener('change', evt => {
				let preferredValue = evt.target.value,
					fallbackValue = inputFallback.value,
					newValue = '';

				if(preferredValue !== '' && fallbackValue === 'inherit') {
					// Fallback of "inherit" doesn't work with a preferred font.
					inputFallback.value = fallbackValue = '';
				}

				if(preferredValue !== '' && fallbackValue !== '') {
					newValue = preferredValue + ', ' + fallbackValue;
				}
				else if(preferredValue !== '') {
					newValue = preferredValue;
				}
				else if(fallbackValue !== '') {
					newValue = fallbackValue;
				}

				document.documentElement.style.setProperty('--' + this.variable, newValue);
			});
			inputFallback.addEventListener('change', evt => {
				let preferredValue = inputFont.value,
					fallbackValue = evt.target.value,
					newValue = '';

				if(fallbackValue === 'inherit') {
					// Fallback of "inherit" doesn't work with a preferred font.
					inputFont.value = preferredValue = '';
				}

				if(preferredValue !== '' && fallbackValue !== '') {
					newValue = preferredValue + ', ' + fallbackValue;
				}
				else if(preferredValue !== '') {
					newValue = preferredValue;
				}
				else if(fallbackValue !== '') {
					newValue = fallbackValue;
				}

				document.documentElement.style.setProperty('--' + this.variable, newValue);
			});

			container.appendChild(inputFont);
			container.appendChild(inputFallback);

			return container;
		}

		/**
		 * Get a list of all web-safe default fonts
		 *
		 * @returns {string[]}
		 */
		static GetFallbackFonts() {
			return [
				'sans-serif',
				'Arial, sans-serif',
				'Arial Black, sans-serif',
				'Verdana, sans-serif',
				'Tahoma, sans-serif',
				'Trebuchet MS, sans-serif',
				'Impact, sans-serif',
				'serif',
				'Times New Roman, serif',
				'Didot, serif',
				'Georgia, serif',
				'American Typewriter, serif',
				'monospace',
				'Lucida Console, monospace',
				'Monaco, monospace',
				'cursive',
				'Bradley Hand, cursive',
				'Brush Script MT, cursive',
				'fantasy',
				'Luminari, fantasy'
			];
		}

		/**
		 * Get a list of all fonts installed on the system
		 *
		 * @returns {string[]}
		 */
		static GetInstalledFonts() {
			let fonts = [],
				documentFonts = document.fonts.entries(),
				font = documentFonts.next(),
				skipFonts = ['Font Awesome 5 Free', 'Font Awesome 5 Brands'];

			while(!font.done) {
				// A single font may contain multiple declarations
				for(let i = 0; i < font.value.length; i++) {
					let fontName = font.value[i].family;

					// Remove quotes around the font names.
					fontName = fontName.replaceAll('"', '');

					if(
						skipFonts.indexOf(fontName) === -1 &&
						fonts.indexOf(fontName) === -1
					) {
						fonts.push(fontName);
					}
				}

				font = documentFonts.next();
			}

			return fonts;
		}
	}

	// @todo font-weight
	// @todo text-align

	/**
	 * @param {VariableStyle[]} options
	 */
	function createStyleEditorUI(options) {
		styleEditorUI = document.createElement('div');
		let editorSection,
			lineSection,
			lineLabel,
			renderedKeys = [];

			/*fontWeights = {
				'inherit': 'Inherit From Parent',
				'100': 'Extra Light (100)',
				'200': 'Light (200)',
				'300': 'Book (300)',
				'400': 'Normal (400)',
				'500': 'Medium (500)',
				'600': 'Semibold (600)',
				'700': 'Bold (700)',
				'800': 'Extra Bold (800)',
				'900': 'Fat (900)'
			},
			fontWeightAliases = {
				'normal': '400',
				'bold': '700',
				'': 'inherit'
			},
			textAligns = {
				'initial': 'Default',
				'left': 'Left Align',
				'right': 'Right Align',
				'center': 'Center Align',
				'justify': 'Justify Text',
				'inherit': 'Inherit From Parent'
			}*/

		styleEditorUI.classList.add('mediamanager-styleeditor-ui');
		styleEditorUI.innerHTML = '<h2>Configurable Styles</h2>' +
			'<div class="editor-section"></div>' +
			'<label><a href="#" class="save button">Save Settings</a></label>';
		document.body.appendChild(styleEditorUI);

		// This is required because the variable styles provide native Elements
		editorSection = styleEditorUI.querySelector('.editor-section');

		// lineSection

		// Parse foreground/background color selections
		for(let i = 0; i < options.length; i++) {
			// Not strictly needed, but makes the logic a little cleaner.
			let optionColor, optionBackground;

			if(options[i].style === 'color' && options[i] instanceof VariableStyleColor) {
				// Search for a complementary background-color
				let complementary = null;
				for(let x = 0; x < options.length; x++) {
					if(
						options[x].style === 'background-color' &&
						options[x].getStateModifier() === options[i].getStateModifier() &&
						options[x] instanceof VariableStyleColor
					) {
						complementary = x;
						/** @var {VariableStyleColor} optionColor */
						optionColor = options[i];
						/** @var {VariableStyleColor} optionBackground */
						optionBackground = options[x];
					}
				}

				if(complementary !== null) {
					let compliancyText = document.createElement('span'),
						compliancyContrast = VariableStyleColor.Contrast(optionColor.getProcessedValue(), optionBackground.getProcessedValue()),
						inputColor = optionColor.getInput(),
						inputBackground = optionBackground.getInput();

					compliancyText.classList.add('compliancy-text');
					compliancyText.innerHTML = '<span class="check-aaa" title="Colors are AAA compliant, nice!"><i class="icon icon-check-double"></i> AAA</span>' +
						'<span class="check-aa" title="Colors are AA compliant, ok, but a 7.0 rating could get you AAA."><i class="icon icon-check"></i> AA</span>' +
						'<span class="check-fail" title="Colors are NOT WCAG compliant for visually impaired users!"><i class="icon icon-times"></i> FAIL</span>';
					
					lineSection = document.createElement('label');
					lineSection.classList.add('editor-group-colors');
					lineSection.title = options[i].variable + ' / ' + options[complementary].variable;

					lineLabel = document.createElement('span');
					lineLabel.classList.add('line-label');
					lineLabel.innerHTML = options[i].getTitle();

					lineSection.appendChild(lineLabel);
					lineSection.appendChild(inputColor);
					lineSection.appendChild(inputBackground);
					lineSection.appendChild(compliancyText);

					// Initial setting
					if(compliancyContrast >= 7.0) {
						compliancyText.classList.add('pass-aaa');
					}
					else if(compliancyContrast >= 4.5) {
						compliancyText.classList.add('pass-aa');
					}
					else {
						compliancyText.classList.add('fail');
					}

					// Change calls, keep the user notified of WCAG rating
					[inputColor, inputBackground].forEach(el => {
						el.addEventListener('change', evt => {
							let compliancyContrast = VariableStyleColor.Contrast(inputColor.value, inputBackground.value);
							if(compliancyContrast >= 7.0) {
								compliancyText.classList.remove('pass-aa', 'fail');
								compliancyText.classList.add('pass-aaa');
							}
							else if(compliancyContrast >= 4.5) {
								compliancyText.classList.remove('pass-aaa', 'fail');
								compliancyText.classList.add('pass-aa');

								$.notify('Selected colors only have a contrast ratio of ' + (Math.round(compliancyContrast * 10) / 10) + ':1, if you can get 7.0:1 you can get AAA compliancy.', 'info');
							}
							else {
								compliancyText.classList.remove('pass-aaa', 'pass-aa');
								compliancyText.classList.add('fail');

								$.notify('Selected colors only have a contrast ratio of ' + (Math.round(compliancyContrast * 10) / 10) + ':1, try to aim for at least 4.5:1');
							}
						});
					});




					editorSection.appendChild(lineSection);
					renderedKeys.push(i);
					renderedKeys.push(complementary);
				}
			}
		}

		// Parse grouped border selections
		for(let i = 0; i < options.length; i++) {
			if(renderedKeys.indexOf(i) !== -1) {
				// Option already rendered, skip
				continue;
			}

			// Not strictly needed, but makes the logic a little cleaner.
			let
				/** @var {VariableStyleColor} optionColor */
				optionColor = null,
				/** @var {VariableStyleSize} optionWidth */
				optionWidth = null,
				/** @var {VariableStyleBorderStyle} optionStyle */
				optionStyle = null,
				optionsFound = [];

			if(
				options[i].style === 'border-color' ||
				options[i].style === 'border-style'
			) {
				// One of the border options was detected, look for the other 2 elements
				for(let x = 0; x < options.length; x++) {
					if(options[x].getStateModifier() === options[i].getStateModifier()) {
						if(options[x].style === 'border-color') {
							optionColor = options[x];
							optionsFound.push(x);
						}
						/*if(options[x].style === 'border-width') {
							optionWidth = options[x];
							optionsFound.push(x);
						}*/
						if(options[x].style === 'border-style') {
							optionStyle = options[x];
							optionsFound.push(x);
						}
					}
				}

				if(optionsFound.length > 1) {
					// More than one border option found, group them!
					// This loop won't run if there's only 1 border option;
					// it would be pointless to group a single option.
					let titleValues = [];

					lineSection = document.createElement('label');
					lineSection.classList.add('editor-group-border');
					lineLabel = document.createElement('span');
					lineLabel.classList.add('line-label');
					lineSection.appendChild(lineLabel);

					if(optionWidth !== null) {
						titleValues.push(optionWidth.variable);
						lineSection.appendChild(optionWidth.getInput());
						if(optionWidth.getStateModifier() !== 'default') {
							lineLabel.innerHTML = 'Border (' + 	optionWidth.getStateModifier() + ')';
						}
						else {
							lineLabel.innerHTML = 'Border';
						}
					}

					if(optionStyle !== null) {
						titleValues.push(optionStyle.variable);
						lineSection.appendChild(optionStyle.getInput());
						if(optionStyle.getStateModifier() !== 'default') {
							lineLabel.innerHTML = 'Border (' + 	optionStyle.getStateModifier() + ')';
						}
						else {
							lineLabel.innerHTML = 'Border';
						}
					}

					if(optionColor !== null) {
						titleValues.push(optionColor.variable);
						lineSection.appendChild(optionColor.getInput());
						if(optionColor.getStateModifier() !== 'default') {
							lineLabel.innerHTML = 'Border (' + 	optionColor.getStateModifier() + ')';
						}
						else {
							lineLabel.innerHTML = 'Border';
						}
					}

					lineSection.title = titleValues.join(' / ');

					editorSection.appendChild(lineSection);
					// Set all detected keys to be skipped, as they've been rendered.
					renderedKeys = renderedKeys.concat(optionsFound);
				}
			}
		}

		// Parse default selections
		for(let i = 0; i < options.length; i++) {
			if(renderedKeys.indexOf(i) !== -1) {
				// Option already rendered, skip
				continue;
			}

			lineSection = document.createElement('label');
			lineSection.classList.add('editor-group-default');
			lineSection.title = options[i].variable;

			lineLabel = document.createElement('span');
			lineLabel.classList.add('line-label');
			lineLabel.innerHTML = options[i].getTitle();

			lineSection.appendChild(lineLabel);
			lineSection.appendChild(options[i].getInput());

			editorSection.appendChild(lineSection);
			renderedKeys.push(i);
		}

		// Update the position.
		//styleEditorUI.style.marginTop = window.pageYOffset + 'px';

		styleEditorUI.querySelector('.save').addEventListener('click', evt => {
			evt.preventDefault();

			let values = {};
			for(let i = 0; i < options.length; i++) {
				values[ options[i].variable ] = options[i].getValue();
			}

			Core.Ajax(
				Core.ROOT_URL + 'mediamanagerstyles/save',
				{
					method: 'POST',
					data: values,
					dataType: 'json',
					callback: ret => {
						if(ret.status === 200) {
							$.notify('Updated styles', 'success');
						}
						else {
							$.notify(ret.message, 'error');
						}
					}
				}
			);

			document.body.removeChild(styleEditorUI);
			document.body.classList.remove('mediamanager-styleeditor-open');
			styleEditorUI = null;
		});
	}

	/**
	 * Calculate the rule weight for a given style rule
	 *
	 * The weight calculation is simplified to this; each...
	 * ID - 10000
	 * class/parameter - 100
	 * element - 1
	 *
	 * While not perfect, (as formal selectors will use a tiered approach), this should be good enough.
	 * This will break down if there are more than 100 selectors on a given rule.
	 *
	 * @param {string} selector
	 * @return {Number}
	 */
	function calculateRuleWeight(selector) {
		let weight = 0,
			weightD = 1,
			weightC = 100,
			weightB = 10000,
			weightA = 1000000,
			parts = selector.split(',');

		// Each rule may contain multiple selectors, separated by a comma, process each individually
		for(let i = 0; i < parts.length; i++) {
			let w = 0,
				rulePieces = parts[i].split(' ');

			for(let x = 0; x < rulePieces.length; x++) {
				let piece = rulePieces[x].trim();

				// The '*' has zero weight and can be omitted
				piece = piece.replace('*', '');

				// '::' selectors can be simplified to ':'
				piece = piece.replace('::', ':');

				if(piece === '') {
					// Skip empty pieces, just in case there is one
					continue;
				}

				if(piece.startsWith('#')) {
					w += weightB
						+ (piece.substr(1).split('#').length - 1) * weightB
						+ (piece.split('.').length - 1) * weightC
						+ (piece.split('[').length - 1) * weightC
						+ (piece.split(':').length - 1) * weightD;
				}
				else if(piece.startsWith('.')) {
					w += weightC
						+ (piece.substr(1).split('.').length - 1) * weightC
						+ (piece.split('#').length - 1) * weightB
						+ (piece.split('[').length - 1) * weightC
						+ (piece.split(':').length - 1) * weightD;
				}
				else if(piece.startsWith('[')) {
					w += weightC
						+ (piece.substr(1).split('[').length - 1) * weightC
						+ (piece.split('#').length - 1) * weightB
						+ (piece.split('.').length - 1) * weightC
						+ (piece.split(':').length - 1) * weightD;
				}
				else if(piece.startsWith(':')) {
					w += weightD
						+ (piece.substr(1).split(':').length - 1) * weightD
						+ (piece.split('#').length - 1) * weightB
						+ (piece.split('[').length - 1) * weightC
						+ (piece.split('.').length - 1) * weightC;
				}
				else {
					w += weightD
						+ (piece.split('#').length - 1) * weightB
						+ (piece.split('[').length - 1) * weightC
						+ (piece.split('.').length - 1) * weightC
						+ (piece.split(':').length - 1) * weightD;
				}
			}

			weight = Math.max(weight, w);
		}

		return weight;
	}

	/**
	 * Select an element to inspect and open the dialog
	 *
	 * @param {Element} inspectElement
	 */
	function selectElement(inspectElement) {
		let targets = [],
			variables = [],
			variableWeights = { },
			// All customizable styles along with the element, [(target, [...]), ...]
			customizableStyles = [],
			manualStyles = [
				'border-color',
				'border-width',
				'border-style',
				'padding',
				'margin'
			];

		// Build a list of parent nodes; we'll do this so we can reverse through them to compile an accurate cascade.
		while(inspectElement !== document.body.parentElement) {
			targets.push(inspectElement);
			inspectElement = inspectElement.parentElement;
		}

		for(let t = 0; t < targets.length; t++) {
			//console.log('Processing CSS rules for', targets[t]);

			let rules = getMatchedCSSRules(targets[t]);

			for(let r = rules.length - 1; r >= 0; r--) {
				let styleValue,
					styleVariable,
					styleKey,
					vs,
					ruleWeight = calculateRuleWeight(rules[r].selectorText),
					foundIndex;

				if(rules[r].parentRule) {
					// This is generally the case for @media queries.
					// here, we want those rules to be slighly better than the default ones, so increase by 1
					ruleWeight += 1;
				}

				// Calculate the rule CSS weight to know if something should override any previously-defined styles
				//console.log('Processing rule ', rules[r]);
				//console.log('Rule weight ', ruleWeight);

				// Some styles are difficult and refuse to expose themselves.
				for(let s = 0; s < manualStyles.length; s++) {
					styleValue = rules[r].style[ manualStyles[s] ]
					//console.log(s, styleValue);

					if(styleValue.indexOf('var(--') !== -1) {
						//console.log(targets[t], manualStyles[s]);

						styleVariable = styleValue.replace(/var\(--([a-zA-Z\-0-9]*).*\)/, '$1').trim();
						vs = VariableStyle.Factory(targets[t], manualStyles[s], styleVariable);
						styleKey = vs.style + '-' + vs.getStateModifier();

						// Only include this variable if not found on a more specific target.
						foundIndex = variables.indexOf(styleKey);
						if(foundIndex === -1) {
							// Simple push; does not exist yet
							customizableStyles.push(vs);
							variables.push(styleKey);
							variableWeights[ styleKey ] = ruleWeight;
						}
						else if(ruleWeight > variableWeights[ styleKey ]) {
							// Already exists, but is a more precise match
							//console.log('Found, but ' + ruleWeight + ' better match than ' + variableWeights[ styleKey ]);
							customizableStyles[ foundIndex ] = vs;
							variableWeights[ styleKey ] = ruleWeight;
						}
						// No else required
					}
				}


				Object.values(rules[r].style).forEach(s => {
					styleValue = rules[r].style[s]
					//console.log(s, styleValue);

					if(typeof(styleValue) !== 'undefined' && styleValue.indexOf('var(--') !== -1) {
						styleVariable = styleValue.replace(/var\(--([a-zA-Z\-0-9]*).*\)/, '$1').trim();
						vs = VariableStyle.Factory(targets[t], s, styleVariable);
						styleKey = vs.style + '-' + vs.getStateModifier();

						// Only include this variable if not found on a more specific target.
						foundIndex = variables.indexOf(styleKey);
						if(foundIndex === -1) {
							// Simple push; does not exist yet
							customizableStyles.push(vs);
							variables.push(styleKey);
							variableWeights[ styleKey ] = ruleWeight;
						}
						else if(ruleWeight > variableWeights[ styleKey ]) {
							// Already exists, but is a more precise match
							customizableStyles[ foundIndex ] = vs;
							variableWeights[ styleKey ] = ruleWeight;
						}
						// No else required
					}
				});
			}

			// Lock the variables in now.
			// This is needed because the processor has detected that the given styles match this attribute already,
			// so even if a parent that may have a more precise match for it has a style, we do not want it to conflict.
			//
			// example:
			// header.page-header { padding: 2em }
			// a { padding: 0 }
			Object.keys(variableWeights).forEach(w => {
				variableWeights[w] = 99999999;
			});
		}

		if(!variables.length) {
			$.notify('Selected element contains no customizable styles', 'error');
			return;
		}

		createStyleEditorUI(customizableStyles);
	}

	document.addEventListener('click', evt => {
		if(evt.target.closest('.mediamanager-styleeditor-ui')) {
			// Click target was on the editor UI itself
			return;
		}

		if(evt.target.closest('.mediamanager-styleeditor-toggle')) {
			// Click target was the fixed controls
			return;
		}

		if(!document.body.classList.contains('mediamanager-styleeditor-enabled')) {
			// Body needs the class to show it's enabled.

			if(styleEditorUI) {
				// Close the currently open window.
				document.body.removeChild(styleEditorUI);
				document.body.classList.remove('mediamanager-styleeditor-open');
				styleEditorUI = null;
			}

			return;
		}

		document.body.classList.remove('mediamanager-styleeditor-enabled');
		document.body.classList.add('mediamanager-styleeditor-open');
		evt.preventDefault();
		selectElement(evt.target);
	});

	document.addEventListener('keyup', evt => {
		// Watch for the ESC key to close the interface without saving
		if(evt.key === 'Escape') {
			document.body.classList.remove('mediamanager-styleeditor-enabled');
			if(styleEditorUI) {
				// Close the currently open window.
				document.body.removeChild(styleEditorUI);
				document.body.classList.remove('mediamanager-styleeditor-open');
				styleEditorUI = null;
			}
		}
	});


	document.addEventListener('contextmenu', evt => {

		if(!document.body.classList.contains('mediamanager-styleeditor-enabled')) {
			// Body needs the class to show it's enabled.
			return;
		}

		document.body.classList.remove('mediamanager-styleeditor-enabled');
		document.body.classList.add('mediamanager-styleeditor-open');
		evt.preventDefault();
		selectElement(evt.target);
	});



	let closeUI = document.createElement('div');
	closeUI.classList.add('mediamanager-styleeditor-toggle');
	closeUI.innerHTML = '<a href="#" class="button toggle"><i class="icon icon-eye-dropper"></i> <span>Inspect Element</span></a>' +
		'<a href="' + Core.ROOT_URL + 'mediamanagerstyles/toggle' + '" class="button"><i class="icon icon-check"></i> <span>Finish Editing</span></a>';
	document.body.appendChild(closeUI);

	closeUI.querySelector('.toggle').addEventListener('click', evt => {
		document.body.classList.toggle('mediamanager-styleeditor-enabled');
		evt.preventDefault();
	});

	document.body.classList.add('mediamanager-styleeditor-active');

	//alert('Inline style editing is enabled!' + "\n\n" + 'To change how something looks, simply click on it and you will be presented with options.');
})();