<?php
/**
 * File for class MediaManagerNavigatorController definition in the coreplus project
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20130603.1641
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Controller;
use Core\Filestore\Directory;
use Core\Filestore\Factory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\FileMetaHelper;
use Core\Filestore\File;
use Core\Forms\Form;
use Core\View;
use MediaManager\Navigator;
use MediaManager\Navigator2;
use Core\AccessDeniedException;
use Core\BadRequestException;
use Core\NotFoundException;
use Core\TemporaryFollowRedirectException;


/**
 * Provides a nautilus/explorer UI of the public files currently on the server.
 *
 * These are files that can be managed by super admins anyway.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for MediaManagerNavigatorController
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class MediaManagerNavigatorController extends  Controller {

	/**
	 * Main view for the navigator.
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function index() {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$navigators = Navigator2::GetAccessableNavigators();

		if($request->hasParameter('navigators')) {
			// Allow the calling script to limit what navigators are provided
			$navigatorsAllowed = explode(',', $request->getParameter('navigators'));
			foreach($navigators as $key => $nav) {
				if(!in_array($nav->getType(), $navigatorsAllowed)) {
					unset($navigators[ $key ]);
				}
			}
		}
		
		if(!count($navigators)) {
			throw new AccessDeniedException();
		}
		
		if($request->getParameter('ajax')) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
			$view->mastertemplate = false;
		}
		elseif($request->getParameter('frame')) {
			$view->mastertemplate = 'blank';
		}
		else{
			$view->mastertemplate = 'admin';
		}

		// @todo Remove this and move it to TinyMCE instead.
		switch($request->getParameter('selectcallback')) {
			case 'tinymce':
				$selectCallback = 'tinymce';
				break;
			default:
				$selectCallback = null;
		}

		if($request->hasParameter('mime')) {
			$mime = $request->getParameter('mime');
			$mime = preg_replace('/[^a-z,\/\*]*/', '', $mime);
		}
		else {
			$mime = '*/*';
		}


		$view->title = 't:STRING_MEDIA_FILE_BROWSER';
		$view->assign('navigators', $navigators);
		$view->assign('select_callback', $selectCallback);
		$view->assign('mime', $mime);
		$view->assign('max_upload', $this->_getMaxUploadSize());
	}

	/**
	 * Loads only the page templates navigator
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function templates() {
		$view = $this->getView();
		$request = $this->getRequest();

		$navigators = [];
		foreach(Navigator2::GetAccessableNavigators() as $nav) {
			// Only load templates.
			if($nav->getType() === 'page') {
				$navigators[] = $nav;
			}
		}

		if(!count($navigators)) {
			throw new AccessDeniedException();
		}

		if($request->getParameter('ajax')) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
			$view->mastertemplate = false;
		}
		elseif($request->getParameter('frame')) {
			$view->mastertemplate = 'blank';
		}
		else{
			$view->mastertemplate = 'admin';
		}

		$view->templatename = 'pages/mediamanagernavigator/index';
		$view->title = 't:STRING_MEDIAMANAGER_TEMPLATES';
		$view->assign('navigators', $navigators);
		$view->assign('select_callback', null);
		$view->assign('mime', '*/*');
		$view->assign('max_upload', $this->_getMaxUploadSize());
	}

	/**
	 * Loads only the site assets navigator
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function assets() {
		$view = $this->getView();
		$request = $this->getRequest();

		$navigators = [];
		foreach(Navigator2::GetAccessableNavigators() as $nav) {
			// Only load templates.
			if($nav->getType() === 'assets') {
				$navigators[] = $nav;
			}
		}

		if(!count($navigators)) {
			throw new AccessDeniedException();
		}

		if($request->getParameter('ajax')) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
			$view->mastertemplate = false;
		}
		elseif($request->getParameter('frame')) {
			$view->mastertemplate = 'blank';
		}
		else{
			$view->mastertemplate = 'admin';
		}

		$view->templatename = 'pages/mediamanagernavigator/index';
		$view->title = 't:STRING_MEDIAMANAGER_ASSETS';
		$view->assign('navigators', $navigators);
		$view->assign('select_callback', null);
	}
	
	/**
	 * File editor for a given path
	 * 
	 * @param string $type Type of browser
	 * @param string $path Path of file or directory to query
	 * 
	 * @throws AccessDeniedException
	 */
	public function file_editor(string $type, string $path) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		// Load the requested navigator; this will handle the bulk of the heavy lifting.
		$nav = Navigator2::Factory($type);
		
		// Ensure user permissions
		if(!$nav->checkCanUpdate()) {
			throw new AccessDeniedException();
		}

		$dir = dirname($path);
		if($dir === '.') {
			$nav->setCWD('');
		}
		else {
			$nav->setCWD($dir . '/');
		}
		
		// Retrieve the File object for the requested path.
		$entry = $nav->getEntry($path);
		
		if($entry === null) {
			if($nav->checkCanUpload()) {
				// If this navigator can upload/create new files, then allow the user to autocreate a new file.
				// This requires the cache to reload for the entries, so we might as well be lazy and perform a full reload.
				$view->mastertemplate = 'admin';
				$view->templatename = 'pages/mediamanagernavigator/file_editor/file_nonexistent';
				$view->assign('type', $type);
				$view->assign('cwd', $nav->getCWD());
				$view->assign('path', basename($path));
				return;
			}

			throw new NotFoundException();
		}
		
		if(!($entry instanceof File)) {
			throw new BadRequestException();
		}

		//$info = $nav->getInfo([$path]);
		//var_dump($info, $entry, $entry->getMimetype()); die();

		$nav->loadEditor($view, $path);
		

		$view->mastertemplate = 'admin';
		$view->addBreadcrumb($nav->getTitle(), $nav->getURL());
		$view->title = $entry->getFilename('');
	}

	/**
	 * Shortcut for file/editor with assets and css/custom.css as the paths.
	 *
	 * @throws AccessDeniedException
	 */
	public function edit_customcss() {
		// Run the parent method first; this will do everything necessary for the actual work
		$this->file_editor('assets', 'css/custom.css');

		// Modify the view slightly to make this its own page
		$view = $this->getView();
		$view->resetBreadcrumbs();
		$view->autoBreadcrumbs('/mediamanagernavigator/assets');
		$view->title = 'Site Custom CSS Editor';
	}

	/**
	 * Shortcut for file/editor with assets and css/custom_print.css as the paths.
	 *
	 * @throws AccessDeniedException
	 */
	public function edit_customprintcss() {
		// Run the parent method first; this will do everything necessary for the actual work
		$this->file_editor('assets', 'css/custom_print.css');

		// Modify the view slightly to make this its own page
		$view = $this->getView();
		$view->resetBreadcrumbs();
		$view->autoBreadcrumbs('/mediamanagernavigator/assets');
		$view->title = 'Site Custom Print CSS Editor';
	}
	
	
	/**
	 * Get a listing of all files and directories contained inside a given media type
	 *
	 * Also return the base permissions for this navigator
	 * 
	 * @param string $type Type of browser
	 * @throws AccessDeniedException
	 */
	public function files(string $type, string $cwd) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		
		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}

		$nav->setCWD($cwd);
		
		$view->jsondata = [
			'results' => $nav->getListing(),
			'permissions' => $nav->getPermissions(),
		];
	}

	/**
	 * Get a listing of all files and directories contained inside a given media type
	 *
	 * Results are returned as JSON
	 *
	 * @param string $type Type of browser
	 * @throws AccessDeniedException
	 */
	public function permissions(string $type, string $cwd) {
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$nav = Navigator2::Factory($type);

		$nav->setCWD($cwd);

		$view->jsondata = [
			'permissions' => $nav->getPermissions(),
		];
	}
	
	/**
	 * Get additional information on a subset of files/paths, usually in a specific directory.
	 * 
	 * The paths must be a JSON array to prevent Firefox's No Script plugin from freaking out.
	 * 
	 * Results are returned as JSON
	 * 
	 * @param string $type  Type of browser
	 * @param string $paths JSON and HTML encoded string of paths to return
	 * @throws AccessDeniedException
	 */
	public function file_info(string $type, string $paths) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		
		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}
		
		$view->jsondata = [
			'results' => $nav->getInfo(json_decode(html_entity_decode($paths), true))
		];
	}
	
	/**
	 * Get full details for a given path.
	 * 
	 * Results are returned as an HTML string
	 * 
	 * @param string $type Type of browser
	 * @param string $cwd
	 * @param string $path Path of file or directory to query
	 * 
	 * @throws AccessDeniedException
	 */
	public function file_details(string $type, string $cwd, string $path) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->mastertemplate = false;
		
		$nav = Navigator2::Factory($type);
		
		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}

		$nav->setCWD($cwd);
		
		$entry = $nav->getEntry($path);
		if(!$entry) {
			throw new NotFoundException();
		}

		$details = $nav->getEntryDetails($path);

		$template = \Core\Templates\Template::Factory('includes/mediamanagernavigator/file_details');
		$template->assign('entry', $entry);
		$template->assign('details', $details);

		$view->jsondata = [
			'body' => $template->fetch(),
			'data' => $details['data'],
			'permissions' => $nav->getPermissions($path),
		];
	}
	
	/**
	 * Upload a new file to the given directory
	 * 
	 * @param string $type Type of navigator for the destination
	 * @param string $path Directory to upload file to
	 * 
	 * @throws AccessDeniedException
	 * @throws BadRequestException
	 */
	public function file_upload(string $type, string $path) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		$nav->setCWD($path);
		
		if(!$nav->checkCanUpload()) {
			throw new AccessDeniedException();
		}
		
		foreach($_FILES as $file) {
			switch($file['error']) {
				case UPLOAD_ERR_INI_SIZE:
					throw new BadRequestException('Uploaded file too large (ini)');
				case UPLOAD_ERR_FORM_SIZE:
					throw new BadRequestException('Uploaded file too large (POST)');
				case UPLOAD_ERR_PARTIAL:
					throw new BadRequestException('Upload did not complete fully, please try again');
				case UPLOAD_ERR_NO_FILE:
					throw new BadRequestException('No file uploaded');
				case UPLOAD_ERR_NO_TMP_DIR:
					throw new BadRequestException('No temp directory suitable for uplaod');
				case UPLOAD_ERR_CANT_WRITE:
					throw new BadRequestException('Unable to write temporary file');
				case UPLOAD_ERR_EXTENSION:
					throw new BadRequestException('Bad extension');
				default:
					$nav->setUploadedFile($file['name'], $file['tmp_name']);
			}
		}
		
		$view->jsondata = ['status' => 200, 'message' => 'Uploaded successfully'];
	}

	/**
	 * Upload a new file to the given directory
	 *
	 * @param string $type Type of navigator for the destination
	 * @param string $path Directory to upload file to
	 *
	 * @throws AccessDeniedException
	 * @throws BadRequestException
	 */
	public function file_upload_chunk(string $type, string $path, string $file, int $size, string $date) {
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$nav = Navigator2::Factory($type);
		$nav->setCWD($path);

		if(!$nav->checkCanUpload()) {
			throw new AccessDeniedException();
		}

		// Generate a temporary filename for this upload
		$filename = 'upload-' . md5($type . $path . $file . $date) . '.tmp';
		$tempFile = FileFactory::File('/tmp/' . $filename);
		if(!$tempFile->exists()) {
			try {
				$tempFile->putContents('');
			}
			catch(Exception $e) {
				$view->setJSONErrorMessage('Unable to save temporary file', 500);
				return;
			}
		}

		// Append this chunk to the temporary file
		file_put_contents($tempFile->getFilename(), $request->getBody(), FILE_APPEND);

		// Check if the file has been completely written,
		// we need to save it to the final destination once that has been completed.
		if($tempFile->getFilesize() >= $size) {
			$navigatorFilePath = $nav->touch($file);
			$destinationFile = $nav->getEntry($navigatorFilePath);

			// Destination will exist because touch will auto-create the file.
			// This is necessary to ensure the specific Navigator can load the entry.
			// However, copyTo will skip overwrite by default.  If the file is empty,
			// it's safe to assume we can overwrite without issue.
			if($destinationFile->getFilesize() === 0) {
				$overwrite = true;
			}
			else {
				$overwrite = false;
			}

			$tempFile->copyTo($destinationFile, $overwrite);

			$view->jsondata = ['status' => 200, 'message' => 'Uploaded successfully'];
		}
		else {
			$view->jsondata = ['status' => 200, 'message' => 'Chunk processed'];
		}
	}

	/**
	 * Main view for the navigator.
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function image(){
		$view = $this->getView();
		$request = $this->getRequest();

		$navigator = new Navigator();
		$navigator->setView($view);
		$navigator->setMode($request->getParameter('mode'));
		$navigator->setAccept('image');
		if($request->getParameter('ajax')){
			$navigator->setBaseURL('/mediamanagernavigator/image?ajax=1');
			$view->mastertemplate = false;
			$view->record = false;
			$view->mode = View::MODE_AJAX;
		}
		elseif($request->getParameter('iframe')){
			$navigator->setBaseURL('/mediamanagernavigator/image?iframe=1');
			$view->mastertemplate = 'blank.tpl';
			$view->record = false;
		}
		else{
			$navigator->setBaseURL('/mediamanagernavigator/image');
			$view->mastertemplate = 'admin';
			//$view->mode = View::MODE_PAGE;
		}

		try{
			$navigator->cd($request->getParameter('dir'));
			if($request->getParameter('controls') !== null){
				$navigator->usecontrols = ($request->getParameter('controls') == 1);
			}
			if($request->getParameter('uploader') !== null){
				$navigator->useuploader = ($request->getParameter('uploader') == 1);
			}
		}
		catch(Exception $e){
			\Core\set_message($e->getMessage(), 'error');
		}

		$view->title = 'Images';
		$navigator->render();
		//var_dump($navigator, $navigator->render()); die();
	}


	/**
	 * Helper function for the directory mkdir command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 *
	 * @core-request-method POST
	 *
	 * @throws BadRequestException
	 */
	public function directory_mkdir(string $type, string $cwd, string $path) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		
		if(!($nav->checkHasAccess() && $nav->checkCanUpload())){
			throw new AccessDeniedException();
		}

		try{
			$nav->setCWD($cwd);
			$newPath = $nav->mkdir($path);
			$view->jsondata = array('status' => 1, 'message' => 'Created directory successfully', 'path' => $newPath);
		}
		catch(Exception $e){
			$view->jsondata = ['status' => 0, 'message' => $e->getMessage()];
		}
	}

	/**
	 * Helper function for the directory rename command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 */
	public function directory_rename(){
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$navigator = new Navigator();
		$navigator->setView($view);

		if(!$request->isPost()){
			$view->jsondata = array('status' => 0, 'message' => 'Invalid request, please ensure it is a POST.');
			return;
		}

		if(!isset($_POST['newdir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Please enter a directory name');
			return;
		}

		if(!isset($_POST['olddir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Previous directory is required!');
			return;
		}

		if(!trim($_POST['newdir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Please enter a directory name');
			return;
		}

		try{
			$navigator->cd($request->getPost('dir'));
			$navigator->rename($_POST['olddir'], $_POST['newdir']);
		}
		catch(Exception $e){
			$view->jsondata = ['status' => 0, 'message' => $e->getMessage()];
			return;
		}

		$view->jsondata = array('status' => 1, 'message' => 'Renamed directory successfully');
	}

	/**
	 * Helper function for the file rename command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 *
	 * @core-request-method POST
	 */
	public function file_rename(string $type, string $cwd, string $path, string $name) {
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->mastertemplate = false;

		$nav = Navigator2::Factory($type);

		if(!$nav->checkHasAccess()) {
			throw new AccessDeniedException();
		}

		$nav->setCWD($cwd);

		$entry = $nav->getEntry($path);
		if(!$entry) {
			throw new NotFoundException();
		}

		$entry->rename(\Core\str_to_url($name, true));

		$view->jsondata = array('status' => 1, 'message' => 'Renamed file successfully');
	}

	/**
	 * Helper function for the file move command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 *
	 * @core-request-method POST
	 */
	public function file_move(string $type, string $path, string $dest) {
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->mastertemplate = false;

		$nav = Navigator2::Factory($type);

		if(!$nav->checkHasAccess()) {
			throw new AccessDeniedException();
		}

		if(!$nav->checkCanUpload()) {
			throw new AccessDeniedException();
		}

		$source = $nav->getEntry($path);
		$destPath = $nav->getEntry($dest);

		if(!$source) {
			throw new NotFoundException();
		}
		if(!$destPath) {
			throw new NotFoundException();
		}

		$source->rename($destPath->getPath() . '/' . $source->getBaseFilename());

		$view->jsondata = array('status' => 1, 'message' => 'Moved file successfully');
	}

	/**
	 * Helper function for the file touch command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 *
	 * @core-request-method POST
	 *
	 * @throws BadRequestException
	 */
	public function file_touch(string $type, string $cwd, string $name) {
		$view = $this->getView();
		$request = $this->getRequest();

		if($request->isAjax()) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
		}

		$nav = Navigator2::Factory($type);

		if(!($nav->checkHasAccess() && $nav->checkCanUpload())){
			throw new AccessDeniedException();
		}

		try {
			$nav->setCWD($cwd);
			$newPath = $nav->touch($name);

			if($request->isAjax()) {
				$view->jsondata = array('status' => 1, 'message' => 'Created file successfully', 'path' => $newPath);
			}
			else {
				throw new \Core\RedirectException('/mediamanagernavigator/file/editor?type=' . $type . '&path=' . $newPath);
			}
		}
		catch(Exception $e) {
			if($request->isAjax()) {
				$view->jsondata = ['status' => 0, 'message' => $e->getMessage()];
			}
			else {
				throw $e;
			}
		}
	}

	/**
	 * Helper function for the file delete command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 */
	public function file_delete(string $type, string $path) {

		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->mastertemplate = false;

		$nav = Navigator2::Factory($type);

		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}

		if(!$nav->checkCanUpload()) {
			throw new AccessDeniedException();
		}

		if(!$request->isPost()){
			throw new BadRequestException();
		}

		$entry = $nav->getEntry($path);

		if($entry === null) {
			throw new NotFoundException();
		}

		$entry->delete();

		$view->jsondata = array('status' => 1, 'message' => 'Deleted entry successfully');
	}

	private function _getMaxUploadSize() : int {
		$max = 10 * (1024 * 1024);

		foreach(['upload_max_filesize', 'post_max_size'] as $key) {
			$value = ini_get($key);

			// Convert this to a number
			if(preg_match('/[0-9\.]*T/', $value)) {
				$value = (int)(substr($value, 0, -1)) * (1024 * 1024 * 1024 * 1024);
			}
			elseif(preg_match('/[0-9\.]*G/', $value)) {
				$value = (int)(substr($value, 0, -1)) * (1024 * 1024 * 1024);
			}
			elseif(preg_match('/[0-9\.]*M/', $value)) {
				$value = (int)(substr($value, 0, -1)) * (1024 * 1024);
			}
			elseif(preg_match('/[0-9\.]*[kK]/', $value)) {
				$value = (int)(substr($value, 0, -1)) * (1024);
			}
			else {
				$value = (int)$value;
			}

			$max = min($max, $value);
		}

		// Ensure a bit of space for headers and additional information.
		$max -= 10240;

		return $max;
	}
}
