<?php
/**
 * Navigation controller, handles the administrative functions for this component.
 *
 * @package Navigation
 * @since 0.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Controller;
use Core\Forms\Form;
use Core\NotFoundException;
use Core\View;


class NavigationController extends Controller {

	public function __construct() {
		$this->accessstring = 'g:admin';
	}

	public function index() {
		$view = $this->getView();

		$models = NavigationModel::Find();
		$pages = PageModel::GetPagesAsOptions();
		$areas = array_merge(\Core\Widget::GetSkinInstallableAreas(), \Core\Widget::GetURLInstallableAreas());

		$view->mastertemplate = 'admin';
		$view->title = 'Navigation Menus';
		$view->assign('models', $models);
		$view->assign('pages', $pages);
		$view->assign('areas', $areas);
	}


	public function edit() {
		$m = new NavigationModel($this->getRequest()->getParameter(0));

		if (!$m->checkExists()) {
			throw new NotFoundException();
		}

		// This has been moved.
		throw new \Core\RedirectException('/navigation#menu:' . $m->id);
	}

	public function create() {
		// This has been moved.
		throw new \Core\RedirectException('/navigation#menu:new');
	}

	public function delete() {
		$view = $this->getView();
		$req  = $this->getRequest();
		$mid  = $req->getParameter(0);
		$m    = new NavigationModel($mid);

		if (!$req->isPost()) {
			throw new \Core\BadRequestException();
		}

		if (!$m->checkExists()) {
			throw new NotFoundException();
		}

		$m->delete();

		throw new \Core\RedirectException('/navigation');
	}

	/**
	 * AJAX call destination to handle saving a menu
	 *
	 * @core-request-method POST
	 * @return string
	 * @throws \Core\ModelValidationException
	 */
	public function savemenu() {
		$request = $this->getRequest();
		$view = $this->getView();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$data = $request->getBody();

		if(!isset($data['id'])) {
			throw new \Core\ModelValidationException('Invalid data provided');
		}

		if(!isset($data['name'])) {
			$data['name'] = 'New Navigation Menu';
		}

		if(!isset($data['entries'])) {
			$data['entries'] = [];
		}

		// Handle the model
		if($data['id'] === 'NEW') {
			$model = new NavigationModel();
		}
		else {
			$model = NavigationModel::Construct($data['id']);
			if(!$model->checkExists()) {
				throw new NotFoundException();
			}
		}

		$model->name = $data['name'];
		$model->save();


		// Save the widget too
		$widget = $model->getLink('Widget');
		$widget->set('title', $model->get('name'));
		$widget->set('baseurl', '/navigation/view/' . $model->get('id'));
		$widget->set('editurl', '/navigation/edit/' . $model->get('id'));
		$widget->set('deleteurl', '/navigation/delete/' . $model->get('id'));
		$widget->save();


		// Save all the entries
		$counter = 0;

		$newMap = [];

		foreach ($data['entries'] as $dat) {
			if($dat['status'] === 'deleted') {
				$entry = NavigationEntryModel::Construct($dat['id']);
				if(!$entry->checkExists()) {
					continue;
				}

				if($entry->navigationid === $model->id) {
					// Security check to only manage entries on this widget.
					$entry->delete();
				}
			}
			elseif($dat['status'] === 'new') {
				// New entry!
				$entry = new NavigationEntryModel();

				$entry->setFromArray([
					'weight'       => ++$counter,
					'navigationid' => $model->id,
					'parentid'     => $newMap[ $dat['parent']] ?? $dat['parent'],
					'type'         => $dat['type'],
					'baseurl'      => $dat['url'],
					'title'        => $dat['title'],
					'target'       => $dat['target'],
				]);

				$entry->save();

				// Save this in the new map for children.
				$newMap[ $dat['id'] ] = $entry->id;
			}
			else {
				// Standard existing entry.
				$entry = NavigationEntryModel::Construct($dat['id']);

				if(!$entry->checkExists()) {
					continue;
				}

				$entry->setFromArray([
					'weight'       => ++$counter,
					'parentid'     => $newMap[ $dat['parent']] ?? $dat['parent'],
					'type'         => $dat['type'],
					'baseurl'      => $dat['url'],
					'title'        => $dat['title'],
					'target'       => $dat['target'],
				]);

				if($entry->navigationid === $model->id) {
					// Security check to only manage entries on this widget.
					$entry->save();
				}

				$newMap[ $dat['id'] ] = $entry->id;
			}
		}

		$view->appendJSONDataKey('location', \Core\resolve_link('/navigation#menu:' . $model->id));
		$view->setJSONSuccessMessage('Saved navigation menu');
	}
}
