<?php
/**
 * Class file for the controller CronController
 *
 * @package Security-Suite
 */

use Core\ConfigHandler;
use Core\Controller;
use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\Forms\Form;
use Core\ListingTable\Table;
use Core\UserAgent;
use Core\View;
use SecuritySuite\SpamCan\SpamCheck;


class SecurityController extends Controller {

	/**
	 * Provide a UI for the simple site password option.
	 */
	public function sitepassword(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('g:admin')){
			return View::STATUS_ACCESSDENIED;
		}

		// Build a form to handle the config options themselves.
		// These will include password strength, whether or not captcha is enabled, etc.
		$configs = [
			'/security/site_password',
		];
		$configform = new Form();
		$configform->set('callsmethod', 'SecurityController::SitePasswordSave');

		foreach($configs as $key){
			$el = ConfigHandler::GetConfig($key)->getAsFormElement();
			// I don't need this, (Everything from this group will be on the root-level form).
			$el->set('group', null);
			$configform->addElement($el);
		}

		$configform->addElement('submit', ['name' => 'submit', 'value' => 'Save Password']);

		$view->title = 'Simple Site Password';
		$view->assign('form', $configform);
		$view->assign('cgimode', stripos(php_sapi_name(), 'cgi') !== false);
	}

	/**
	 * View a specific cron execution and its details.
	 */
	public function view(){
		$view = $this->getView();
		$request = $this->getRequest();
		$view->mode = View::MODE_PAGEORAJAX;

		if(!\Core\user()->checkAccess('p:/security/viewlog')){
			return View::STATUS_ACCESSDENIED;
		}

		$logid = $request->getParameter(0);
		$log = SecurityLogModel::Construct($logid);
		if(!$log->exists()){
			return View::STATUS_NOTFOUND;
		}

		if($log->get('user_id')){
			$userobject = UserModel::Construct($log->get('user_id'));
			$user = $userobject->getDisplayName();
		}
		else{
			$user = null;
		}

		if($log->get('affected_user_id')){
			$userobject = UserModel::Construct($log->get('affected_user_id'));
			$affected_user = $userobject->getDisplayName();
		}
		else{
			$affected_user = null;
		}

		$view->addBreadcrumb('Security Log', '/security/log');
		$view->title = 'Details';
		$view->assign('entry', $log);
		$view->assign('user', $user);
		$view->assign('affected_user', $affected_user);
	}

	/**
	 * Display a list of blacklisted IP addresses and subnets.
	 */
	public function blacklistip(){
		$view = $this->getView();
		$request = $this->getRequest();

		$listing = new Table();
		$listing->setModelName('IpBlacklistModel');
		$listing->addColumn('ip_addr');
		$listing->addColumn('comment');
		$listing->addColumn('expires');
		$listing->addColumn('created');
		$listing->setDefaultSort('created', 'DESC');

		$listing->loadFromRequest($request);

		$view->addControl([
			'title' => 'Ban IP...',
			'icon' => 'add',
			'link' => '/security/blacklistip/add'
		]);
		$view->title = 'Blacklisted IP addresses';
		$view->assign('listings', $listing);
	}

	/**
	 * Quick-add an IP address to the blacklist.
	 */
	public function blacklistip_add(){
		$view = $this->getView();
		$request = $this->getRequest();

		$ban = new IpBlacklistModel();
		$ban->set('ip_addr', $request->getParameter('ip_addr'));
		$ban->set('message', 'Your IP address has been blocked from this site by the administrator!');

		$form = new Form();
		$form->set('callsmethod', 'SecurityController::SaveBlacklistIp');
		$form->addModel($ban, 'model');
		$form->addElement('submit', ['name' => 'submit', 'value' => 'Ban IP!']);

		$view->title = 'Ban IP';
		$view->assign('form', $form);
	}

	/**
	 * Quick-add an IP address to the blacklist.
	 */
	public function blacklistip_edit(){
		$view = $this->getView();
		$request = $this->getRequest();

		$ban = new IpBlacklistModel($request->getParameter(0));

		if(!$ban->exists()){
			return View::STATUS_NOTFOUND;
		}

		$form = new Form();
		$form->set('callsmethod', 'SecurityController::SaveBlacklistIp');
		$form->addModel($ban, 'model');
		$form->addElement('submit', ['name' => 'submit', 'value' => 'Edit Banned IP!']);

		$view->title = 'Edit Banned IP';
		$view->assign('form', $form);
	}

	public function blacklistip_delete(){
		$view = $this->getView();
		$request = $this->getRequest();

		$ban = new IpBlacklistModel($request->getParameter(0));

		if(!$ban->exists()){
			return View::STATUS_NOTFOUND;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$ban->delete();
		\Core\set_message('Removed ban successfully', 'success');
		\Core\go_back();
	}

	/**
	 * Save the site password.
	 *
	 * @param Form $form
	 *
	 * @return bool
	 */
	public static function SitePasswordSave(Form $form){
		$pass = $form->getElement('config[/security/site_password]')->get('value');

		ConfigHandler::Set('/security/site_password', $pass);

		return true;
	}

	/**
	 * @param Form $form
	 * @return false|string
	 */
	public static function SaveBlacklistIp(Form $form){
		try{
			$ban = $form->getModel('model');

			// First thing... check and make sure that this directive won't block out the current user!
			$longip = ip2long(REMOTE_IP);
			for($i=32; $i>0; $i--){
				$mask = ~((1 << (32 - $i)) - 1);
				$join = long2ip($longip & $mask) . '/' . $i;
				if($join == $ban->get('ip_addr')){
					\Core\set_message('Corwardly refusing to ban an IP range that will blacklist your current connection!', 'error');
					return false;
				}
			}

			// The expires value will probably come in as a date string :/
			if($ban->get('expires')){
				$date = new DateTime($ban->get('expires'));
				$ban->set('expires', $date->getFormatted('U', Timezone::TIMEZONE_GMT));
			}

			$ban->save();
			\Core\set_message('Banned IP range ' . $ban->get('ip_addr'), 'success');
			return 'back';
		}
		catch(Exception $e){
			\Core\set_message($e->getMessage());
			return false;
		}
	}
}
