<?php


namespace SecuritySuite\Akismet;


use Core\ConfigHandler;
use Core\Core;
use Core\Filestore\Backends\FileRemote;
use Core\Router;
use Core\UserAgent;

class AkismetCheck {
	private $_useragent;
	private $_ip;
	private $_referrer;
	private $_type = 'contact-form';
	private $_email;
	private $_content;

	private $_key;
	private $_url;

	public function __construct() {
		$request = Router::$DefaultRequest;

		$this->_useragent = $request->useragent;
		$this->_ip = REMOTE_IP;
		$this->_referrer = $request->referrer;

		$this->_key = ConfigHandler::Get('/security/akismet/key');
		$this->_url = ConfigHandler::Get('/security/akismet/url');
	}

	/**
	 * Check if Akismet is installed and ready to use.
	 *
	 * Requires the key and host URL to be set.
	 *
	 * @return bool
	 */
	public function checkIsReady() : bool {
		return $this->_key && $this->_url;
	}

	/**
	 * Check if this is a valid key and URL
	 *
	 * @return bool
	 * @throws \Exception
	 */
	public function checkKey() : bool {
		$file = new FileRemote('https://rest.akismet.com/1.1/verify-key');
		$file->setMethod('post');
		$file->cacheable = false;
		$file->setPayload([
			'key' => $this->_key,
			'blog' => $this->_url,
		]);

		// Akismet prefers user agents in a different format, so change that.
		$file->setRequestHeader(
			'User-Agent',
			'Core Plus/' . Core::GetComponent('core')->getVersion() . ' | ' . 'Security Suite/' . Core::GetComponent('security-suite')->getVersion()
		);

		$contents = $file->getContents();

		return trim($contents) === 'valid';
	}

	/**
	 * Set the Email of the submitter
	 *
	 * @param string $email
	 */
	public function setEmail(string $email) {
		$this->_email = $email;
	}

	/**
	 * Set the content for this message
	 *
	 * @param string $content
	 */
	public function setContent(string $content) {
		$this->_content = $content;
	}

	/**
	 * Check if this message is spam
	 *
	 * @return bool
	 * @throws \Exception
	 */
	public function checkIsSpam() : bool {
		$file = new FileRemote('https://' . $this->_key . '.rest.akismet.com/1.1/comment-check');
		$file->setMethod('post');
		$file->cacheable = false;

		$data = [
			'blog'         => $this->_url,
			'user_ip'      => $this->_ip,
			'user_agent'   => $this->_useragent,
			'referrer'     => $this->_referrer,
			'comment_type' => $this->_type,
			'blog_lang'    => 'en', // @todo Grab these tags from the I18N system
		];

		if($this->_email) {
			$data['comment_author_email'] = $this->_email;
		}

		if($this->_content) {
			$data['comment_content'] = $this->_content;
		}

		$file->setPayload($data);

		// Akismet prefers user agents in a different format, so change that.
		$file->setRequestHeader(
			'User-Agent',
			'Core Plus/' . Core::GetComponent('core')->getVersion() . ' | ' . 'Security Suite/' . Core::GetComponent('security-suite')->getVersion()
		);

		$contents = $file->getContents();

		return trim($contents) === 'true';
	}


}