<?php
/**
 * File for class SpamCheck definition in the coreplus project
 * 
 * @package SecuritySuite\SpamCan
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140411.1017
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace SecuritySuite\SpamCan;
use Core\ConfigHandler;
use Core\Datamodel\Dataset;
use GeoLookup\Lookup;
use SecuritySuite\Akismet\AkismetCheck;


/**
 * A short teaser of what SpamCheck does.
 *
 * More lengthy description of what SpamCheck does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for SpamCheck
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @package SecuritySuite\SpamCan
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class SpamCheck {
	/** @var string The source IP address of this spam test */
	private $_ip;

	/** @var string The full textual contents of this spam test */
	private $_content;

	/** @var string Email address of the submitter, when available */
	private $_email;

	/** @var string UserAgent of the submitting user; NOT USED CURRENTLY */
	private $_userAgent;

	private $_results = null;

	private $_threshold;

	public function __construct(string $content = '') {
		$this->_content = $content;
		$this->_ip = REMOTE_IP;
		$this->_threshold = ConfigHandler::Get('/security/spam_threshold');
	}

	public function addContent(string $content) {
		$this->_content .= $content . "\n\n";
	}

	public function setEmail(string $email) {
		$this->_email = $email;
	}

	public function setUserAgent(string $userAgent) {
		$this->_userAgent = $userAgent;
	}

	public function setIP(string $ip) {
		$this->_ip = $ip;
	}

	public function checkIsSpam() : bool {
		return $this->getScore() > $this->_threshold;
	}

	public function checkIsBlockable() : bool {
		return $this->getScore() > ($this->_threshold * 2);
	}

	/**
	 * Get all results for this message
	 *
	 * @return SpamCheckResult[]
	 */
	public function getResults() : array {
		if($this->_results === null) {
			$this->_runChecks();
		}

		return $this->_results;
	}

	/**
	 * Get the score of spam level
	 *
	 * @return float
	 */
	public function getScore() : float {
		if($this->_results === null) {
			$this->_runChecks();
		}

		$result = 0.0;
		foreach($this->_results as $r) {
			/** @var SpamCheckResult $r */
			$result += $r->score;
		}
		return $result;
	}

	/**
	 * Extract all possible keywords from the content along with their current scores.
	 *
	 * @return array
	 */
	public function getKeywords(){
		$s = strtolower($this->_content);

		$s = strip_tags($s);

		$s = preg_replace('/[^a-z0-9 ]/', '', $s);

		$stopwords = \Core\get_stop_words();

		// Add on a few more custom stop words that don't necessarily belong in upstream.
		$stopwords[] = 'id';
		$stopwords[] = 'like';

		$exploded = explode(' ', $s);
		$nt = [];
		foreach($exploded as $i => $w){
			if($w == ''){
				continue;
			}
			if(in_array($w, $stopwords)){
				continue;
			}

			$nt[] = $w;

			if(isset($exploded[$i+1])){
				$nt[] = $w . ' ' . $exploded[$i+1];
			}
			if(isset($exploded[$i+2])){
				$nt[] = $w . ' ' . $exploded[$i+1] . ' ' . $exploded[$i+2];
			}

		}
		$nt = array_unique($nt);

		sort($nt);

		// @todo Reimplement this without requiring a database
		//$all = Dataset::Init()->select('*')->table('spam_ham_keyword')->execute();
		$all = [];
		// Convert this to something I can quickly check through.
		$keywords = [];
		foreach($all as $row){
			$keywords[ $row['keyword'] ] = $row['score'];
		}

		$ret = [];

		foreach($nt as $w){
			if(isset($keywords[ $w ])){
				$score = $keywords[ $w ];
			}
			else{
				$score = 0;
			}

			$ret[] = [
				'keyword' => $w,
				'score' => $score,
			];
		}

		return $ret;
	}

	/**
	 * Run this test and get the results.
	 *
	 */
	private function _runChecks() {
		$this->_results = [];

		// HTML tags indicates +1
		$this->_checkHTML();

		// Each link indicates +2
		$this->_checkLinks();

		// This can be +10
		$this->_checkEmailDomain();

		// Each keyword indicates an arbitruary value, (setting by the admin)
		//$this->_checkKeywords();

		// Akismet indicates +10
		$this->_checkAkismet();

		// Various source contries have different weights
		$this->_checkGeoLocation();
	}

	/**
	 * Check if the content has HTML content.
	 */
	private function _checkHTML() {
		if(strpos($this->_content, '<') !== false && strpos($this->_content, '>') !== false){
			$this->_addResult('Has HTML Content', 1);
		}
	}

	/**
	 * Check if there are links in this message
	 */
	private function _checkLinks() {
		// Since all links require :// in them, (the protocol), we'll use that as the basis of the check.
		$score = preg_match_all('#://#', $this->_content);

		if($score == 1){
			$this->_addResult('Has ' . $score . ' link', $score * 2);
		}
		elseif($score > 1){
			$this->_addResult('Has ' . $score . ' links', $score * 2);
		}
	}

	/**
	 * Check the email domain against a list of free email providers
	 *
	 * List compiled from Okutbay
	 *
	 * @link https://gist.github.com/okutbay/5b4974b70673dfdcc21c517632c1f984
	 */
	private function _checkEmailDomain() {
		if($this->_email) {
			$pos = strpos($this->_email, '@');
			if(!$pos) {
				// Does not appear to be an email.
				return;
			}
			$email = strtolower(trim(substr($this->_email, $pos)));

			$res = preg_grep('/' . $email . '/', file(ROOT_PDIR . 'components/security-suite/resources/free_email_provider_domains.txt'));

			if($res) {
				$this->_addResult($this->_email . ' listed in spam email database', 10);
			}
		}
	}

	/**
	 * Check they keywords against the local database of spam keywords
	 */
	private function _checkKeywords() {
		$keywords = $this->getKeywords();

		foreach($keywords as $dat){
			if($dat['score'] != 0){
				$this->_addResult('Contains the word or phrase "' . $dat['keyword'] . '"', $dat['score']);
			}
		}
	}

	/**
	 * Check against Akismet, when enabled.
	 */
	private function _checkAkismet() {
		$akismet = new AkismetCheck();
		if(!$akismet->checkIsReady()) {
			// If Akismet is not ready, (ie: configured), then nothing to do.
			return;
		}

		$akismet->setEmail($this->_email);
		$akismet->setContent($this->_content);

		if($akismet->checkIsSpam()) {
			$this->_addResult('Akismet reported this message as spam', 10);
		}
	}

	/**
	 * Geolocation checks, currently just grabs the country and marks as appropriate.
	 */
	private function _checkGeoLocation() {
		if($this->_ip && \Core\Core::IsComponentAvailable('geographic-codes')) {
			$lookup = Lookup::ByIP($this->_ip);

			switch($lookup->country) {
				case 'RU':
					$this->_addResult('Source is Russia', 10);
					break;
				case 'DE':
					$this->_addResult('Source is Germany', 1);
					break;
				case 'FR':
					$this->_addResult('Source is France', 1);
					break;
				case 'RO':
					$this->_addResult('Source is Romania', 5);
			}
		}
	}

	private function _addResult($comment, $score) {
		$r = new SpamCheckResult();
		$r->score = (float)$score;
		$r->comment = $comment;
		$this->_results[] = $r;
	}
}

class SpamCheckResult {
	public $score;
	public $comment;
}