"use strict";

(() => {

	/**
	 * Simple function to create a message box with the given message and status
	 *
	 * @param message
	 * @param type
	 *
	 * @return string
	 */
	function create_message_html(message, type) {
		return '<p class="message-' + type + '">' + message + '</p>';
	}

	/**
	 * Simple function to create an icon and optionally apply spin
	 *
	 * @param icon
	 * @param spin
	 * @returns {string}
	 */
	function create_icon_html(icon, spin) {
		let classes = ['icon', 'icon-' + icon];

		if(typeof spin !== 'undefined' && spin === true) {
			classes.push('icon-spin');
		}

		return '<i class="' + classes.join(' ') + '"></i>';
	}

	/**
	 * Create the actual table HTML
	 *
	 * @param table
	 * @param data
	 */
	function render_table_html(table, data) {
		const $data  = table.querySelector('.listing-table-data-container'),
			$optionButton = table.querySelector('.open-options'),
			$filterButton = table.querySelector('.open-filters'),
			$setViewTableButton = table.querySelector('.set-view-table'),
			$setViewCollectionButton = table.querySelector('.set-view-collection'),
			$setViewReportButton = table.querySelector('.set-view-report'),
			$downloadButton = table.querySelector('.download-link');

		let url = new URLSearchParams(window.location.search);

		$data.innerHTML = data.body;

		// Enable this once a good solution for nested tables on the same page is implemented.
		url.set(table.dataset.tableName, JSON.stringify(data.parameters.url));
		window.history.replaceState(null, document.title, '?' + url.toString());

		//window.history.replaceState(null, document.title, '?' + data.parameters.url);
		//console.log(data.parameters.url);
		//console.log(url.toString());

		if($data.querySelector('.listing-table-options') && $optionButton) {
			$optionButton.classList.remove('disabled');
		}
		else if($optionButton) {
			$optionButton.classList.add('disabled');
		}

		if(data.filters_set && $filterButton) {
			$filterButton.classList.remove('disabled');
			$filterButton.classList.add('active');
		}
		else if(data.has_filters && $filterButton) {
			$filterButton.classList.remove('disabled', 'active');
			$filterButton.classList.remove('active');
		}
		else if($filterButton) {
			$filterButton.classList.add('disabled');
			$filterButton.classList.remove('active');
		}

		if($downloadButton) {
			if(data.download === 'csv') {
				$downloadButton.classList.remove('disabled');
				$downloadButton.querySelector('i').className = 'icon icon-file-excel';
			}
			else if(data.download === 'pdf') {
				$downloadButton.classList.remove('disabled');
				$downloadButton.querySelector('i').className = 'icon icon-file-pdf';
			}
			else {
				$downloadButton.classList.add('disabled');
				$downloadButton.querySelector('i').className = 'icon icon-download';
			}
		}


		if($setViewTableButton) {
			if(data.parameters.view === 'table') {
				$setViewTableButton.classList.add('active');
			}
			else {
				$setViewTableButton.classList.remove('active');
			}
		}

		if($setViewCollectionButton) {
			if(data.parameters.view === 'collection') {
				$setViewCollectionButton.classList.add('active');
			}
			else {
				$setViewCollectionButton.classList.remove('active');
			}
		}

		if($setViewReportButton) {
			if(data.parameters.view === 'report') {
				$setViewReportButton.classList.add('active');
			}
			else {
				$setViewReportButton.classList.remove('active');
			}
		}

		// Sync the parameter data just for reference.
		for(let paramKey in data.parameters) {
			let keyName = paramKey.replace(/_/g, '-'),
				keyValue;

			if(typeof(data.parameters[paramKey]) === 'undefined') {
				keyValue = '';
			}
			else if(data.parameters[paramKey] === null) {
				keyValue = '';
			}
			else {
				keyValue = data.parameters[paramKey];
			}

			table.setAttribute('data-' + keyName, keyValue);
		}

		// Notify any listening system that this table has loaded.
		table.dispatchEvent(new Event('load'));
		window.dispatchEvent(new Event('change'));
	}

	/**
	 * Parse a URL and merge it with given parameters
	 *
	 * @param {string}      url        Base URL of this table
	 * @param {string}      tableName  Table name to request, (for multiple tables)
	 * @param {string|null} parameters Additional parameters to send as a GET request
	 * @param {string}      format     Format extension to request, defaults to json
	 *
	 * @returns {string}
	 */
	function parse_url(url, tableName, parameters, format) {
		if(url.indexOf('?') !== -1) {
			// Trim this off the path; it may be provided if the incoming URL had a query string.
			url = url.substr(0, url.indexOf('?'));
		}

		// Prepend the root url on this.
		url = Core.ROOT_URL + url.substr(1);


		// All requests will have an extension, defaults to a JSON request
		url += '.' + (format || 'json');

		// Always add the table name, for reference on multiple tables
		url += '?table=' + tableName;

		if(parameters !== null) {
			if(parameters.indexOf('://') !== -1 && parameters.indexOf('?') !== -1) {
				// Parameters probably are fully resolved
				parameters = parameters.substr(parameters.indexOf('?'));
			}

			// Add the requested parameters to this lookup.
			if(parameters.indexOf('?') === 0) {
				url += '&' + parameters.substr(1);
			}
			else {
				url += '&' + parameters;
			}
		}

		return url;
	}

	/**
	 * Load (or reload) table data
	 *
	 * @param table
	 * @param parameters
	 * @param background
	 */
	function load_table(table, parameters, background) {
		const url    = parse_url(table.dataset.tableBaseurl, table.dataset.tableName, parameters),
			$data    = table.querySelector('.listing-table-data-container'),
			$refresh = table.querySelector('.auto-refresh-link i');

		if($refresh) {
			$refresh.classList.add('icon-spin');
		}

		if(!background) {
			// Set the text as a loading template
			$data.innerHTML = '<div class="listing-table-loading">' + create_icon_html('spinner', true) + '</div>';

			let filtersButton = table.querySelector('.open-filters'),
				optionsButton = table.querySelector('.open-options'),
				downloadButton = table.querySelector('.download-link');

			// Clear the dynamic sections for active loading
			if(table.dialog && table.dialog.isOpen) {
				table.dialog.close();
			}

			if(filtersButton) {
				filtersButton.classList.remove('active');
				filtersButton.classList.add('disabled');
			}

			if(optionsButton) {
				optionsButton.classList.remove('active');
				optionsButton.classList.add('disabled');
			}

			if(downloadButton) {
				downloadButton.classList.add('disabled');
			}

			// HACK
			// Force tables with large amounts of columns/data to be contained within their parent.
			// This must be done prior to loading the data, as we need to determine the base width.
			table.style.maxWidth = table.parentNode.clientWidth + 'px';
		}
		
		fetch(url, {method: 'GET'})
			.then(res => {
				return res.json();
			})
			.then(data => {
				render_table_html(table, data);

				if($refresh) {
					$refresh.classList.remove('icon-spin');
				}

				if(typeof(Core.DevTools) !== 'undefined' && typeof(data._debug) !== 'undefined') {
					Core.DevTools.SetPageDetails(data._debug.page);
				}
			})
			.catch(e => {
				$data.innerHTML = create_message_html(e.message, 'error');
				console.log(e);

				if($refresh) {
					$refresh.classList.remove('icon-spin');
				}
			});
	}

	function matches() {
		if(arguments.length < 2) {
			throw 'matches requires at least 2 arguments, the target followed by the matches';
		}

		const target = arguments[0];

		for(let i = 1; i < arguments.length; i++) {
			let match = target.closest(arguments[i]);
			if(match !== null) {
				// Matched one of the arguments!
				return match;
			}
		}

		return null;
	}

	function register_table_events(table) {
		let refreshTimer = null;

		// Assign a tracker so we know these event has been registered.
		// This prevents double registration.
		table.dataset.eventsRegistered = '1';

		// Assign the events for this table
		table.addEventListener('click', evt => {
			let match;

			if((match = matches(evt.target, '.listing-table-pager a')) !== null) {
				load_table(table, match.href, false);
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.print-link')) !== null) {
				window.print();
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.sort-icons .sort-icon')) !== null) {
				load_table(
					table,
					'sortkey=' + match.dataset.key + '&sortdir=' + match.dataset.dir,
					false
				);
			}


			if((match = matches(evt.target, '.apply-filters', '.apply-columns')) !== null) {
				let form = match.closest('form');

				load_table(
					table,
					(new URLSearchParams(new FormData(form))).toString(),
					false
				);
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.reset-filters')) !== null) {
				let form = match.closest('form');

				form.reset();
				load_table(table, match.href,false);
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.table-change-view')) !== null) {
				evt.preventDefault();

				if(match.classList.contains('disabled')) {
					return;
				}

				load_table(table, match.href,false);
			}


			if((match = matches(evt.target, '.auto-refresh-link')) !== null) {
				if(match.classList.contains('active')) {
					match.classList.remove('active');

					if(refreshTimer !== null) {
						clearInterval(refreshTimer);
						refreshTimer = null;
					}
				}
				else {
					match.classList.add('active');

					load_table(table, null,true);

					refreshTimer = setInterval(table => {
						load_table(table, null,true);
					}, 30000, table);
				}

				evt.preventDefault();
			}


			if((match = matches(evt.target, '.bulk-select')) !== null) {

				if(table.querySelectorAll('.bulk-select:checked').length > 0) {
					table.classList.add('bulk-action-selected');
				}
				else {
					table.classList.remove('bulk-action-selected');
				}
			}


			if((match = matches(evt.target, '.bulk-select-all')) !== null) {
				let isChecked = match.checked;


				table.querySelectorAll('.bulk-select').forEach(el => {
					el.checked = isChecked;
				});

				if(isChecked) {
					table.classList.add('bulk-action-selected');
				}
				else {
					table.classList.remove('bulk-action-selected');
				}
			}


			if((match = matches(evt.target, '.report-view')) !== null) {
				let form = match.closest('form');

				load_table(
					table,
					'report=' + match.dataset.report,
					false
				);
				evt.preventDefault();
			}

			if((match = matches(evt.target, '.open-filters')) !== null) {
				evt.preventDefault();

				if(match.classList.contains('disabled')) {
					return;
				}

				if(!table.dialog) {
					table.dialog = new Core.ModalDialog();
				}

				table.dialog.setFrame(table.querySelector('.listing-table-filters'));
				table.dialog.open();
			}

			if((match = matches(evt.target, '.open-options')) !== null) {
				evt.preventDefault();

				if(match.classList.contains('disabled')) {
					return;
				}

				if(!table.dialog) {
					table.dialog = new Core.ModalDialog();
				}

				table.dialog.setFrame(table.querySelector('.listing-table-options'));
				table.dialog.open();
			}

			if((match = matches(evt.target, '.download-link')) !== null) {
				if(match.classList.contains('disabled')) {
					evt.preventDefault();
				}
			}

			if((match = matches(evt.target, '.bulk-action')) !== null) {
				evt.preventDefault();

				if(match.classList.contains('disabled')) {
					return;
				}

				let form = table.querySelector('.listing-table-data form');
				if(!form) {
					// No form present, nothing to submit
					return;
				}

				form.action = match.dataset.href;
				form.submit();
			}
		});


		// "SUBMIT" Event Listener
		table.addEventListener('submit', evt => {
			let match;

			if((match = matches(evt.target, '.listing-table-filters form')) !== null) {
				load_table(
					table,
					(new URLSearchParams(new FormData(match))).toString(),
					false
				);

				evt.preventDefault();
			}
		});


		// "LOAD" Event Listener
		table.addEventListener('load', evt => {
			if(evt.target.dataset.view === 'collection') {
				// Work tasks support collapsible fields.
				evt.target.querySelectorAll('.collection-item.collapsed').forEach(item => {
					let collapsibles = item.querySelectorAll('.collapsible'),
						expander = item.querySelector('.collection-item-expand-hint'),
						collapser = item.querySelector('.collection-item-collapse-hint'),
						hasCollapsible = false;

					collapsibles.forEach(el => {
						if(el.scrollHeight > el.clientHeight) {
							hasCollapsible = true;
						}
					});

					if(!hasCollapsible) {
						item.classList.remove('collapsed');
					}
					else {
						expander.addEventListener('click', evt => {
							let item = evt.target.closest('.collection-item');
							item.classList.remove('collapsed');
							item.classList.add('expanded');

							evt.preventDefault();
						});

						collapser.addEventListener('click', evt => {
							let item = evt.target.closest('.collection-item');
							item.classList.add('collapsed');
							item.classList.remove('expanded');

							evt.preventDefault();
						});
					}
				});
			}

			evt.target.querySelectorAll('a[data-target="background"]').forEach(el => {
				// This link is being submitted
				el.submit = () => {
					$.notify('Loading...', 'info');
				};

				// The response of this link came back.
				el.load = data => {
					if(data.status === 200) {
						$.notify(data.message, 'success');
						load_table(table, null,true);
					}
					else {
						$.notify(data.message, 'error');
					}
				};
			});
		});


		// Bind actions onto the table for use by external actions
		table.load_table = (url, background) => {
			load_table(table, url, background);
		}
	}

	/**
	 * Check the page for any listing table and ensure it's setup.
	 */
	function check_page_for_tables() {
		let tables = document.querySelectorAll('.listing-table');

		tables.forEach(table => {
			if(table.dataset.eventsRegistered !== '1') {
				// Only register new tables.
				register_table_events(table);

				// Load the table initially with default parameters.
				load_table(table, null, false);
			}
		});
	}


	// Bind to the "change" event for background data
	window.addEventListener('change', evt => {
		check_page_for_tables();
	});

	// Hack pt2 for fixed width tables
	// on resize, force a reload of all tables to trigger their resize algorithm to run
	let documentWidth = document.querySelector('.outer-wrapper').clientWidth;

	window.addEventListener('resize', evt => {
		if(document.querySelector('.outer-wrapper').clientWidth !== documentWidth) {
			documentWidth = document.querySelector('.outer-wrapper').clientWidth;
			let tables = document.querySelectorAll('.listing-table');

			tables.forEach(table => {
				load_table(table, null, false);
			});
		}
	});

	// First load too!
	check_page_for_tables();
})();
