"use strict";

Core.PaneViewer = function(pane) {

	this.pane = pane;
	this.links = pane.querySelector('.pane-viewer-links');
	this.name = pane.dataset.pane;
	this.selected = localStorage.getItem('core-pane-' + this.name);
	this.renderArea = pane.querySelector('.pane-viewer-window');
	this.scripts = [];
	this.inlineScripts = [];
	this.scriptsLoaded = 0;
	this.loadingHTML = '<div class="pane-loading"><i class="icon icon-spin icon-spinner"></i></div>';


	/**
	 * Event listener to watch for clicks on the pane links
	 */
	this.links.addEventListener('click', evt => {
		let match = evt.target.closest('a');

		if(match !== null) {
			this.activate_link(match);
			evt.preventDefault();
		}
	});


	/**
	 * Activate a selected link and load the target into the pane
	 *
	 * @param link
	 */
	this.activate_link = link => {

		let parameters = {};

		this.renderArea.innerHTML = this.loadingHTML;

		// Remember this selected link for the next page reload
		localStorage.setItem('core-pane-' + this.name, link.href);

		// Iterate over the set of links contained in the links region and set the correct one to active.
		this.links.querySelectorAll('a').forEach(el => {
			if(el === link) {
				// The target link gets the active flag.
				link.classList.add('active');
			}
			else {
				el.classList.remove('active');
			}
		});

		// Cleanup previously loaded scripts so the new page scripts will execute
		for(let i = 0; i < this.scripts.length; i++) {
			document.body.removeChild(this.scripts[i]);
		}


		// Scripts are handled internally.
		parameters.loadScripts = false;

		parameters.callback = (data, contentType) => {
			if(contentType === 'application/json' && typeof(data.body) !== 'undefined') {
				// Body is expected to be HTML data encoded in the JSON stream.
				this.renderArea.innerHTML = data.body;
			}
			else if(contentType === 'text/html') {
				this.renderArea.innerHTML = data;
			}
			else if(contentType === 'text/plain') {
				this.renderArea.innerHTML = '<pre>' + data + '</pre>';
			}
			else {
				throw new Error('What am I supposed to do with the response data?');
			}

			if(typeof(data.scripts) !== 'undefined') {
				// Everything should be loaded now, include the scripts that may be required.
				this.setScripts(data.scripts);
			}

			if(typeof(data.scripts_inline) !== 'undefined') {
				// Also include any inline scripts
				this.setInlineScripts(data.scripts_inline);
			}


			if(
				typeof(data.scripts) === 'undefined' &&
				typeof(data.scripts_inline) === 'undefined'
			) {
				// No scripts were provided, but a callback was requested, so call that manually
				window.dispatchEvent(new Event('change'));
			}
		};

		Core.Ajax(link.href, parameters).catch(e => {
			this.renderArea.innerHTML = '<p class="message-error">Sorry, but something unexpected happened.  Press ESCAPE to close.</p>';
			console.error(e);
		});
	};


	/**
	 * Set any required script for this dialog, usually set from JSON data.scripts
	 *
	 * @param {array} scripts
	 */
	this.setScripts = scripts => {
		let currentDocumentScriptURLs = [],
			toAdd = [];

		// Check the current scripts, we don't want to add duplicates.
		for(let i = 0; i < document.scripts.length; i++) {
			if(document.scripts.item(i).src) {
				currentDocumentScriptURLs.push(document.scripts.item(i).src);
			}
		}

		// Check the incoming scripts and queue only those needed.
		for(let i = 0; i < scripts.length; i++) {
			if(currentDocumentScriptURLs.indexOf(scripts[i]) === -1) {
				toAdd.push(scripts[i]);
			}
		}

		this.scripts = [];
		this.scriptsLoaded = 0;

		// We now have a queue of scripts to append.
		for(let i = 0; i < toAdd.length; i++) {
			let scriptNode = document.createElement('script');
			scriptNode.src = toAdd[i];
			scriptNode.addEventListener('load', evt => {
				this.scriptsLoaded += 1;

				if(this.scriptsLoaded === this.scripts.length) {
					// All scripts are loaded.
					this.loadInlineScripts();
				}
			});

			document.body.appendChild(scriptNode);

			// Remember this so we can perform cleanup
			this.scripts.push(scriptNode);
		}
	};


	/**
	 * Set any required script for this dialog, usually set from JSON data.scripts
	 *
	 * These are INLINE scripts, and as such do not have a src to track,
	 * so they better be able to run multiple times on the same page!
	 *
	 * @param {array} scripts
	 */
	this.setInlineScripts = scripts => {
		this.inlineScripts = scripts;
		if(this.scriptsLoaded === this.scripts.length) {
			// All scripts are loaded or no scripts are required.
			this.loadInlineScripts();
		}
	};

	/**
	 * Load all inline scripts requested from this nested page
	 *
	 * Will automatically trigger the 'change' event on the window
	 */
	this.loadInlineScripts = () => {
		for(let i = 0; i < this.inlineScripts.length; i++) {
			eval(this.inlineScripts[i]);
		}

		window.dispatchEvent(new Event('change'));
	};


	/**
	 * Initialize the pane
	 * @type {null}
	 */
	let selectedLink = null;
	if(this.selected) {
		// If there's not one selected, then activate the first link.
		// Find the selected link and try to activate it.
		this.links.querySelectorAll('a').forEach(link => {
			if(link.href === this.selected) {
				this.activate_link(link);
				selectedLink = link;
				return false;
			}
		});
	}

	// Fallback if no link is selected or the previously selected link is no longer present
	if(selectedLink === null) {
		this.activate_link(this.links.querySelector('a'));
	}
};


(() => {
	document.querySelectorAll('.pane-viewer').forEach(pane => {
		pane.PaneViewer = new Core.PaneViewer(pane);
	});
})();