/**
 * Javascript helpers for editing forms with page elements integrated
 */

Core.PageEditor = {};

/**
 * Setup the meta author tag functionality
 *
 * @param jQuery $element Element object to setup against
 */
Core.PageEditor.metaauthor = function($parent) {
	var
		$element  = $parent.find('input[type=text]'),
		elementid = $element.attr('id'),
		hiddenid  = elementid.replace('formpagemetaauthorinput', 'formhiddeninput') + "id",
		$hidden   = $('#' + hiddenid);


	// Initial statuses
	if($hidden.val()) {
		$parent.addClass('valid');
	}

	// Gogo autocomplete!
	$element.autocomplete({
		source: Core.ROOT_URL + 'form/pagemetas/autocompleteuser.ajax',
		minLength: 2,
		search: function(event, ui) {
			// Called when searching has been initiated

			$parent.removeClass('valid').addClass('loading');

			// When searching, clear out the previous value, if any set.
			$hidden.val('');
		},
		response: function( event, response ) {
			// Called when a response from the server was received.

			$parent.removeClass('loading');
		},
		select: function( event, ui ) {
			if(ui.item){
				// This is a bit different because the value is actually going to a different field.
				$hidden.val(ui.item.id);
				$element.val(ui.item.label);
				$parent.addClass('valid');

				// The return false is to prevent jqueryui from setting the value to the id of the user.
				// I want the label instead, (set above).
				return false;
			}
		}
		// ui-autocomplete-loading
	});
};

/**
 * Process the value of a given box, and provide a UI prompt on its status.
 *
 * @param jQuery $parent
 */
Core.PageEditor.seotitle = function($parent) {
	var
		$element  = $parent.find('input[type=text]'),
		run = function(){
			var val = $element.val(),
				words, stopWordCount;

			if(val.length > 69) {
				$element.notify('SEO Title is longer than 69 characters!', 'error');
				return;
			}

			if(val.length < 20) {
				$element.notify('SEO Title should be longer than 20 characters!', 'error');
				return;
			}

			words = val.split(' ');
			stopWordCount = 0;
			for(i in words) {
				if(Core.StopWords.indexOf(words[i]) !== -1) {
					stopWordCount++;
				}
			}
			if(stopWordCount > 2) {
				$element.notify('SEO Title should not contain very many "stop" words, you have ' + stopWordCount + ' stop words!', 'warning');
				return;
			}

			$element.notify('SEO Title looks good!', 'success');
		};

	$element.on('keyup', run);
	$element.on('click', run);
};

/**
 * Handle the logic required for insertable groups, including adding new, moving, etc.
 */
Core.PageEditor.InsertableGroups = class {

	/**
	 *
	 * @param {Node} groupRoot
	 */
	constructor(groupRoot) {
		this.inputs = { };
		this.elementHTML = '';
		this.counter = 0;
		this.renderArea;
		this.descriptor = groupRoot.dataset.descriptor;

		let createButton,
			groupName = groupRoot.dataset.group;

		groupRoot.classList.add('page-editor-insertable-group');

		groupRoot.querySelectorAll('.formelement').forEach(el => {
			this.elementHTML += el.outerHTML;

			el.querySelectorAll('input,textarea,select').forEach(input => {
				if(input.name) {
					let pos = input.name.indexOf('__INSERTABLEGROUPTEMPLATE'),
						inputKey;

					// Trim anything before the template declaration off.
					// the remainder is the actual element key
					inputKey = input.name.substr(pos + 27);
					// (sans the trailing ']' character)
					inputKey = inputKey.substr(0, inputKey.length - 1);

					// Support nested elements
					// This will convert them to a single string.
					inputKey = inputKey.replace(new RegExp(/[\[\]]/g), '/').replace(new RegExp(/\/+/g), '/');

					// Store the actual input elements so we know where to do the value replacements.
					this.inputs[ inputKey ] = input.name;
				}
			});
			el.parentElement.removeChild(el);
		});

		// Create markup to hold the generated elements and the group controls
		this.renderArea = document.createElement('div');
		this.renderArea.classList.add('insertable-groups', 'collection-list');
		groupRoot.appendChild(this.renderArea);

		// Create a button the user can click to add a new group
		createButton = document.createElement('a');
		createButton.classList.add('insertablegroup-add', 'button');
		createButton.innerHTML = '<i class="icon icon-add"></i> <span>Add ' + this.descriptor + '</span>';
		groupRoot.appendChild(createButton);

		createButton.addEventListener('click', evt => {
			this.addGroup(null);
			evt.preventDefault();
		});

		// Scan if there are current elements to create
		if(
			typeof(window.PageEditorGroups) !== 'undefined' &&
			typeof(window.PageEditorGroups[groupName]) !== 'undefined'
		) {
			for(let i = 0; i < window.PageEditorGroups[groupName].length; i++) {
				this.addGroup(window.PageEditorGroups[groupName][i]);
			}
		}
	}

	addGroup(groupData) {
		let group = document.createElement('div'),
			groupBody = document.createElement('div'),
			groupExpand = document.createElement('div'),
			groupCollapse = document.createElement('div'),
			groupControls = document.createElement('div'),
			html = this.elementHTML;

		group.classList.add('insertable-group', 'collection-item', 'collapsed');
		this.renderArea.appendChild(group);

		groupBody.classList.add('collection-item-description', 'collapsible');

		html = html.replace(
			new RegExp('__INSERTABLEGROUPTEMPLATE', 'g'),
			this.counter
		);
		html = html.replace(
			new RegExp('__insertablegrouptemplate', 'g'),
			this.counter
		);
		html = html.replace(
			new RegExp('insertablegrouptemplate', 'g'),
			this.counter
		);


		groupBody.innerHTML = html;
		group.appendChild(groupBody);

		groupExpand.classList.add('collection-item-expand-hint');
		groupExpand.innerHTML = 'Show';
		group.appendChild(groupExpand);

		groupCollapse.classList.add('collection-item-collapse-hint');
		groupCollapse.innerHTML = 'Hide';
		group.appendChild(groupCollapse);

		groupControls.classList.add('collection-item-footnote');
		groupControls.innerHTML = '<div class="button-group">' +
			'<a href="#" class="insertablegroup-move-top button"><i class="icon icon-angle-double-up"></i> <span>Move First</span></a>' +
			'<a href="#" class="insertablegroup-move-up button"><i class="icon icon-arrow-up"></i> <span>Move Up</span></a>' +
			'<a href="#" class="insertablegroup-move-down button"><i class="icon icon-arrow-down"></i> <span>Move Down</span></a>' +
			'<a href="#" class="insertablegroup-move-last button"><i class="icon icon-angle-double-down"></i> <span>Move Last</span></a>' +
			'<a href="#" class="insertablegroup-remove button"><i class="icon icon-trash"></i> <span>Remove ' + this.descriptor + '</span></a>' +
			'</div>';
		group.appendChild(groupControls);


		groupControls.querySelector('.insertablegroup-move-down').addEventListener('click', evt => {
			this.moveDown(evt.target.closest('.insertable-group'));
			evt.preventDefault();
		});

		groupControls.querySelector('.insertablegroup-move-up').addEventListener('click', evt => {
			this.moveUp(evt.target.closest('.insertable-group'));
			evt.preventDefault();
		});

		groupControls.querySelector('.insertablegroup-move-top').addEventListener('click', evt => {
			this.moveTop(evt.target.closest('.insertable-group'));
			evt.preventDefault();
		});

		groupControls.querySelector('.insertablegroup-move-last').addEventListener('click', evt => {
			this.moveLast(evt.target.closest('.insertable-group'));
			evt.preventDefault();
		});

		groupControls.querySelector('.insertablegroup-remove').addEventListener('click', evt => {
			this.removeGroup(evt.target.closest('.insertable-group'));
			evt.preventDefault();
		});


		if(groupData) {
			this.populateData(groupBody, groupData, '');
		}

		this.updatePositions();

		// test
		groupBody.querySelectorAll('input,select,textarea').forEach(el => {
			try {
				el.dispatchEvent(new Event('loaded', {bubbles: true, cancelable: true}));
			}
			catch (e) {
				console.log('Failed to dispatch "loaded" event on input, probably fine.', e);
			}
		})

		this.counter++;
	};

	moveUp(group) {
		let parent = group.parentNode,
			prevNode = group.previousElementSibling;

		if(prevNode) {
			parent.insertBefore(group, prevNode);
			this.updatePositions();
		}
	}

	moveTop(group) {
		let parent = group.parentNode,
			prevNode = group.previousElementSibling;

		while(prevNode.previousElementSibling) {
			prevNode = prevNode.previousElementSibling;
		}

		if(prevNode) {
			parent.insertBefore(group, prevNode);
			this.updatePositions();
		}
	}

	moveDown(group) {
		let parent = group.parentNode,
			nextNode = group.nextElementSibling;

		if(nextNode) {
			parent.insertBefore(nextNode, group);
			this.updatePositions();
		}
	}

	moveLast(group) {
		let parent = group.parentNode;

		parent.insertBefore(group, null);
		this.updatePositions();
	}

	removeGroup(group) {
		group.parentNode.removeChild(group);
		this.updatePositions();
	}

	updatePositions() {
		let x = 0,
			groups = this.renderArea.querySelectorAll('.insertable-group'),
			glen = groups.length;

		groups.forEach(g => {
			if(x === 0) {
				g.querySelector('.insertablegroup-move-up').classList.add('disabled');
				g.querySelector('.insertablegroup-move-top').classList.add('disabled');
			}
			else {
				g.querySelector('.insertablegroup-move-up').classList.remove('disabled');
				g.querySelector('.insertablegroup-move-top').classList.remove('disabled');
			}

			if(x + 1 === glen) {
				g.querySelector('.insertablegroup-move-down').classList.add('disabled');
				g.querySelector('.insertablegroup-move-last').classList.add('disabled');
			}
			else {
				g.querySelector('.insertablegroup-move-down').classList.remove('disabled');
				g.querySelector('.insertablegroup-move-last').classList.remove('disabled');
			}

			x++;
		});
	}

	populateData(group, data, prefix) {
		Object.keys(data).forEach(key => {
			let keyName;

			if(prefix) {
				keyName = prefix + '/' + key;
			}
			else {
				keyName = key;
			}

			if(data[key] instanceof Object) {
				// Nesting!
				this.populateData(group, data[key], keyName);
			}
			else if(data[key] !== null) {
				if(typeof(this.inputs[ keyName ]) !== 'undefined') {
					let inputName = this.inputs[ keyName ];
					inputName = inputName.replace(
						new RegExp('__INSERTABLEGROUPTEMPLATE', 'g'),
						this.counter
					);

					group.querySelectorAll('[name="' + inputName + '"]').forEach(input => {
						try {
							input.value = data[ key ];
							// Notify that this input has been changed; just in case something cares.
							input.dispatchEvent(new Event('change', {bubbles: true, cancelable: true}));
						}
						catch (e) {
							console.log('Unable to set value on input, this is probably fine', input, e);
						}
					});
				}
			}
		});
	}
};

// @todo Finish porting away from jQuery
// Setup all the various default form elements, just for convenience.
$(function(){
	Core.PageEditor.metaauthor($('.formpagemetaauthorinput'));
	Core.PageEditor.seotitle($('.formtextinput-page-metas-title'));
});

(() => {
	document.querySelectorAll('fieldset.insertablegroup').forEach(el => {
		new Core.PageEditor.InsertableGroups(el);
	});
})();