<?php
/**
 * Main Controller parent for the 2.1 API version.
 *
 * @package Core
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Exception;
use PageModel;

class Controller {

	/**
	 * @var Request Request object that contains all parameters from the user agent
	 */
	public $request;

	/**
	 * The page model for the current page if set.
	 * 
	 * THIS MAY NOT BE SET!  If there is no associated PageModel registered for this page,
	 * this will be NULL.
	 *
	 * @var PageModel
	 */
	public $model;

	/**
	 * @var View View object that gets sent to the user agent on completion
	 */
	public $view;

	/**
	 * Shared access string for this controller.
	 *
	 * Optional, if set to non-null, it will be checked before any method is called.
	 *
	 * @var string
	 */
	public $accessstring = null;


	/**
	 * Get the page request for the current page.
	 *
	 * @return Request
	 */
	protected function getRequest() : Request {
		return $this->request;
	}
	
	/**
	 * Auto-run method for any method in this Controller.
	 * 
	 * If this throws an exception, the requested method is NOT called.
	 * 
	 * This is useful if you need to do common checks across all methods in your Controller
	 * before the method is allowed to be ran.
	 */
	public function __common() {
		
	}

	/**
	 * Set the page request for this page.  Only really useful in the Router
	 *
	 * @since 2.4.1
	 * @param Request $request
	 */
	public function setRequest(Request $request){
		$this->request = $request;
	}

	/**
	 * Internal function for setting the view object for this controller initially.
	 *
	 * Really only useful internally in the Router
	 *
	 * @param View $view
	 */
	public function setView(View $view){
		$this->view = $view;
	}

	/**
	 * Get the view for this controller.
	 * Up to the extending Controller to use this object is it wishes.
	 *
	 * @return View
	 */
	public function getView() : View {
		return $this->view;
	}

	/**
	 * Function that is called to get the controls to the current view.
	 *
	 * This function can either return an array of controls to be added, or just add them directly to the view.
	 *
	 * @return array|null
	 */
	public function getControls(){
		return null;
	}

	/**
	 * Replace this controller's view with a different one.
	 *
	 * This is useful for controllers that intercept a page request and replace their own content.
	 *
	 * @param View $newview
	 */
	protected function overwriteView($newview) {
		// Reset some of the new view's information.
		$newview->status = View::STATUS_OK;

		$this->_view = $newview;
	}

	/**
	 * Get the page model for the current page.
	 *
	 * @throws NotFoundException If there is no associated PageModel, (which is common)
	 * 
	 * @return PageModel
	 */
	public function getPageModel() : PageModel {
		if($this->model === null){
			throw new NotFoundException();
		}
		else{
			return $this->model;
		}
	}

	/**
	 * Set a JSON error message and optionally redirect if the page is not an ajax request.
	 *
	 * @param $code
	 * @param $message
	 * @param $redirect
	 *
	 * @return int
	 */
	public function sendJSONError($code, $message, $redirect){
		$view    = $this->getView();
		$request = $this->getPageRequest();

		if($request->isAjax()){
			$view->mode = View::MODE_PAGEORAJAX;
			$view->jsondata = ['status' => $code, 'message' => $message];
			$view->status = $code;
		}
		else{
			\Core\set_message($message, 'error');
			if($redirect){
				\Core\redirect($redirect);
			}
			else{
				\Core\go_back();
			}
		}
	}


	/**
	 * Set the access string for this view and do the access checks against the
	 * currently logged in user.
	 *
	 * Will also set the access string on the PageModel, since it needs to be reflected in the database.
	 *
	 * @since 2012.01
	 * @version 2.1
	 *
	 * @param string $accessstring
	 *
	 * @return boolean True or false based on access for current user.
	 */
	protected function setAccess($accessstring) {
		// Update the model
		//$this->getPageModel()->set('access', $accessstring);
		return (\Core\user()->checkAccess($accessstring));
	}

	/**
	 * Set the content of the view being returned.
	 *
	 * Important for JSON, XML, and other types.
	 *
	 * @param string $ctype
	 */
	protected function setContentType($ctype) {
		$this->getView()->contenttype = $ctype;
	}

	protected function setTemplate($template) {
		$this->getView()->templatename = $template;
	}
}
