<?php
/**
 * [PAGE DESCRIPTION HERE]
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Core\Utilities\Profiler\ProfilerEvent;

class HookHandler implements SingletonInterface {

	private static $RegisteredHooks = array();

	private static $Instance = null;

	private static $EarlyRegisteredHooks = array();

	private function __construct() {

	}

	public static function Singleton() {
		if (is_null(self::$Instance)) self::$Instance = new self();
		return self::$Instance;
	}

	/**
	 * Attach a call onto an existing hook.
	 *
	 * @param string       $hookName The name of the hook to bind to.
	 * @param string|array $callFunction The function to call.
	 *
	 * @return void
	 */
	public static function AttachToHook($hookName, $callFunction) {
		$hookName = strtolower($hookName); // Case insensitive will prevent errors later on.
		\Core\log_trace('Registering function ' . $callFunction . ' to hook ' . $hookName);
		//if(!isset(HookHandler::$RegisteredHooks[$hookName])) HookHandler::$RegisteredHooks[$hookName] = array();
		if (!isset(HookHandler::$RegisteredHooks[$hookName])) {

			// This hook registration may have happened before the hook is 
			// actually registered... throw this into a stack for later.
			if (!isset(self::$EarlyRegisteredHooks[$hookName])) self::$EarlyRegisteredHooks[$hookName] = array();
			self::$EarlyRegisteredHooks[$hookName][] = array('call' => $callFunction);

			return;
		}
		HookHandler::$RegisteredHooks[$hookName]->attach($callFunction);
	}

	/**
	 * Register a hook object with the global HookHandler object.
	 *
	 * Allows for abstract calling of the hook.
	 *
	 * @param Hook $hook
	 *
	 * @return void
	 */
	public static function RegisterHook(Hook $hook) {
		$name = $hook->getName();

		if(isset(HookHandler::$RegisteredHooks[$name])) {
			// Skip hooks that have already been registered,
			// will happen if the calling script is called multiple times
			// or during package upgrades
			return;
		}

		\Core\log_trace('Registering new hook [' . $name . ']');

		HookHandler::$RegisteredHooks[$name] = $hook;
		//var_dump(self::$EarlyRegisteredHooks);
		// Attach any bindings that may have existed.
		if (isset(self::$EarlyRegisteredHooks[$name])) {
			foreach (self::$EarlyRegisteredHooks[$name] as $b) {
				$hook->attach($b['call']);
			}

			unset(self::$EarlyRegisteredHooks[$name]);
		}
	}

	/**
	 * Shorthand for registering a new hook, usually dynamically within the code.
	 *
	 * @param string $name
	 * @param string $description
	 * @param string $return
	 */
	public static function RegisterNewHook(string $name, string $description = '', string $return = 'bool') {
		if(isset(HookHandler::$RegisteredHooks[$name])) {
			// Skip hooks that have already been registered, will happen if the calling script is called multiple times.
			return;
		}

		$hook = new Hook($name);

		if($description) {
			$hook->description = $description;
		}

		switch($return) {
			case 'bool':
			case 'void':
			case 'array':
			case 'string':
				$hook->returnType = $return;
				break;
			default:
				\Core\log_warning('Invalid return type requested for ' . $name . ', please ensure $return is one of [bool, void, array, string]');
				$hook->returnType = 'void';
		}

		HookHandler::RegisterHook($hook);
	}

	/**
	 * Dispatch an event, optionally passing 1 or more parameters.
	 *
	 * @param string  $hookName The name of the hook to dispatch
	 * @param mixed[] $args Some hooks support a variety of variables passed
	 *
	 * @return mixed
	 */
	public static function DispatchHook($hookName, ...$args) {
		// Prevent any hooks from being dispatched until Core is ready!
		if(!Core::GetComponent()) return null;

		$hookName = strtolower($hookName); // Case insensitive will prevent errors later on.
		
		if (!isset(HookHandler::$RegisteredHooks[$hookName])) {
			//\Core\log_notice('Tried to dispatch an undefined hook ' . $hookName);
			return null;
		}

		$hook   = HookHandler::$RegisteredHooks[$hookName];
		$result = call_user_func_array(array(&$hook, 'dispatch'), $args);

		return $result;
	}

	/**
	 * Simple function to return all hooks currently registered on the system.
	 *
	 * @return array
	 */
	public static function GetAllHooks() {
		return self::$RegisteredHooks;
	}

	/**
	 * Get the hook by name
	 *
	 * @param $hookname
	 *
	 * @return null|Hook
	 */
	public static function GetHook($hookname){
		return isset(self::$RegisteredHooks[$hookname]) ? self::$RegisteredHooks[$hookname] : null;
	}

	/**
	 * Just a simple debugging function to print out a list of the currently
	 * registered hooks on the system.
	 */
	public static function PrintHooks() {
		echo '<dl class="xdebug-var-dump">';
		foreach (self::$RegisteredHooks as $h) {
			echo '<dt>' . $h->name . '</dt>';
			if ($h->description) echo '<dd>' . $h->description . '</dd>';
			echo "<br/>\n";
		}
		echo '</dl>';
	}
}


/**
 * The actual hook object that will have the events attached to it.
 * Also allows for extra information
 *
 */
class Hook {

	/**
	 * Only a false return will trigger a failed hook deployment.
	 */
	const RETURN_TYPE_BOOL = 'bool';

	/**
	 * No return types are given any concern, and no feedback is available.
	 */
	const RETURN_TYPE_VOID = 'void';

	/**
	 * The return statuses from hooks are expected to be an array,
	 *
	 * Multiple calls are merged together in the overall result.
	 */
	const RETURN_TYPE_ARRAY = 'array';

	/**
	 * A non-empty return status from one hook will return immediately and stop execution of any other hook attachments.
	 *
	 * This does not constitute a failed hook, simply that it no longer needs to proceed, ie: useful for form submissions
	 * where the returning URL may be overwritten.  In this case, it's simply FIFO.
	 */
	const RETURN_TYPE_STRING = 'string';

	/**
	 * The name of this hook.  MUST be system unique.
	 * @var string
	 */
	public $name;

	/*
	 * Description of this hook, provides some human-friendly information about what it does and how it's used.
	 * @var string
	 */
	public $description;

	/**
	 * The return type of this hook, MUST be one of the RETURN_TYPE_* strings.
	 * @var string
	 */
	public $returnType = self::RETURN_TYPE_BOOL;

	/**
	 * An array of bound function/methods to call when this event is dispatched.
	 * @var array <<array>>
	 */
	private $_bindings = array();

	/**
	 * Instantiate a new generic hook object and register it with the global HookHandler.
	 *
	 * @param string $name
	 */
	public function __construct($name) {
		$this->name = $name;
		HookHandler::RegisterHook($this);
	}

	public function attach($function) {
		//echo "Binding event " . $function . " to " . $this->getName() . "<br/>";
		$this->_bindings[] = array('call' => $function);
	}

	/**
	 * Dispatch the event, calling any bound functions.
	 *
	 * @param mixed[] $args
	 *
	 * @return mixed
	 */
	public function dispatch(...$args) {
		$profiler = new ProfilerEvent('Hook Dispatch ' . $this->name);
		$profiler->start();
		
		// This sets up the original variable and the default return status.
		switch($this->returnType){
			case self::RETURN_TYPE_BOOL:
				// Default status for void, either no calls made or all returned successfully.
				$return = true;
				break;
			case self::RETURN_TYPE_ARRAY:
				$return = array();
				break;
			case self::RETURN_TYPE_VOID:
				$return = null;
				break;
			case self::RETURN_TYPE_STRING:
				$return = false;
				break;
		}

		foreach ($this->_bindings as $call) {
			// If the type is set to something and the call type is that
			// OR
			// The call type is not set/null.
			$result = $this->callBinding($call, ...$args);

			switch($this->returnType){
				case self::RETURN_TYPE_BOOL:
					// This will allow a hook to prevent continuation of a script.
					if ($result === false){
						$profiler->stop();
						return false;
					}
					break;
				case self::RETURN_TYPE_ARRAY:
					if(is_array($result)){
						$return = array_merge($return, $result);
					}
					break;
				case self::RETURN_TYPE_VOID:
					// I DON'T CARE!  :p
					break;
				case self::RETURN_TYPE_STRING:
					if(is_scalar($result) && $result != ''){
						// It's a non-empty string!
						$profiler->stop();
						return $result;
					}
					break;
			}

		}

		$profiler->stop();
		return $return;
	}

	/**
	 * Actually execute a binding call and get back the result.
	 *
	 * @param $call
	 * @param $args
	 *
	 * @return null|boolean|array
	 */
	public function callBinding($call, ...$args){
		// If the type is set to something and the call type is that
		// OR
		// The call type is not set/null.
		$profiler = new ProfilerEvent('Binding Call ' . $call['call']);
		$profiler->start();

		if(strpos($call['call'], '::') !== false){
			// Make sure this class exists first.
			// This way I can throw a pretty error instead of the default "Class not found".
			$parts = explode('::', $call['call']);
			if(!class_exists($parts[0])){
				trigger_error('The hook [' . $this->name . '] has an invalid call binding, the class [' . $parts[0] . '] does not appear to exist.', E_USER_NOTICE);
				$result = null;
			}
			else{
				$result = call_user_func_array($call['call'], $args);
			}
		}
		else{
			$result = call_user_func_array($call['call'], $args);
		}

		// If it's expecting an array, make sure it's an array!
		if($this->returnType == self::RETURN_TYPE_ARRAY && !is_array($result)){
			$result = array();
		}

		$profiler->stop();
		return $result;
	}

	public function __toString() {
		return $this->getName();
	}

	public function getName() {
		return strtolower($this->name);
	}

	public function getBindingCount(){
		return sizeof($this->_bindings);
	}

	/**
	 * Get the array of bindings currently attached to this hook
	 * @return array
	 */
	public function getBindings(){
		return $this->_bindings;
	}
}


HookHandler::Singleton();


// Create some pre-system hooks that are used before the component system is functional.
//HookHandler::RegisterNewHook('/core/db/ready', 'Called immediately after the database is available, but before any component is registered.');;
