<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core;

use Core\Datamodel\DatasetFrom;
use Core\Datamodel\DatasetSelect;
use ReflectionClass;
use Core\Datamodel\Dataset;
use Core\Datamodel\DatasetStream;
use Core\Datamodel\DatasetWhereClause;
use Core\Search\Helper as SearchHelper;
use Exception;

/**
 * Factory utility for models.
 *
 * This class provides an interface for searching for and counting models.  Generally, there are shortcut functions
 * available on the Model class that utilize this class, namely Count, Find, and FindRaw.
 */
class ModelFactory {

	/**
	 * Which DataModelInterface should this model execute its operations with.
	 * 99.9% of the time, it's fine to leave this as null, which will use the
	 * system DMI.  If however you want to utilize a Model with Memcache,
	 * (say for session information), it can be useful.
	 *
	 * @var DMI_Backend
	 */
	public $interface = null;

	/**
	 * What model is this a factory of?
	 * @var string
	 */
	private $_model;

	/**
	 * Contains the dataset object for this search.
	 *
	 * @var Dataset
	 */
	private $_dataset;

	/**
	 * @var DatasetStream The stream object used in getNext.
	 */
	private $_stream;
	
	private $_searchQuery;


	/**
	 * @param string $model
	 */
	public function __construct($model) {

		$this->_model = $model;

		$m              = $this->_model;
		$this->_dataset = new Dataset();
		$this->_dataset->table($m::GetTablename());
		$this->_dataset->select('*');
	}

	/**
	 * On clone, make a deep copy of this object!
	 *
	 * This is required so that the WHERE clause does not get copied by memory space.
	 * Otherwise altering the clone dataset will modify the original dataset!
	 */
	public function __clone() {
		if($this->_dataset){
			$this->_dataset = clone $this->_dataset;
		}

		if($this->_stream) {
			$this->_stream = clone $this->_stream;
		}
	}

	/**
	 * Where clause for the search, passed directly to the dataset object.
	 */
	public function where() {
		call_user_func_array([$this->_dataset, 'where'], func_get_args());
	}

	/**
	 * Perform a grouped WHERE clause for this lookup
	 */
	public function whereGroup() {
		call_user_func_array([$this->_dataset, 'whereGroup'], func_get_args());
	}

	/**
	 * Set the ordering for results
	 */
	public function order() {
		call_user_func_array([$this->_dataset, 'order'], func_get_args());
	}

	/**
	 * Limit the results
	 */
	public function limit() {
		call_user_func_array([$this->_dataset, 'limit'], func_get_args());
	}
	
	/**
	 * Perform a search instead of a direct lookup, allows for more flexible searches.
	 */
	public function search(string $query) {
		SearchHelper::AddSearchToDataset($query, $this->_dataset);
	}

	/**
	 * Get the result or results from this factory.
	 *
	 * If limit is set to 1, either a Model or null is returned.
	 * Else, an array of Models is returned, be it populated or empty.
	 *
	 * @return array|null|Model
	 */
	public function get() {
		$this->_performMultisiteCheck();
		$rs = $this->_dataset->execute($this->interface)->getResults();

		$ret = [];
		foreach ($rs as $row) {
			$model = new $this->_model();
			$model->_loadFromRecord($row);
			$ret[] = $model;
		}


		// Only return the model if "1" was requested as the limit.
		if ($this->_dataset->_limit == 1) {
			return (sizeof($ret)) ? $ret[0] : null;
		}
		else {
			return $ret;
		}
	}

	/**
	 * Get the results from this factory as a raw associative array.
	 *
	 * @return array|
	 */
	public function getRaw() {
		$this->_performMultisiteCheck();
		$rs = $this->_dataset->execute($this->interface);

		if($rs->_results !== null) {
			return $rs->_results;
		}
		elseif($rs->_data !== null) {
			return $rs->_data;
		}
		else {
			return [];
		}
	}

	/**
	 * Similar to get(), but only one model at a time is rendered and returned.
	 *
	 * This is ideal for large datasets and limited amounts of memory.
	 *
	 * When the end of the stream is hit, null is returned.
	 *
	 * @return Model|null
	 */
	public function getNext(){
		if($this->_stream === null){
			// Setup this stream object!
			$this->_performMultisiteCheck();
			$this->_stream = new DatasetStream($this->_dataset);
		}

		$next = $this->_stream->getRecord();

		if($next === null){
			// End of the stream has been reached!
			return null;
		}

		/** @var Model $model */
		$model = new $this->_model();
		$model->_loadFromRecord($next);

		return $model;
	}

	/**
	 * Memory-efficient generator wrapper when the functionality of getNext is useful,
	 * while needed to pass a simple array.
	 *
	 * @return \Generator|Model[]
	 */
	public function generator() : \Generator {
		$this->_performMultisiteCheck();
		$rs = $this->_dataset->execute($this->interface)->getResults();

		foreach ($rs as $row) {
			$model = new $this->_model();
			$model->_loadFromRecord($row);

			yield $model;
		}
	}

	/**
	 * Get a count of how many records are in this factory
	 * (without counting the records one by one)
	 *
	 * @return int
	 */
	public function count() {
		$this->_performMultisiteCheck();

		// Create a clone so it doesn't affect the original dataset.
		$clone = clone $this->_dataset;

		$selects = [];

		// We want to preserve any formulas here, they may be useful.
		foreach($clone->_selects as $select) {
			if($select->formula) {
				$selects[] = $select;
			}
		}

		try {
			if(count($selects)) {
				// This is a complex query, so use an embedded query instead.
				$clone->limit(null)->order(null);
				$countQuery = Dataset::Init()->count();
				$from = new DatasetFrom();
				$from->subquery = $clone;
				$from->alias = 'subquery';
				$countQuery->from($from);

				$ret = $countQuery->executeAndGet();
			}
			else {
				// Normal query, we can safely modify
				$clone
					->select(null)
					->count()
					->limit(null)
					->order(null);
				$ret = $clone->executeAndGet();
			}
		}
		catch(Exception $ex) {
			\Core\log_exception($ex);
			$ret = 0;
		}

		// Formula-based queries may have multiple columns, check for that.
		// This is required for search systems, as the logic to process the search is performed
		// on the database server, so that counts and pagination matches expected results.
		if(is_int($ret)) {
			return $ret;
		}
		elseif(is_array($ret)) {
			$count = 0;
			foreach($ret as $row) {
				$count += $row['COUNT'] ?? 0;
			}
			return $count;
		}
		else {
			\Core\log_error('Unsupported return from ModelFactory->count!');
			return 0;
		}
	}

	/**
	 * Get the raw dataset object for this factory.
	 * This can sometimes be useful for advanced manipulations of the low-level object.
	 *
	 * @since 2.1
	 * @return Dataset
	 */
	public function getDataset(){
		return $this->_dataset;
	}
	
	/**
	 * Helper method to return the name of the model this Factory is representing.
	 * 
	 * @return string
	 */
	public function getModelName(){
		return $this->_model;
	}

	/**
	 * Internal function to do the multisite check on the model.
	 * If the model supports a site attribute and none requested, then set it to the current site.
	 */
	private function _performMultisiteCheck(){
		
		$m = $this->_model;
		$ref = new ReflectionClass($m);
		$schema = $ref->getMethod('GetSchema')->invoke(null);

		// Is there a site property?  If not I don't even care.
		if(
			isset($schema->definitions['site']) &&
			$schema->definitions['site']->type === Model::ATT_TYPE_SITE &&
			Core::IsComponentAvailable('multisite') &&
			MultiSiteHelper::IsEnabled()
		){
			// I want to look it up because if the script actually set the site, then
			// it evidently wants it for a reason.
			$siteexact = (sizeof($this->_dataset->getWhereClause()->findByField('site')) > 0);
			$idexact = false;

			// The primary check will allow a model to be instantiated with the exact primary key string.
			$pri = isset($schema->indexes['primary']) ? $schema->indexes['primary'] : null;
			
			if($pri){
				$allids = true;
				foreach($pri as $k){
					if(sizeof($this->_dataset->getWhereClause()->findByField($k)) == 0){
						$allids = false;
						break;
					}
				}
				if($allids) $idexact = true;
			}

			if(!($siteexact || $idexact)){
				$w = new DatasetWhereClause();
				$w->setSeparator('or');
				$w->addWhere('site = ' . MultiSiteHelper::GetCurrentSiteID());
				$w->addWhere('site = -1');
				$this->_dataset->where($w);

				//$this->_dataset->where('site = ' . MultiSiteHelper::GetCurrentSiteID());
			}
		}
	}


	/**
	 * @since 2.4.0
	 * @param $model
	 *
	 * @return ModelSchema
	 */
	public static function GetSchema($model){
		$s = new ModelSchema($model);
		return $s;
	}
}