<?php
/**
 * The main object responsible for setting up the page request and getting the data corresponding to it.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @since 1.9
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Core\Filestore\FileHelper;
use PageModel;

class Request {
	
	/**
	 * @var string Delete the requested resource
	 */
	const METHOD_DELETE = 'DELETE';
	
	/**
	 * @var string Request method for standard GET request
	 */
	const METHOD_GET = 'GET';
	
	/**
	 * @var string -NOTSUPPORTED-
	 */
	const METHOD_HEAD = 'HEAD';
	
	/**
	 * @var string -NOTSUPPORTED-
	 */
	const METHOD_OPTIONS = 'OPTIONS';
	
	/**
	 * @var string Request method for standard form submission
	 */
	const METHOD_POST = 'POST';
	
	/**
	 * @var string Create a new object
	 */
	const METHOD_PUT = 'PUT';

	/**
	 * @var string Modify an object
	 */
	const METHOD_PATCH = 'PATCH';
	
	/**
	 * @var string -NOTSUPPORTED-
	 */
	const METHOD_TRACE = 'TRACE';

	/**
	 * @var array List of content types accepted by the browser
	 */
	public $contentTypes = [];

	/**
	 * @var array List of languages accepted by the browser
	 */
	public $acceptLanguages = [];

	/**
	 * @var string Request method used, one of the PageRequest::METHOD_* strings
	 */
	public $method;

	/**
	 * @var string Full string of the incoming user agent
	 */
	public $useragent;
	
	/**
	 * @var string Full string of the path + query string requested
	 */
	public $uriRaw;
	
	/**
	 * @var string URI with GET parameters and extension trimmed off
	 */
	public $uriBase;
	
	/**
	 * @var string Protocol of the requested connection, (HTTP/1.1 or HTTP/1.0 usually)
	 */
	public $protocol;

	/**
	 * @var array Array of the GET parameters on this request.
	 */
	public $parameters = [];

	/**
	 * Content type requested for the response based on the extension and environment
	 *
	 * This is **NOT** the Content-Type header that specifies the submission content type,
	 * but rather the requested response content type from the client!
	 *
	 * @var string
	 */
	public $ctype = View::CTYPE_HTML;

	/**
	 * The **submission DATA** content type, either application/json, application/x-www-form-urlencoded, etc.
	 *
	 * This specifies the REQUEST content type used in decoding the inbound request.
	 *
	 * @var string
	 */
	public $requestContentType;

	/**
	 * @var string Extension of the file requested; usually html, but could be anything.
	 */
	public $ext = 'html';
	
	/** @var string HTTP Referrer (aka HTTP_REFERER) of this page. */
	public $referrer;

	private $_body;

	public function __construct(string $uri = null) {
		$this->_resolveURI($uri);
		$this->protocol = EXEC_MODE === 'WEB' ?
			($_SERVER['SERVER_PROTOCOL'] ?? 'HTTP/1.0') :
			'CLI';

		if(isset($_SERVER['HTTP_REFERER'])) {
			$this->referrer = $_SERVER['HTTP_REFERER'] ?? '';
		}

		$this->_resolveRequestMethod();
		$this->_resolveAcceptHeader();
		$this->_resolveUAHeader();
		$this->_resolveLanguageHeader();
	}

	/**
	 * Check to see if the page request prefers a particular type of content type request.
	 * This is useful for allowing JSON requests on a per-case basis in the controller.
	 *
	 * @param string $type
	 *
	 * @return bool
	 */
	public function prefersContentType($type) {
		// First, find the current.
		$current     = 0;
		$currentmain = substr($this->ctype, 0, strpos($this->ctype, '/'));
		foreach ($this->contentTypes as $t) {
			if ($t['type'] == $this->ctype || ($t['type'] == $t['group'] . '/*' && $t['group'] == $currentmain)) {
				$current = max($current, $t['weight']);
			}
		}

		// Now that I have the current weight...
		$typeweight = 0;
		$typemain   = substr($type, 0, strpos($type, '/'));
		foreach ($this->contentTypes as $t) {
			if ($t['type'] == $type || ($t['type'] == $t['group'] . '/*' && $t['group'] == $typemain)) {
				$typeweight = max($typeweight, $t['weight']);
			}
		}

		// Now I have the weight values, (if any), of both current and requested.
		return ($typeweight > $current);
	}

	/**
	 * Shortcut function to return just the base url
	 *
	 * @return string
	 */
	public function getBaseURL() : string {
		return $this->uriBase;
	}

	/**
	 * Set all parameters for this view
	 * @param $params
	 */
	public function setParameters($params) {
		$this->parameters = $params;
	}

	/**
	 * Set a single parameter, useful for overriding.
	 *
	 * @param $key
	 * @param $value
	 */
	public function setParameter($key, $value){
		$this->parameters[$key] = $value;
	}

	/**
	 * Get all parameters from the GET variables.
	 *
	 * "Core" parameters are returned on a 0-based index, whereas named GET variables are returned with their respective name.
	 *
	 * @return array
	 */
	public function getParameters() {
		return $this->parameters;
	}

	/**
	 * Get a single parameter from the GET variables.
	 *
	 * @param $key string|int The parameter to request
	 *
	 * @return null|string|array
	 */
	public function getParameter($key) {
		return isset($this->parameters[ $key ]) ? $this->parameters[ $key ] : null;
	}

	/**
	 * Get a single parameter as a string from the GET/search string
	 *
	 * @param string $key
	 * @return string|null
	 */
	public function getParameterString(string $key) : ?string {
		return isset($this->parameters[ $key ]) ? $this->parameters[ $key ] : null;
	}

	/**
	 * Get a single parameter as an int from the GET/search string
	 *
	 * @param string $key
	 * @return int|null
	 */
	public function getParameterInt(string $key) : ?int {
		return isset($this->parameters[ $key ]) ? (int)$this->parameters[ $key ] : null;
	}

	/**
	 * Get a single parameter as a float from the GET/search string
	 *
	 * @param string $key
	 * @return float|null
	 */
	public function getParameterFloat(string $key) : ?float {
		return isset($this->parameters[ $key ]) ? (float)$this->parameters[ $key ] : null;
	}

	/**
	 * Check if this request has the given parameter
	 *
	 * @param $key string|int The parameter to request
	 *
	 * @return null|string
	 */
	public function hasParameter($key) : bool {
		return isset($this->parameters[ $key ]);
	}

	/**
	 * Just a shortcut function to make things consistent; returns a given POST variable.
	 * If the parameter does not exist, null is simply returned.
	 *
	 * It is still better to use the form system, as that has data sanitization and everything built in,
	 * but this allows a lower-level of access to the variables without resorting to raw access.
	 *
	 * @param $key string|null The POST variable to get
	 * @return mixed|null
	 */
	public function getPost($key = null){
		// Damn nested data.... :/
		$src = &$_POST;

		// Were all post parameters requested?
		// This isn't common, but can be useful in debugging.
		if($key === null){
			return $src;
		}
		elseif(strpos($key, '[') !== false){
			// User requested nested data, so look for it.
			$matches = [];
			preg_match_all('/\[([^\]]*)\]/', $key, $matches);

			// Build the array of map, starting with the first parent, (which will be omitted in the preg match).
			$keys = array_merge(
				[ substr($key, 0, strpos($key, '[')) ],
				$matches[1]
			);

			foreach($keys as $k) {
				if(isset($src[ $k ])) {
					$src =& $src[ $k ];
				}
				else {
					return null;
				}
			}

			// The last element will be the key the user requested.
			return $src;

			/*
			$k1 = substr($key, 0, strpos($key, '['));
			$key = substr($key, strlen($k1) + 1, -1);
			$src = &$_POST[$k1];
			*/
		}
		else {
			return (isset($src[$key])) ? $src[$key] : null;
		}
	}

	/**
	 * Retrieve a POST variable as a string
	 *
	 * Will return NULL if the key is not present
	 *
	 * @param string $key
	 * @return string|null
	 */
	public function getPostString(string $key) : ?string {
		$val = $this->getPost($key);

		return $val !== null ? (string)filter_var($val, FILTER_SANITIZE_STRING) : null;
	}

	/**
	 * Retrieve a POST variable as an int
	 *
	 * Will return NULL if the key is not present
	 *
	 * @param string $key
	 * @return string|null
	 */
	public function getPostInt(string $key) : ?int {
		$val = $this->getPost($key);

		return $val !== null ? (int)filter_var($val, FILTER_SANITIZE_NUMBER_INT) : null;
	}

	/**
	 * Shortcut for getting cookie
	 * 
	 * @param null|string $key
	 *
	 * @return null|string|array
	 */
	public function getCookie($key = null){
		if($key === null){
			// return all
			return $_COOKIE;
		}
		elseif(isset($_COOKIE[$key])){
			return $_COOKIE[$key];
		}
		else{
			return null;
		}
	}

	/**
	 * Get the raw data payload requested by the client
	 *
	 * If Content-Type is application/json, the data is parsed as a JSON string and the data is returned.
	 * Please note, in order for JSON requests to be
	 */
	public function getBody() {
		if($this->_body === null) {
			$raw = file_get_contents('php://input');

			if($this->requestContentType === View::CTYPE_JSON) {
				// JSON encoded strings get decoded automatically with PHP's built-in JSON decoder
				$this->_body = json_decode($raw, true);
			}
			elseif($this->requestContentType === 'application/x-www-form-urlencoded') {
				$data = [];
				// We just use PHP's built-in parser for URL encoded form data
				parse_str($raw, $data);
				$this->_body = $data;
			}
			else {
				// text/plain or other, it's just text.
				$this->_body = $raw;
			}

			// @todo Support multipart/form-data
		}

		return $this->_body;
	}

	/**
	 * Simple check to see if the page request is a POST method.
	 *
	 * Returns true if it is POST, false if anything else.
	 *
	 * @return bool
	 */
	public function isPost() : bool {
		return ($this->method === self::METHOD_POST);
	}

	/**
	 * Simple check to see if the page request is a GET method.
	 *
	 * Returns true if it is GET, false if anything else.
	 *
	 * @return bool
	 */
	public function isGet() : bool {
		return ($this->method === self::METHOD_GET);
	}

	public function isDelete() : bool {
		return ($this->method === self::METHOD_DELETE);
	}

	public function isPut() : bool {
		return ($this->method === self::METHOD_PUT);
	}

	public function isHead() : bool {
		return ($this->method === self::METHOD_HEAD);
	}
	
	public function isPatch() : bool {
		return ($this->method === self::METHOD_PATCH);
	}

	/**
	 * Simple check to see if the page request is a json content type.
	 *
	 * @return bool
	 */
	public function isJSON() : bool {
		return ($this->ctype === View::CTYPE_JSON);
	}

	/**
	 * Simple check to guess if the page request was an ajax-based request.
	 *
	 * @return bool
	 */
	public function isAjax() : bool {
		return (
			strtolower($_SERVER['HTTP_X_REQUESTED_WITH'] ?? 'standard') === 'xmlhttprequest'
		);
	}

	/**
	 * Get the user agent for this request.
	 *
	 * @return \Core\UserAgent
	 */
	public function getUserAgent(){
		return new \Core\UserAgent($this->useragent);
	}

	/**
	 * Get the referrer of this request, based on $_SERVER information.
	 *
	 * @return string
	 */
	public function getReferrer(){
		// Not available?.. Just return the root URL.
		return $this->referrer;
	}

	/**
	 * Get the user's preferred language set from either the browser of the LANG cookie.
	 * 
	 * This just returns the language portion, NOT the full string.
	 * 
	 * @return string
	 */
	public function getPreferredLanguage(){
		if(!is_array($this->acceptLanguages)){
			return 'en'; // ???
		}
		
		// The first is the preferred.
		$preferred = $this->acceptLanguages[0];
		// I just want the part before the '_'.
		// The string is in the format of "en_US"
		if(($key = strpos($preferred, '_')) !== false){
			return substr($preferred, 0, $key);
		}
		else{
			return $preferred;
		}
	}

	/**
	 * Get the user's preferred language+locale set from either the browser of the LANG cookie.
	 *
	 * @return string
	 */
	public function getPreferredLocale(){
		if(!is_array($this->acceptLanguages)){
			return 'en_US'; // ???
		}
		else{
			// The first is the preferred.
			return $this->acceptLanguages[0];	
		}
	}
	
	private function _resolveURI($uri){
		$this->uriRaw = $uri;
		
		// Use built-in functions whenever possible.
		$parts = parse_url($uri);
		
		$path = isset($parts['path']) ? $parts['path'] : '/';
		
		$this->parameters = [];
		if(isset($parts['query'])){
			parse_str($parts['query'], $this->parameters);
		}
		
		// Extract or guess the requested file extension and requested content type for the content
		// This has to do with the type of content requested, not the actual request itself!
		if(strpos($path, '.') !== false){
			$this->ext = pathinfo($path, PATHINFO_EXTENSION);

			if(strlen($this->ext)) {
				$path = substr($path, 0, -1-strlen($this->ext));
			}
		}
		elseif(EXEC_MODE === 'CLI'){
			$this->ext = 'txt';
		}
		else{
			$this->ext = 'html';
		}

		// The requested content type is based off the extension, as that's all we have to work with.
		$this->ctype = FileHelper::GuessMimeFromExt($this->ext);
		// Invalid mimetype?  Default to an HTML file.
		if(!$this->ctype){
			$this->ctype = 'text/html';
		}


		// Resolve the request content type based on either the header value or the extension when not available.
		if(isset($_SERVER['CONTENT_TYPE'])) {
			$this->requestContentType = filter_var($_SERVER['CONTENT_TYPE'], FILTER_SANITIZE_STRING);
		}

		
		// These default to the incoming path, and get resolved if necessary by the Router.
		$this->uriBase = $path;
	}

	/**
	 * Sets the method of this Request
	 */
	private function _resolveRequestMethod() {
		$method = isset($_SERVER['REQUEST_METHOD']) ? $_SERVER['REQUEST_METHOD'] : 'GET';
		// Make sure it's a valid METHOD... don't know what else it could be, but...
		switch ($method) {
			case self::METHOD_DELETE:
			case self::METHOD_GET:
			case self::METHOD_HEAD:
			case self::METHOD_POST:
			case self::METHOD_PATCH:
			case self::METHOD_PUT:
				$this->method = $method;
				break;
			default:
				$this->method = self::METHOD_GET;
		}
	}

	/**
	 * Sets the contentTypes of this Request
	 */
	private function _resolveAcceptHeader() {
		// I need to ensure there's at least a default.
		$header = $_SERVER['HTTP_ACCEPT'] ?? (EXEC_MODE === 'WEB' ? 'text/html' : 'text/plain');

		// As per the Accept HTTP 1.1 spec, all accepts MUST be separated with a comma.
		$header = explode(',', $header);

		// Clear the array
		$this->contentTypes = array();

		// There are a couple special-case exceptions that must go first.
		if ($this->ctype === View::CTYPE_JSON) {
			// JSON is dependent on either the config being true or an appropriate header.
			//if (ALLOW_NONXHR_JSON || $this->isAjax()) {
				$this->contentTypes[] = array(
					'type'   => View::CTYPE_JSON,
					'weight' => 1.0
				);
			//}
			//else {
				// DENIED :p
			//	$this->ctype = View::CTYPE_HTML;
			//}
		}

		// And set each one.
		foreach ($header as $h) {
			if (strpos($h, ';') === false) {
				$weight  = 1.0; // Do 1.0 to ensure it's parsed as a float and not an int.
				$content = $h;
			}
			else {
				list($content, $weight) = explode(';', $h);
				// Trim off the "q=" bit.
				$weight = floatval(substr($weight, 3));
			}

			$this->contentTypes[] = array(
				'type'   => $content,
				'weight' => $weight
			);
		}

		// And finally, run through all the content types and make them a little easier to parse.
		foreach ($this->contentTypes as $k => $v) {
			$this->contentTypes[$k]['group'] = substr($v['type'], 0, strpos($v['type'], '/'));
		}
	}

	/**
	 * Sets the acceptLanguages value of this Request
	 */
	private function _resolveLanguageHeader() {
		// I need to ensure there's at least a fallback.
		// This will prefer the browser's language set, but fallback to en if not set.
		$header = $_SERVER['HTTP_ACCEPT_LANGUAGE'] ?? (EXEC_MODE === 'CLI' ? $_SERVER['LANG'] ?? 'en' : 'en');
		
		// As per the Accept HTTP 1.1 spec, all accepts MUST be separated with a comma.
		$header = explode(',', $header);

		// Clear the array
		$this->acceptLanguages = array();
		$langs = [];

		// Set at least one default as a fallback!
		$langs['en'] = 0.0;

		// And set each one.
		foreach ($header as $h) {
			if (strpos($h, ';') === false) {
				$weight  = 1.0; // Do 1.0 to ensure it's parsed as a float and not an int.
				$content = $h;
			}
			else {
				list($content, $weight) = explode(';', $h);
				// Trim off the "q=" bit.
				$weight = floatval(substr($weight, 3));
			}

			$content = str_replace('-', '_', $content);

			$langs[$content] = $weight;
		}

		// If there is a LANG cookie set for this site, then prefer that over the browser preference.
		// This must be done after the header checks as it is meant to overwrite the headers with a higher precedence.
		if(filter_has_var(INPUT_COOKIE, 'LANG')){
			$langs[ filter_input(INPUT_COOKIE, 'LANG', FILTER_SANITIZE_STRING) ] = 2.0;
		}

		// Sort the languages by weight.
		arsort($langs);
		foreach($langs as $l => $w){
			// Ensure that it's a valid string before accepting the input!
			// Remember, this value does come from the browser which is an untrusted source.
			// This regex matches any two or three lower-case letters optionally followed by an underscore and two capital letters.
			if(preg_match('/^[a-z]{2,3}(_[A-Z]{2})?.*/', $l) === 1){
				$this->acceptLanguages[] = preg_replace('/^([a-z]{2,3}(_[A-Z]{2})?).*/', '\1', $l);
			}
		}
	}

	private function _resolveUAHeader() {
		$ua              = (isset($_SERVER['HTTP_USER_AGENT'])) ? $_SERVER['HTTP_USER_AGENT'] : '';
		$this->useragent = $ua;
	}
}
