<?php


namespace Core;


use Core\Forms\SystemInput;
use Core\ListingTable\Filter;

/**
 * Class RestApi provides helper support for handling REST requests and integration with Models
 *
 * Basic operation of the REST API helper system is:
 *
 * 1. Grab environment variables and setup filters
 * 2. Load Model data from database, optionally based on filters
 * 3. Execute the request
 * 4. Commit the changes to the database
 *
 * Each step can be overridden by extending this class to customize functionality
 * during each step.
 *
 * The default actions for each step:
 *
 * ## Construct
 *
 * Constructor doesn't change behaviour based on request type, as it's not available yet.
 * Instead, it sets all links as necessary to the request and calls setupFilters
 *
 * @se RestApi::_setupFilters
 *
 *
 * ## Load
 *
 * ### GET
 *
 * Link the Model or ModelFactory and populates Filter parameters if necessary
 *
 * @see RestApi::_loadGet
 * @see RestApi::_loadExistingModel
 * @see RestApi::_loadFactory
 * @see RestApi::_loadFiltersFromRequest
 *
 * ### PATCH
 *
 * Link an existing Model
 *
 * @see RestApi::_loadPatch
 * @see RestApi::_loadExistingModel
 *
 * ### PUT
 *
 * Create a new Model
 *
 * @see RestApi::_loadPut
 * @see RestApi::_loadNewModel
 *
 * ### DELETE
 *
 * Link an existing Model
 *
 * @see RestApi::_loadDelete
 * @see RestApi::_loadExistingModel
 *
 *
 * ## Execute
 *
 * ### GET
 *
 * Process the filters from the request when no single Model ID is requested
 *
 * @see RestApi::_executeGet
 * @see RestApi::_processFilters
 *
 * ### PATCH
 *
 * Loads the incoming data from the request and sets them on the existing Model
 *
 * @see RestApi::_executePatch
 * @see RestApi::_loadDataFromRequest
 *
 * ### PUT
 *
 * Loads the incoming data from the request and sets them on the new Model
 *
 * @see RestApi::_executePut
 * @see RestApi::_loadDataFromRequest
 *
 * ### DELETE
 *
 * No actions performed
 *
 * @see RestApi::_executeDelete
 *
 *
 * ## Commit
 *
 * ### GET
 *
 * No actions performed
 *
 * ### PATCH
 *
 * Saves the existing Model
 *
 * @see RestApi::_commitPatch
 * @see Model::save
 *
 * ### PUT
 *
 * Saves the new Model
 *
 * @see RestApi::_commitPut
 * @see Model::save
 *
 * ### DELETE
 *
 * Deletes the existing Model
 *
 * @see RestApi::_commitDelete
 * @see Model::delete
 *
 *
 * ## Results
 *
 * Provides a status code, message, and maybe either result or results based on:
 *
 * GET provides 'result' or 'results' based if a single or multiple records were requested
 * PUT and PATCH provides 'result' with the new Model data
 * DELETE has no results
 *
 * @see RestApi::_resultsPut
 * @see RestApi::_resultsDelete
 * @see RestApi::_resultsPatch
 * @see RestApi::_resultsGet
 *
 *
 * @package Core
 */
class RestApi {
	/**
	 * Inbound request where all the parameters are pulled
	 * @var Request
	 */
	protected $_request;
	/**
	 * Fully resolved name of the class to utilize for lookups and operations
	 * @var string
	 */
	protected $_modelName;
	/**
	 * Single Model instance, used in single GET, PATCH, PUT, and DELETE.
	 *
	 * @var Model
	 */
	protected $_model;
	/**
	 * Factory of Models for a given instance, used in multiple GET
	 *
	 * @var ModelFactory
	 */
	protected $_factory;

	/**
	 * List of filters available
	 * @var Filter[]
	 */
	protected $_filters = [];


	public function __construct(Request $request, string $modelClass) {
		$this->_request = $request;
		$this->_modelName = $modelClass;

		// Setup initial data for this REST API
		$this->_setupFilters();
	}

	/**
	 * Shortcut to handle all operations for this REST request in one go
	 *
	 * Calls load, execute, commit, and results
	 *
	 * @throws NotFoundException
	 * @throws AccessDeniedException
	 * @throws BadRequestException
	 * @throws MethodNotAllowedException
	 *
	 */
	public function process() : array {
		$this->load();
		$this->execute();
		$this->commit();
		return $this->results();
	}

	/**
	 * @throws MethodNotAllowedException
	 */
	public function load() {
		switch($this->_request->method) {
			case Request::METHOD_GET:
				$this->_loadGet();
				break;
			case Request::METHOD_PATCH:
				$this->_loadPatch();
				break;
			case Request::METHOD_DELETE:
				$this->_loadDelete();
				break;
			case Request::METHOD_PUT:
				$this->_loadPut();
				break;
			default:
				throw new MethodNotAllowedException();
		}
	}

	/**
	 * Run whatever VERB was requested.
	 *
	 * It's assumed that the Controller has verified that the user has access to perform this action by this stage!
	 *
	 * @throws MethodNotAllowedException
	 */
	public function execute() {
		switch($this->_request->method) {
			case Request::METHOD_GET:
				$this->_executeGet();
				break;
			case Request::METHOD_PATCH:
				$this->_executePatch();
				break;
			case Request::METHOD_DELETE:
				$this->_executeDelete();
				break;
			case Request::METHOD_PUT:
				$this->_executePut();
				break;
			default:
				throw new MethodNotAllowedException();
		}
	}

	/**
	 * Run whatever VERB was requested.
	 *
	 * It's assumed that the Controller has verified that the user has access to perform this action by this stage!
	 *
	 * @throws MethodNotAllowedException
	 */
	public function commit() {
		switch($this->_request->method) {
			case Request::METHOD_GET:
				// No action needed, GETs do not commit any data.
				break;
			case Request::METHOD_PATCH:
				$this->_commitPatch();
				break;
			case Request::METHOD_DELETE:
				$this->_commitDelete();
				break;
			case Request::METHOD_PUT:
				$this->_commitPut();
				break;
			default:
				throw new MethodNotAllowedException();
		}
	}

	/**
	 * @return array
	 *
	 * @throws MethodNotAllowedException
	 */
	public function results() : array {
		switch($this->_request->method) {
			case Request::METHOD_GET:
				return $this->_resultsGet();
			case Request::METHOD_PATCH:
				return $this->_resultsPatch();
			case Request::METHOD_DELETE:
				return $this->_resultsDelete();
			case Request::METHOD_PUT:
				return $this->_resultsPut();
			default:
				throw new MethodNotAllowedException();
		}
	}

	public function getModel() : ?Model {
		return $this->_model;
	}

	public function getFactory() : ?ModelFactory {
		return $this->_factory;
	}

	/**
	 * Get all filters registered on this Table
	 *
	 * @return Filter[]
	 */
	public function getFilters() : array {
		return $this->_filters;
	}

	/**
	 * Add a new filter for this listing table.
	 *
	 * @param string $type
	 * @param array $attributes
	 *
	 * @return Filter
	 */
	public function addFilter(string $type, array $attributes) : Filter {
		$filter = new Filter($type, $attributes);

		$this->_filters[ $filter->getName() ] = $filter;

		return $filter;
	}

	/**
	 * Setup all default filters for this Table
	 */
	protected function _setupFilters() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}

	/**
	 * Process all the attached filters from the Request.
	 */
	protected function _processFilters() {
		if(!$this->_factory) {
			throw new \Exception('Filters are only available with GET lookups');
		}

		$filters = $this->getFilters();
		$dataset = $this->_factory->getDataset();

		foreach($filters as $filter) {
			$filter->attachToDataset($dataset);
		}
	}

	protected function _loadGet() {
		if($this->_request->getParameter(0)) {
			// GET with a single parameter to be requested
			$this->_loadExistingModel();
		}
		else {
			// GET with multiple parameters, will return all of this type
			$this->_loadFiltersFromRequest();
			$this->_loadFactory();
		}
	}

	protected function _loadDelete() {
		// PATCH and DELETE both use an existing Model
		$this->_loadExistingModel();
	}

	protected function _loadPut() {
		// PUT uses a new Model
		$this->_loadNewModel();
	}

	protected function _loadPatch() {
		// PATCH and DELETE both use an existing Model
		$this->_loadExistingModel();
	}

	protected function _executeGet() {
		if($this->_factory) {
			// Factories are set when no specific item is requested, ie: a search.
			// Here, we use filters to process the search.
			$this->_processFilters();
		}
		// No additional action needed, single Model lookups should be retrieved from the initial loadExistingModel.
	}

	protected function _executeDelete() {
		// Nothing to do, the action is performed on commit.
	}

	protected function _executePut() {
		$this->_loadDataFromRequest();
	}

	protected function _executePatch() {
		$this->_loadDataFromRequest();
	}

	protected function _commitDelete() {
		$this->_model->delete();
	}

	protected function _commitPut() {
		$this->_model->save();
	}

	protected function _commitPatch() {
		$this->_model->save();
	}

	protected function _resultsGet() : array {
		if($this->_model) {
			return [
				'status' => 200,
				'message' => 'Performed GET successfully on ' . $this->_model->getLabel(),
				'result' => $this->_model->asArray(),
			];
		}
		else {
			$results = [];
			foreach($this->_factory->generator() as $model) {
				if($model->checkCanAccess()) {
					$results[] = $model->asArray();
				}
			}

			return [
				'status' => 200,
				'message' => 'Performed GET successfully',
				'results' => $results,
			];
		}
	}

	protected function _resultsDelete() : array {
		return [
			'status' => 200,
			'message' => 'Performed DELETE successfully',
		];
	}

	protected function _resultsPut() : array {
		return [
			'status' => 200,
			'message' => 'Performed PUT successfully on ' . $this->_model->getLabel(),
			'result' => $this->_model->asArray(),
		];
	}

	protected function _resultsPatch() : array {
		return [
			'status' => 200,
			'message' => 'Performed PATCH successfully on ' . $this->_model->getLabel(),
			'result' => $this->_model->asArray(),
		];
	}

	/**
	 * Load the filter parameters from the request
	 *
	 * Used by GET lookups with multiple objects requested
	 */
	protected function _loadFiltersFromRequest() {
		$filters = $this->getFilters();

		foreach($filters as $filter) {
			$filterName = $filter->getName();

			if($this->_request->hasParameter($filterName)) {
				$filter->setValue($this->_request->getParameter($filterName));
			}
		}
	}

	/**
	 * Load the actual data from the request into the Model
	 * @throws BadRequestException
	 */
	protected function _loadDataFromRequest() {
		// We use the Form subsystem to determine what values can be set
		$elements = $this->_model->asFormArray();

		// Pull the data from the request body
		$data = $this->_request->getBody();

		// Full sets check if required values are missing
		$isFull = $this->_request->isPut();

		foreach($elements as $key => $element) {
			if($element instanceof SystemInput) {
				// System inputs are not user editable
				continue;
			}

			if($element->get('access') && !\Core\user()->checkAccess($element->get('access'))) {
				// This element has an access string but the user cannot modify the value.
				continue;
			}

			if(!isset($data[$key])) {
				// Key not present, is this required?
				if($isFull && $element->get('required')) {
					throw new BadRequestException('Field ' . $element->get('title') . ' is required');
				}

				continue;
			}

			// Set the value on the Model
			try {
				$this->_model->set($key, $data[$key]);
			}
			catch(ModelValidationException $ex) {
				throw new BadRequestException($ex->getMessage());
			}
		}
	}

	protected function _loadFactory() {
		try {
			$this->_factory = new ModelFactory($this->_modelName);
		}
		catch(\Exception $ex) {
			throw new NotFoundException($this->_modelName . ' class does not exist');
		}
	}

	/**
	 * @throws NotFoundException
	 * @throws AccessDeniedException
	 */
	protected function _loadExistingModel() {
		try {
			$ref = new \ReflectionClass($this->_modelName);
			$this->_model = $ref->getMethod('Construct')->invoke(null, $this->_request->getParameter(0));
		}
		catch(\Exception $ex) {
			throw new NotFoundException($this->_modelName . ' class does not exist');
		}

		// We can also check if it exists and the user has access.
		if(!$this->_model->checkExists()) {
			throw new NotFoundException();
		}

		if(!$this->_model->checkCanAccess()) {
			throw new AccessDeniedException();
		}
	}

	/**
	 * @throws NotFoundException
	 */
	protected function _loadNewModel() {
		try {
			$ref = new \ReflectionClass($this->_modelName);
			$this->_model = $ref->newInstance();
		}
		catch(\Exception $ex) {
			throw new NotFoundException($this->_modelName . ' class does not exist');
		}
	}
}