<?php
/**
 * User Agent object, fancy alternative PHP's native get_browser function.
 *
 * @package    Core
 * @author     Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license    http://www.gnu.org/licenses/lgpl.html GNU Lesser General Public License
 * @link       http://php.net/manual/en/function.get-browser.php
 * @link       http://tempdownloads.browserscap.com/
 */

namespace Core;

use Core\Filestore\Contents\ContentGZ;
use Core\Filestore\Factory;
use Core\Renderer\ImageRenderer;
use DeviceDetector\DeviceDetector;
use Exception;

/**
 * User Agent object, fancy alternative PHP's native get_browser function.
 *
 * <h3>Usage Exaples</h3>
 *
 * <h4>Standard Usage</h4>
 * <p>Just instantiate a new UserAgent object and inspect it.</p>
 * <code>
 * $ua = new \Core\UserAgent();
 *
 * // Prints "Firefox", "Chrome", "Safari", etc.
 * echo $ua->browser;
 *
 * // If you're on a mobile device, this will print true.
 * print_r($ua->isMobile());
 * </code>
 *
 * <h4>Specify a User Agent</h4>
 * <p>A user agent can also be specified in the constructor.  This is useful for log scans or other lookups.</p>
 *
 * <code>
 * // In this example, the UA is a mobile android device, so...
 * $someuastring = 'Mozilla/5.0 (Android; Mobile; rv:18.0) Gecko/18.0 Firefox/18.0';
 * $ua = new \Core\UserAgent($someuastring);
 *
 * // true
 * print_r($ua->isMobile());
 *
 * // "Android"
 * print_r($ua->platform);
 * </code>
 *
 * @package  Core
 * @author   Charlie Powell <charlie@evalagency.com>
 * @link     http://php.net/manual/en/function.get-browser.php
 * @link     http://tempdownloads.browserscap.com/
 *
 *
 */
class UserAgent {
	/**
	 * @var string Full string of the useragent
	 */
	public $useragent = '';

	/**
	 * @var null|string The name of the user agent, ie: "Firefox", "Chrome", etc.
	 */
	public $browser = null;

	/**
	 * @var null|string Browser type
	 */
	public $browserType = null;

	/**
	 * @var null|string Full version of the user agent
	 */
	public $browserVersion = null;

	/**
	 * @var null|int Major version of the user agent
	 */
	public $browserVersionMajor = null;

	/**
	 * @var null|int Minor version of the user agent
	 */
	public $browserVersionMinor = null;

	/**
	 * @var null|string Operating System, (or platform), of the host.
	 * Example: "Linux", "Windows", "MacOSX", etc.
	 */
	public $platform = null;

	/**
	 * @var null|string Operating system version,
	 */
	public $platformVersion = null;

	/**
	 * @var null|string OS architecture, eg: x86, powerpc, risc, etc.
	 */
	public $platformArchitecture = null;

	/**
	 * @var null|string Device name or model number
	 */
	public $device = null;

	/**
	 * @var null|string Manufacturer
	 */
	public $deviceManufacturer = null;

	/**
	 * @var null|string Rendering engine, ie: "Gecko", "WebKit", "Trident", etc.
	 */
	public $engine = null;

	/**
	 * @var null|string Version of the rendering engine
	 */
	public $engineVersion = null;

	/**
	 * @var null|string Category of this bot
	 */
	public $botCategory = null;

	/**
	 * @var null|string Vendor or publisher of this bot
	 */
	public $botVendor = null;

	/**
	 * @var null|string Informative URL about this bot
	 */
	public $botInfoURL = null;

	private $_platformIcon = null;
	private $_browserIcon = null;


	/** @deprecated */
	public $parent                       = null;
	/** @deprecated */
	public $comment                      = null;
	/** @deprecated */
	public $browser_short_name           = null;
	/** @deprecated */
	public $version                      = 0.0;
	/** @deprecated */
	public $major_ver                    = 0;
	/** @deprecated */
	public $minor_ver                    = 0;
	/** @deprecated */
	public $platform_version             = null;
	/** @deprecated */
	public $platform_architecture        = null;
	/** @deprecated */
	public $platform_bits                = null;
	/** @deprecated */
	public $platform_short_name          = null;
	/** @deprecated */
	public $frames                       = false;
	/** @deprecated */
	public $iframes                      = false;
	/** @deprecated */
	public $tables                       = false;
	/** @deprecated */
	public $cookies                      = false;
	/** @deprecated */
	public $javascript                   = false;
	/** @deprecated */
	public $is_mobile_device             = false;
	/** @deprecated */
	public $css_version                  = 0;
	/** @deprecated */
	public $device_name                  = null;
	/** @deprecated */
	public $device_maker                 = null;
	/** @deprecated */
	public $rendering_engine_name        = null;
	/** @deprecated */
	public $rendering_engine_version     = null;
	/** @deprecated */
	public $platform_description         = null;
	/** @deprecated */
	public $alpha                        = false;
	/** @deprecated */
	public $beta                         = false;
	/** @deprecated */
	public $win16                        = false;
	/** @deprecated */
	public $win32                        = false;
	/** @deprecated */
	public $win64                        = false;
	/** @deprecated */
	public $vbscript                     = false;
	/** @deprecated */
	public $activex_controls             = false;
	/** @deprecated */
	public $is_syndication_reader        = false;
	/** @deprecated */
	public $crawler                      = false;
	/** @deprecated */
	public $aol_version                  = null;

	/**
	 * @var array Cache container for the Construct static method.
	 */
	protected static $_Cache = [];

	/**
	 * Constructor with a user agent string.
	 *
	 * If no string is provided, the current user's string is used.
	 *
	 * @param string|null $useragent
	 */
	public function __construct(string $useragent = null) {
		if($useragent === null){
			$this->useragent = $_SERVER['HTTP_USER_AGENT'] ?: '';
		}
		else{
			$this->useragent = $useragent;
		}

		if(function_exists('yaml_parse_file')) {
			$this->_loadFromPiwik();
		}
	}

	/**
	 * Guesses if this user request was a bot.
	 *
	 * @return boolean
	 */
	public function isBot() : bool {
		return $this->botCategory !== null;
	}

	/**
	 * Simple check if this is a mobile user agent
	 *
	 * @deprecated
	 * @return bool
	 */
	public function isMobile() {
		return false;
	}

	/**
	 * Get this user agent as an associative array.
	 *
	 * * useragent
	 * * browser_name
	 * * browser_type
	 * * browser_version
	 * * platform_name
	 * * platform_version
	 * * platform_architecture
	 * * device_name
	 * * device_manufacturer
	 *
	 * @return array
	 */
	public function asArray() : array {
		return [
			'useragent' => $this->useragent,
			'browser_name' => $this->browser,
			'browser_type' => $this->browserType,
			'browser_version' => $this->browserVersion,
			'platform_name' => $this->platform,
			'platform_version' => $this->platformVersion,
			'platform_architecture' => $this->platformArchitecture,
			'device_name' => $this->device,
			'device_manufacturer' => $this->deviceManufacturer,
		];
	}

	/**
	 * Get the info for this user agent as a pretty HTML string.
	 * 
	 * @return string HTML String
	 */
	public function getAsHTML() : string {
		$parts = [];
		$browser = $this->_getAsHTMLBrowser();
		$platform = $this->_getAsHTMLPlatform();
		$device = $this->_getAsHTMLDevice();
		
		if($browser){
			$parts[] = $browser;	
		}
		if($device){
			$parts[] = $device;
		}
		if($platform){
			$parts[] = $platform;
		}
		
		return implode('&nbsp;&nbsp;', $parts);
	}

	/**
	 * Get a pseudo-unique identifier for this user agent.
	 *
	 * This is useful for identifying a certain type of UA for caching reasons.
	 *
	 * @param bool $as_array
	 *
	 * @return array|string
	 */
	public function getPseudoIdentifier(bool $as_array = false) {
		$a = [];

		if($this->browser !== null) {
			$a[] = 'ua-browser-' . strtolower(str_replace(' ', '-', $this->browser));
		}

		if($this->engine !== null) {
			$a[] = 'ua-engine-' . strtolower($this->engine);
		}

		if($this->browserVersionMajor !== null) {
			$a[] = 'ua-browser-version-' . $this->browserVersionMajor;
		}

		if($this->platform !== null) {
			$a[] = 'ua-platform-' . strtolower(str_replace(' ', '-', $this->platform));
		}

		// Does the user want an array or a flat hash?
		if($as_array) {
			return $a;
		}
		else {
			return strtolower(implode(';', $a));
		}
	}

	public function getBrowserIcon() : ?string {
		if($this->_browserIcon === false) {
			// Already checked; doesn't exist
			return null;
		}

		$base = 'assets/images/apps/';

		if($this->browser === null) {
			$this->_browserIcon = false;
			return null;
		}

		$attempts = [];

		$browserName = strtolower(str_replace(' ', '-', $this->browser));
		$browserName = preg_replace('/[^a-z0-9\-]/', '', $browserName);

		if($this->browserVersionMajor !== null) {
			$browserVersion = strtolower(str_replace(' ', '-', $this->browserVersionMajor));
			$browserVersion = preg_replace('/[^a-z0-9]/', '', $browserVersion);

			$attempts[] = $base . $browserName . '_' . $browserVersion . '.svg';
			$attempts[] = $base . $browserName . '_' . $browserVersion . '.png';
		}

		$attempts[] = $base . $browserName . '.svg';
		$attempts[] = $base . $browserName . '.png';

		foreach($attempts as $attempt) {
			$file = Factory::File($attempt);
			if($file->exists()) {
				$this->_browserIcon = $file->getFilename(false);
				return $this->_browserIcon;
			}
		}

		// None found?
		$this->_browserIcon = false;
		return null;
	}

	public function getPlatformIcon() : ?string {
		if($this->_platformIcon === false) {
			// Already checked; doesn't exist
			return null;
		}

		$base = 'assets/images/os/';

		if($this->platform === null) {
			$this->_platformIcon = false;
			return null;
		}

		$attempts = [];

		$platformName = strtolower(str_replace(' ', '-', $this->platform));
		$platformName = preg_replace('/[^a-z0-9\-]/', '', $platformName);

		if($this->platformVersion !== null) {
			$platformVersion = strtolower(str_replace(' ', '-', $this->platformVersion));
			$platformVersion = preg_replace('/[^a-z0-9]/', '', $platformVersion);

			$attempts[] = $base . $platformName . '_' . $platformVersion . '.svg';
			$attempts[] = $base . $platformName . '_' . $platformVersion . '.png';
		}

		$attempts[] = $base . $platformName . '.svg';
		$attempts[] = $base . $platformName . '.png';

		foreach($attempts as $attempt) {
			$file = Factory::File($attempt);
			if($file->exists()) {
				$this->_platformIcon = $file->getFilename(false);
				return $this->_platformIcon;
			}
		}

		// None found?
		$this->_platformIcon = false;
		return null;
	}
	
	private function _loadFromPiwik() {
		// Piwik's DeviceDetector is available, use that instead! :)
		$dd = new DeviceDetector($this->useragent);
		$dd->setYamlParser(new PiwikYamlParser());
		$dd->parse();

		$client = $dd->getClient();
		$bot = $dd->getBot();
		$os = $dd->getOs();
		$device = $dd->getDevice();

		if($bot !== null) {
			$this->crawler = true;
			$this->botCategory = $bot['category'] ?: 'Other bot';
			$this->botInfoURL = $bot['url'] ?: null;
			if(isset($bot['producer'])) {
				$this->botVendor = $bot['producer']['name'] ?: null;
			}
			$this->browser = $bot['name'];
		}

		if($client !== null) {
			$this->browser = $client['name'];
			$this->browserType = $client['type'] ?? null;
			$this->browserVersion = $client['version'] ?? null;
			$this->version = $client['version'] ?? 0.0;
			$this->engine = $client['engine'] ?? null;
			$this->engineVersion = $client['engine_version'] ?? null;
		}

		if($os !== null) {
			$this->platform = $os['name'] ?? null;
			$this->platformVersion = $os['version'] ?? null;
			$this->platformArchitecture = $os['platform'] ?? null;
		}

		if($device !== null) {
			$this->deviceManufacturer = $dd->getBrandName() ?: null;
			$this->device = $dd->getDeviceName() ?: null;
		}


		if(strpos($this->useragent, 'Core Plus') !== false) {
			$this->browser = 'Core Plus';
			$this->browserVersion = preg_replace('#.*Core Plus ([0-9]+\.[0-9]+).*#', '$1', $this->useragent);
			$this->browserType = 'library';
		}

		$this->_fixVersion();
	}
	
	private function _fixVersion(){
		if($this->browserVersion !== null) {
			$this->browserVersionMajor = (int)substr($this->browserVersion, 0, strpos($this->browserVersion, '.'));
			$this->browserVersionMinor = (int)substr($this->browserVersion, strpos($this->browserVersion, '.')+1);
		}
	}

	/**
	 * Get the browser component of this useragent as pretty HTML.
	 * 
	 * @return string
	 */
	private function _getAsHTMLBrowser() : string {
		$icon = $this->getBrowserIcon();
		$title = '';
		$text  = '';
		$out   = '';
		$class = 'useragent-pretty-browser';
		
		// The text is just browser and version if set.
		$text .= $this->browser;
		if($this->browserVersion !== null) {
			$text .= ' ' . $this->browserVersion;
		}
		
		// The browser title should have the full UA attached to it.
		$title .= htmlentities($this->useragent);
		
		// Start compiling this attribute together.
		$out .= '<span class="' . $class . '" title="' . $title . '">';
		if($icon) {
			$iconImage = new ImageRenderer();
			$iconImage->setSource($icon);
			$iconImage->setDimensions('32x32');
			$out .= $iconImage->fetch();
		}
		$out .= $text;
		$out .= '</span>';
		
		return $out;
	}

	/**
	 * Get the platform/OS component of this useragent as pretty HTML.
	 *
	 * @return string
	 */
	private function _getAsHTMLPlatform() : string {
		$icon = $this->getPlatformIcon();
		$title = '';
		$text  = '';
		$out   = '';
		$class = 'useragent-pretty-platform';

		if($this->platform === null) {
			return '';
		}

		$title .= $this->platform;
		$text .= $this->platform;
		if($this->platformVersion !== null) {
			$title .= ' ' . $this->platformVersion;
			$text .= ' ' . $this->platformVersion;
		}
		if($this->platformArchitecture !== null) {
			$title .= ' (' . $this->platformArchitecture . ')';
		}

		// Start compiling this attribute together.
		$out .= '<span class="' . $class . '" title="' . $title . '">';
		if($icon) {
			$iconImage = new ImageRenderer();
			$iconImage->setSource($icon);
			$iconImage->setDimensions('32x32');
			$out .= $iconImage->fetch();
		}
		$out .= $text;
		$out .= '</span>';

		return $out;
	}

	/**
	 * Get the device component of this useragent as pretty HTML.
	 *
	 * @return string
	 */
	private function _getAsHTMLDevice() : string {
		// Piwik can provide pretty icons!
		$icon = null;
		//$icon  = (Core::IsComponentAvailable('piwik-analytics') && $this->device_maker) ? 'assets/images/brand/' . $this->device_maker . '.gif' : null;
		$title = '';
		$text  = '';
		$out   = '';
		$class = 'useragent-pretty-device';


		if($this->device === null && $this->deviceManufacturer === null) {
			return '';
		}
		elseif($this->device === 'desktop') {
			return '';
		}
		elseif($this->device !== null && $this->deviceManufacturer !== null) {
			$title .= $this->deviceManufacturer . ' ' . $this->device;
			$text .= $this->device;
		}
		elseif($this->deviceManufacturer !== null) {
			$title .= $this->deviceManufacturer;
			$text .= $this->deviceManufacturer;
		}
		else {
			$title .= $this->device;
			$text .= $this->device;
		}

		// Start compiling this attribute together.
		$out .= '<span class="' . $class . '" title="' . $title . '">';
		if($icon){
			$out .= '<img src="' . \Core\resolve_asset($icon) . '"/> ';
		}
		$out .= $text;
		$out .= '</span>';

		return $out;
	}

	/**
	 * Construct a UserAgent object with full caching enabled.
	 *
	 * @param string|null $useragent
	 *
	 * @return UserAgent
	 */
	public static function Construct(string $useragent = null) : UserAgent {
		if($useragent === null){
			$useragent = isset($_SERVER['HTTP_USER_AGENT']) ? $_SERVER['HTTP_USER_AGENT'] : '';
		}

		$cachekey = 'useragent-constructor-' . md5($useragent);
		$cache = Cache::Get($cachekey);
		if(!$cache){
			$cache = new UserAgent($useragent);
			Cache::Set($cachekey, $cache, SECONDS_ONE_WEEK);
		}

		return $cache;
	}
}
