<?php
/**
 * Provides all elements required to connect the Controller's data and logic back to the browser in the necessary format.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @since 1.9
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */
namespace Core;

use Core\CLI\CLI;
use Exception;
use Core\Templates\Template;
use Core\Utilities\Profiler\ProfilerEvent;

class View {

	/* Errors that may be occurred with views */
	const STATUS_OTHER                       = 1;
	const STATUS_OK                          = 200;  // Request OK
	const STATUS_CREATED                     = 201;  // Resource has been created
	const STATUS_ACCEPTED                    = 202;  // Request accepted for further processing
	const STATUS_NOCONTENT                   = 204;  // Request OK but no content returned
	const STATUS_RESETCONTENT                = 205;  // Request OK but no content returned, please reset the View
	const STATUS_MULTIPLECHOICES             = 300;  // Multiple formats available for the user agent to select from
	const STATUS_MOVEDPERMANENTLY            = 301;  // This and all future requested should use new URI
	const STATUS_FOUND                       = 302;  // Standard ambigious "temporary redirect"
	const STATUS_SEEOTHER                    = 303;  // POST/PUT/DELETE Data received; perform GET to see results
	const STATUS_NOTMODIFIED                 = 304;  // Not modified since last request; no need to redownload
	const STATUS_TEMPORARYREDIRECT           = 307;  // Resource temporarily moved; please resubmit POST/PUT/DELETE request there
	const STATUS_PERMANENTREDIRECT           = 308;  // Resource permanently moved; please resubmit POST/PUT/DELETE request there
	const STATUS_BADREQUEST                  = 400;  // Section 10.4.1: Bad Request
	const STATUS_UNAUTHORIZED                = 401;  // Section 10.4.2: Unauthorized
	const STATUS_PAYMENTREQUIRED             = 402;  // Section 10.4.3: Payment Required
	const STATUS_ACCESSDENIED                = 403;  // Section 10.4.4: Forbidden
	const STATUS_NOTFOUND                    = 404;  // Section 10.4.5: Not Found
	const STATUS_METHODNOTALLOWED            = 405;  // Section 10.4.6: Method Not Allowed
	const STATUS_NOTACCEPTABLE               = 406;  // Section 10.4.7: Not Acceptable
	const STATUS_PROXYAUTHENTICATIONREQUIRED = 407;  // Section 10.4.8: Proxy Authentication Required
	const STATUS_REQUESTTIMEOUT              = 408;  // Section 10.4.9: Request Time-out
	const STATUS_CONFLICT                    = 409;  // Section 10.4.10: Conflict
	const STATUS_GONE                        = 410;  // Section 10.4.11: Gone
	const STATUS_LENGTHREQUIRED              = 411;  // Section 10.4.12: Length Required
	const STATUS_PRECONDITIONFAILED          = 412;  // Section 10.4.13: Precondition Failed
	const STATUS_ENTITYTOOLARGE              = 413;  // Section 10.4.14: Request Entity Too Large
	const STATUS_URITOOLARGE                 = 414;  // Section 10.4.15: Request-URI Too Large
	const STATUS_UNSUPPORTEDMEDIATYPE        = 415;  // Section 10.4.16: Unsupported Media Type
	const STATUS_RANGENOTSATISFIABLE         = 416;  // Section 10.4.17: Requested range not satisfiable
	const STATUS_EXPECTATIONFAILED           = 417;  // Section 10.4.18: Expectation Failed
	const STATUS_TEAPOT                      = 418;  // I'm a fucking teapot motherfucker!
	const STATUS_ENHANCEYOURCALM             = 420;  // Too many requests in too short of time, chiiiiiiill
	const STATUS_ILLEGALYCENSORED            = 451;  // Content not allowed to be served due to legal reasons
	const STATUS_SERVERERROR                 = 500;  // Generic server error

	/*** Modes for handling the rendering of the view ***/

	/**
	 * Standard page inside the master page template
	 */
	const MODE_PAGE = 'page';
	/**
	 * Widget page inside a given widget container
	 */
	const MODE_WIDGET = 'widget';
	/**
	 * No automatic output whatsoever
	 * Useful for file downloads and completely custom views.
	 */
	const MODE_NOOUTPUT = 'nooutput';
	/**
	 * Render a template, but do not wrap in a master page template
	 * This is useful if you want the power of a template, but for an
	 * ajax or otherwise custom response.
	 */
	const MODE_AJAX = 'ajax';
	/**
	 * Detect if the page request is a standard or ajax, toggle between respectively
	 * If the page is loaded via an ajax request, no master template is used.
	 * otherwise, a standard page load will render with the master page template.
	 */
	const MODE_PAGEORAJAX = 'pageorajax';

	/**
	 * Allow the mode to be set to a streamlined version of PAGE.
	 * This does not include debugging information and the like.
	 */
	const MODE_EMAILORPRINT = 'print';

	/**
	 * "Streaming" content mode, supports some basic templating on the skin, but no page templates.
	 *
	 * Supports automatic resolution for WEB and CLI mode.
	 */
	const MODE_STREAM = 'stream';

	//const MODE_JSON       = 'json';

	/* Content types for this view */
	const CTYPE_HTML  = 'text/html';
	const CTYPE_PLAIN = 'text/plain';
	const CTYPE_JSON  = 'application/json';
	const CTYPE_XML   = 'application/xml';
	const CTYPE_ICS   = 'text/calendar';
	const CTYPE_RSS   = 'application/rss+xml';
	const CTYPE_CSV   = 'text/csv';


	public $status;
	private $_template;

	/**
	 * @var string Title for this View, (DOES support I18N via "t:" prefixes)
	 */
	public $title;

	/**
	 * The access string for this page.
	 * @var string
	 */
	public $access;

	/**
	 * The template to render this view with.
	 * Should be the partial path of the template, including /pages/
	 *
	 * @example /pages/mycomponent/view.tpl
	 * @var string
	 */
	public $templatename;

	/**
	 * The content type of this view.
	 * Generally set from the controller.
	 *
	 * This is sent to the browser if it's a page-type view as the Header: Content-Type field.
	 *
	 * @var string
	 */
	public $contenttype;


	/**
	 * The master template to render this view with.
	 * Should be just the filename, as it will be located automatically.
	 *
	 * @example index.tpl
	 * @var string
	 */
	public $mastertemplate;
	public $breadcrumbs = array();

	/**
	 * The controls for ths view
	 *
	 * @var ViewControls
	 */
	public $controls;

	/**
	 * The mode of this View.
	 * Greatly affects the rendering result, since this can be a full page or a single widget.
	 *
	 * MUST be one of the valid View::MODE_* strings!
	 *
	 * @var string
	 */
	public $mode;

	/**
	 * An array, object, string, or other data that is sent to the browser via json_encode if content type is set to JSON.
	 *
	 * @var mixed
	 */
	public $jsondata = null;

	/**
	 * Set this to a non-null value to set the http-equiv="last-modified" metatag.
	 *
	 * Also handles the Header: Last-Modified field.
	 *
	 * @var null|int
	 */
	public $updated = null;

	/**
	 * Any "other" string to put into the head.  This can include <link> tags, or any other tag not defined otherwise.
	 *
	 * @var array
	 */
	public $head = array();

	/**
	 * @var ViewMetas Handler for meta fields
	 */
	public $meta;

	/**
	 * Array of scripts to load in the head and foot of the document, respectively.
	 *
	 * @var array
	 */
	public $scripts = array('head' => array(), 'foot' => array());

	/**
	 * Array of stylesheets to load in the head of the document.
	 *
	 * @var array
	 */
	public $stylesheets = array();

	/**
	 * Set to true to allow the page template to run with errors.
	 *
	 * By default, all errors are caught and the system template overrides the page template. This is
	 * a security precaution to prevent the template from being rendered when access is denied.
	 *
	 * If HOWEVER that is the preferred behaviour, ie: user logins, set this to true to allow the page template
	 * to be used in rendering.
	 *
	 * @var bool
	 */
	public $allowerrors = false;

	/**
	 * Set to true to require this page to be viewed as SSL.
	 * Obviously if SSL is not enabled on this site, this has no effect.
	 *
	 * @var bool
	 */
	public $ssl = false;

	/**
	 * Set to false to skip this View from being recorded in analytical tools and navigation.  Useful for JSON or POST pages.
	 *
	 * @var bool
	 */
	public $record = true;

	/**
	 * Reference baseurl set by some systems, NOT USED INTERNALLY!
	 * 
	 * @var null|string
	 */
	public $baseurl = null;

	/**
	 * When fetching the body numerous times, the contents may be cached here to speed up fetchBody().
	 *
	 * @var null|string
	 */
	protected $_bodyCache = null;

	/**
	 * When fetching this entire skin+body numerous times, (ie: cache uses), the entire HTML may be cached here
	 * to be used by cache.
	 *
	 * @var null|string
	 */
	private $_fetchCache = null;

	/**
	 * An array of the body classes and their values.
	 *
	 * These automatically get appended in the <body> tag of the skin, providing the skin supports it.
	 *
	 * @var array
	 */
	public $bodyclasses = [];

	/**
	 * @var array Associative array of attributes for the <html> tag.
	 */
	public $htmlAttributes = [];

	/**
	 * @var array Associative array of custom headers to send to the browser automatically.
	 */
	public $headers = [];

	/**
	 * @var bool Set to false to prevent this page from being cacheable by Core.
	 */
	protected $cacheable = true;

	/**
	 * For widget and sub-page based views, they need to have a parent to render specific elements to,
	 * namely classes, styles, scripts, and the like.
	 *
	 * @var null|View
	 */
	public $parent = null;

	private $_footContent = [];

	private $_bodyContent = [];

	public function __construct() {
		$this->status = View::STATUS_OK;
		$this->mode  = View::MODE_PAGE;
		$this->controls = new ViewControls();
		$this->meta = new ViewMetas();
	}

	/**
	 * Get the template responsible for rendering this View's body content.
	 *
	 * Based on templatename.
	 * 
	 * @throws \Exception
	 *
	 * @return Template
	 */
	public function getTemplate() : Template {
		if (!$this->_template) {
			$this->_template = new Template();
		}

		return $this->_template;
	}

	/**
	 * Resolve template filenames for the View based on common parameters.
	 *
	 * @return string|null
	 */
	private function getTemplateName() : ?string {
		if($this->mode === View::MODE_NOOUTPUT) {
			// Pages that are set to have no output will *ALWAYS* have no template.
			// This is used by files that are sent to the useragent without templates.
			return null;
		}

		if($this->contenttype === View::CTYPE_JSON && $this->jsondata !== null) {
			// JSON pages that have JSON data set will not have any templates; the data is already set.
			return null;
		}

		if($this->templatename === false || $this->templatename === null) {
			// This View was configured to not use any templates, so don't return anything.
			return null;
		}

		return $this->templatename;
	}

	/**
	 * Retrieve the skin name for this View, as calculated from various conditions and settings
	 *
	 * @return string|null
	 */
	public function getSkinName() : ?string {
		if($this->mode === View::MODE_NOOUTPUT) {
			// Pages that are set to have no output will *ALWAYS* have no template.
			// This is used by files that are sent to the useragent without templates.
			return null;
		}

		if($this->mastertemplate === false || $this->mastertemplate === null) {
			// If no skin is set, then don't try to resolve anything.
			return null;
		}

		if($this->contenttype === View::CTYPE_JSON && $this->jsondata !== null) {
			// JSON pages that have JSON data set will not have any templates; the data is already set.
			return null;
		}

		// Block to handle default skins
		if($this->mastertemplate === 'admin'){
			// Configurable admin-level skin
			// This one uses the legacy configuration parameter
			return 'skins/' . ConfigHandler::Get('/theme/default_admin_template');
		}
		elseif($this->mastertemplate === 'default'){
			// Configurable default site skin
			// This one uses the legacy configuration parameter
			return 'skins/' . ConfigHandler::Get('/theme/default_template');
		}
		elseif($this->mastertemplate === 'email') {
			// Emails are no longer configurable, it was a silly option.
			// It is however determined based on the content type.
			// Here we support 3 modes, plain text, html, and none.
			switch($this->contenttype) {
				case View::CTYPE_PLAIN:
					return 'skins/_email.txt';
				case View::CTYPE_HTML:
					return 'skins/_email';
				default:
					return null;
			}
		}
		elseif($this->mastertemplate === 'blank') {
			// Another special term to refer to the _blank.[ext] template
			return 'skins/_blank';
		}
		elseif(strpos($this->mastertemplate, 'skins/') === false) {
			// Skin isn't resolved, try to map it to a common alias
			// This uses the expandable configuration parameter for any number of skin variations as per defined in components
			$c = ConfigHandler::Get('/theme/siteskin/' . $this->mastertemplate);
			if(!$c) {
				// Simply remap invalid aliases to the default skin.
				return 'skins/' . ConfigHandler::Get('/theme/default_template');
			}
			else {
				return 'skins/' . $c;
			}
		}
		else {
			return $this->mastertemplate;
		}
	}

	/**
	 * Get the title of this page, (with automatic i18n translation)
	 *
	 * @return string
	 */
	public function getTitle() : string {
		if(strpos($this->title, 't:') === 0){
			return t(substr($this->title, 2));
		}
		else{
			return $this->title === null ? '' : $this->title;
		}
	}

	/**
	 * Override a template, useful for forcing a different template type for this view.
	 *
	 * @param $template Core\Templates\TemplateInterface
	 * @return bool False on failure, True on success.
	 */
	public function overrideTemplate($template){
		if(!is_a($template, 'TemplateInterface')){
			return false;
		}

		if($template == $this->_template){
			return false;
		}

		if($this->_template !== null){
			foreach($this->_template->getTemplateVars() as $k => $v){
				$template->assign($k, $v);
			}
		}

		$this->_template = $template;
		return true;
	}

	/**
	 * Assign a variable to this view
	 *
	 * @param $key string
	 * @param $val mixed
	 */
	public function assign($key, $val) {
		$this->getTemplate()->assign($key, $val);
	}

	/**
	 * Alias of assign
	 *
	 * @param $key string
	 * @param $val mixed
	 */
	public function assignVariable($key, $val) {
		$this->assign($key, $val);
	}

	/**
	 * Get a variable that was set with "assign()"
	 *
	 * @param string $key
	 *
	 * @return mixed
	 */
	public function getVariable($key) {
		// Damn smarty and its being more difficult...
		try {
			return $this->getTemplate()->getVariable($key);
		}
		catch(Exception $ex) {
			return null;
		}
	}

	public function fetchBody() : ?string {

		if($this->_bodyCache !== null) {
			return $this->_bodyCache;
		}
		
		// If there is set to be no system content, don't even bother with anything here!
		if ($this->mode === View::MODE_NOOUTPUT) {
			return null;
		}

		$template = $this->getTemplateName();
		

		if($this->contenttype === View::CTYPE_JSON && $this->jsondata !== null){
			// If this View contains JSON data and is set to display as such...

			// Allow any last minute script to run something
			// This is done before the JSON data is encoded because some scripts may need to alter the data therein.
			HookHandler::DispatchHook('/core/page/rendering/json', $this);

			$this->_bodyCache = json_encode($this->jsondata);
			return $this->_bodyCache;
		}
		elseif($this->contenttype === View::CTYPE_JSON && $this->jsondata === null && $template !== null) {
			// The View was requested to be JSON data and a template was requested, but no JSON data set
			// Here we'll create some automatic data for the view.
			// This is a bit of a patch to allow traditional page snippets to be retrieved as JSON data.

			$data = [
				'status'  => $this->status,
				'message' => 'Content Retrieved',
				'body'    => $this->getTemplate()->fetch($template),
				'scripts' => [],
				'scripts_inline' => [],
			];

			// Allow any last minute script to run something
			HookHandler::DispatchHook('/core/page/rendering', $this);

			foreach(array_merge($this->scripts['head'], $this->scripts['foot']) as $script) {
				$match = [];
				preg_match('#<script [^>]*src=["\']([a-zA-Z0-9.:/~&_=\-?]*)["\'].*#', $script, $match);
				if(count($match) === 2) {
					$data['scripts'][] = $match[1];
					continue;
				}

				preg_match('#<script[^>]*>(.*)</script>#s', $script, $match);
				if(count($match) === 2) {
					$data['scripts_inline'][] = $match[1];
					continue;
				}
			}

			foreach($this->_footContent as $content) {
				$data['body'] .= $content;
			}

			// Store the rendered data into the JSON object so any supplemental scripts know about it.
			$this->jsondata = $data;

			// Allow JSON-specific scripts to tap into this too.
			// This is used in conjunction with the standard rendering above, as the source content
			// is derived from a template, and additional scripts may utilize functionality for the base HTML content.
			HookHandler::DispatchHook('/core/page/rendering/json', $this);

			$this->_bodyCache = json_encode($this->jsondata);
			return $this->_bodyCache;
		}
		elseif($template) {
			$t = $this->getTemplate();
			$html = $t->fetch($template);

			$this->_bodyCache = $html;
			return $html;
		}
		else {
			// No template available, nothing to render.
			return null;
		}
	}

	/**
	 * Fetch this view as an HTML string.
	 * @return mixed|null|string
	 */
	public function fetch() {

		if($this->_fetchCache !== null){
			// w00t ;)
			return $this->_fetchCache;
		}
		
		$profiler = new ProfilerEvent('Core Application View Fetch - ' . $this->templatename);
		$profiler->start();

		$body = $this->fetchBody();
		$profiler->stop();
		
		if(EXEC_MODE === 'CLI') {
			// At this moment, only a single CLI skin is supported.
			// There's one set, but it's probably the web skin, so we'll remap it now.
			$skinFile = Template::ResolveFile('skins/_generic.txt');
		}
		elseif(($skin = $this->getSkinName()) !== null) {
			try{
				$skinFile = Template::ResolveFile($skin);
			}
			catch (Exception $ex) {
				// Fallback to the Core generic skin
				$type = EXEC_MODE === 'WEB' ? '.html' : '.txt';
				$skinFile = Template::ResolveFile('skins/_generic' . $type);
			}
		}
		else {
			$skinFile = null;
		}

		if(!$skinFile) {
			// No skin template, nothing to do.
			return $body;
		}

		$template = Template::Factory($skinFile);

		
		if($this->mode === View::MODE_PAGE) {
			$template->assign('messages', Core::GetMessages());

			// Last chance for scripts to jump in!
			HookHandler::DispatchHook('/core/page/rendering', $this);
		}
		elseif($this->mode === View::MODE_EMAILORPRINT) {
			// Last chance for scripts to jump in!
			HookHandler::DispatchHook('/core/page/rendering/email', $this);
		}

		$template->assign('body_classes', strtolower(implode(' ', $this->bodyclasses)));
		$template->assign('body', $body);
		$template->assign('title', $this->getTitle());
		$template->assign('breadcrumbs', $this->getBreadcrumbs());
		$template->assign('controls', $this->controls);
		//$template->assign('messages', Core::GetMessages());
		$template->assign('footer_legal', Core::GetLegalFooterContent());
		// This logic is needed for the SEO title, since that's usually completely human unfriendly.
		$this->meta['title']->setTitle($this->getTitle());
		$template->assign('seotitle', $this->meta['title']);
		
		// Attach any theme config options; as they are expected to be used along with the skin!
		foreach(ConfigHandler::GetAll('/theme/*') as $key => $val){
			$template->assign( str_replace('/', '_', substr($key, 1)), $val);
		}

		$data = $template->fetch();


		
		// Replace the </head> tag with the head data from the current page
		// and the </body> with the foot data from the current page.
		// This is needed to be done at this stage because some element in
		// the template after rendering may add additional script to the head.
		// Also tack on any attributes for the <html> tag.
		if(
			($this->contenttype === View::CTYPE_HTML && $this->mode === View::MODE_PAGE) ||
			($this->contenttype === View::CTYPE_HTML && $this->mode === View::MODE_EMAILORPRINT)
		) {
			$data = preg_replace('#<html#', '<html ' . $this->getHTMLAttributes(), $data, 1);
			
			if(preg_match('#<head>#i', $data)){
				// I need to do preg_replace because I only want to replace the FIRST instance of <head>
				$data = preg_replace('#<head>#i', "<head>\n" . $this->_getHeadContent(), $data, 1);
			}

			if(preg_match('#<body[^>]*>#i', $data)) {
				$data = preg_replace('#(<body[^>]*>)#i', '$1' . "\n" . $this->_getBodyContent(), $data);
			}
			
			if(preg_match('#</body>#i', $data)){
				// I need to use strrpos because I only want the LAST instance of </body>
				$match = strrpos($data, '</body>');

				$foot = $this->_getFootContent();

				$data = substr_replace($data, $foot . "\n" . '</body>', $match, 7);
			}
		}
		
		$this->_fetchCache = $data;
		return $data;
	}

	/**
	 * Render this view and send all appropriate headers to the browser, (if applicable)
	 *
	 * @return void
	 */
	public function render() {
		
		// Redirect statuses should execute immediately and not try to call any templates.
		if(
			$this->status === View::STATUS_MOVEDPERMANENTLY ||
			$this->status === View::STATUS_FOUND ||
			$this->status === View::STATUS_SEEOTHER ||
			$this->status === View::STATUS_TEMPORARYREDIRECT ||
			$this->status === View::STATUS_PERMANENTREDIRECT
		){
			$this->sendHeaders();
			echo '<html><body><a href="' . $this->headers['Location'] . '">Please click here to continue.</a></body></html>';
			return;
		}
		
		if(
			$this->status === View::STATUS_NOTMODIFIED
		){
			$this->sendHeaders();
			echo '<html><body>Resource not modified.</body></html>';
			return;
		}

		// Before I go about fetching anything, enable UTF-8 to ensure proper i18n!
		if ($this->contenttype && $this->contenttype === View::CTYPE_HTML) {
			$this->addMeta('http-equiv="Content-Type" content="text/html;charset=UTF-8"');
		}


		// Streaming content will post directly to the page and skip the Router,
		// so we need to check that here and adjust the skin if necessary.
		if($this->mode === View::MODE_STREAM) {
			$this->mastertemplate = 'skins/_stream' . (EXEC_MODE === 'WEB' ? '.html' : '.txt');
			$this->templatename = null;
		}


		$data = $this->fetch();

		// Be sure to send the content type and status to the browser, (if it's a page)
		// This is done after document fetching to allow supplemental scripts to add headers
		// during execution, but before rendering.
		$this->sendHeaders();


		if(SSL_MODE != SSL_MODE_DISABLED){
			// If SSL is required by the controller and it's available, redirect there!
			// This could also be set by something during document fetching.
			if($this->ssl && !SSL){
				$u = ROOT_URL_SSL . substr(REL_REQUEST_PATH, 1);
				
				// Smarty likes to be over zealous and send headers prematurely...
				if(!headers_sent()){
					header('Location: ' . $u );	
				}
				die('<html><body onload="window.location = \'' . $u . '\'" >This page requires SSL, please <a href="' . $u . '">Click Here to continue</a>.</body></html>');
			}
			// If SSL is set to be ondemand and the page does not have it set but it's enabled, redirect to the non-SSL version.
			elseif(!$this->ssl && SSL && SSL_MODE == SSL_MODE_ONDEMAND){
				$u = ROOT_URL_NOSSL . substr(REL_REQUEST_PATH, 1);

				// Smarty likes to be over zealous and send headers prematurely...
				if(!headers_sent()){
					header('Location: ' . $u );
				}
				die('<html><body onload="window.location = \'' . $u . '\'" >This page does not require SSL, please <a href="' . $u . '">Click Here to continue</a>.</body></html>');
			}
			// Else, SSL_MODE_ALLOWED doesn't care if SSL is enabled or not!
		}

		//echo mb_convert_encoding($data, 'UTF-8', 'auto');
		//echo mb_convert_encoding($data, 'HTML-ENTITIES', 'auto');
		if(EXEC_MODE === 'CLI' && $this->mode === View::MODE_STREAM) {
			// Streaming output on the CLI means the specific application is configured to expect this.
			CLI::Clear();
		}

		echo $data;
	}

	/**
	 * Add a breadcrumb to the end of the breadcrumb stack.
	 *
	 * @param      $title
	 * @param null $link
	 */
	public function addBreadcrumb($title, $link = null) {
		
		if($link && strpos($link, '://') === false) {
			// Allow a non-resolved link to be passed in.
			$map = new BaseURLMap($link);
			$baseurl = $map->getBaseURL();
			$link = \Core\resolve_link($baseurl);
		}
		else{
			$baseurl = null;
		}
		
		// New support for i18n strings!
		if(strpos($title, 't:') === 0){
			$title = t(substr($title, 2));
		}

		$this->breadcrumbs[] = array(
			'title'   => $title,
			'link'    => $link,
			'baseurl' => $baseurl,
		);
	}

	/**
	 * Add automatic breadcrumbs for this page based on the baseURL of the target page
	 *
	 * If sub-page C is being displayed, the parentURL of that page should be requested here.
	 * If that page and any parentURL page thereof exist, the breadcrumbs will be added accordingly.
	 *
	 * **WARNING** this ONLY WORKS WITH REGISTERED PAGES!
	 *
	 * This method will do nothing if called with an invalid, non-existent, or dynamic URL.
	 * 
	 * @param string $link
	 */
	public function autoBreadcrumbs(string $link) {
		try {
			$page = \PageModel::GetByURI($link);

			if($page->checkExists()) {
				$linkTree = $page->getParentTree();

				foreach($linkTree as $parentPage) {
					$this->addBreadcrumb($parentPage->getTitle(), $parentPage->getResolvedURL());
				}

				$this->addBreadcrumb($page->getTitle(), $page->getResolvedURL());
			}
		}
		catch(Exception $e) {
			\Core\log_exception($e);
		}
	}

	/**
	 * Override and replace the breadcrumbs with an array.
	 *
	 * @param $array
	 */
	public function setBreadcrumbs($array) {
		// Array should be an array of either link => title keys or pages.
		$this->breadcrumbs = array();

		// If null is passed in, just leave them blank.
		// This is useful for implementing completely custom breadcrumbs.
		if (!$array) return;

		foreach ($array as $k => $v) {
			if ($v instanceof PageModel) $this->addBreadcrumb($v->get('title'), $v->getResolvedURL());
			else $this->addBreadcrumb($v, $k);
		}
	}

	/**
	 * Remove and reset any breadcrumbs set
	 *
	 */
	public function resetBreadcrumbs() {
		$this->breadcrumbs = [];
	}

	/**
	 * Append a key onto the JSON payload to be sent back to the browser.
	 *
	 * Will automatically set the View MODE and View CONTENT_TYPE values for JSON/AJAX
	 *
	 * @param string $key  Data key
	 * @param mixed $value Data payload
	 */
	public function appendJSONDataKey(string $key, $value) {
		if(!is_array($this->jsondata)) {
			$this->jsondata = [];
		}

		$this->jsondata[ $key ] = $value;

		$this->mode = self::MODE_AJAX;
		$this->contenttype = self::CTYPE_JSON;
	}

	/**
	 * Append an array of data onto the JSON payload to be sent back to the browser.
	 *
	 * Will automatically set the View MODE and View CONTENT_TYPE values for JSON/AJAX
	 * @param array $payload Key/Value pair of data
	 */
	public function appendJSONData(array $payload) {
		if(!is_array($this->jsondata)) {
			$this->jsondata = [];
		}

		$this->jsondata = array_merge($this->jsondata, $payload);

		$this->mode = self::MODE_AJAX;
		$this->contenttype = self::CTYPE_JSON;
	}

	public function setJSONSuccessMessage(string $message) {
		$this->appendJSONData([
			'status' => 200,
			'message' => $message,
		]);
	}

	public function setJSONErrorMessage(string $message, int $code = 400) {
		$this->appendJSONData([
			'status' => $code,
			'message' => $message,
		]);
	}

	/**
	 * Get this view's breadcrumbs as an array
	 *
	 * @return array
	 */
	public function getBreadcrumbs() {
		$crumbs = $this->breadcrumbs;
		if ($this->title){
			$crumbs[] = [
				'title'   => $this->getTitle(),
				'link'    => null,
				'baseurl' => null,
			];
		}

		// Remove duplicates
		// This can happen when the developer sets a breadcrumb manually AND the system adds the same breadcrumb,
		// (since breadcrumbs can be managed automatically by Core).
		$seen = [];
		foreach($crumbs as $k => $dat){
			if(in_array($dat['link'], $seen)){
				unset($crumbs[$k]);
			}
			else{
				$seen[] = $dat['link'];
				
				// Also ensure that the title for this entry has i18n translations.
				if(substr($dat['title'], 0, 2) == 't:'){
					$crumbs[$k]['title'] = t(substr($dat['title'], 2));
				}
			}
		}

		return array_values($crumbs);
	}
	
	/**
	 * Add a control into the page template.
	 *
	 * Useful for embedding functions and administrative utilities inline without having to adjust the
	 * application template.
	 *
	 * @param string|array|Model|ViewControl $title  The title to set for this control
	 * @param string             $link   The link to set for this control
	 * @param string|array       $class  The class name or array of attributes to set on this control
	 *                            If this is an array, it should be an associative array for the advanced parameters
	 */
	public function addControl($title, $link = null, $class = 'edit') {

		if($title instanceof Model){
			// Allow a raw Model to be sent in as the control subject.
			// This is a shortcut for Controllers much like the {controls} smarty function has.
			$this->controls = ViewControls::DispatchModel($title);
		}
		elseif($title instanceof ViewControl) {
			// Support overloading
			$this->controls[] = $title;
		}
		elseif(is_array($title)) {
			trigger_error('Passing an array into addControl has been deprecated, please use the ViewControl::Factory instead', E_USER_DEPRECATED);
			$this->controls[] = ViewControl::Factory($title);
		}
		else {
			trigger_error('Passing a string into addControl has been deprecated, please use the ViewControl::Factory instead', E_USER_DEPRECATED);
			$this->addControlTLI($title, $link, $class);
		}
	}

	/**
	 * Add a control to this view via its title, link, and icon parameters.
	 *
	 * This provides support for Core-2018 and prior.
	 *
	 * @param string $title
	 * @param string $link
	 * @param string|null $icon
	 */
	public function addControlTLI(string $title, string $link, string $icon = null) {
		$control = new ViewControl();

		// Completely associative-array based version!
		if(func_num_args() == 1 && is_array($title)){
			foreach($title as $k => $v){
				$control->set($k, $v);
			}
		}
		else{
			// Advanced method, allow for associative arrays.
			if(is_array($icon)){
				foreach($icon as $k => $v){
					$control->set($k, $v);
				}
			}
			// Default method; just a string for the class name.
			else{
				$control->icon = $icon;
			}

			$control->title = $title;
			$control->link = \Core\resolve_link($link);
		}

		$this->controls[] = $control;
	}

	/**
	 * Add an array of controls at once, useful in conjunction with the model->getControlLinks method.
	 *
	 * If a Model is provided as the subject, that is used as the subject and all system hooks apply thereof.
	 *
	 * @param array|Model $controls
	 */
	public function addControls($controls){
		if($controls instanceof Model){
			// Allow a raw Model to be sent in as the control subject.
			// This is a shortcut for Controllers much like the {controls} smarty function has.
			$this->controls = ViewControls::DispatchModel($controls);
			return;
		}

		foreach($controls as $c){
			if(is_array($c)) {
				$this->addControl(ViewControl::Factory($c));
			}
			else {
				$this->addControl($c);
			}
		}
	}

	/**
	 * Set the access string for this view and do the access checks against the
	 * currently logged in user.
	 *
	 * If the user does not have access to the resource, $this->status is set to 403.
	 *
	 * (if you only want to set the access string, please just use $view->access = 'your_string';)
	 *
	 * @since 2011.08
	 *
	 * @param string $accessstring
	 *
	 * @return boolean True or false based on access for current user.
	 */
	public function setAccess($accessstring) {
		$this->access = $accessstring;

		return $this->checkAccess();
	}
	
	/**
	 * Set (overwrite) the cached body for this View.
	 * 
	 * Generally this is **NOT** needed, but can be useful in certain situations such as emails.
	 * 
	 * @param string $body
	 */
	public function setBody(string $body) {
		$this->_bodyCache = $body;
	}

	/**
	 * Check the access currently set on the view against the currently logged in user.
	 *
	 * If the user does not have access to the resource, $this->status is set to 403.
	 *
	 * @since 2011.10
	 * @return boolean
	 */
	public function checkAccess() {
		// And do some logic to see if the current user can access this resource.
		// This is more of a helper function to Controllers.
		$u = \Core\user();
		if ($u->checkAccess($this->access)) {
			// yay.
			return true;
		}
		else {
			$this->status = View::STATUS_ACCESSDENIED;
			return false;
		}
	}

	/**
	 * Get if this View is cacheable
	 *
	 * @return bool
	 */
	public function isCacheable(){
		return $this->cacheable;
	}


	


	/**
	 * Add a script to the global View object.
	 *
	 * This will be rendered when a page-level view is rendered.
	 *
	 * @param string $script
	 * @param string $location
	 */
	public function addScript($script, $location = 'head') {
		if (strpos($script, '<script') === false) {
			// Resolve the script and wrap it with a script block.
			$script = '<script type="text/javascript" src="' . \Core\resolve_asset($script) . '"></script>';
		}

		
		// I can check to see if this script has been loaded before.
		if (in_array($script, $this->scripts['head'])) return;
		if (in_array($script, $this->scripts['foot'])) return;

		// No? alright, add it to the requested location!
		if ($location == 'head'){
			$this->scripts['head'][] = $script;
		}
		else{
			$this->scripts['foot'][] = $script;
		}
	}

	/**
	 * Append arbitrary HTML content at the end of the body.
	 * 
	 * @param string $content
	 */
	public function appendBodyContent(string $content) {
		if (!in_array($content, $this->_footContent)){
			$this->_footContent[] = $content;
		}
	}

	/**
	 * Prepend arbitrary HTML content at the start of the body.
	 *
	 * @param string $content
	 */
	public function prependBodyContent(string $content) {
		if (!in_array($content, $this->_bodyContent)){
			$this->_bodyContent[] = $content;
		}
	}

	/**
	 * Add a linked stylesheet file to the global View object.
	 *
	 * @param string $link The link of the stylesheet
	 * @param string $media Media to display the stylesheet with.
	 */
	public function addStylesheet($link, $media = "all") {
		if (strpos($link, '<style') === 0) {
			// This is a style tag, not a stylesheet.  Use that method instead.
			$this->addStyle($link);
			return;
		}

		if (strpos($link, '<link') === false) {

			// Is this a CSS file or a LESS file?
			if(strripos($link, '.less') == strlen($link)-5 ){
				$rel = 'stylesheet/less';
				Core::_AttachLessJS();
			}
			else{
				$rel = 'stylesheet';
			}

			// Resolve the script and wrap it with a script block.
			$link = '<link type="text/css" href="' . \Core\resolve_asset($link) . '" media="' . $media . '" rel="' . $rel . '"/>';
		}

		// I can check to see if this script has been loaded before.
		if (!in_array($link, $this->stylesheets)){
			$this->stylesheets[] = $link;
		}
	}

	/**
	 * Add an inline style to the global View object.
	 *
	 * @param string $style The contents of the <style> tag.
	 */
	public function addStyle($style) {
		if(strpos($style, '<link ') === 0){
			// This is a full stylesheet, not a style!
			$this->addStylesheet($style);
			return;
		}

		if (strpos($style, '<style') === false) {
			// Every style must be wrapped in style tags, so do that if not already done.
			$style = '<style>' . $style . '</style>';
		}

		// Don't forget to include the less compiler if it was requested!
		if(strpos($style, 'rel="stylesheet/less"') !== false){
			Core::_AttachLessJS();
		}
		

		// I can check to see if this script has been loaded before.
		if (!in_array($style, $this->stylesheets)){
			$this->stylesheets[] = $style;
		}
	}

	/**
	 * Set an HTML attribute
	 *
	 * @param string $attribute key
	 * @param string $value     value
	 */
	public function setHTMLAttribute($attribute, $value) {
		$this->htmlAttributes[$attribute] = $value;
	}

	/**
	 * Get the HTML attributes as either a string or an array.
	 *
	 * These attributes are from the <html> tag.
	 *
	 * @param bool $asarray Set to false for a string, true for an array.
	 *
	 * @return array|string
	 */
	public function getHTMLAttributes($asarray = false) {
		$atts = $this->htmlAttributes;

		if ($asarray) {
			return $atts;
		}
		else {
			$str = '';
			foreach ($atts as $k => $v) $str .= " $k=\"" . str_replace('"', '&quot;', $v) . "\"";
			return trim($str);
		}
	}

	/**
	 * Add a meta name, (and value), to this view.
	 *
	 * @param string $key
	 * @param string $value
	 */
	public function addMetaName($key, $value) {
		$this->meta[$key] = $value;
	}

	/**
	 * Add a full meta string to the head of this view.
	 *
	 * This should be formatted as &lt;meta name="blah" content="foo"/&gt;, (or however as necessary).
	 *
	 * @param $string
	 */
	public function addMeta($string) {
		if (strpos($string, '<meta') === false){
			$string = '<meta ' . $string . '/>';
		}
		$this->head[] = $string;
	}

	/**
	 * Add a full string to the head of this view.
	 *
	 * @param $string
	 */
	public function addHead($string){
		$this->head[] = $string;
	}

	/**
	 * Method to append a header onto the array of headers to be sent to the browser when render is called.
	 *
	 * @param string $key
	 * @param string $value
	 */
	public function addHeader($key, $value){
		$this->headers[$key] = $value;
	}

	/**
	 * Disable the View cache on this object and any/all parents
	 *
	 * Will also request that robots do not cache the page
	 */
	public function disableCache() {
		$this->cacheable = false;

		if($this->parent){
			$this->parent->disableCache();
		}

		$this->meta->setMeta('robots', 'noarchive');
	}

	public function disableIndexable() {
		$this->meta->setMeta('robots', 'noindex');
	}
	
	/**
	 * Sends the assigned View headers (and dynamically generated ones), to the user agent.
	 * 
	 * If headers already sent nothing is done.
	 */
	public function sendHeaders(){
		if(!(
			$this->mode === View::MODE_PAGE || 
			$this->mode === View::MODE_PAGEORAJAX || 
			$this->mode === View::MODE_AJAX || 
			$this->mode === View::MODE_NOOUTPUT || 
			$this->mode === View::MODE_EMAILORPRINT
		)){
			// Only full pages support headers.
			return;
		}
		
		if(headers_sent()){
			// Headers already sent, don't try to resend.
			return;
		}
		
		$this->_generateHeaders();
		http_response_code($this->status);
		foreach($this->headers as $k => $v){
			header($k . ': ' . $v);
		}
	}
	
	/**
	 * Generate any headers required for this View.
	 */
	private function _generateHeaders(){
		// Be sure to send the content type and status to the browser, (if it's a page)
		$codes = [
			200 => 'OK',
			201 => 'Created',
			202 => 'Accepted',
			204 => 'No Content',
			205 => 'Reset Content',
			300 => 'Multiple Choices',
			301 => 'Moved Permanently',
			302 => 'Found',
			303 => 'See Other',
			304 => 'Not Modified',
			400 => 'Bad Request',
			401 => 'Unauthorized',
			402 => 'Payment Required',
			403 => 'Forbidden',
			404 => 'Not Found',
			405 => 'Method Not Allowed',
			406 => 'Not Acceptable',
			407 => 'Proxy Authentication Required',
			408 => 'Request Timeout',
			409 => 'Conflict',
			410 => 'Gone',
			411 => 'Length Required',
			412 => 'Precondition Failed',
			413 => 'Payload Too Large',
			414 => 'URI Too Long',
			415 => 'Unsupported Media Type',
			416 => 'Range Not Satisfiable',
			417 => 'Expectation Failed',
			418 => 'I\'m A Teapot',
			420 => 'Enhance Your Calm',
			451 => 'Unavailable For Legal Reasons',
			500 => 'Internal Server Error',
		];
		
		if(isset($codes[ $this->status ])){
			$this->addHeader('Status', $this->status . ' ' . $codes[ $this->status ]);
			$this->addHeader('X-Status-Cat', 'https://http.cat/' . $this->status . '.jpg');
		}
		else{
			$this->addHeader('Status', '520 Unknown Error');
		}
		
		if($this->contenttype === View::CTYPE_HTML){
			$this->addHeader('Content-Type', 'text/html; charset=UTF-8');
		}
		elseif($this->contenttype){
			$this->addHeader('Content-Type', $this->contenttype);
		}

		$this->addHeader('X-Content-Encoded-By', 'Core Plus' . (DEVELOPMENT_MODE ? ' ' . Core::GetComponent('core')->getVersion() : ''));
		
		if(ConfigHandler::Get('/core/security/x-frame-options')){
			$this->addHeader('X-Frame-Options', ConfigHandler::Get('/core/security/x-frame-options'));
		}

		if(ConfigHandler::Get('/core/security/csp-frame-ancestors')){
			$this->addHeader('Content-Security-Policy', 'frame-ancestors \'self\' ' . ConfigHandler::Get('/core/security/content-security-policy'));
		}

		if($this->updated !== null){
			$this->addHeader('Last-Modified', Date\DateTime::FormatString($this->updated, Date\DateTime::FORMAT_RFC2822, Date\Timezone::TIMEZONE_GMT));
		}
	}
	
	/**
	 * Get the content to be inserted into the <head> tag for this view.
	 *
	 * @return string
	 */
	private function _getHeadContent(){
		$minified = ConfigHandler::Get('/core/markup/minified');

		// Some of the automatic settings only get set if no errors.
		if($this->status === View::STATUS_OK){
			
			// Allow the title to be set from the base property for convenience
			if($this->title){
				$this->meta['title']->setTitle($this->title);
			}
		}
		
		
		$data = array_merge(
			$this->meta->fetch(), 
			$this->head, 
			$this->stylesheets, 
			$this->scripts['head']
		);

		if ($minified) {
			$out = implode('', $data);
		}
		else {
			$out = '<!-- BEGIN Automatic HEAD generation -->' . "\n\n" . implode("\n", $data) . "\n\n" . '<!-- END Automatic HEAD generation -->';
		}

		return trim($out);
	}

	/**
	 * Get the content to be inserted just after the <body> tag for this view.
	 *
	 * @return string
	 */
	private function _getBodyContent() : string {
		$minified = ConfigHandler::Get('/core/markup/minified');

		// This section only contains scripts right now.
		if($minified){
			$data = $this->_bodyContent;
		}
		else{
			$data = array_merge(
				['<!-- BEGIN BODY INSERTIONS -->'],
				$this->_bodyContent,
				['<!-- END BODY INSERTIONS -->']
			);
		}


		if ($minified) {
			$out = implode('', $data);
		}
		else {
			$out = implode("\n", $data);
		}

		return trim($out);
	}

	/**
	 * Get the content to be inserted just before the </body> tag for this view.
	 *
	 * @return string
	 */
	private function _getFootContent() : string {
		$minified = ConfigHandler::Get('/core/markup/minified');

		// This section only contains scripts right now.
		if($minified){
			$data = array_merge($this->_footContent, $this->scripts['foot']);
		}
		else{
			$data = array_merge(
				['<!-- BEGIN FOOT INSERTIONS -->'],
				$this->_footContent,
				$this->scripts['foot'],
				['<!-- END FOOT INSERTIONS -->']
			);
		}


		if ($minified) {
			$out = implode('', $data);
		}
		else {
			$out = implode("\n", $data);
		}

		return trim($out);
	}

	/**
	 * Internal only method to sync another view's metadata into this one.
	 *
	 * This is to allow Views to have certain attributes bubble up to the top-most view to be rendered out correctly.
	 *
	 * @param View $view
	 */
	protected function _syncFromView(View $view){
		if($view === $this){
			// !??!
			return;
		}

		foreach($view->head as $h){
			$this->addHead($h);
		}

		foreach($view->meta as $m){
			$this->addMeta($m);
		}

		foreach($view->scripts['head'] as $s){
			$this->addScript($s, 'head');
		}
		foreach($view->scripts['foot'] as $s){
			$this->addScript($s, 'foot');
		}
		foreach($view->stylesheets as $s){
			$this->addStyle($s);
		}

		if($view->ssl){
			// Allow children to mandate that a given page be presented in SSL if available.
			$this->ssl = true;
		}

		// Simple merges are possible here!
		$this->bodyclasses += $view->bodyclasses;
		$this->htmlAttributes += $view->htmlAttributes;
	}
}
