<?php
/**
 * Systems for the meta data and tags on Views.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @since 2.4.0
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use ArrayAccess;
use Iterator;
use Core\ViewMeta\Base as MetaBase;

/**
 * The main controller for a set of controls, can be instantiated with either page level or inline operations,
 * it doesn't care which.
 */
class ViewMetas implements Iterator, ArrayAccess {

	private $_links = array();

	private $_pos = 0;

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Return the current element
	 * @link http://php.net/manual/en/iterator.current.php
	 * @return mixed Can return any type.
	 */
	public function current() {
		return $this->_links[$this->_pos];
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Move forward to next element
	 * @link http://php.net/manual/en/iterator.next.php
	 * @return void Any returned value is ignored.
	 */
	public function next() {
		++$this->_pos;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Return the key of the current element
	 * @link http://php.net/manual/en/iterator.key.php
	 * @return mixed scalar on success, or null on failure.
	 */
	public function key() {
		return $this->_pos;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Checks if current position is valid
	 * @link http://php.net/manual/en/iterator.valid.php
	 * @return boolean The return value will be casted to boolean and then evaluated.
	 * Returns true on success or false on failure.
	 */
	public function valid() {
		return isset($this->_links[$this->_pos]);
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Rewind the Iterator to the first element
	 * @link http://php.net/manual/en/iterator.rewind.php
	 * @return void Any returned value is ignored.
	 */
	public function rewind() {
		$this->_pos = 0;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Whether a offset exists
	 * @link http://php.net/manual/en/arrayaccess.offsetexists.php
	 *
	 * @param mixed $offset <p>
	 * An offset to check for.
	 * </p>
	 *
	 * @return boolean true on success or false on failure.
	 * </p>
	 * <p>
	 * The return value will be casted to boolean if non-boolean was returned.
	 */
	public function offsetExists($offset) {
		if(!isset($this->_links[$offset])){
			// The key doesn't exist, then false.
			return false;
		}

		// Also if the key is explictly empty...
		/** @var ViewMeta $meta */
		$meta = $this->_links[$offset];
		if($meta->content === null){
			return false;
		}

		// Otherwise...
		return true;

		//return array_key_exists($offset, $this->_links);
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to retrieve
	 * @link http://php.net/manual/en/arrayaccess.offsetget.php
	 * @param mixed $offset <p>
	 * The offset to retrieve.
	 * </p>
	 * @return mixed Can return all value types.
	 */
	public function offsetGet($offset) {
		if(!isset($this->_links[ $offset ])){
			$this->_links[ $offset ] = MetaBase::Factory($offset);
		}
		
		return $this->_links[$offset];
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to set
	 * @link http://php.net/manual/en/arrayaccess.offsetset.php
	 * @param mixed $offset <p>
	 * The offset to assign the value to.
	 * </p>
	 * @param mixed $value <p>
	 * The value to set.
	 * </p>
	 * @return void
	 *
	 * @throws Exception
	 */
	public function offsetSet($offset, $value) {
		if($offset === null){
			// The user just wants the next available one.

			if($this->valid()){
				$this->next();

			}
			$offset = $this->key();
		}
		
		if(!isset($this->_links[ $offset ])){
			$this->_links[ $offset ] = MetaBase::Factory($offset);
		}
		
		$this->_links[ $offset ]->setContent($value);
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to unset
	 * @link http://php.net/manual/en/arrayaccess.offsetunset.php
	 * @param mixed $offset <p>
	 * The offset to unset.
	 * </p>
	 * @return void
	 */
	public function offsetUnset($offset) {
		unset($this->_links[$offset]);
	}

	/**
	 * Add an array of links to this control set.
	 *
	 * @param array $links
	 */
	public function addLinks(array $links){
		foreach($links as $l){
			$this[] = $l;
		}
	}

	/**
	 * Get this meta set as an array of key-indexed elements
	 *
	 * Each array key is the keyname of the meta tag, (description, keywords, etc).
	 *
	 * @return array
	 */
	public function fetch(){
		$data = array();
		foreach($this->_links as $l){
			/** @var ViewMeta $l */
			$ea = $l->fetch();
			if(is_array($ea) && sizeof($ea)){
				$data = array_merge($data, $l->fetch());
			}
		}
		return $data;
	}
	
	public function setMeta($name, $value, $valueKey = null) {
		if(!isset($this->_links[ $name ])){
			$this->_links[ $name ] = MetaBase::Factory($name);
		}
		
		$this->_links[ $name ]->setContent($value, $valueKey);
	}
}
