<?php
/**
 * File for the class UserGroupAdminController
 *
 * @package Core\Cache
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Cache;

/**
 * Simple file-only caching system for Core
 *
 * Will use file modified time set to the future to track expiration date.
 * This is to support systems where extended attributes are not supported
 * and exec() is not permitted.
 *
 * If the zlib library is available, gzip compression will be used to save disk space.
 *
 * Usage:
 *
 * ```.code-php
 * $cache = new \Core\Cache\File();
 * $data = $cache->read('some-key');
 * $cache->write('some-key', 'some-data');
 *
 * // Expiration dates can be used
 * $cache->write('some-key', 'some-data', SECONDS_ONE_HOUR);
 * ```
 */
class File implements CacheInterface {

	private $_dir;
	private $_gzip;

	/**
	 * Construct a new cache object with the given key and the given expiration.
	 */
	public function __construct() {
		$this->_dir = TMP_DIR . 'cache/';
		$this->_gzip = (extension_loaded('zlib'));
	}
	
	/**
	 * Writes the requested key and data optionally with an expiration date.
	 * 
	 * @param string $key  Key of the cache object to write
	 * @param mixed $data  The data to cache
	 * @param int|null $expires Expiration in number of seconds from now
	 * 
	 * @return bool Whether the operation was successful.
	 */
	public function write(string $key, $data, int $expires = null) : bool {
		$data = ($this->_gzip) ? 
			gzcompress(serialize($data)) :
			serialize($data);
		
		$file = $this->_resolveKeyToFile($key);
		if(file_put_contents($file, $data) === false){
			return false;
		}

		if($expires === null) {
			// Files require *some* sort of expiration.
			$expires = SECONDS_ONE_MONTH;
		}

		// Use mtime instead of extended attributes because xattr support cannot be guaranteed.
		touch($file, $expires + time());
		// exec('setfattr -n user.expires -v ' . ($expires + time()) . ' ' . escapeshellarg($file) . ' 2>/dev/null');

		return true;
	}

	/**
	 * Reads a cache entry based on its key
	 * 
	 * @param string $key  Key of the cache object to write
	 * 
	 * @returns mixed Either the content of the cache object or null if it doesn't exist
	 */
	public function read(string $key) {
		$file = $this->_resolveKeyToFile($key);
		
		if(!file_exists($file)){
			return null;
		}
		
		if($this->_isExpired($file)){
			// auto-cleanup for expired records.
			$this->delete($key);
			return null;
		}
		
		$data = $this->_gzip ? 
			unserialize(gzuncompress(file_get_contents($file))) :
			unserialize(file_get_contents($file));

		if ($data === false) {
			/*
				This should only happen when someone changes the gzip settings and there is
				existing data or someone has been mucking about in the cache folder manually.
				Delete the bad entry since the file cache doesn't clean up after itself and
				then return false so fresh data will be retrieved.
			 */
			$this->delete($key);
			return null;
		}

		return $data;
	}

	/**
	 * Delete the cache entry matching the key
	 * 
	 * @param string $key Key to delete
	 * 
	 * @return boolean Whether the operation was successful.
	 */
	public function delete(string $key) : bool {
		$file = $this->_resolveKeyToFile($key);
		if (file_exists($file)) {
			return unlink($file);
		}

		return false;
	}

	/**
	 * Method: flush()
	 *  Invalidate all items in the cache
	 *
	 * @since 2011.07.28
	 * @access public
	 * @return boolean Whether the operation was successful.
	 */
	public function flush() : bool {
		$dir = opendir($this->_dir);
		
		if(!$dir){
				return true;
		}
		while(($file = readdir($dir)) !== false){
			if($file == '.' || $file == '..') continue;

			unlink($this->_dir . $file);
		}
		closedir($dir);

		return true;
	}
	
	public function ls() : array {
		// @todo Not supported yet.
		
		return [];
	}
	
	public function isReady() : bool {
		return ( is_dir($this->_dir) || mkdir($this->_dir) );
	}
	
	/**
	 * Resolve a key to a filename, this is because the key could contain strings that filesystems don't like.
	 * 
	 * @param string $key
	 * 
	 * @return string Fully resolved file path
	 */
	private function _resolveKeyToFile(string $key) : string {
		return $this->_dir . 
			preg_replace('/[^a-zA-Z0-9_-]/', '', $key) .
			'.cache';
	}

	/**
	 * Checks whether the cache object is expired or not.
	 *
	 * @returns boolean Whether the cache is expired or not.
	 */
	private function _isExpired(string $filename) : bool {

		// Use mtime instead of extended attributes because xattr support cannot be guaranteed.
		if(file_exists($filename)) {
			return filemtime($filename) <= time();
		}
		else {
			return true;
		}
		// $expires = exec('getfattr --only-values -n user.expires ' . escapeshellarg($filename) . ' 2>/dev/null');
	}
}
