<?php
/**
 * Dataset
 *
 * @package Core\Datamodel
 * @since 0.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Datamodel;

class Dataset implements \Iterator{

	/**
	 * Mode for altering the structure of the datastore
	 */
	const MODE_ALTER = 'alter';
	/**
	 * Mode for getting data from the datastore
	 */
	const MODE_GET = 'get';
	/**
	 * Mode for inserting data into the datastore
	 *
	 * INSERT INTO (`key`, `key2) VALUES ('val1', 'val2');
	 */
	const MODE_INSERT = 'insert';
	/**
	 * Mode for bulk inserting data into the datastore
	 *
	 * INSERT INTO (`key`, `key2) VALUES ('val1', 'val2'), ('val1', 'val2'), ('val1', 'val2'), ('val1', 'val2')...;
	 */
	const MODE_BULK_INSERT = 'bulk_insert';
	/**
	 * Mode for updating data in the datastore
	 */
	const MODE_UPDATE = 'update';
	/**
	 * Mode for either updating or inserting data in the datastore
	 */
	const MODE_INSERTUPDATE = 'insertupdate';
	/**
	 * Mode for deleting data from the datastore
	 */
	const MODE_DELETE = 'delete';
	/**
	 * Mode for counting records in the datastore
	 */
	const MODE_COUNT = 'count';


	/** @var DatasetSelect[] */
	public $_selects = null;

	public $_groupby = null;

	/**
	 * From / Table to retrieve data
	 *
	 * @var DatasetFrom[]
	 */
	public $_tables = null;

	/**
	 * The root where clause for this dataset
	 * @var null|DatasetWhereClause
	 */
	public $_where = null;

	public $_having = null;

	/**
	 * Set to an explict mode if required and supported by the datamodel backend.
	 * 
	 * @var string|null
	 */
	public $_mode = null;

	/**
	 * @var array
	 */
	public $_sets = array();

	public $_idcol = null;

	public $_idval = null;

	public $_limit = null;

	public $_offset = null;

	public $_order = false;

	/**
	 * @var DatasetJoin[]
	 */
	public $_joins = [];

	public $_data = null;

	/** @var null|\Generator */
	public $_results = null;

	public $num_rows = null;

	public $_inserts = null;

	public $_updates = null;
	
	public $_deletes = null;
	
	/** @var bool Tracker for if this is a bulk operation, (not all DMIs may support this!) */
	public $_isBulk = false;

	/**
	 * Column renames used in the alter mode
	 * @var null|array
	 */
	public $_renames = null;

	/**
	 * Set to true to return only unique records, ala SELECT DISTINCT
	 *
	 * @var bool
	 */
	public $uniquerecords = false;

	public function __construct(){

	}

	/**
	 * On clone, make a deep copy of this object!
	 * 
	 * This is required so that the WHERE clause does not get copied by memory space.
	 * Otherwise altering the clone dataset will modify the original dataset!
	 */
	public function __clone() {
		if(is_array($this->_selects)) {
			foreach($this->_selects as $idx => $select) {
				$this->_selects[ $idx ] = clone $select;
			}
		}

		if($this->_where){
			$this->_where = clone $this->_where;
		}

		if($this->_having) {
			$this->_having = clone $this->_having;
		}
	}

	/**
	 * Set the columns to select.
	 *
	 * Argument can be the following:
	 * null: reset the array to blank
	 * single value: add the value to the columns
	 * array of values: add each value to the columns
	 * multiple arguments: add each value to the columns
	 *
	 * @param null|string[]|DatasetSelect[] $selects
	 *
	 * @return Dataset
	 *
	 * @throws DMIQueryException
	 */
	public function select(...$selects) : Dataset {

		if(count($selects) === 0) {
			throw new DMIQueryException (
				'Invalid amount of parameters requested for Dataset->select'
			);
		}

		// Allow null to clear out the selects.
		if(count($selects) === 1 && $selects[0] === null){
			$this->_selects = null;
			
			// Allow chaining
			return $this;
		}

		if($this->_selects === null){
			$this->_selects = [];
		}


		foreach($selects as $a){
			if($a instanceof DatasetSelect) {
				$this->_selects[] = $a;
			}
			elseif(is_array($a)){
				\Core\log_warning('Passing an array into Dataset->select is deprecated (core-7.x)');

				// It should just be a flat array, ie: array('cola', 'colb', 'colc', ...);
				foreach($a as $str){
					if($str instanceof DatasetSelect) {
						$this->_selects[] = $str;
					}
					else {
						$this->_selects[] = new DatasetSelect($str);
					}
				}
			}
			elseif(strpos($a, ',') !== false){
				\Core\log_warning(
					'Passing a string of comma-separated items into Dataset->select is deprecated (core-7.x)'
				);

				// User submitted a comma-separated list... damn them
				$parts = explode(',', $a);
				foreach($parts as $str){
					$this->_selects[] = new DatasetSelect($str);
				}
			}
			else{
				// Just a regular column, yay
				$this->_selects[] = new DatasetSelect($a);
			}
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Group results based on columns
	 *
	 * Argument can be the following:
	 * null: reset the array to blank
	 * single value: add the value to the columns
	 * array of values: add each value to the columns
	 * multiple arguments: add each value to the columns
	 *
	 * @param null|string[]|DatasetSelect[] $columns
	 *
	 * @return Dataset
	 *
	 * @throws DMIQueryException
	 */
	public function groupBy(...$columns) : Dataset {

		if(count($columns) === 0){
			throw new DMIQueryException (
				'Invalid amount of parameters requested for Dataset::groupBy()'
			);
		}

		// Allow null to clear out the grouping.
		if(count($columns) === 1 && $columns[0] === null){
			$this->_groupby = null;

			// Allow chaining
			return $this;
		}

		if($this->_groupby === null){
			$this->_groupby = [];
		}


		foreach($columns as $a){
			if($a instanceof DatasetSelect) {
				$this->_groupby[] = $a;
			}
			elseif(is_array($a)){
				\Core\log_warning(
					'Passing an array into Dataset->groupBy is deprecated (core-7.x)'
				);

				// It should just be a flat array, ie: array('cola', 'colb', 'colc', ...);
				foreach($a as $str){
					$this->_groupby[] = new DatasetSelect($str);
				}
			}
			elseif(strpos($a, ',') !== false){
				\Core\log_warning(
					'Passing a string of comma-separated items into Dataset->groupBy is deprecated (core-7.x)'
				);

				// User submitted a comma-separated list... damn them
				$parts = explode(',', $a);
				foreach($parts as $str){
					$this->_groupby[] = new DatasetSelect($str);
				}
			}
			else{
				// Just a regular column, yay
				$this->_groupby[] = new DatasetSelect($a);
			}
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Insert a single key, value or group of keys as an associative array
	 *
	 * The first parameter allows a string, NULL, or an array of elements
	 *
	 * ## Basic Usage
	 *
	 * Insert a single key and value pair
	 *
	 * ```php
	 * ->insert('key', 'value');
	 * ```
	 *
	 * ## Multiple
	 *
	 * Insert multiple at one time
	 *
	 * ```php
	 * ->insert(['key1' => 'value1', 'key2' => 'value2']);
	 * ```
	 *
	 * ## Reset Inserts
	 *
	 * Resets the internal insert tracker to default
	 *
	 * ```php
	 * ->insert(null);
	 * ```
	 *
	 * @param string|array|null $key_or_array
	 * @param mixed|null $value
	 *
	 * @throws DMIQueryException
	 * 
	 * @return Dataset
	 */
	public function insert($key_or_array, $value = null) : Dataset {

		if(is_array($key_or_array)) {
			foreach($key_or_array as $k => $v) {
				$this->_inserts[$k] = $v;
			}
			$this->_isBulk = false;
		}
		elseif($key_or_array === null) {
			$this->_inserts = null;
		}
		else {
			$this->_inserts[ $key_or_array ] = $value;
			$this->_isBulk = false;
		}

		return $this;
	}

	/**
	 * Request a bulk insert, ONE RECORD AT A TIME
	 *
	 * @param array|null $data Key/Value array of the data to bulk insert as a new record
	 *
	 * @return Dataset
	 */
	public function bulkInsert(?array $data) : Dataset {

		if($data === null) {
			// Reset
			$this->_inserts = null;
			$this->_isBulk = false;
		}
		else {
			// Push this data to the stack of inserts to perform.
			$this->_inserts[] = $data;
			$this->_isBulk = true;
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Request a bulk insert, FOR ALL DATA
	 *
	 * This is a **DESTRUCTIVE** call, as it will overwrite ALL data currently queued for insert!
	 *
	 * This is suitable to be called if you have an entire dataset as a 0-indexed array
	 * ready to be dropped in.
	 *
	 * @param array $data INDEXED Array containing key/value pairs of data to insert
	 *
	 * @return Dataset
	 */
	public function bulkInsertData(array $data) : Dataset {
		// This is a destructive insert, so blow away all the insert data queued thus far.
		$this->_inserts = $data;
		$this->_isBulk = true;

		// Allow chaining
		return $this;
	}

	/**
	 * Update a key to a given value, or set of data
	 *
	 * ## Basic Usage
	 *
	 * Update a single key and value pair
	 *
	 * ```php
	 * ->update('key', 'value');
	 * ```
	 *
	 * ## Multiple
	 *
	 * Update multiple at one time
	 *
	 * ```php
	 * ->update(['key1' => 'value1', 'key2' => 'value2']);
	 * ```
	 *
	 * ## Reset Update
	 *
	 * Resets the internal update tracker to default
	 *
	 * ```php
	 * ->update(null);
	 * ```
	 *
	 * @param string|array|null $key_or_array
	 * @param mixed|null $value
	 *
	 * @throws DMIQueryException
	 * 
	 * @return Dataset
	 */
	public function update($key_or_array, $value = null) : Dataset {
		if(is_array($key_or_array)) {
			foreach($key_or_array as $k => $v) {
				$this->_updates[$k] = $v;
			}
		}
		elseif($key_or_array === null) {
			$this->_updates = null;
		}
		else {
			$this->_updates[ $key_or_array ] = $value;
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Rename a column in this dataset, primarlly an administrative / installer function.
	 * 
	 * @throws DMIException
	 * 
	 * @return Dataset
	 */
	public function renameColumn() : Dataset {
		$n = func_num_args();

		if($n != 2){
			throw new DMIQueryException ('Invalid amount of parameters requested for Dataset::renameColumn(), ' . $n . ' provided, exactly 2 expected');
		}

		$oldname = func_get_arg(0);
		$newname = func_get_arg(1);

		if($this->_renames === null){
			$this->_renames = [];
		}
		
		$this->_renames[$oldname] = $newname;

		// Supported for legacy systems.
		$this->_mode = Dataset::MODE_ALTER;

		// Allow chaining
		return $this;
	}

	/**
	 * Delete an entire record or specific key from the store, (on supported DMIs).
	 *
	 * For noSQL type databases, (LDAP, Mongo), this operation can delete a specific key from an object.
	 * Otherwise, simply calling it will request the entire record/object to be deleted.
	 *
	 * @return Dataset
	 *
	 * @throws DMIQueryException
	 */
	public function delete() : Dataset {
		$n = func_num_args();

		if($n == 0 ) {
			// Simple call, allowed to delete an entire object/record.
			$this->_setDelete('*', '*');
		}
		elseif($n == 1) {
			$a = func_get_arg(0);
			
			if(is_array($a)) {
				foreach($a as $k => $v) {
					$this->_setDelete($k, $v);
				}
			}
			elseif($a === null) {
				// Reset
				$this->_deletes = null;
			}
			else {
				// Delete all values that match this key
				$this->_setDelete($a, '*');
			}
		}
		elseif($n === 2) {
			$k = func_get_arg(0);
			$v = func_get_arg(1);
			$this->_setDelete($k, $v);
		}
		else {
			throw new DMIQueryException (
				'Invalid amount of parameters requested for Dataset->delete, ' . $n . ' provided, exactly 0, 1, or 2 expected'
			);
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Set this dataset to only return the count of records.
	 *
	 * @return Dataset
	 */
	public function count() : Dataset {
		$s = new DatasetSelect();
		$s->function = 'COUNT';
		$s->alias = 'COUNT';

		if(!is_array($this->_selects)) {
			$this->_selects = [];
		}

		$this->_selects[] = $s;

		// Allow chaining
		return $this;
	}

	public function setID($key, $val = null) : Dataset {
		$this->_idcol = $key;
		$this->_idval = $val;

		if($val) {
			$this->whereEquals($key, $val);
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Get the ID of the inserted column; useful for auto-incs.
	 * 
	 * Will return null if there is no ID set.
	 * 
	 * @return null|int
	 */
	public function getID(){
		return $this->_idval;
	}

	/**
	 * Get the mode for this query
	 * 
	 * Pulled dynamically based on what parameters have been requested.
	 * 
	 * @return string
	 */
	public function getMode() : ?string {
		if($this->_isBulk && $this->_inserts !== null){
			// Bulk insert mode!
			return self::MODE_BULK_INSERT;
		}
		
		if(
			($this->_inserts !== null && $this->_updates !== null) ||
			($this->_inserts !== null && $this->_deletes !== null) ||
			($this->_updates !== null && $this->_deletes !== null)
		){
			// This is a combined operation which includes at least two of: delete, insert, or update.
			return self::MODE_INSERTUPDATE;
		}
		
		if($this->_selects !== null && sizeof($this->_selects) == 1 && $this->_selects[0] == '__COUNT__'){
			// Simple count mode.
			return self::MODE_COUNT;
		}
		
		if($this->_selects !== null){
			// Simple select mode
			return self::MODE_GET;
		}
		
		if($this->_inserts !== null){
			// Simple insert mode
			return self::MODE_INSERT;
		}
		
		if($this->_updates !== null){
			// Simple update mode
			return self::MODE_UPDATE;
		}
		
		if($this->_renames !== null){
			return self::MODE_ALTER;
		}
		
		if($this->_deletes !== null){
			// Is this a full delete or a key delete?
			if(sizeof($this->_deletes) == 1 && isset($this->_deletes['*']) && $this->_deletes['*'] == '*'){
				return self::MODE_DELETE;
			}
			else{
				// It's a complex delete, (unsupported by traditional SQL engines).
				return self::MODE_INSERTUPDATE;
			}
		}

		// Something needs to be returned.
		return null;
	}

	/**
	 * Get the columns to insert
	 * @return null|array
	 */
	public function getInserts(){
		return $this->_inserts;
	}

	/**
	 * Get the columns to update along with their new values.
	 * 
	 * @return null|array
	 */
	public function getUpdates(){
		return $this->_updates;
	}

	/**
	 * Get the columns to delete and/or the values on the key to delete.
	 * 
	 * @return null|array
	 */
	public function getDeletes(){
		return $this->_deletes;
	}

	/**
	 * Get the results in a Generator-friendly format
	 * @return \Generator|array
	 */
	public function getResults() {
		return $this->_results ?? [];
	}

	/**
	 *
	 * @param string $tablename
	 * @return Dataset
	 */
	public function table(string $tablename, string $alias = null) : Dataset {
		if(!is_array($this->_tables)) {
			$this->_tables = [];
		}

		$table = new DatasetFrom();
		$table->table = $tablename;
		$table->alias = $alias;

		$this->_tables[] = $table;

		// Allow chaining
		return $this;
	}

	public function from(DatasetFrom $from) : Dataset {
		if(!is_array($this->_tables)) {
			$this->_tables = [];
		}

		$this->_tables[] = $from;

		// Allow chaining
		return $this;
	}

	/**
	 * @param bool $unique
	 *
	 * @return Dataset
	 */
	public function unique($unique = true) : Dataset{
		$this->uniquerecords = $unique;

		// Allow chaining
		return $this;
	}

	/**
	 * @return DatasetWhereClause
	 */
	public function getWhereClause() : DatasetWhereClause {
		// Make sure that the root where clause exists first!
		if($this->_where === null){
			$this->_where = new DatasetWhereClause('root');
		}

		return $this->_where;
	}

	/**
	 * @return DatasetWhereClause
	 */
	public function getHavingClause() {
		// Make sure that the root where clause exists first!
		if($this->_having === null){
			$this->_having = new DatasetWhereClause('root');
		}

		return $this->_having;
	}

	/**
	 * Set or add to the where clause for this query.
	 *
	 * Argument passed in can be a multitude of options:
	 * key/value paired array:
	 *
	 *
	 * ## Supported Formats:
	 *
	 * ### Basic Single Call
	 *
	 * Just a regular string for the where statement
	 *
	 * ```
	 * $dataset->where('key = some value');
	 * $dataset->where('key > 123');
	 * $dataset->where('key LIKE something%foo');
	 * ```
	 *
	 * ### Combined Calls
	 *
	 * Multiple parameters can be combined on a single call for convenience
	 *
	 * ```
	 * $dataset->where('key = some value', 'updated > 123', 'name LIKE %blah');
	 * ```
	 *
	 * ### Array Call
	 *
	 * A single array can be passed in for simple "equal" comparisons.
	 *
	 * ```
	 * $dataset->where(['key1' => 'value1', 'key2' => 'value2']);
	 * ```
	 *
	 *
	 * ### Array Call Combined
	 *
	 * A single array can be passed in with full statements, but this is pointless, do not do this.
	 *
	 * ```
	 * $dataset->where(['key = some value', 'updated > 123', 'name LIKE %blah']);
	 * ```
	 *
	 * @return Dataset
	 */
	public function where() : Dataset {
		$args = func_get_args();

		if(count($args) === 1 && is_array($args[0])) {
			// The block to handle any single-parameter-array call.

			if(count($args[0]) > 0) {
				// Support passing in ['something' => [1, 2, 3]]
				// If this is not done, then it gets wrapped in multiple arrays
				// The count is required to not create a blank WHERE statement
				$this->getWhereClause()->addWhere($args[0]);
			}
		}
		else {
			$this->getWhereClause()->addWhere($args);
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Basic WHERE clause with a single key and single value
	 *
	 * @param string $key
	 * @param string|int|float|null|array $value
	 *
	 * @return Dataset
	 */
	public function whereEquals(string $key, $value) : Dataset {

		$this->getWhereClause()->whereEquals($key, $value);

		// Allow chaining
		return $this;
	}

	/**
	 * Basic WHERE clause with a single key and single value
	 *
	 * @param string $key
	 * @param string|int|float|null|array $value
	 *
	 * @return Dataset
	 */
	public function whereNotEquals(string $key, $value) : Dataset {

		$this->getWhereClause()->whereNotEquals($key, $value);

		// Allow chaining
		return $this;
	}

	/**
	 * Allow for grouping of groups of where clauses.
	 *
	 * This is useful for statements such as
	 * WHERE (this = 1 OR that = 1) AND something = blah;
	 *
	 * The where clause can either be a single array, a single string, or a list of arguments
	 *
	 * @param string $separator 'AND', 'OR'
	 * @param array|string $wheres
	 * @return Dataset
	 */
	public function whereGroup(string $separator, ...$wheres) : Dataset {
		if(count($wheres) === 1 && is_array($wheres[0])) {
			// A single array of multiple elements was passed in instead of single elements, convert that data.
			$wheres = $wheres[0];
		}

		$clause = new DatasetWhereClause();
		$clause->setSeparator($separator);
		$clause->addWhere($wheres);

		// Since everything will be under the root node anyway...
		$this->getWhereClause()->addWhere($clause);

		// Allow chaining
		return $this;
	}

	/**
	 * Chainable constructor for a new nested WHERE clause
	 *
	 * @param string $separator
	 * @return DatasetWhereClause
	 */
	public function newWhereGroup(string $separator = 'AND') : DatasetWhereClause {
		$clause = new DatasetWhereClause();
		$clause->setSeparator($separator);

		// Since everything will be under the root node anyway...
		$this->getWhereClause()->addWhere($clause);

		// Allow chaining
		return $clause;
	}
	
	/**
	 * Allow for a preformatted WHERE clause to be passed in.
	 * 
	 * This provides support for advanced WHERE operations when necessary.
	 * 
	 * @param string $clause
	 * 
	 * @return Dataset
	 */
	public function wherePreformatted(string $clause) {
		$where = new DatasetWhere();
		$where->preformatted = $clause;
		
		// Since everything will be under the root node anyway...
		$this->getWhereClause()->addWhere($where);
		
		// Allow chaining
		return $this;
	}

	/**
	 * Set the HAVING clause on this query.
	 *
	 * This behaves similar to WHERE, so it uses the same underlying system.
	 *
	 * @param string|DatasetWhere $where
	 *
	 * @return Dataset
	 */
	public function having($where) : Dataset {
		$this->getHavingClause()->addWhere($where);

		// Allow chaining
		return $this;
	}

	/**
	 * Set the limit for this dataset.
	 *
	 * @param int|null $limit
	 * @param int|null $offset
	 *
	 * @return Dataset
	 */
	public function limit(int $limit = null, int $offset = null) : Dataset {
		$this->_limit = $limit;
		$this->_offset = $offset;

		// Allow chaining
		return $this;
	}

	/**
	 * @throws DMIQueryException
	 * @return Dataset
	 */
	public function order() : Dataset {
		$n = func_num_args();
		if($n == 1) {
			$this->_order = func_get_arg(0);
		}
		elseif($n == 2) {
			$this->_order = func_get_arg(0) . ', ' . func_get_arg(1);
		}
		else {
			throw new DMIQueryException('Invalid amount of parameters requested for Dataset::order()');
		}

		// Allow chaining
		return $this;
	}

	/**
	 * Add a pre-created Join statement
	 *
	 * @param DatasetJoin $join
	 * @return Dataset
	 */
	public function join(DatasetJoin $join) : Dataset {
		$this->_joins[] = $join;

		// Allow chaining
		return $this;
	}

	/**
	 * Add a basic LEFT JOIN on this dataset
	 *
	 * @param string $table      Joining table name
	 * @param string $on         WHERE clause for this join, *USES KEYS FOR REFERENCES*
	 * @param string|null $alias Optional alias for this join
	 *
	 * @return $this
	 */
	public function joinLeft(string $table, string $on, string $alias = null) : Dataset {
		$this->_join('LEFT', $table, $on, $alias);
		
		// Allow chaining
		return $this;
	}

	/**
	 * Add a basic LEFT JOIN on this dataset
	 *
	 * @param string $table      Joining table name
	 * @param string $on         WHERE clause for this join, *USES KEYS FOR REFERENCES*
	 * @param string|null $alias Optional alias for this join
	 *
	 * @return $this
	 */
	public function joinRight(string $table, string $on, string $alias = null) : Dataset {
		$this->_join('RIGHT', $table, $on, $alias);

		// Allow chaining
		return $this;
	}

	/**
	 * Add a basic LEFT JOIN on this dataset
	 *
	 * @param string $table      Joining table name
	 * @param string $on         WHERE clause for this join, *USES KEYS FOR REFERENCES*
	 * @param string|null $alias Optional alias for this join
	 *
	 * @return $this
	 */
	public function joinInner(string $table, string $on, string $alias = null) : Dataset {
		$this->_join('INNER', $table, $on, $alias);

		// Allow chaining
		return $this;
	}

	/**
	 * Add a basic LEFT JOIN on this dataset
	 *
	 * @param string $table      Joining table name
	 * @param string $on         WHERE clause for this join, *USES KEYS FOR REFERENCES*
	 * @param string|null $alias Optional alias for this join
	 *
	 * @return $this
	 */
	public function joinOuter(string $table, string $on, string $alias = null) : Dataset {
		$this->_join('OUTER', $table, $on, $alias);

		// Allow chaining
		return $this;
	}


	/**
	 *
	 * @param BackendInterface $interface
	 * @return Dataset
	 */
	public function execute($interface = null){
		// Default to the system interface.
		if(!$interface){
			$dmi = DMI::GetSystemDMI();
			$interface = $dmi->connection();
		}

		// This actually goes the other way, as the interface has the logic.
		$interface->execute($this);

		if($this->_results === null && $this->_data === null && $this->_mode == Dataset::MODE_GET ){
			// It's been executed, so data should at least be something.
			$this->_data = [];
			reset($this->_data);
		}

		// Allow Chaining
		return $this;
	}

	/**
	 * Execute this query and return the records or record, based on requested criteria.
	 *
	 * If limit == 1 and only one select was issued, that singular value or null is returned.
	 * If limit == 1 and more than one select was issued, an associative array is returned.
	 * If select contains 1 key and it's not "*", an indexed array is returned containing all results.
	 * Otherwise, an array of associative arrays is returned.
	 *
	 * !WARNING!  This negates benefits from the embedded Generator support!
	 *            Only use this method for relatively small datasets.
	 *
	 * If multiple records (ie: limit is not set to 1), are requested,
	 * they can optionally be grouped by a requested key present in the return.
	 *
	 * ie: if "id", "name", and "status" are the select keys,
	 * 'id' can be passed as the $keyMap argument to return results
	 * with 'id' as the array key, with 'name' and 'status' as elements in each element.
	 *
	 * If only two keys are selected, say 'id' and 'status', then the value of each key
	 * will simply be the other key's value.
	 *
	 * @param null|DMI $interface
	 * @param null|string $keyMap Optional keymap for results to be grouped under
	 *
	 * @return array|null|mixed
	 */
	public function executeAndGet($interface = null, $keyMap = null) {
		$this->execute($interface);

		if(is_array($this->_selects)) {
			$selectCount = count($this->_selects);
		}
		else {
			$selectCount = 0;
		}


		if($this->_limit == 1 && $this->num_rows == 1){
			// Return a single key's value or null if not found.
			if($this->_results !== null) {
				$row = $this->_results->current();
			}
			elseif($this->_data !== null) {
				$row = $this->_data[0];
			}

			// LIMIT 1 was requested and only 1 row was returned, return that single record value.
			if(count($this->_selects) === 1 && $this->_selects[0]->column !== '*'){
				$sel = $this->_selects[0];

				$k = $sel->alias ?? $sel->column;

				return $row[$k] ?? null;
			}
			else{
				// Return a single record
				return $row;
			}
		}
		elseif($this->_limit == 1 && $this->num_rows == 0){
			// Only one record was selected but no records found.
			// Return as close to what was expected as possible.
			if(count($this->_selects) == 1 && $this->_selects[0]->column !== '*'){
				// Single key requested, return a single value.
				return null;
			}
			else{
				// Multiple keys, blank array returned.
				return [];
			}
		}
		elseif($selectCount === 1 && $this->_selects[0]->function === 'COUNT') {
			// Simple count was requested, return just that!
			$row = null;

			if($this->_results !== null) {
				$row = $this->_results->current();
			}
			elseif($this->_data !== null) {
				$row = $this->_data[0];
			}

			if($row === null) {
				return 0;
			}
			else {
				return (int)$row['COUNT'] ?? 0;
			}
		}
		elseif( $selectCount === 2 && $this->_selects[0]->function === 'COUNT' && is_scalar($this->_selects[1]->column) ) {
			// Simple COUNT request with a single GROUP BY combiner.
			// This should be returned as an associative row with each grouped column as the key of the count.
			$ret = [];

			if($this->_results !== null) {
				$rows = $this->_results;
			}
			elseif($this->_data !== null) {
				$rows =& $this->_data;
			}
			else {
				$rows = [];
			}

			foreach($rows as $d) {
				// Index them because I don't care what they're called.
				$indexed = array_values($d);
				$ret[ $indexed[1] ] = (int)$indexed[0];
			}
			return $ret;
		}
		elseif( $selectCount === 2 && $this->_selects[1]->function === 'COUNT' && is_scalar($this->_selects[0]->column) ) {
			// Simple COUNT request with a single GROUP BY combiner (REVERSED).
			// This should be returned as an associative row with each grouped column as the key of the count.
			$ret = [];

			if($this->_results !== null) {
				$rows = $this->_results;
			}
			elseif($this->_data !== null) {
				$rows =& $this->_data;
			}
			else {
				$rows = [];
			}

			foreach($rows as $d) {
				// Index them because I don't care what they're called.
				$indexed = array_values($d);
				$ret[ $indexed[0] ] = (int)$indexed[1];
			}
			return $ret;
		}
		elseif( $selectCount === 1 && $this->_selects[0]->column !== '*'){
			// Only one column was selected, just return an array of that column instead of an indexed array containing a 1-record associative array.
			$ret = [];

			if($this->_results !== null) {
				$rows = $this->_results;
			}
			elseif($this->_data !== null) {
				$rows =& $this->_data;
			}
			else {
				$rows = [];
			}

			$k = $this->_selects[0]->alias ? $this->_selects[0]->alias : $this->_selects[0]->column;
			foreach($rows as $d){
				$ret[] = isset($d[$k]) ? $d[$k] : null;
			}
			return $ret;
		}
		else{
			$ret = [];

			if($keyMap !== null && $selectCount === 2) {
				// Special case where a keymap was requested and only 2 columns are present.
				if($this->_selects[0]->alias === $keyMap || $this->_selects[0]->column === $keyMap) {
					$keyMap = $this->_selects[0]->alias ?? $this->_selects[0]->column;
					$keyValue = $this->_selects[1]->alias ?? $this->_selects[1]->column;
				}
				elseif($this->_selects[1]->alias === $keyMap || $this->_selects[1]->column === $keyMap) {
					$keyValue = $this->_selects[0]->alias ?? $this->_selects[0]->column;
					$keyMap = $this->_selects[1]->alias ?? $this->_selects[1]->column;
				}
				else {
					// Requested keymap not present, don't try to perform any mapping
					$keyMap = null;
					$keyValue = null;
				}
			}
			else {
				$keyValue = null;
			}

			if($this->_results !== null) {
				$rows = $this->_results;
			}
			elseif($this->_data !== null) {
				$rows =& $this->_data;
			}
			else {
				$rows = [];
			}

			foreach($rows as $d) {

				if($keyMap !== null && isset($d[ $keyMap ])) {
					if($keyValue !== null) {
						$ret[ $d[ $keyMap ] ] = $d[ $keyValue ];
					}
					else {
						$ret[ $d[ $keyMap ] ] = $d;
					}
				}
				else {
					$ret[] = $d;
				}
			}
			return $ret;
		}
	}

	/**
	 * Add a basic LEFT JOIN on this dataset
	 *
	 * @param string $table      Joining table name
	 * @param string $on         WHERE clause for this join, *USES KEYS FOR REFERENCES*
	 * @param string|null $alias Optional alias for this join
	 *
	 * @return $this
	 */
	private function _join(string $type, string $table, string $on, string $alias = null) {
		$join = new DatasetJoin();
		$join->type  = $type;
		$join->table = $table;
		$join->alias = $alias;

		$join->on = new DatasetWhereClause('root');
		$onClause = new DatasetWhere();
		$onClause->parseWhere($on, true);
		$join->on->addWhere($onClause);

		$this->_joins[] = $join;
	}
	
	/**
	 * Internal method to add a delete record to the list of to-be-deleted keys.
	 * 
	 * Usually this is just "*", but can be a specific key+value
	 * for noSQL/LDAP databases.
	 * 
	 * @param string $key
	 * @param mixed  $value
	 */
	private function _setDelete(string $key, $value) {
		if($key === '*' && $value === '*') {
			// Delete EVERYTHING
			$this->_deletes = [];
			$this->_deletes['*'] = '*';
		}
		elseif(!isset($this->_deletes[ $key ])) {
			// Not set yet
			$this->_deletes[$key] = $value;
		}
		elseif(is_array($this->_deletes[$key]) ) {
			// An existing array, append!
			$this->_deletes[$key][] = $value;
		}
		elseif(isset($this->_deletes[$key])) {
			// Not an array, but already present, convert and append.
			$this->_deletes[$key] = [
				$this->_deletes[$key],
				$value
			];
		}
	}

	/****  Iterator Methods *****/

	function rewind() {
		if($this->_results !== null) {
			// Migrate this data to legacy format first
			$this->_loadResultsToData();
		}

		if($this->_data !== null) {
			// Traditional fully loaded array
			reset($this->_data);
		}
	}

	function current() {
		if($this->_results !== null) {
			// Migrate this data to legacy format first
			$this->_loadResultsToData();
		}

		// If no data was selected before... I need to execute the query!
		if($this->_data === null) {
			$this->execute();
		}

		if($this->_data !== null) {
			$k = key($this->_data);
			return $this->_data[$k] ?? null;
		}
		else {
			return null;
		}
	}

	function key() {
		if($this->_results !== null) {
			// Migrate this data to legacy format first
			$this->_loadResultsToData();
		}

		// If no data was selected before... I need to execute the query!
		if($this->_data === null) $this->execute();

		return key($this->_data);
	}

	function next() {
		if($this->_results !== null) {
			// Migrate this data to legacy format first
			$this->_loadResultsToData();
		}

		// If no data was selected before... I need to execute the query!
		if($this->_data === null) $this->execute();

		next($this->_data);
	}

	function valid() {
		if($this->_results !== null) {
			// Migrate this data to legacy format first
			$this->_loadResultsToData();
		}

		// If no data was selected before... I need to execute the query!
		if($this->_data === null) $this->execute();

		return isset($this->_data[key($this->_data)]);
	}

	/**
	 * Internal method to map the results to a flat data
	 *
	 * Provides backwards compatibilty with accessing the array as just an array.
	 */
	private function _loadResultsToData() {
		if($this->_results !== null) {
			$this->_data = [];

			foreach($this->_results as $row) {
				$this->_data[] = $row;
			}

			$this->_results = null;
		}
	}




	/****  Static Methods  *****/

	/**
	 * Simple constructor that allows chaining.
	 * @return Dataset
	 */
	public static function Init(){

		// Allow chaining
		return new self();
	}
}
