<?php


namespace Core\Datamodel;


use Core\HealthCheckResult;

class SchemaComparison {
	private $_application;
	private $_database;
	private $_dbConnection;
	private $_checkKeys = ['type', 'typecast', 'required', 'maxlength', 'options', 'default', 'null', 'encoding'];


	public function __construct(Schema $application = null, Schema $database = null) {
		$this->_application = $application;
		$this->_database = $database;
		$this->_dbConnection = \Core\db();
	}

	/**
	 * Get a flat list of all columns defined in both Schemas merged together
	 *
	 * @return string[]
	 */
	public function getColumnNames() : array {
		if($this->_application && $this->_database) {
			return array_unique(
				array_keys($this->_application->definitions) + array_keys($this->_database->definitions)
			);
		}
		elseif($this->_application) {
			return array_keys($this->_application->definitions);
		}
		elseif($this->_database) {
			return array_keys($this->_database->definitions);
		}
		else {
			return [];
		}
	}

	/**
	 * Get all tracked column parameters and their values
	 *
	 * Each key checked will be the index of the returned associative array, with
	 * sub-index 0: application value (as string)
	 * sub-index 1: database value (as string)
	 * sub-index 2: comparison matches if true (as bool)
	 *
	 * @param string $columnName
	 * @return array
	 */
	public function getColumnParameters(string $columnName) : array {
		$ret = [];
		$app = $this->_application && isset($this->_application->definitions[$columnName]) ?
			$this->_application->definitions[$columnName] :
			null;
		$db = $this->_database && isset($this->_database->definitions[$columnName]) ?
			$this->_database->definitions[$columnName] :
			null;

		foreach($this->_checkKeys as $key) {
			if($app && $db) {
				$ret[$key] = [
					$this->_describeParameterValue($app->$key),
					$this->_describeParameterValue($db->$key),
					($app->$key === $db->$key)
				];
			}
			elseif($app) {
				$ret[$key] = [
					$this->_describeParameterValue($app->$key),
					'NOT PRESENT',
					false
				];
			}
			elseif($db) {
				$ret[$key] = [
					'NOT PRESENT',
					$this->_describeParameterValue($db->$key),
					false
				];
			}
			else {
				$ret[$key] = [
					'NOT PRESENT',
					'NOT PRESENT',
					true
				];
			}
		}

		return $ret;
	}

	/**
	 * Get the column strings,
	 *
	 * index 0: application
	 * index 1: database
	 *
	 * @param string $columnName
	 * @return string[]
	 */
	public function getColumnStrings(string $columnName) : array {
		$app = $this->_application && isset($this->_application->definitions[$columnName]) ?
			$this->_application->definitions[$columnName] :
			null;
		$db = $this->_database && isset($this->_database->definitions[$columnName]) ?
			$this->_database->definitions[$columnName] :
			null;

		if($app && $db) {
			return [
				$this->_dbConnection->parseColumnString($app),
				$this->_dbConnection->parseColumnString($db),
			];
		}
		elseif($app) {
			return [
				$this->_dbConnection->parseColumnString($app),
				'',
			];
		}
		elseif($db) {
			return [
				'',
				$this->_dbConnection->parseColumnString($db),
			];
		}
		else {
			return ['', ''];
		}
	}

	/**
	 * Get the column comment string, pulled from either the application or database
	 *
	 * @param string $columnName
	 * @return string
	 */
	public function getColumnComment(string $columnName) : string {
		$app = $this->_application && isset($this->_application->definitions[$columnName]) ?
			$this->_application->definitions[$columnName] :
			null;
		$db = $this->_database && isset($this->_database->definitions[$columnName]) ?
			$this->_database->definitions[$columnName] :
			null;

		if($app) {
			return $app->comment;
		}
		elseif($db) {
			return $db->comment;
		}
		else {
			return '';
		}
	}

	/**
	 * Get health checks for a column that apply to the column as a whole
	 *
	 * These are high level notices generally, like missing default value, not present in A or B, etc.
	 *
	 * @param string $columnName
	 * @return HealthCheckResult[]
	 */
	public function getColumnGeneralHealthChecks(string $columnName) : array {
		$ret = [];
		$app = $this->_application && isset($this->_application->definitions[$columnName]) ?
			$this->_application->definitions[$columnName] :
			null;
		$db = $this->_database && isset($this->_database->definitions[$columnName]) ?
			$this->_database->definitions[$columnName] :
			null;

		// Check for if the column is present in both sources
		if($app && $db) {
			$ret[] = HealthCheckResult::ConstructGood('Column is present in application and database.');
		}
		elseif($app && !$db) {
			$ret[] = HealthCheckResult::ConstructError(
				'Column is present in application but not database!',
				'Reinstalling/Resyncing Core should fix this issue.'
			);
		}
		elseif(!$app && $db) {
			$ret[] = HealthCheckResult::ConstructWarn('Column is present in database but not application!',
				'This may happen when a component which adds functionality to another is uninstalled.
				These fields are not automatically removed to preserve any existing data, but can be safely removed if no longer needed.'
			);
		}

		// Check the default value
		if($app) {
			if($app->default !== false) {
				$ret[] = HealthCheckResult::ConstructGood('Column has a default value');
			}
			elseif(isset($this->_application->indexes['primary']) && in_array($columnName, $this->_application->indexes['primary'])) {
				$ret[] = HealthCheckResult::ConstructGood('Column has no default value, but is the primary key');
			}
			else {
				$ret[] = HealthCheckResult::ConstructWarn(
					'Column has no default value',
					'Not having a default value could cause issues if this column is later removed from the application
					without being removed from the database.'
				);
			}
		}
		elseif($db) {
			if($db->default === false) {
				$ret[] = HealthCheckResult::ConstructError(
					'Column has no default value',
					'Not having a default value will cause issues since this column has been removed from the application
					without being removed from the database.'
				);
			}
		}

		// Check that the default is not set to NULL if NULL is disallowed
		if($app) {
			if($app->default === null && !$app->null) {
				$ret[] = HealthCheckResult::ConstructError(
					'Column default set to NULL but NULL is not allowed'
				);
			}
		}
		elseif($db) {
			if($db->default === null && !$db->null) {
				$ret[] = HealthCheckResult::ConstructError(
					'Column default set to NULL but NULL is not allowed'
				);
			}
		}

		return $ret;
	}

	/**
	 * Get the count of all issues on this Schema comparison
	 *
	 * @return int
	 */
	public function getIssuesCount() : int {
		$issues = 0;

		foreach($this->getColumnNames() as $columnName) {
			foreach($this->getColumnGeneralHealthChecks($columnName) as $check) {
				if(
					$check->result === HealthCheckResult::RESULT_WARN ||
					$check->result === HealthCheckResult::RESULT_ERROR
				) {
					$issues++;
				}
			}
			foreach($this->getColumnParameters($columnName) as $parameter) {
				if(!$parameter[2]) {
					$issues++;
				}
			}
		}

		return $issues;
	}

	private function _describeParameterValue($value) : string {
		if($value === null) {
			return 'NULL';
		}
		elseif($value === false) {
			return 'FALSE';
		}
		elseif($value === true) {
			return 'TRUE';
		}
		elseif(is_scalar($value) && $value === '') {
			return '<em>string</em> (empty string)';
		}
		elseif(is_int($value)) {
			return '<em>int</em> ' . $value;
		}
		elseif(is_scalar($value)) {
			return '<em>string</em> ' . $value;
		}
		else {
			return 'unknown';
		}
	}
}