<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Filestore;

use Core\ConfigHandler;
use Core\Logger\LogSinkInterface;
use Core\Logger\LogEntry;

/**
 * Description of FileLogger
 *
 * @author charlie
 */
class FileLogger implements LogSinkInterface {
	
	/**
	 * Cache of files to use for logging.
	 * 
	 * This is used so that the file object does not need to be instantiated on each
	 * and every individual log entry.
	 * 
	 * @var array <<File>>
	 */
	private $_logFiles = [];
	
	/**
	 * Get the default level for this sink
	 * 
	 * @return int
	 */
	public function getDefaultLevel() : int {
		$level = ConfigHandler::Get('/core/logs/level');

		if($level === null) {
			// Level is not set yet, this usually happens on installation.
			// We still want something.
			$level = LOG_LEVEL_INFO;
		}
		
		return (int)$level;
	}
	
	/**
	 * Get the maximum level (lowest priority), that this sink supports.
	 * 
	 * @return int
	 */
	public function getMaximumLevel() : int{
		return LOG_LEVEL_TRACE; // Supports full trace levels.
	}
	
	/**
	 * Record an entry into this sink's destination.
	 * 
	 * @param \Core\Logger\LogEntry $entry
	 * 
	 * @throws \Exception
	 * 
	 */
	public function log(LogEntry $entry){
		$outfile = $this->_getLogFile($entry->type);
		
		// Generate a nice line header to prepend on the line.
		$header = '[' . \Core\Date\DateTime::Now('r', \Core\Date\Timezone::TIMEZONE_DEFAULT) . '] ';
		$multiHeader = $header;
		
		// Convert the level of this entry to something that admins may understand.
		switch($entry->level){
			case LOG_LEVEL_CRITICAL:
				$level = '<CRIT>';
				break;
			case LOG_LEVEL_HIGH:
				$level = '<HIGH>';
				break;
			case LOG_LEVEL_INFO:
				$level = '<INFO>';
				break;
			case LOG_LEVEL_DEBUG:
				$level = '<DBUG>';
				break;
			case LOG_LEVEL_TRACE:
				$level = '<TRCE>';
				break;
			default:
				$level = '<UNKN>';
		}
		$header .= $level . ' ';

		if($entry->code){
			$header .= $entry->code . ': ';
		}

		// Build the environmental string
		$header .= ((EXEC_MODE === 'WEB') ? REMOTE_IP : '[CLI]') . ' ';

		if(isset($_SERVER['REQUEST_URI'])) {
			$header .= $_SERVER['REQUEST_URI'] . ' ';
		}

		if(isset($_SERVER['HTTP_USER_AGENT'])) {
			$header .= '"' . $_SERVER['HTTP_USER_AGENT'] . '" ';
		}

		foreach($entry->tags as $k => $v) {
			$header .= '[' . $k . ': ' . $v . '] ';
		}

		$logfh = fopen($outfile, 'a');
		if(!$logfh){
			throw new \Exception('Unable to open ' . $outfile . ' for appending!');
		}
		
		// Convert the message to something that is text-only friendly.
		$lines = $this->_convertMessageToText($entry->message);
		$c = 0;
		foreach($lines as $line){
			$c++;
			if($c == 1) {
				fwrite($logfh, $header . $line . "\n");
			}
			else {
				fwrite($logfh, $multiHeader . $line . "\n");
			}
		}
		
		if($entry->caller) {
			fwrite($logfh, $multiHeader . $entry->caller . "\n");
		}
		
		if($entry->trace) {
			$c = count($entry->trace);
			for($i = 0; $i <= 4 && $i < $c; $i++) {
				fwrite($logfh, $multiHeader . $entry->trace[$i]['formatted'] . "\n");
			}
		}
		
		fclose($logfh);
	}

	/**
	 * Manual logger to record custom data to a log file
	 *
	 * NOT usually used, please use log() in most cases.
	 *
	 * @param string $line
	 * @param string $type
	 *
	 * @throws \Exception
	 */
	public function logLine(string $line, string $type) {
		$this->logLines([ $line ], $type);
	}

	/**
	 * Manual logger to record custom data to a log file
	 *
	 * NOT usually used, please use log() in most cases.
	 *
	 * @param string[] $lines
	 * @param string $type
	 *
	 * @throws \Exception
	 */
	public function logLines(array $lines, string $type) {
		$outfile = $this->_getLogFile($type);

		// Generate a nice line header to prepend on the line.
		$header = '[' . \Core\Date\DateTime::Now('r', \Core\Date\Timezone::TIMEZONE_DEFAULT) . '] ';

		$logfh = fopen($outfile, 'a');
		if(!$logfh){
			throw new \Exception('Unable to open ' . $outfile . ' for appending!');
		}

		foreach($lines as $line) {
			fwrite($logfh, $header . $line . "\n");
		}

		fclose($logfh);
	}
	
	private function _convertMessageToText(string $message) : array {
		// Convert <br/> and <br> to newlines
		$message = str_replace('<br/>', "\n", $message);
		$message = str_replace('<br>', "\n", $message);
		
		// Remove HTML strings
		$message = preg_replace('/<[^>]*>/im', '', $message);
		
		// And convert HTML codes to their character.
		$message = html_entity_decode($message);
		
		// Return an array so the logger can iterate over each and prepend the header.
		return explode("\n", $message);
	}
	
	/**
	 * Get the necessary File to handle logging of the requested type.
	 * 
	 * This will return the fully resolved filename to write to.
	 * 
	 * @param string $type
	 * 
	 * @throws \Exception
	 * 
	 * @return string
	 */
	private function _getLogFile(string $type): string {
		if(!isset($this->_logFiles[ $type ])){
			
			// Ensure the directory exists and is setup.
			if(!is_dir(ROOT_PDIR . 'logs/')){
				$this->_setupDirectory();
			}
			
			// Make sure it contains only valid characters!
			$filebase = preg_replace('/[^a-z0-9\-]/', '', str_replace(' ', '-', strtolower($type)));

			if(!$filebase){
				$filebase = 'unknown';
			}

			$filename = 'logs/' . $filebase . '.log';
			
			// Ensure that something is here for appending.
			if(!file_exists(ROOT_PDIR . $filename)){
				touch(ROOT_PDIR . $filename);
			}
			
			// I can now load the new file and cache it.
			$this->_logFiles[ $type ] = ROOT_PDIR . $filename;
		}
		
		return $this->_logFiles[ $type ];
	}
	
	/**
	 * Try to auto-create the logs directory.
	 * 
	 * @throws \Exception
	 */
	private function _setupDirectory(){
		// Will attempt an auto-creation of the directory and .htaccess file.
		if(!is_writable(ROOT_PDIR)){
			// Not writable, no log!
			throw new \Exception('Unable create logs/ directory, please ensure that ' . ROOT_PDIR . ' can be written to or create logs/ manually.');
		}
		
		mkdir(ROOT_PDIR . 'logs/');
		
		// Use Core's built-in function to install the necessary .htaccess in this directory.
		// @todo retool this
		//\Core\install_directory_protection(ROOT_PDIR . 'logs/', 'private');
	}
}
