<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Filestore\Contents;

use Core\Filestore\ContentsInterface;
use Core\Filestore\Factory;
use Core\Filestore\File;
use Core\Filestore\FileHelper;
use Core\Parser\JPEGExif\JPEGExifParser;


/**
 * Description of ContentImage
 *
 * @author charlie
 */
class ContentImage implements ContentsInterface {
	
	private $_file;
	
	private $_currentWidth;
	private $_currentHeight;
	private $_type;
	
	private $_resizeWidth;
	private $_resizeHeight;
	private $_resizeMode;
	
	/**
	 * @var int Normal resizing; aspect ratio maintained and largest dimension used
	 */
	const RESIZE_MODE_NORMAL   = 1;
	/**
	 * @var int Smallest resizing; aspect ratio maintained but smallest matching dimension used
	 */
	const RESIZE_MODE_FILL = 2;
	/**
	 * @var int Force new size regardless of aspect ratio
	 */
	const RESIZE_MODE_ABSOLUTE = 3;
	/**
	 * @var int Only increase small images, (if larger, leave alone)
	 */
	const RESIZE_MODE_INCREASE = 4;
	/**
	 * @var int Only decrease large images, (if smaller, leave alone)
	 */
	const RESIZE_MODE_DECRAESE = 5;
	
	public function __construct(File $file) {
		$this->_file = $file;
		
		if($this->_file->exists()){
			$this->_loadCurrentDimensions();
		}
	}
	
	/**
	 * Set some requested resize dimensions for calculations on this image
	 * 
	 * If dimensions is a single int, that is used as the calculated width and height.
	 * If dimensions is a string containing NxN, that is transposed to width x height.
	 * If dimensions contains a mode operator, that is used in the calculations.
	 * 
	 * @param string|int $dimensions Requested dimensions to check against
	 * 
	 * @return $this
	 */
	public function setResizeDimensions($dimensions) : ContentImage {
		if (is_numeric($dimensions)) {
			$dimensions = $dimensions . 'x' . $dimensions;
		}
		
		
		if(strpos($dimensions, '^') !== false){
			// Fit the smallest dimension instead of the largest, (useful for overflow tricks)
			$this->_resizeMode = self::RESIZE_MODE_FILL;
			$dimensions = str_replace('^', '', $dimensions);
		}
		elseif(strpos($dimensions, '!') !== false){
			// Absolutely resize, regardless of aspect ratio
			$this->_resizeMode = self::RESIZE_MODE_ABSOLUTE;
			$dimensions = str_replace('!', '', $dimensions);
		}
		elseif(strpos($dimensions, '>') !== false){
			// Only increase images.
			$this->_resizeMode = self::RESIZE_MODE_INCREASE;
			$dimensions = str_replace('>', '', $dimensions);
		}
		elseif(strpos($dimensions, '<') !== false){
			// Only decrease images.
			$this->_resizeMode = self::RESIZE_MODE_DECRAESE;
			$dimensions = str_replace('<', '', $dimensions);
		}
		else{
			// Default mode
			$this->_resizeMode = self::RESIZE_MODE_NORMAL;
		}
		
		
		// Split on the "x" and that should give me the width/height.
		$vals   = explode('x', strtolower($dimensions));
		$this->_resizeWidth  = (int)$vals[0];
		$this->_resizeHeight = (int)$vals[1];
		
		// Allow chaining
		return $this;
	}
	
	
	/**
	 * Check if a resize is necessary given the resize dimensions
	 * 
	 * @return bool
	 */
	public function isResizeNecessary() : bool {
		if(
			$this->_resizeHeight === null || 
			$this->_resizeMode === null
		){
			// No resize data available or not set yet; no resizing ability present.
			return false;
		}
		
		$dims = $this->getResizeDimensions();
		return ($dims[0] !== $this->_currentWidth || $dims[1] !== $this->_currentHeight);
	}
	
	/**
	 * Get the resized dimensions for this image
	 * 
	 * @return array <<int,int>> List of width x height of the resized image
	 */
	public function getResizeDimensions() : array {
		/**
		 * Ratio of width conversion
		 * if current is 1000 and new is 100,  ratio is 0.1
		 * if current is 1000 and new is 2000, ratio is 2.0
		 */
		$ratioW = $this->_resizeWidth / $this->_currentWidth;
		
		/**
		 * Ratio of height conversion
		 * if current is 1000 and new is 100,  ratio is 0.1
		 * if current is 1000 and new is 2000, ratio is 2.0
		 */
		$ratioH = $this->_resizeHeight / $this->_currentHeight;
		
		/**
		 * Key of difference values between the differences of width and differences of height
		 * 
		 * used to determine which dimension to use directly and which to calculate.
		 */
		$ratioDiff = $ratioW <=> $ratioH;
		
		if(
			$this->_resizeMode === self::RESIZE_MODE_ABSOLUTE
		){
			// Absolute size was requested!
			return [$this->_resizeWidth, $this->_resizeHeight];
		}
		elseif(
			($this->_resizeMode === self::RESIZE_MODE_DECRAESE && $ratioW >= 1 && $ratioH >= 1) ||
			($this->_resizeMode === self::RESIZE_MODE_INCREASE && $ratioW <= 1 && $ratioH <= 1)
		){
			// Image already smaller/larger in all dimensions, no resizing needed.
			return [ $this->_currentWidth, $this->_currentHeight ];
		}
		elseif(
			($this->_resizeMode === self::RESIZE_MODE_FILL     && $ratioDiff === -1) ||
			($this->_resizeMode === self::RESIZE_MODE_NORMAL   && $ratioDiff === 1) ||
			($this->_resizeMode === self::RESIZE_MODE_DECRAESE && $ratioDiff === 1 && $ratioH < 1) ||
			($this->_resizeMode === self::RESIZE_MODE_INCREASE && $ratioDiff === 1 && $ratioH > 1)
		){
			// Resize diff is greater on height, so use width
			// Ensure that width is at least 1px.
			return [ 
				max(1, (int)round($this->_currentWidth * $ratioH)),
				$this->_resizeHeight
			];
		}
		elseif(
			($this->_resizeMode === self::RESIZE_MODE_FILL     && $ratioDiff === 1) ||
			($this->_resizeMode === self::RESIZE_MODE_NORMAL   && $ratioDiff === -1) ||
			($this->_resizeMode === self::RESIZE_MODE_DECRAESE && $ratioDiff === -1 && $ratioW < 1) ||
			($this->_resizeMode === self::RESIZE_MODE_INCREASE && $ratioDiff === -1 && $ratioW > 1)
		){
			// Resize diff is greater on width, so use height
			// Ensure that height is at least 1px.
			return [
				$this->_resizeWidth,
				max(1, (int)round($this->_currentHeight * $ratioW))
			];
		}
		else{
			// Resize diff is the same, no aspect ratio conversion needed
			return [ $this->_resizeWidth, $this->_resizeHeight ];
		}
	}
	
	/**
	 * Get the file *WITHOUT* resizing the actual contents.
	 * 
	 * This file is NOT guaranteed to exist!
	 * 
	 * @return File
	 */
	public function getResizeFile() : File {
		list($width, $height) = $this->getResizeDimensions();
		$mimetype = $this->_file->getMimetype();
		
		$ext = $this->_file->getExtension();
		// Ensure that an extension is used if none present, (may happen with temporary files).
		if(!$ext){
			$ext = FileHelper::GuessExtFromMime($mimetype);
		}
		$fileBase = \Core\str_to_url($this->_file->getBasename(true));

		// If this file already has a dimension hash attached, drop that off the end.
		if(preg_match('/-[0-9]*x[0-9]*$/', $fileBase)){
			$fileBase = preg_replace('/-[0-9]*x[0-9]*$/', '', $fileBase);
		}

		// Ensure this isn't too long, just in case.
		if(strlen($fileBase) > 42){
			$fileBase = substr($fileBase, 0, 42);
		}

		$modes = [
			0 => '-na',
			1 => '',
			2 => '-f',
			3 => '-a',
			4 => '-i',
			5 => '-d',
		];
		$mode = $modes[ $this->_resizeMode ];

		$sizeKey = '-' . $width . 'x' . $height . $mode;
		
		// The basename is for SEO purposes, that way even resized images still contain the filename.
		// The hash is just to ensure that no two files conflict, ie: /public/a/file1.png and /public/b/file1.png
		//  might conflict without this hash.
		// Finally, the width and height dimensions are there just because as well; it gives more of a human
		//  touch to the file. :p
		// Also, keep the original file extension, this way PNGs remain PNGs, GIFs remain GIFs, JPEGs remain JPEGs.
		// This is critical particularly when it comes to animated GIFs.
		$key = $fileBase . '-' . $this->_file->getHash() . $sizeKey . '.' . $ext;
		
		
		// New destination for the resulting resized file.
		return Factory::File('public/tmp/' . $key);
	}
	
	public function getDimensions() : string {
		return $this->_currentWidth . 'x' . $this->_currentHeight;
	}
	
	/**
	 * Get the exif data of this image.
	 *
	 * @return JPEGExifParser
	 */
	public function getExif() : JPEGExifParser {
		$parser = new JPEGExifParser();
		$parser->parseFile($this->_file);
		return $parser;
	}
	
	/**
	 * Resize this image and save the output as another File object.
	 *
	 * This is used on conjunction with getPreview* and getQuickPreview.
	 * QuickPreview creates the destination file in the correct directory
	 * and getPreview* methods request the actual resizing.
	 */
	public function resize() : File {

		if(!$this->isResizeNecessary()){
			// No resize needed, just return the original File.
			return $this->_file;
		}
		
		// Use the helper method to get the File object.
		// This will contain all the necessary file keys for the requested resize mode and size.
		$file = $this->getResizeFile();
		
		list($width, $height) = $this->getResizeDimensions();
		$mimetype = $this->_file->getMimetype();

		// TIME SAVER!
		// Check if the file exists and has content in it.
		// This allows the modified time to be accurate, saving client time
		// AND saves server time in resizing files that have already been resized.
		// The filesize check is a safegaurd against errors in the resizing that may result in a 0-byte file.
		if($file->exists() && $file->getFilesize() > 0) {
			return $file;
		}

		// Make sure the directory of the destination file exists!
		// By touching the file, Core will create all parent directories as necessary.
		$file->putContents('');

		
		if($mimetype === 'image/gif' && exec('which convert 2>/dev/null')){
			// The GIF resizer handles EVERYTHING :)
			// Granted of course, that imagemagick's convert is available on the server.
			// This has the cool effect of preserving GIF animation.
			
			$resize = escapeshellarg($width . 'x' . $height);
			exec('convert ' . escapeshellarg($this->_file->getFilename()) . ' -resize ' . $resize . ' ' . escapeshellarg($file->getFilename()));

			return $file;
		}
		
		// Traditional resizing logic.
		$srcWidth = $this->_currentWidth;
		$srcHeight = $this->_currentHeight;
		
		switch ($mimetype) {
			case 'image/jpeg':
				$thumbType = 'JPEG';
				if($width <= 200 && $height <= 200 && function_exists('exif_thumbnail')){
					// Try to write out from the thumbnail img instead of the full size.
					// This is done to increase server performance.
					// eg: resizing a 5MB JPEG can take upwards of 50-100ms,
					// whereas the embedded thumbnail will take only 2-10ms.
					// Not to mention professional JPEG management tools such as PS and Gimp
					// produce marginally higher-quality thumbnails than GD will.
					// (The resulting filesize is negligible.)
					// Of course if the requested image is larger than a thumbnail size, (200x200 in this case),
					// using the thumbnail is counter-productive!
					$thumbWidth = $width;
					$thumbHeight = $height;
					$img = exif_thumbnail($this->_file->getFilename(), $thumbWidth, $thumbHeight, $thumbType);
					if($img){
						\Core\log_trace('JPEG has thumbnail data of ' . $thumbWidth . 'x' . $thumbHeight . '!');
						$file->putContents($img);
						$srcWidth = $thumbWidth;
						$srcHeight = $thumbHeight;
						$img = imagecreatefromjpeg($file->getFilename());
					}
					else{
						$img = imagecreatefromjpeg($this->_file->getFilename());
					}
				}
				else{
					$img = imagecreatefromjpeg($this->_file->getFilename());
				}

				break;
			case 'image/png':
				$img = imagecreatefrompng($this->_file->getFilename());
				break;
			case 'image/gif':
				$img = imagecreatefromgif($this->_file->getFilename());
				break;
			default:
				// Hmmm...
				\Core\log_trace('Resizing complete (failed, not sure what it was)');
				return $this->_file;
		}
		
		// Do some trickery here as opposed than just resizing the image to perserve transparency.
		$img2 = imagecreatetruecolor($width, $height);
		imagealphablending($img2, false);
		imagesavealpha($img2, true);
		imagealphablending($img, true);
		// Assign a transparency color.
		//$trans = imagecolorallocatealpha($img2, 0, 0, 0, 0);
		//imagefill($img2, 0, 0, $trans);
		imagecopyresampled($img2, $img, 0, 0, 0, 0, $width, $height, $srcWidth, $srcHeight);
		imagedestroy($img);


		switch ($mimetype) {
			case 'image/jpeg':
				imagejpeg($img2, $file->getFilename(), 60);
				\Core\log_trace('Resizing complete (via imagejpeg)');
				break;
			case 'image/png':
				imagepng($img2, $file->getFilename(), 9);
				\Core\log_trace('Resizing complete (via imagepng)');
				break;
			case 'image/gif':
				imagegif($img2, $file->getFilename());
				\Core\log_trace('Resizing complete (via imagegif)');
				break;
			default:
				// Hmmm...
				\Core\log_trace('Resizing complete (failed, not sure what it was)');
				return $this->_file;
		}
		
		// Newly resized image!
		return $file;
	}
	
	private function _loadCurrentDimensions(){
		$data = getimagesize($this->_file->getFilename());
		if($data !== false){
			$this->_currentWidth  = (int)$data[0];
			$this->_currentHeight = (int)$data[1];
			$this->_type          = (int)$data[2];
		}
	}
}
