<?php
/**
 * Description of ContentTGZ
 *
 * Provides useful extra functions that can be done with a GZipped file.
 *
 * @package
 * @since 0.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */


namespace Core\Filestore\Contents;

use Core\Filestore\Directory;
use Core\Filestore\Factory;
use Core\Filestore\ContentsInterface;
use Core\Filestore\File;
use Exception;
use mikehaertl\shellcommand\Command;

class ContentTAR implements ContentsInterface {
	protected $_file;
	protected $_compressionFlag;

	public function __construct(File $file) {
		$this->_file = $file;
	}

	/**
	 * Extract this archive to a requested directory.
	 *
	 * @param string|Directory $dst Destination to extract the archive to.
	 *
	 * @throws Exception
	 *
	 * @return Directory
	 */
	public function extract($destdir) : Directory {
		if($destdir instanceof Directory) {
			$d = $destdir;
		}
		else {
			// This will ensure that the destdir is properly resolved.
			$d = Factory::Directory($destdir);
		}

		if(!$d->exists()) {
			$d->mkdir();
		}

		/*
		if(!$d->isWritable()) {
			throw new Exception($d->getPath() . ' is not writable, unable to extract tarball.');
		}*/

		$cmd = $this->_getCommand();

		// Command to perform
		$cmd->addArgKey('--extract');

		// Destination
		$cmd->addArgKeyValue('--directory', $d->getPath());

		$cmd->execute();

		if($cmd->getError()) {
			throw new Exception($cmd->getError());
		}
		else {
			return $d;
		}
	}

	/**
	 * Extract a single file from the tarball
	 *
	 * The contents of the file is returned unless destination is requested instead.
	 *
	 * @param string $filename
	 * @param string|null $destination
	 *
	 * @return string
	 *
	 * @throws Exception
	 */
	public function extractSingleFile(string $filename, string $destination = null) {
		$cmd = $this->_getCommand();

		// Command to perform
		$cmd->addArgKey('--extract');

		// Requested file to extract
		$cmd->addArgValue($filename);

		if($destination) {
			// Redirect stdout on the shell to a destination file.
			// We do this here because the extracted file could be HUGE, and we don't want to
			// hassle the web server with having to load all that data just to write it back out
			// to disk.
			$cmd->addArgKey('-O > ' . escapeshellarg($destination));
		}
		else {
			// No destination was requested, so retrieve the contents of the file directly.
			$cmd->addArgKey('-O');
		}

		$cmd->execute();

		if($cmd->getExitCode() !== 0) {
			throw new Exception($cmd->getError());
		}
		else {
			return $cmd->getOutput(false);
		}
	}

	public function listfiles() : array {
		$cmd = $this->_getCommand();

		// Command to perform
		$cmd->addArgKey('--list');

		$cmd->execute();

		return explode("\n", $cmd->getOutput());
	}

	/**
	 * Get the base of the command, with the file and compression mode already set.
	 *
	 * @return Command
	 */
	protected function _getCommand() : Command {
		$cmd = new Command('tar');

		// Compression option
		// bzip2
		// xz (.tar.xz)
		// lzip
		// lzma
		// lzop
		// gzip (.tgz, .tar.gz)
		// compress (.tar.Z)
		if($this->_compressionFlag) {
			$cmd->addArgKey('--' . $this->_compressionFlag);
		}

		// Source file option
		$cmd->addArgKeyValue('--file', $this->_file->getLocalFilename());

		return $cmd;
	}
}

