<?php


namespace Core\Filestore\Contents;


use Core\Filestore\ContentsInterface;
use Core\Filestore\File;
use mikehaertl\shellcommand\Command;

class ContentVideo implements ContentsInterface {

	private $_file;
	private $_metadata;

	public function __construct(File $file) {
		$this->_file = $file;

		$this->_metadata = $this->_getData();

		//var_dump($this->_metadata['streams']); die();
	}

	/**
	 * Get the duration of this video
	 *
	 * Format will be HH:MM:SS.ss or NULL if not available
	 *
	 * @return string|null
	 */
	public function getDuration() {
		return (isset($this->_metadata['root']['duration'])) ? $this->_metadata['root']['duration'] : null;
	}

	/**
	 * Get the overall bitrate of this video
	 *
	 * @return string|null
	 */
	public function getBitrate() {
		return (isset($this->_metadata['root']['bitrate'])) ? $this->_metadata['root']['bitrate'] : null;
	}

	public function getStreams() {
		return $this->_metadata['streams'];
	}

	private function _getData() : array {
		// This can only run if ffmpeg is installed
		$ret = null;
		$out = null;
		exec('which ffmpeg', $out, $ret);

		if($ret !== 0) {
			return [
				'root' => [],
				'streams' => [],
			];
		}

		$command = new Command('ffmpeg');
		$command->addArg('-i', $this->_file->getFilename());
		$command->execute();

		// FFMPEG will print results to stderr
		$output = explode("\n", $command->getError(true));
		$parse = false;

		// Strip the last line out; this will always be "At least one output file must be specified"
		// because we're only collecting meta data.
		array_pop($output);

		$streams = [];
		$root = [];
		$last =& $root;
		$inMeta = false;

		foreach($output as $line) {
			// Do not start processing output until the Input is detected.
			if(strpos($line, 'Input #0') === 0) {
				$parse = true;
				continue;
			}

			if(!$parse) {
				continue;
			}

			if(trim($line) === 'Metadata:') {
				$inMeta = true;
				$last['meta'] = [];
			}
			elseif(strpos(trim($line), 'Duration: ') === 0) {
				// Duration: 01:12:55.21, start: 0.000000, bitrate: 1416 kb/s
				$inMeta = false;
				$parts = array_map('trim', explode(',', $line));
				foreach($parts as $part) {
					$key = strtolower(trim(substr($part, 0, strpos($part, ':'))));
					$val = trim(substr($part, strpos($part, ':') + 1));

					switch($key) {
						case 'duration':
							$root['duration'] = $val;
							break;
						case 'bitrate':
							$root['bitrate'] = $val;
							break;
					}
				}
			}
			elseif(strpos(trim($line), 'Stream #') === 0) {
				// Stream #0:0(und): Video: h264 (High) (avc1 / 0x31637661), yuv420p(tv, bt709), 1280x720, 1158 kb/s, 23.98 fps, 23.98 tbr, 24k tbn, 48k tbc (default)
				$inMeta = false;
				$line = substr(trim($line), 8);

				$streamID = substr($line, 0, strpos($line, ' ') - 1);
				$line = trim(substr($line, strlen($streamID) + 1));
				// Trim the comment
				$streamID = preg_replace('/\([^\)]*\)/', '', $streamID);

				$type = strtolower(substr($line, 0, strpos($line, ':')));
				$line = trim(substr($line, strlen($type) + 1));

				$codec = substr($line, 0, strpos($line, ','));
				$line = trim(substr($line, strlen($codec) + 1));

				$format = null;
				$dimensions = null;
				$bitrate = null;
				$channels = null;
				$frequency = null;

				if($type === 'video') {
					// Next section is format and resolution for video
					$parts = [];
					preg_match('/^([a-zA-Z0-9 ,\(\)]*), ([0-9]*x[0-9]*),.*/', $line, $parts);
					$format = $parts[1];
					$dimensions = $parts[2];
					$line = substr($line, strlen($format) + strlen($dimensions) + 4);
				}

				// Trim the (default) flag off the end if it's present.
				$line = trim(str_replace('(default)', '', $line));

				$flags = array_map('trim', explode(',', $line));

				foreach($flags as $flag) {
					if(preg_match('/^[0-9\.]* [kKmMgGbB]*\/s$/', $flag)) {
						$bitrate = $flag;
					}
					elseif($type === 'video' && strpos($flag, ' fps') !== false) {
						$frequency = $flag;
					}
					elseif($type === 'audio' && strpos($flag, ' Hz') !== false) {
						$frequency = $flag;
					}
					elseif($type === 'audio' && ($flag === 'stereo' || $flag === 'mono')) {
						$channels = $flag;
					}
				}

				$stream = [
					'id' => $streamID,
					'type' => $type,
					'codec' => $codec,
					'format' => $format,
					'dimensions' => $dimensions,
					'bitrate' => $bitrate,
					'frequency' => $frequency,
					'channels' => $channels,
				];

				$streams[] = $stream;
				$last =& $streams[ count($streams) - 1 ];
			}
			elseif($inMeta) {
				$parts = explode(':', $line);
				$last['meta'][ trim($parts[0]) ] = trim($parts[1]);
			}
		}

		return [
			'root' => $root,
			'streams' => $streams,
		];
	}
}