<?php
/**
 * File for class Key definition in the coreplus project
 * 
 * @package Core\GPG
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140319.1923
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\GPG;


/**
 * A short teaser of what Key does.
 *
 * More lengthy description of what Key does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for Key
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @package Core\GPG
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class Key {

	/**
	 * o = Unknown (this key is new to the system)
	 * i = The key is invalid (e.g. due to a missing self-signature)
	 * d = The key has been disabled (deprecated - use the 'D' in field 12 instead)
	 * r = The key has been revoked
	 * e = The key has expired
	 * - = Unknown validity (i.e. no value assigned)
	 * q = Undefined validity
	 * '-' and 'q' may safely be treated as the same value for most purposes
	 * n = The key is valid
	 * m = The key is marginal valid.
	 * f = The key is fully valid
	 * u = The key is ultimately valid.  This often means that the secret key is available, but any key may be marked as ultimately valid.
	 *
	 * @var string
	 */
	public $validity;

	/**
	 * @var int
	 */
	public $encryptionBits;

	/**
	 * @var string
	 */
	public $encryptionType;

	/**
	 * @var string The full ID of this key
	 */
	public $id;

	/**
	 * @var string The short ID of this key
	 */
	public $id_short;

	/**
	 * @var string Serial number of this key
	 */
	public $serial;

	/**
	 * @var string The fingerprint of this key
	 */
	public $fingerprint;

	/**
	 * @var int Created date in UTC timestamp
	 */
	public $created;

	/**
	 * @var int Expiration date in UTC timestamp
	 */
	public $expires;
	
	/**
	 * @var int Timestamp of the last update in UTC timestamp
	 */
	public $updated;
	
	

	const ENCRYPTION_TYPE_RSA = 'RSA';
	const ENCRYPTION_TYPE_DSA = 'DSA';
	const ENCRYPTION_TYPE_ELGAMAL = 'Elgamal';

	/**
	 * Check and see if this key is currently valid and not expired nor revoked.
	 *
	 * @return bool
	 */
	public function isValid(){
		/*
		 * o = Unknown (this key is new to the system)
		 * i = The key is invalid (e.g. due to a missing self-signature)
		 * d = The key has been disabled (deprecated - use the 'D' in field 12 instead)
		 * r = The key has been revoked
		 * e = The key has expired
		 * - = Unknown validity (i.e. no value assigned)
		 * q = Undefined validity
		 * '-' and 'q' may safely be treated as the same value for most purposes
		 * n = The key is valid
		 * m = The key is marginal valid.
		 * f = The key is fully valid
		 * u = The key is ultimately valid.  This often means that the secret key is available, but any key may be marked as ultimately valid.
		 */

		switch($this->validity){
			case 'i':
			case 'd':
			case 'D':
			case 'r':
			case 'e':
				return false;
			default:
				return true;
		}
	}
	
	protected function _loadFromLine(ListKeysOutputLine $line){
		if($line->id) {
			$this->_setID($line->id);
		}
		
		if($line->encryption_type) {
			$this->_setEncryptionType($line->encryption_type);
		}
		
		if($line->uid) {
			$this->_setUID($line->uid);
		}
		
		if($line->sig) {
			$this->_setCertification($line->sig);
		}
		
		$ints = [
			'created'         => 'created',
			'expires'         => 'expires',
			'updated'         => 'updated',
			'encryption_bits' => 'encryptionBits',
		];
		
		$strs = [
			'capabilities'     => 'capabilities',
			'compliance_flags' => 'complianceFlags',
			'curve'            => 'curve',
			'hash_algorithm'   => 'hashAlgorithm',
			'serial'           => 'serial',
			'validity'         => 'validity',
		];
		
		foreach($ints as $lineKey => $thisKey){
			if($line->$lineKey !== null){
				$this->$thisKey = (int)$line->$lineKey;
			}
		}
		
		foreach($strs as $lineKey => $thisKey){
			if($line->$lineKey !== null){
				$this->$thisKey = $line->$lineKey;
			}
		}
	}
	
	protected function _setCertification(string $level) {
		if($level !== ''){
			$code = substr($level, 0, 2);
			switch($code){
				case 10:
					$this->certification = UIDSig::CERTIFY_NONE;
					break;
				case 11:
					$this->certification = UIDSig::CERTIFY_PERSONA;
					break;
				case 12:
					$this->certification = UIDSig::CERTIFY_CASUAL;
					break;
				case 13:
				case 18:
					$this->certification = UIDSig::CERTIFY_EXTENSIVE;
					break;
			}
		}
	}
	
	protected function _setEncryptionType(string $type) {
		switch($type){
			case '1':
				$this->encryptionType = Key::ENCRYPTION_TYPE_RSA;
				break;
			case '16':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
			case '17':
				$this->encryptionType = Key::ENCRYPTION_TYPE_DSA;
				break;
			case '20':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
		}
	}
	
	protected function _setID(string $id) {
		$this->id = $id;
		$this->id_short = substr($id, -8);
	}
	
	protected function _setUID(string $uid) {
		$split = GPG::ParseAuthorString($uid);
		
		if($this instanceof PublicKey) {
			$uid = new UID();
			$uid->comment  = $split['comment'];
			$uid->email    = $split['email'];
			$uid->fullname = $split['name'];
			$uid->validity = $this->validity;
			$uid->id_short = $this->id_short;
			$uid->id       = $this->id;
			
			$this->uids[] = $uid;
		}
		elseif($this instanceof UID) {
			$this->comment  = $split['comment'];
			$this->email    = $split['email'];
			$this->fullname = $split['name'];
		}
	}
	
	public static function FactoryFromLine(ListKeysOutputLine $line) : Key {
		
		switch($line->type){
			case 'pub':
				$key = new PublicKey();
				break;
			case 'sig':
				$key = new UIDSig();
				break;
			case 'ssb':
			case 'sub':
				$key = new Key();
				break;
			case 'sec':
				$key = new SecretKey();
				break;
			case 'uid':
				$key = new UID();
				break;
			default:
				throw new \Exception('Unsupported line type, ' . $line->type);
		}
		
		$key->_loadFromLine($line);
		
		return $key;
	}
}