<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\GPG;

/**
 * Provides rendering of the output of GPG --list-keys --with-colons
 * 
 * @see https://git.gnupg.org/cgi-bin/gitweb.cgi?p=gnupg.git;a=blob_plain;f=doc/DETAILS
 * @author charlie
 */
class ListKeysOutputLine {
	
	/**
	 * @var string Type of record 
	 * 
	 * - pub :: Public key
	 * - crt :: X.509 certificate
	 * - crs :: X.509 certificate and private key available
	 * - sub :: Subkey (secondary key)
	 * - sec :: Secret key
	 * - ssb :: Secret subkey (secondary key)
	 * - uid :: User id
	 * - uat :: User attribute (same as user id except for field 10).
	 * - sig :: Signature
	 * - rev :: Revocation signature
	 * - fpr :: Fingerprint (fingerprint is in field 10)
	 * - pkd :: Public key data [*]
	 * - grp :: Keygrip
	 * - rvk :: Revocation key
	 * - tfs :: TOFU statistics [*]
	 * - tru :: Trust database information [*]
	 * - spk :: Signature subpacket [*]
	 * - cfg :: Configuration data [*]
	 */
	public $type;
	
	/**
	 * @var string Raw output of this line
	 */
	public $raw;
	
	/**
	 * @var string Key capabilities
	 * 
	 * The defined capabilities are:
	 * 
	 * - e :: Encrypt
	 * - s :: Sign
	 * - c :: Certify
	 * - a :: Authentication
	 * - ? :: Unknown capability
	 * 
	 * A key may have any combination of them in any order.  In addition
	 * to these letters, the primary key has uppercase versions of the
	 * letters to denote the _usable_ capabilities of the entire key, and
	 * a potential letter 'D' to indicate a disabled key.
	 */
	public $capabilities;
	
	/**
	 * @var string Compliance flags
	 * 
	 * Space separated list of asserted compliance modes and
	 * screening result for this key.
	 * 
	 * Valid values are:
	 * 
	 * - 8  :: The key is compliant with RFC4880bis
	 * - 23 :: The key is compliant with compliance mode "de-vs".
	 * - 6001 :: Screening hit on the ROCA vulnerability.
	 */
	public $compliance_flags;
	
	/**
	 * @var string Creation date
	 * 
	 * The creation date of the key is given in UTC.  For UID and UAT
	 * records, this is used for the self-signature date.  Note that the
	 * date is usually printed in seconds since epoch, however, we are
	 * migrating to an ISO 8601 format (e.g. "19660205T091500").  This is
	 * currently only relevant for X.509.  A simple way to detect the new
	 * format is to scan for the 'T'.  Note that old versions of gpg
	 * without using the =--fixed-list-mode= option used a "yyyy-mm-tt"
	 * format.
	 */
	public $created;
	
	/**
	 * @var string Curve name
	 * 
	 * For pub, sub, sec, and ssb records this field is used for the ECC curve name.
	 */
	public $curve;
	
	/**
	 * @var string Public key algorithm
	 * 
	 * The values here are those from the OpenPGP specs or if they are
	 * greather than 255 the algorithm ids as used by Libgcrypt.
	 */
	public $encryption_type;
	
	/**
	 * @var string Key length
	 * 
	 * The length of key in bits.
	 */
	public $encryption_bits;
	
	/**
	 * @var string Expiration date
	 * 
	 *  Key or UID/UAT expiration date or empty if it does not expire.
	 */
	public $expires;
	
	/**
	 * @var string Hash algorithm
	 * 
	 * For sig records, this is the used hash algorithm.
	 */
	public $hash_algorithm;
	
	/**
	 * @var string KeyID
	 * 
	 * This is the 64 bit keyid as specified by OpenPGP and the last 64
	 * bit of the SHA-1 fingerprint of an X.509 certifciate.
	 */
	public $id;
	
	/**
	 * @var string Issuer certificate fingerprint or other info
	 * 
	 * Used in FPR records for S/MIME keys to store the fingerprint of
	 * the issuer certificate.  This is useful to build the certificate
	 * path based on certificates stored in the local key database it is
	 * only filled if the issuer certificate is available. The root has
	 * been reached if this is the same string as the fingerprint. The
	 * advantage of using this value is that it is guaranteed to have
	 * been built by the same lookup algorithm as gpgsm uses.
	 * 
	 * For "uid" records this field lists the preferences in the same way
	 * gpg's --edit-key menu does.
	 * 
	 * For "sig" records, this is the fingerprint of the key that issued
	 * the signature.  Note that this is only filled in if the signature
	 * verified correctly.  Note also that for various technical reasons,
	 * this fingerprint is only available if --no-sig-cache is used.
	 */
	public $issuer;
	
	/**
	 * @var string Origin
	 * 
	 * The origin of the key or the user ID.  This is an integer
	 * optionally followed by a space and an URL.  This goes along with
	 * the previous field.  The URL is quoted in C style.
	 */
	public $origin;
	
	/**
	 * @var string
	 */
	public $serial;
	
	/**
	 * @var string Signature class
	 * 
	 * Signature class as per RFC-4880.  This is a 2 digit hexnumber
	 * followed by either the letter 'x' for an exportable signature or
	 * the letter 'l' for a local-only signature.  The class byte of an
	 * revocation key is also given here, 'x' and 'l' is used the same
	 * way.  This field if not used for X.509.
	 */
	public $sig;
	
	/**
	 * @var string S/N of a token
	 * 
	 * Used in sec/ssb to print the serial number of a token (internal
	 * protect mode 1002) or a '#' if that key is a simple stub (internal
	 * protect mode 1001).  If the option --with-secret is used and a
	 * secret key is available for the public key, a '+' indicates this.
	 */
	public $token_sn;
	
	/**
	 * @var string
	 */
	public $trust;
	
	/**
	 * @var string User-ID
	 * 
	 * The value is quoted like a C string to avoid control characters (the colon is quoted =\x3a=).
	 */
	public $uid;
	
	/**
	 * @var string Last update
	 * 
	 * The timestamp of the last update of a key or user ID.  The update
	 * time of a key is defined a lookup of the key via its unique
	 * identifier (fingerprint); the field is empty if not known.  The
	 * update time of a user ID is defined by a lookup of the key using a
	 * trusted mapping from mail address to key.
	 */
	public $updated;
	
	/**
	 * @var string Validity
	 *             This is a letter describing the computed validity of a key.
	 * 
	 * - o :: Unknown (this key is new to the system)
	 * - i :: The key is invalid (e.g. due to a missing self-signature)
	 * - d :: The key has been disabled
	 *    (deprecated - use the 'D' in field 12 instead)
	 * - r :: The key has been revoked
	 * - e :: The key has expired
	 * - - :: Unknown validity (i.e. no value assigned)
	 * - q :: Undefined validity.  '-' and 'q' may safely be treated as
	 * 	   the same value for most purposes
	 * - n :: The key is not valid
	 * - m :: The key is marginal valid.
	 * - f :: The key is fully valid
	 * - u :: The key is ultimately valid.  This often means that the
	 * 	   secret key is available, but any key may be marked as
	 * 	   ultimately valid.
	 * - w :: The key has a well known private part.
	 * - s :: The key has special validity.  This means that it might be
	 * 	   self-signed and expected to be used in the STEED system.
	 */
	public $validity;
	
	public $debugMapping = [];
	
	protected static $_FprMap = [
		'unused', // 0.
		'unused', // 1.
		'unused', // 2.
		'unused', // 3.
		'unused', // 4.
		'unused', // 5.
		'unused', // 6.
		'unused', // 7.
		'unused', // 8.
		'id', // 9.
		'unused', // 10.
		'unused', // 11.
		'unused', // 12.
	];
	
	protected static $_PubMap7 = [
		'type',            // 0.  Type of record
		'id',              // 1.  KeyID
		'encryption_type', // 2.  Algorithm:
		'encryption_bits', // 3.  length of key in bits.
		'created',         // 4.  Creation Date (in UTC).
		'unused',          // 5.  ???
		'unused',          // 6.  ???
	];
	
	protected static $_PubMap13 = [
		'type',            // 0.  Type of record
		'validity',        // 1.  A letter describing the calculated validity.
		'encryption_bits', // 2.  length of key in bits.
		'encryption_type', // 3.  Algorithm:
		'id',              // 4.  KeyID
		'created',         // 5.  Creation Date (in UTC).
		'expires',         // 6.  Key or user ID/user attribute expiration date or empty if none.
		'serial',          // 7.  Used for serial number in crt records (used to be the Local-ID).
		'unused',          // 8.  Ownertrust (primary public keys only)
		'uid',             // 9.  User-ID.
		'unused',          // 10. Signature class as per RFC-4880.
		'unused',          // 11. Key capabilities
		'unused',          // 12. Used in FPR records for S/MIME keys to store the fingerprint of the issuer certificate.
	];
	
	protected static $_PubMap21 = [
		'type',             // 0.  Type of record
		'validity',         // 1.  A letter describing the calculated validity.
		'encryption_bits',  // 2.  length of key in bits.
		'encryption_type',  // 3.  Algorithm:
		'id',               // 4.  KeyID
		'created',          // 5.  Creation Date (in UTC).
		'expires',          // 6.  Key or user ID/user attribute expiration date or empty if none.
		'serial',           // 7.  Used for serial number in crt records (used to be the Local-ID).
		'trust',            // 8.  Ownertrust (primary public keys only)
		'sig',              // 9.  User-ID.
		'unused',           // 10. Signature class as per RFC-4880.
		'capabilities',     // 11. Key capabilities
		'issuer',           // 12. Issuer certificate fingerprint or other info
		'unused',           // 13. Flag field
		'token_sn',         // 14. S/N of a token
		'hash_algorithm',   // 15. Hash algorithm
		'curve',            // 16. Curve name
		'compliance_flags', // 17. Compliance flags
		'updated',          // 18. Last update
		'origin',           // 19. Origin
	];
	
	protected static $_SigMap17 = [
		'type',             // 0.  Type of record
		'validity',         // 1.  A letter describing the calculated validity.
		'encryption_bits',  // 2.  length of key in bits.
		'encryption_type',  // 3.  Algorithm:
		'id',               // 4.  KeyID
		'created',          // 5.  Creation Date (in UTC).
		'expires',          // 6.  Key or user ID/user attribute expiration date or empty if none.
		'serial',           // 7.  Used for serial number in crt records (used to be the Local-ID).
		'trust',            // 8.  Ownertrust (primary public keys only)
		'uid',              // 9.  User-ID.
		'sig',              // 10. Signature class as per RFC-4880.
		'capabilities',     // 11. Key capabilities
		'issuer',           // 12. Issuer certificate fingerprint or other info
		'unused',           // 13. Flag field
		'token_sn',         // 14. S/N of a token
		'hash_algorithm',   // 15. Hash algorithm
		'curve',            // 16. Curve name
		'compliance_flags', // 17. Compliance flags
		'updated',          // 18. Last update
		'origin',           // 19. Origin
	];
	
	protected static $_UidMap21 = [
		'type',             // 0.  Type of record
		'validity',         // 1.  A letter describing the calculated validity.
		'encryption_bits',  // 2.  length of key in bits.
		'encryption_type',  // 3.  Algorithm:
		'id',               // 4.  KeyID
		'created',          // 5.  Creation Date (in UTC).
		'expires',          // 6.  Key or user ID/user attribute expiration date or empty if none.
		'serial',           // 7.  Used for serial number in crt records (used to be the Local-ID).
		'trust',            // 8.  Ownertrust (primary public keys only)
		'uid',              // 9.  User-ID.
		'unused',           // 10. Signature class as per RFC-4880.
		'capabilities',     // 11. Key capabilities
		'issuer',           // 12. Issuer certificate fingerprint or other info
		'unused',           // 13. Flag field
		'token_sn',         // 14. S/N of a token
		'hash_algorithm',   // 15. Hash algorithm
		'curve',            // 16. Curve name
		'compliance_flags', // 17. Compliance flags
		'updated',          // 18. Last update
		'origin',           // 19. Origin
	];
	
	public function __construct(string $line) {
		$this->raw  = $line;
		$this->type = substr($line, 0, 3);
		$parts      = substr_count($line, ':') + 1;
		
		switch($this->type){
			case 'gpg':
				// This is a GPG comment, skip it.
				return;
			case 'tru':
				// This is a trust definition, skip it for now.
				return;
			case 'fpr':
				$this->_parseWith(self::$_FprMap);
				break;
			case 'pub':
				if($parts === 21){
					$this->_parseWith(self::$_PubMap21);
				}
				elseif($parts === 19 || $parts === 16 || $parts === 13){
					// Public keys on newer versions of GPG have 19 parts.
					// PUblic keys on 2.1.23 and newer have 21 parts.
					// Secret keys often have 16 parts, but the 13 length can handle it.
					$this->_parseWith(self::$_PubMap13);
				}
				elseif($parts === 11){
					// Public keys BEFORE being imported, eg: examine, will have only 11 parts.
					// This also includes integrated UID data.
					$this->_parseWith(self::$_PubMap13);
				}
				else {
					$this->_parseWith(self::$_PubMap7);
				}
				break;
			case 'uid':
				if($parts === 21){
					$this->_parseWith(self::$_UidMap21);
				}
				break;
			case 'sig':
				$this->_parseWith(self::$_SigMap17);
				break;
			case 'sub':
				$this->_parseWith(self::$_PubMap21);
				break;
			case 'sec':
				$this->_parseWith(self::$_PubMap21);
				break;
			default:
				// @todo
				return;
		}
	}
	
	private function _parseWith(array $map) {
		$parts = explode(':', $this->raw);
		
		foreach($map as $idx => $key){
			if($key === 'unused' || !isset($parts[ $idx ]) || $parts[ $idx ] === ''){
				// Skip unused fields and fields that cannot be mapped to the data
				
				$this->debugMapping[] = $key === 'unused' ? '-skipped-' : ($key . ' EMTPY');
				continue;
			}
			
			$this->debugMapping[] = $key . ' [' . $parts[ $idx ] . ']';
			$this->$key = $parts[ $idx ];
		}
	}
}
