<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Installer\Checks;
use Core\Installer\InstallHelper;

/**
 * The Apache server check is only really useful when running from the CLI.
 * 
 * It will ensure that apache is installed and the necessary PHP bindings are in place.
 *
 * @author charlie
 */
class ApacheServerCheck implements CheckInterface {
	/**
	 * Get the title of this check
	 * @return string
	 */
	public function getTitle() : string {
		return 'Apache Server';
	}
	
	/**
	 * Run the test for this check
	 * @return bool True if passed, False otherwise
	 */
	public function test() : bool {
		
		if(
			$this->_testBinary() &&
			$this->_testPHPModuleInstalled() &&
			$this->_testPHPModuleEnabled() &&
			$this->_testRewriteModuleEnabled() &&
			$this->_testUserdirModuleEnabled() &&
			$this->_testUserdirModuleConfigured()
		){
			return true;
		}
		elseif(EXEC_MODE === 'CLI'){
			// AUTO FIX!
			$ret = null;
			$out = [];
			exec($this->getFix(), $out, $ret);
			return ($ret === 0);
		}
		else{
			return false;
		}
	}

	public function testExtended(): string {
		return $this->test() ? self::STATUS_PASSED : self::STATUS_FAILED;
	}

	/**
	 * Get the fix for this problem that instructs the user how to fix it
	 * @return string
	 */
	public function getDetails(): string {
		return 'Apache web server is required';
	}

	/**
	 * Get additional details about this check, but not necessarily about the fix.
	 * @return string
	 */
	public function getInformation() : string {
		return '';
	}
	
	/**
	 * Get the scripted fix for this problem that provides the user how to fix it
	 * @return string
	 */
	public function getFix() : string {
		
		$fixes      = [];
		$restart    = false;
		$installCMD = InstallHelper::GetPackageInstallCommand();
		
		if(!$this->_testBinary()){
			$fixes[] = $installCMD . ' apache2';
		}
		
		if(!$this->_testPHPModuleInstalled()){
			$fixes[] = $installCMD . ' libapache2-mod-php7.2';
		}
		
		if(!$this->_testPHPModuleEnabled()){
			$fixes[] = 'a2enmod php7.2';
			$restart = true;
		}
		
		if(!$this->_testRewriteModuleEnabled()){
			$fixes[] = 'a2enmod rewrite';
			$restart = true;
		}
		
		if(!$this->_testUserdirModuleEnabled()){
			$fixes[] = 'a2enmod userdir';
			$restart = true;
		}
		
		if(!$this->_testUserdirModuleConfigured()){
			$fixes[] = "sed -i 's:^[ ]*php_admin_flag:#       php_admin_flag:' /etc/apache2/mods-enabled/php7.0.conf";
			$fixes[] = "sed -i 's:AllowOverride .*:AllowOverride all:' /etc/apache2/mods-enabled/userdir.conf";
			$restart = true;
		}
		
		
		if($restart){
			$fixes[] = 'systemctl restart apache2';
		}
		
		return implode(NL, $fixes);
	}
	
	private function _testBinary() : bool {
		$family = InstallHelper::GetOSFamily();
		
		if($family === 'debian'){
			return (
				is_dir('/etc/apache2') &&
				trim(exec('which apache2ctl')) !== ''
			);
		}
		
		// @todo Support other server families
		return false;
	}
	
	private function _testPHPModuleInstalled() : bool {
		$family = InstallHelper::GetOSFamily();
		
		if($family === 'debian'){
			return ( file_exists('/usr/lib/apache2/modules/libphp7.2.so') );
		}
		
		// @todo Support other server families
		return false;
	}
	
	private function _testPHPModuleEnabled() : bool {
		$family = InstallHelper::GetOSFamily();
		
		if($family === 'debian'){
			return ( file_exists('/etc/apache2/mods-enabled/php7.2.conf') );
		}
		
		// @todo Support other server families
		return false;
	}
	
	private function _testRewriteModuleEnabled() : bool {
		$family = InstallHelper::GetOSFamily();
		
		if($family === 'debian'){
			return ( file_exists('/etc/apache2/mods-enabled/rewrite.load') );
		}
		
		// @todo Support other server families
		return false;
	}
	
	private function _testUserdirModuleEnabled() : bool {
		
		if(strpos(ROOT_PDIR, '/home/') === false){
			// Root directory not in a /home directory, no need for this check!
			return true;
		}
		
		$family = InstallHelper::GetOSFamily();
		
		if($family === 'debian'){
			return ( file_exists('/etc/apache2/mods-enabled/userdir.conf') );
		}
		
		// @todo Support other server families
		return false;
	}
	
	private function _testUserdirModuleConfigured() : bool {
		
		if(strpos(ROOT_PDIR, '/home/') === false){
			// Root directory not in a /home directory, no need for this check!
			return true;
		}
		
		$family = InstallHelper::GetOSFamily();
		
		if($family === 'debian'){
			if(trim(exec("egrep '^#[ ]*php_admin_flag' /etc/apache2/mods-available/php7.2.conf")) !== ''){
				// This indicates that the PHP flag is disabled, (default for Debian).
				return false;
			}
			if(trim(exec("egrep 'AllowOverride all' /etc/apache2/mods-available/userdir.conf")) === ''){
				// This indicates that the PHP flag is disabled, (default for Debian).
				return false;
			}
			
			return true;
		}
		
		// @todo Support other server families
		return false;
	}
}
