<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Installer\Checks;
use Core\Installer\InstallHelper;


/**
 * Description of MemoryLimitCheck
 *
 * @author charlie
 */
class WriteableDirectoryCheck implements CheckInterface {
	private $_path;
	
	public function __construct($path){
		$this->_path = $path;
	}
	
	/**
	 * Get the title of this check
	 * @return string
	 */
	public function getTitle() : string {
		return 'Directory ' . basename($this->_path) . ' writeable';
	}

	/**
	 * Get additional details about this check, but not necessarily about the fix.
	 * @return string
	 */
	public function getInformation() : string {
		return '';
	}
	
	/**
	 * Run the test for this check
	 * @return bool True if passed, False otherwise
	 */
	public function test() : bool {
		if(EXEC_MODE === 'CLI'){
			return $this->_testCLI();
		}
		elseif(
			is_dir($this->_path) && 
			!is_writable($this->_path)
		){
			// Special case if the directory exists but is not writeable.
			// This can happen if the parent files/ is writeable but the child
			// files/whatever is owned by a different user.
			return false;
		}
		elseif(
			is_dir($this->_path) && 
			is_writable($this->_path)
		){
			return true;
		}
		elseif(
			is_dir(dirname($this->_path)) && 
			is_writable(dirname($this->_path))
		){
			return true;
		}
		elseif(
			is_dir(dirname(dirname($this->_path))) && 
			is_writable(dirname(dirname($this->_path)))
		){
			return true;
		}
		else{
			return false;
		}
	}

	public function testExtended(): string {
		return $this->test() ? self::STATUS_PASSED : self::STATUS_FAILED;
	}
	
	/**
	 * Get additional details that may help the user fix the problem
	 * @return string
	 */
	public function getDetails() : string {
		$user = InstallHelper::GetUser();
		$test = $this->test();

		if($test) {
			return $user . ' can write to the directory ' . $this->_path;
		}
		else {
			return 'Ensure that the user "' . $user . '" can write to the directory ' . $this->_path;
		}
	}
	
	/**
	 * Get the fix for this problem that instructs the user how to fix it
	 * @return string
	 */
	public function getFix(): string {
		$user = InstallHelper::GetUser();
		
		// The user string depends on if the site is in DEV mode.
		// If so, the permissions should be less restrictive.
		$userStr = (defined('DEVELOPMENT_MODE') && DEVELOPMENT_MODE) ? $user : $user . ':' . $user;
		
		if(is_dir($this->_path) && !is_writable($this->_path)){
			// File exists but just is not writeable.
			return 'sudo chown -R ' . $userStr . ' ' . escapeshellarg($this->_path);
		}
		else{
			// Directory doesn't exist
			return
				'mkdir -p ' . escapeshellarg($this->_path) . ' && ' . NL .
				'sudo chown -R ' . $userStr . ' ' . escapeshellarg($this->_path);
		}
	}
	
	/**
	 * Test this directory when in CLI mode.
	 * 
	 * This has the additional checks for testing if the web user can write to it!
	 */
	private function _testCLI() : bool {
		// CLI mode we'll just do it!
		if(is_dir($this->_path)){
			chmod($this->_path, 0700);
		}
		else{
			// Most directories are only two-levels deep, so ensure the parent directory has the correct permissions.
			if(!is_dir(dirname($this->_path))){
				mkdir(dirname($this->_path), 0700, true);
				chown(dirname($this->_path), InstallHelper::GetWebServerUser());
				chgrp(dirname($this->_path), InstallHelper::GetWebServerGroup());
			}
			mkdir($this->_path, 0700);
		}
		
		chown($this->_path, InstallHelper::GetWebServerUser());
		chgrp($this->_path, InstallHelper::GetWebServerGroup());
		
		return true;
	}
}
