<?php
/**
 * File for class Table definition in the coreplus project
 *
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140406.2004
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\ListingTable;

use Core\Datamodel\DatasetWhereClause;
use Core\Date\DateTime;
use Core\Forms\Form;
use Core\ListingTable\Renderer\CollectionRenderer;
use Core\ListingTable\Renderer\TableRenderer;
use Core\ModelFactory;
use Core\Router;
use Core\Templates\Template;
use Core\Request;
use Core\UserAgent;
use Core\View;
use Core\ViewControl;
use Core\ViewControls;
use Exception;
use Core\FilterForm;
use Generator;
use Iterator;
use Core\Model;
use ReflectionClass;
use Core\Core;
use Core\Utilities\Profiler\Profiler;
use Core\Utilities\Profiler\ProfilerEvent;
use Core\Session;


/**
 * Listing table, but with manually set records.
 *
 * Each record must be an associative array with the desired keys.
 *
 * Only a single rendering method is supported, and it must be set from the external template.
 *
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class RecordTable implements TableInterface {

	public $testmode = false;

	/** @var Column[] List of columns present*/
	protected $_columns = [];

	/** @var Filter[] List of filters present */
	protected $_filters = [];

	/** @var Request */
	protected $_request;

	/** @var Action[] */
	protected $_actions = [];

	/** @var ViewControls */
	protected $_controls = null;

	/** @var Report[] */
	protected $_reports = [];

	protected $_defaultLimit = 25;

	protected $_limit;
	protected $_page;
	protected $_sort;

	/** @var string Table name override, defaults to the Model name */
	protected $_tableName;

	public function __construct() {
		// Setup all default columns, (used for extended classes)
		$this->_setupColumns();

		// Setup all default filters, (used for extended classes)
		$this->_setupFilters();

		// Setup all actions
		$this->_setupActions();

		// Setup all controls
		$this->_setupControls();
	}

	// <editor-fold desc="Action & Control Methods">

	/**
	 * Add a control to this table to display inline
	 *
	 * @param ViewControl $control
	 */
	public function addControl(ViewControl $control) {
		if($this->_controls === null) {
			$this->_controls = new ViewControls();
		}

		$control->data('table-source', $this->getTableName());

		$this->_controls->addLink($control);
	}

	/**
	 * Add a control to this page
	 *
	 * Does the same thing as $view->addControl, but
	 * useful if you need to link it to the table to preserve
	 * functionality such as 'table-only-with-results'
	 *
	 * @param ViewControl $control
	 */
	public function addControlToPage(ViewControl $control) {

		$control->data('table-source', $this->getTableName());

		Router::$DefaultView->addControl($control);
	}

	public function addAction(Action $action) {
		$this->_actions[] = $action;
	}

	public function getControls() : ViewControls {
		if($this->_controls === null) {
			$this->_controls = new ViewControls();
		}

		return $this->_controls;
	}

	/**
	 * Setup default controls for this Table
	 */
	protected function _setupControls() {
		$this->addControl(
			ViewControl::Factory()
				->title('t:STRING_PRINT_PAGE')
				->icon('print')
				->classname('print-link')
				->data('table-only-with-results', true)
		);

		$this->addControl(
			ViewControl::Factory()
				->title('t:STRING_AUTO_REFRESH')
				->icon('sync')
				->classname('auto-refresh-link')
		);
	}

	/**
	 * Setup default actions for this Table
	 */
	protected function _setupActions() {
		// All tables support JSON loading, this provides support for retrieving the results.
		$json = new Action();
		$json->call = [$this, 'renderAsJSON'];
		$json->ctype = View::CTYPE_JSON;

		$this->addAction($json);
	}

	/**
	 * Process all actions provided and execute any that match the request
	 */
	protected function _processActions() {
		foreach($this->_actions as $action) {
			if($action->checkMatches($this->_request)) {
				$action->execute();
				return;
			}
		}
	}

	// </editor-fold>

	// <editor-fold desc="Column Methods">

	/**
	 * Add a new Column onto this Listing Table
	 *
	 * @param string $key
	 * @param string $title
	 *
	 * @return Column
	 */
	public function addColumn(string $key, string $title) : Column {
		$column = new Column();
		$column->title($title)
			->name($key)
			->key($key);

		$this->_columns[ $key ] = $column;

		return $column;
	}

	/**
	 * Retrieve a single Column by its key name
	 *
	 * If the column does not exist, an exception will be thrown.
	 *
	 * @param string $key
	 * @return Column
	 *
	 * @throws Exception
	 */
	public function getColumn(string $key) : Column {
		if(!isset($this->_columns[ $key ] )) {
			throw new Exception('Column "' . $key . '" is not defined on table ' . $this->getTableName());
		}

		return $this->_columns[ $key ];
	}

	/**
	 * Get all columns declared on this table
	 *
	 * @return Column[]
	 */
	public function getColumns() : array {
		return $this->_columns;
	}

	/**
	 * Get all VISIBLE columns declared on this table
	 *
	 * @return Column[]
	 */
	public function getColumnsVisible() : array {
		$ret = [];
		foreach($this->_columns as $c) {
			if($c->visible) {
				$ret[] = $c;
			}
		}

		return $ret;
	}

	/**
	 * Set the visible columns on this table
	 *
	 * @param string[] $columns
	 */
	public function setVisibleColumns(array $columns) {
		$tableName = $this->getTableName();

		$cols = [];
		foreach($columns as $column) {
			try {
				// Will ensure this column actually exists.
				$this->getColumn($column);
				$cols[] = $column;
			}
			catch(Exception $ex) {
				// pass
			}
		}

		if(count($cols) === 0) {
			// No columns displayed, resets the view
			Session::Delete('tablecolumns/' . $tableName);
		}
		else {
			Session::Set('tablecolumns/' . $tableName, $cols);
		}
	}

	/**
	 * Setup all default columns for this Table
	 */
	protected function _setupColumns() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}

	// </editor-fold>


	// <editor-fold desc="Filter Methods">

	/**
	 * Add a new filter for this listing table.
	 *
	 * @param string $type
	 * @param array $atts
	 *
	 * @return Filter
	 */
	public function addFilter(string $type, array $attributes) : Filter {
		$filter = new Filter($type, $attributes);

		$this->_filters[ $filter->getName() ] = $filter;

		return $filter;
	}

	/**
	 * Get a single filter by its key name
	 *
	 * @param string $key
	 * @return Filter
	 *
	 * @throws Exception
	 */
	public function getFilter(string $key) : Filter {
		if(!isset($this->_filters[ $key ] )) {
			throw new Exception('Filter "' . $key . '" is not defined on table ' . $this->getModelName());
		}

		return $this->_filters[ $key ];
	}

	/**
	 * Get all filters registered on this Table
	 *
	 * @return Filter[]
	 */
	public function getFilters() : array {
		return $this->_filters;
	}

	/**
	 * Get the processed value for the requested filter
	 *
	 * If the filter does not exist, NULL is returned.
	 *
	 * @param string $filter
	 *
	 * @return mixed
	 */
	public function getFilterValue(string $filter) {
		try {
			return $this->getFilter($filter)->getProcessedValue();
		}
		catch(Exception $ex) {
			return null;
		}
	}

	/**
	 * Set a group of filter values and save them in the session
	 *
	 * @param array $values
	 */
	public function setFilterValues(array $values) {
		$allValues = [];
		$tableName = $this->getTableName();

		foreach($values as $f => $v){
			try {
				$filter = $this->getFilter($f);
				$filter->setValue($v);

				$allValues[ $filter->getName() ] = $filter->getRawValue();
			}
			catch(Exception $ex) {
				// pass
			}
		}

		// Save these in the session
		Session::Set('tablefilters/' . $tableName, $allValues);
	}

	/**
	 * Check if there are filter values currently set from the user
	 *
	 * @return bool
	 */
	public function checkFiltersSet() : bool {
		$tableName = $this->getTableName();

		return count(Session::Get('tablefilters/' . $tableName, [])) > 0;
	}

	/**
	 * Setup all default filters for this Table
	 */
	protected function _setupFilters() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}

	/**
	 * Process all the attached filters from the Request.
	 */
	protected function _processFilters() {
		$filters = $this->getFilters();

		foreach($filters as $filter) {
			$linkType = $filter->getLinkType();
			if($linkType === '') {
				// Skip filters that do not have a standard link type.
				continue;
			}

			$value = $filter->getProcessedValue();
			if($value === null) {
				// Skip filters without a value set
				continue;
			}

			$statements = [];
			$names = $filter->getLinkNames();
			foreach($names as $n){
				switch($linkType){
					case Filter::LINK_TYPE_STANDARD:
					case Filter::LINK_TYPE_GT:
					case Filter::LINK_TYPE_GE:
					case Filter::LINK_TYPE_LT:
					case Filter::LINK_TYPE_LE:
						$statements[] = $n . ' ' . $linkType . ' ' . $value;
						break;
					case Filter::LINK_TYPE_STARTSWITH:
						$statements[] = $n . ' LIKE ' . $value . '%';
						break;
					case Filter::LINK_TYPE_CONTAINS:
						$statements[] = $n . ' LIKE %' . $value . '%';
						break;
					case Filter::LINK_TYPE_SEARCH:
						// search is the oddball out; it doesn't generate a statement.
						$this->_factory->search($value);
						break;
				}
			}

			if(count($statements) == 1) {
				// Single query, just append normally
				$this->_factory->where($statements[0]);
			}
			elseif(count($statements) > 1) {
				// Generate a subwhere for this filter to use OR internally.
				try {
					$subwhere = new DatasetWhereClause();
					$subwhere->setSeparator('OR');
					foreach($statements as $s){
						$subwhere->addWhere($s);
					}
					$this->_factory->where($subwhere);
				}
				catch(Exception $ex) {
					// pass
				}
			}
		}
	}

	// </editor-fold>


	// <editor-fold desc="Report Methods">

	public function addReport(Report $report) {
		// Populate the report with the necessary metadata from here
		$report->setFactory($this->getModelFactory());
		$report->setFilters($this->getFilters());
		$report->testmode = $this->testmode;

		// Add it to the list of reports
		$this->_reports[ $report->getName() ] = $report;

		if(Core::IsLibraryAvailable('HTMLtoPDF')) {
			// Create a "View PDF" action
			$pdf = new Action();
			$pdf->matches = [
				'report' => $report->getName(),
				'format' => 'pdf'
			];
			$pdf->call = [$report, 'renderAsPDF'];
			$this->addAction($pdf);
		}

		// Create a "View HTML" action
		$html  = new Action();
		$html->matches = [
			'report' => $report->getName()
		];
		$pdf->call = [$report, 'renderAsHTML'];
		$this->addAction($html);

		if($this->_request !== null) {
			\Core\log_notice('Table->setCSVDownloadable called after request loaded!  Support for this will be removed in future versions.');
			$this->_processActions();
		}
	}

	// </editor-fold>

	// <editor-fold desc="CSV Methods">

	/**
	 * Mark this table as downloadable via CSV
	 *
	 * Will setup the action and control automatically.
	 */
	public function setCSVDownloadable() {
		$control = new ViewControl();
		$control->title('t:STRING_DOWNLOAD_CSV')
			->link('.csv')
			->icon(ViewControl::ICON_DOWNLOAD)
			->data('table-only-with-results', true);

		$action = new Action();
		$action->ctype = View::CTYPE_CSV;
		$action->call = [$this, 'renderAsCSV'];

		$this->addControl($control);
		$this->addAction($action);

		if($this->_request !== null) {
			\Core\log_notice('Table->setCSVDownloadable called after request loaded!  Support for this will be removed in future versions.');
			$this->_processActions();
		}
	}

	public function renderAsCSV() {
		$profiler = new ProfilerEvent('Listing Table Render (JSON) - ' . $this->getModelName());
		$profiler->start();

		$view  = Router::$DefaultView;


		$filename = \Core\str_to_url($this->getModelName()) . '-' . date('Y-m-d@Hi') . '.csv';


		if($this->testmode && DEVELOPMENT_MODE){
			// Allow rendering as a test output for easy debugging for the developer.
			// For security reasons, this is only allowed when in DEVELOPMENT_MODE!
			$view->contenttype = 'text/plain';
		}
		else{
			$view->contenttype = 'text/csv';
			$view->addHeader('Content-Disposition', 'attachment; filename=' . $filename);
		}


		// This method will stream output directly to the browser
		$view->mode = View::MODE_NOOUTPUT;
		$view->render();

		$this->_factory->limit(9999);


		// Render comment records to assist the reader of the file.
		// This is not supported by many systems, but should still render as a record for their viewing pleasure.
		$comments = [];

		$comments[] = ['#COMMENT', 'Generated', DateTime::Now('FDT') . ' by ' . \Core\user()->get('email')];
		foreach($this->_filters as $filter) {
			$value = $filter->getRawValue();

			if($value) {
				$comments[] = ['#COMMENT', 'Filter ' . $filter->getLabel(), $value];
			}
		}

		foreach($comments as $c){
			$this->_sendCSVRecord($c);
		}


		// Build the CSV header to send, (the first record).
		$header = [];
		foreach($this->_columns as $c){
			/** @var Column $c */
			$header[] = $c->getTitle();
		}
		$this->_sendCSVRecord($header);


		// Retrieve all the data and push it to the browser.
		foreach($this->_factory->generator() as $model) {
			$data = [];
			foreach($this->_columns as $c){
				/** @var Column $c */
				$data[] = $model->get($c->renderkey, View::CTYPE_PLAIN);
			}

			$this->_sendCSVRecord($data);
		}
	}

	/**
	 * Send an indexed array to the browser as a valid CSV record.
	 *
	 * To be used in conjunction with sendCSVHeader().
	 *
	 * All scalar data in the array will be sanitized automatically.
	 *
	 * @param array $data
	 */
	protected function _sendCSVRecord($data){
		// CSV data is.... difficult.
		foreach($data as $k => $v){

			$v = trim($v);

			$escape = false;
			if($v == ''){
				$escape = false;
			}
			elseif(strpos($v, '"') !== false){
				// Strings with a quote in them get the quote escaped out and then wrapped in quotes.
				$escape = true;
				$v = str_replace('"', '\\"', $v);
			}
			elseif(strpos($v, ' ') !== false){
				// Strings with a space in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, ',') !== false){
				// Strings with a comma in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, "\n") !== false){
				// Strings with a newline in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, "\r") !== false){
				// Strings with a newline in the content get wrapped in quotes.
				$escape = true;
			}

			if($escape){
				// Wrap the formatted output with quotes
				$data[$k] = '"' . $v . '"';
			}
			else{
				// Just in case the trim() did something!
				$data[$k] = $v;
			}
		}

		// Now...
		echo implode(',', $data) . "\r\n";
	}

	/**
	 * Process requests for CSV versions
	 *
	 * Will send the results directly to the user agent; no additional action needed.
	 */
	private function _processCSVParameter() {
		if(
			$this->_request !== null &&
			$this->_request->ctype === View::CTYPE_CSV &&
			$this->_hasCSV
		) {
			$this->getFilters()->setPage(1);
			$this->setLimit(9999);
			$this->render('csv');
		}
	}

	// </editor-fold>


	// <editor-fold desc="Record / Dataset Methods">


	public function getRecords() : Generator {
		yield [];
	}


	/**
	 * Get the sort string, formatted as "[key] [dir]"
	 *
	 * @return string
	 */
	public function getSort() : string {
		$tableName = $this->getTableName();

		return Session::Get('tablesort/' . $tableName, '');
	}

	/**
	 * Set the sort key and direction and save them
	 *
	 * @param string $key       The key of the column to sort by
	 * @param string $direction "DESC" or "ASC" for descending or ascending sort
	 *
	 * @throws Exception
	 */
	public function setSort(string $key, string $direction = 'DESC') {
		// Will ensure this column exists
		$this->getColumn($key);
		$tableName = $this->getTableName();

		switch(strtolower($direction)) {
			case 'asc':
			case 'up':
				$direction = 'ASC';
				break;
			default:
				$direction = 'DESC';
		}

		$value = $key . ' ' . $direction;

		Session::Set('tablesort/' . $tableName, $value);
	}

	/**
	 * Get the display limit, ie: number of results per page
	 *
	 * @return int
	 */
	public function getLimit() : int {
		$tableName = $this->getTableName();

		return Session::Get('tablelimit/' . $tableName, $this->_defaultLimit);
	}

	/**
	 * Set the limit of results to show before pagination kicks in.
	 *
	 * @param $limit
	 */
	public function setLimit(int $limit) {
		$tableName = $this->getTableName();

		Session::Set('tablelimit/' . $tableName, min($limit, 500));
	}

	public function getPage() : int {
		$tableName = $this->getTableName();

		return Session::Get('tablepage/' . $tableName, 1);
	}

	/**
	 * Set the current Page number
	 *
	 * @param $page
	 */
	public function setPage(int $page) {
		$tableName = $this->getTableName();

		Session::Set('tablepage/' . $tableName, $page);
	}


	/**
	 * Get the total number of records based on current filtering
	 *
	 * @return int
	 */
	public function getTotalCount() : int {
		return 0;
	}

	/**
	 * Convenience function to paginate records based on current rules
	 *
	 * @param array $records
	 *
	 * @return array
	 */
	protected function _paginateRecords(array &$records) : array {
		$limit  = $this->getLimit();
		$offset = ($this->getPage() - 1) * $limit;

		return array_slice($records, $offset, $limit);
	}

	// </editor-fold>


	// <editor-fold desc="General Methods">

	/**
	 * Load/set the Request object to retrieve user-supplied parameters from.
	 *
	 * By default, this will use the default system request.
	 *
	 * @param Request $request
	 */
	public function loadFromRequest(Request $request = null) {
		if($request === null) {
			// Pull this from the router instead.
			$this->_request = \Core\Router::$DefaultRequest;
		}
		else {
			$this->_request = $request;
		}

		$tableName = $this->getTableName();

		if($this->_request->getParameter('table') === $tableName) {
			// Only process this request if it was the specific table requested.
			// On most pages, there will only be one, but this is required for pages that have multiple listing tables.

			// Process all the GET parameters to retrieve/set the current request data
			$this->_processRequestParameters();
		}

		// Process all the SESSION data for stored parameters
		$this->_processSessionParameters();

		if($this->_request->getParameter('table') === $tableName) {
			// Only process this request if it was the specific table requested.
			// On most pages, there will only be one, but this is required for pages that have multiple listing tables.

			// Give the filters a chance to be processed, this will update the factory as necessary
			$this->_processFilters();

			// Check for any action that may have been requested.
			$this->_processActions();
		}
	}

	/**
	 * Get the table name, (just the linked Model name lowercase.
	 *
	 * @return string
	 */
	public function getTableName() : string {
		return $this->_tableName ?? '';
	}

	/**
	 * Set the table name, useful for multiple listing tables on the same page.
	 *
	 * @param string $name
	 */
	public function setTableName(string $name) {
		$this->_tableName = $name;
	}

	/**
	 * Render this table skeleton
	 *
	 * @return string
	 * @throws Exception
	 */
	public function render() : string {
		$profiler = new ProfilerEvent('Record Table Render - ' . $this->getTableName());
		$profiler->start();

		$request    = Router::$DefaultRequest;
		$view       = Router::$DefaultView;
		$filters    = $this->getFilters();
		$attributes = [];
		$controls   = $this->getControls();
		$tableName  = $this->getTableName();

		// @todo Support supplemental classes
		$tableclasses = ['listing-table'];
		$attributes['class'] = implode(' ', $tableclasses);

		// @todo Support supplemental data tags
		$attributes['data-table-name'] = $tableName;
		$attributes['data-table-baseurl'] = $request->getBaseURL();


		$template = Template::Factory('includes/listingtable/table');


		//$template->assign('view_type', $viewType);
		//$template->assign('view_type_forced', $viewTypeForced);
		$template->assign('filters', $filters);
		$template->assign('attributes', $attributes);
		$template->assign('table', $this);
		$template->assign('controls', $controls);
		$template->assign('reports', $this->_reports);
		$template->assign('base_url', Router::$DefaultRequest->getBaseURL());
		$template->assign('limit', $this->getLimit());
		$template->assign('columns', $this->getColumns());
		$template->assign('pdf_support', Core::IsLibraryAvailable('HTMLtoPDF'));
		//$template->assign('bulk_actions', $this->_bulkActions);

		$html = $template->fetch();

		// Don't forget the necessary scripts!
		//\Core\view()->addScript('assets/js/core.listingtable.js', 'foot');

		$profiler->stop();
		return $html;
	}

	public function renderAsJSON() {
		$profiler = new ProfilerEvent('Record Table Render (JSON) - ' . $this->getTableName());
		$profiler->start();

		$view  = Router::$DefaultView;
		$page  = $this->getPage();
		$total = $this->getTotalCount();
		$limit = $this->getLimit();
		$pages = (int) ceil($total / $limit);
		$sort  = $this->getSort();

		// Set the appropriate column sort.
		if($sort !== '') {
			list($sortkey, $sortdir) = explode(' ', $sort);
			try {
				$col = $this->getColumn($sortkey);
				$col->currentSort = $sortdir;
			}
			catch(Exception $ex) {
				// pass
			}
		}

		$renderer = new TableRenderer($this);

		$view->contenttype = View::CTYPE_JSON;
		$view->jsondata = $renderer->render();

		$profiler->stop();
	}

	/**
	 * Process all the standard request parameters,
	 * generally for things like resets, view changes, etc.
	 */
	private function _processRequestParameters() {
		$tableName = $this->getTableName();
		$req = $this->_request;

		if($req->hasParameter('reset')) {
			// Reset the view to defaults
			Session::Delete('tablefilters/' . $tableName);
			Session::Delete('tablepage/' . $tableName);
		}

		if(
			$req->hasParameter('sortkey')&&
			$req->hasParameter('sortdir')
		) {
			try {
				$this->setSort(
					$req->getParameter('sortkey'),
					$req->getParameter('sortdir')
				);
			}
			catch(Exception $ex) {
				// pass
			}
		}

		if($req->hasParameter('page')) {
			$this->setPage((int) $req->getParameter('page'));
		}

		if($req->hasParameter('limit')) {
			// Update the limit AND page number.
			$this->setPage(1);
			$this->setLimit((int) $req->getParameter('limit'));
		}

		if($req->hasParameter('filter')) {
			$this->setFilterValues($req->getParameter('filter'));

			// Reset the page to 1 as well
			$this->setPage(1);
		}

		if($req->hasParameter('column')) {
			$this->setVisibleColumns($req->getParameter('column'));
		}
	}

	/**
	 * Process all the Session data and populate the necessary destinations.
	 */
	private function _processSessionParameters() {
		$tableName = $this->getTableName();

		$filters = Session::Get('tablefilters/' . $tableName, []);
		foreach($filters as $k => $v) {
			try {
				$this->getFilter($k)->setValue($v);
			}
			catch(Exception $ex) {
				// pass
			}
		}

		$columns = Session::Get('tablecolumns/' . $tableName, []);
		if(count($columns) > 0) {
			foreach($this->getColumns() as $column) {
				if(in_array($column->name, $columns)) {
					$column->show();
				}
				else {
					$column->hidden();
				}
			}
		}
	}

	// </editor-fold>

	/**
	 * Check if this table has bulk actions
	 *
	 * @return bool
	 */
	public function checkHasBulkActions(): bool {
		return false;
	}
}
