<?php

/*
 * @copyright Copyright (C) 2009-2017  charlie
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Logger;

use Core\Utilities\Profiler\Profiler;
use Exception;

/**
 * Description of Logger
 *
 * @author charlie
 */
class Logger {
	private static $_Sinks = [];
	
	/**
	 * Register a new Sink destination for the Logger to use.
	 * 
	 * @param string $alias
	 * @param \Core\Logger\LogSinkInterface $destination
	 */
	public static function RegisterSink(string $alias, LogSinkInterface $destination){
		if(
			isset(self::$_Sinks[ $alias ]) &&
			is_int(self::$_Sinks[ $alias ]['level'])
		){
			// This sink already has a requested level assigned, that's fine!
			// Ensure it's within an acceptable range. 
			// min+max will ensure a number between 0 and (sink-max-level).
			self::$_Sinks[ $alias ]['level'] = min(
				$destination->getMaximumLevel(),
				max(0, self::$_Sinks[ $alias ]['level'])
			);
			self::$_Sinks[ $alias ]['sink'] = $destination;
		}
		else{
			// Just a normal registration with no prior setting.
			self::$_Sinks[ $alias ] = [
				'level' => $destination->getDefaultLevel(),
				'sink' => $destination,
			];
		}
	}
	
	/**
	 * Update a registered sink with a new maximum logging level.
	 * 
	 * This can also be called BEFORE the sink is registered!
	 * 
	 * @param string $alias
	 * @param int $newLevel
	 */
	public static function SetSinkLevel(string $alias, int $newLevel) {
		if(
			isset(self::$_Sinks[ $alias ]) &&
			self::$_Sinks[ $alias ]['sink'] instanceof LogSinkInterface
		){
			// Ensure it's within an acceptable range. 
			// min+max will ensure a number between 0 and (sink-max-level).
			self::$_Sinks[ $alias ]['level'] = min(
				self::$_Sinks[ $alias ]['sink']->getMaximumLevel(),
				max(0, $newLevel)
			);
		}
		else{
			// Sink not registered yet, that's OK!
			// We support this so an external script can say that XYZ should be at
			// so-and-so logging level at the start.
			// Example: CLI system management scripts!
			self::$_Sinks[ $alias ] = [
				'level' => $newLevel,
				'sink'  => null,
			];
		}
	}
	
	/**
	 * Record a given LogEntry to the various destinations that have been registered.
	 * 
	 * @param \Core\Logger\LogEntry $entry
	 */
	public static function Log(LogEntry $entry){
		
		if(!$entry->message){
			// Don't try to record a log entry with no message!
			return;
		}
		
		foreach(self::$_Sinks as $alias => $dat){
			if(
				$entry->level <= $dat['level'] &&
				$dat['sink'] instanceof LogSinkInterface
			){
				try{
					$dat['sink']->log($entry);
				}
				catch (Exception $ex){
					if(DEVELOPMENT_MODE){
						// Just write to stdout in this case.
						// notifies the developer that they have a problem with their code.
						echo "Unable to record log entry to sink " . $alias . " due to " . $ex->getMessage() . "\n";
					}
				}
			}
		}
	}
}
