<?php
/**
 * @package Core\Templates\Smarty
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

/**
 * Primary method for a block of user-customizable content inside a template.
 *
 * Insertables are the core method of injecting blocks of user-customizable content into a template.
 *
 * An insertable must be on a template that has a registered page URL, as the baseurl is what is tracked as one of the main primary keys.
 * The other PK is the insertable's name, which must be unique on that one template.
 *
 * #### Smarty Parameters
 *
 *  * name
 *    * The key name of this input value, must be present and unique on this template.
 *  * assign
 *    * Assign the value instead of outputting to the screen.
 *  * title
 *    * When editing the insertable, the title displayed along side the input field.
 *  * type
 *
 * #### Example Usage
 *
 * <pre>
 * {insertable name="body" title="Body Content"}
 * <p>
 * This is some example content!
 * </p>
 * {/insertable}
 * </pre>
 *
 * <pre>
 * {insertable name="img1" title="Large Image" assign="img1"}
 * {img src="`$img1`" placeholder="generic" dimensions="800x400"}
 * {/insertable}
 * </pre>
 *
 * @param array       $params  Associative (and/or indexed) array of smarty parameters passed in from the template
 * @param string|null $content Null on opening pass, rendered source of the contents inside the block on closing pass
 * @param Smarty      $smarty  Parent Smarty template object
 * @param boolean     $repeat  True at the first call of the block-function (the opening tag) and
 * false on all subsequent calls to the block function (the block's closing tag).
 * Each time the function implementation returns with $repeat being TRUE,
 * the contents between {func}...{/func} are evaluated and the function implementation
 * is called again with the new block contents in the parameter $content.
 *
 * @return string
 */
function smarty_block_insertablegroup($params, $content, $smarty, &$repeat) {

	global $SMARTY_IN_INSERTABLEGROUP;

	if($repeat) {
		// First pass, set the global variable to indicate any embedded insertables *NOT* to render.
		$SMARTY_IN_INSERTABLEGROUP = true;
		return;
	}
	else {
		$SMARTY_IN_INSERTABLEGROUP = false;
	}

	// On the second pass, we'll actually pull in all the insertables for this group and provide them
	// as variables on the page.

	// I need to use the parent to lookup the current base url.
	$page = Core\Router::$DefaultPageModel;

	if(!isset($params['name'])) {
		// If no name was requested, there's nothing to link for this insertable.
		return '';
	}
	
	if(!$page) {
		// This page doesn't exist, (or otherwise is a dynamic-only page)
		return '';
	}

	/** @var InsertableModel[] $insertables */
	$groupData = InsertableModel::Construct($page->get('site'), $page->get('baseurl'), $params['name']);

	if($groupData->checkExists()) {
		$smarty->assign($params['name'], $groupData->value);
	}
	else {
		$smarty->assign($params['name'], []);
	}
}