<?php
/**
 * Model for UserGroupModel
 * 
 * @package User
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Model;
use Core\Core;
use Core\View;
use Core\I18N\I18NLoader;

/**
 * Class UserGroupModel
 *
 * @property string $context     If this group is for a specific context, the Model base name is here.
 * @property int    $created     Creation timestamp, populated automatically
 * @property bool   $default
 * @property string $description
 * @property string $id          Primary ID, as a UUID-formatted string
 * @property string $login_url   Optional login URL destination for members
 * @property string $name
 * @property array  $permissions json-encoded array of permissions this group has
 * @property int    $site        The site id in multisite mode, (or -1 if global)
 * @property int    $updated     Last update timestamp, populated automatically
 */
class UserGroupModel extends Model {

	public static $ExportGroup = 't:STRING_SITE_USERS';

	
	public static $Schema = array(
		'id' => [
			'type' => Model::ATT_TYPE_UUID,
			'required' => true,
			'null' => false,
		],
		'site' => [
			'type' => Model::ATT_TYPE_SITE,
		],
		'name' => [
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 48,
			'null' => false,
			'required' => true,
			'validation' => ['this', '_validateName'],
			'form' => [
				'group' => 't:STRING_BASIC',
				'grouptype' => 'tabs',
			],
		],
		'description' => [
			'type' => Model::ATT_TYPE_TEXT,
			'form' => [
				'group' => 't:STRING_BASIC',
				'grouptype' => 'tabs',
			],
		],
		'context' => [
			'type' => Model::ATT_TYPE_STRING,
			'default' => '',
			'comment' => 'If this group is for a specific context, the Model base name is here.',
			'form' => [
				'type' => 'disabled',
				'description' => 'If this group will be used to apply permissions in specific contexts, select it here. The specific context object is selected on the respective user edit page under their groups.',
			]
		],
		'permissions' => [
			'type' => Model::ATT_TYPE_DATA,
			'encoding' => Model::ATT_ENCODING_JSON,
			'formtype' => 'disabled',
			'comment' => 'json-encoded array of permissions assigned to this group',
			'formatter' => ['this', 'getPermissionsFormatted'],
		],
		'default' => [
			'type' => Model::ATT_TYPE_BOOL,
			'default' => false,
			'form' => [
				'group' => 't:STRING_BASIC',
				'grouptype' => 'tabs',
			],
		],
		'login_url' => [
			'type' => Model::ATT_TYPE_STRING,
			'null' => true,
			'default' => null,
			'form' => [
				'group' => 'User Behaviour',
				'grouptype' => 'tabs',
				'type' => 'select',
				'source' => ['PageModel', 'GetPagesAsOptions', null, '-- No Redirect Page --'],
			],
			'comment' => 'Optional login URL destination for members',
		],
		'created' => [
			'type' => Model::ATT_TYPE_CREATED,
			'null' => false,
		],
		'updated' => [
			'type' => Model::ATT_TYPE_UPDATED,
			'null' => false,
		],
	);
	
	public static $Indexes = array(
		'primary' => array('id'),
		'unique:name' => array('site', 'name'),
	);

	public function __construct(...$keys){

		$this->_linked['UserUserGroup'] = [
			'link' => Model::LINK_HASMANY,
			'on' => ['group_id' => 'id'],
		];

		parent::__construct(...$keys);
	}

	/**
	 * Get all the permissions this group as assigned to it.
	 *
	 * @return array
	 */
	public function getPermissions() : array {
		$p = $this->get('permissions');
		
		return is_array($p) ? $p : [];
	}
	
	public function getPermissionsFormatted(string $format = null) : string {
		$perms = $this->getPermissions();
		$strs  = [];
		
		if($format === View::CTYPE_HTML) {
			foreach($perms as $perm) {
				$strs[] = t( 'STRING_PERMISSION_' . I18NLoader::KeyifyString($perm) );
			}
		}
		else{
			$strs = $perms;
		}
		
		sort($strs);
		
		return count($strs) ? implode(', ', $strs) : t('STRING_NO_PERMISSIONS_SET');
	}
	
	public function getControlLinks(): array {
		
		$ret     = [];
		$manager = \Core\user()->checkAccess('p:/user/groups/manage');
		$id      = $this->get('id');
		
		if($manager) {
			$ret[] = [
				'title' => 't:STRING_EDIT',
				'link'  => '/usergroupadmin/update/' . $id,
				'icon'  => 'edit',
			];
			
			$ret[] = [
				'title'   => 't:STRING_DELETE',
				'confirm' => 'Confirm deletion of this group?',
				'link'    => '/usergroupadmin/delete/' . $id,
				'icon'    => 'delete',
			];
		}
		
		$auto = parent::getControlLinks();
		
		return array_merge($ret, $auto);
	}

	/**
	 * Get all members in this group
	 *
	 * @return UserModel[]
	 */
	public function getMembers() : array {
		$userFactory = new \Core\ModelFactory('UserModel');
		// Grab the source dataset, we'll modify it to allow an efficient lookup
		// (ie: without loading the UUG table, and then the U table)
		$factoryDS = $userFactory->getDataset();
		// Reset some fields to empty, we'll redeclare them with our own parameters.
		$factoryDS->_selects = null;
		$factoryDS->_tables = null;
		$factoryDS->select('user.*')
			->table('user', 'user')
			->joinRight('user_user_group', 'uug.user_id = id', 'uug')
			->whereEquals('uug.group_id', $this->id);

		return $userFactory->get();
	}

	public function setPermissions(array $permissions){
		// Verify that the permissions being set are valid permissions in the system
		$allperms = Core::GetPermissions();
		
		$perms = array_values(
			array_intersect($allperms, $permissions)
		);
		
		$this->set('permissions', $perms);
	}

	public function _validateName($key){
		// there are a few reserved keywords for the name.
		switch(strtolower($key)){
			case 'admin':
			case 'anonymous':
			case 'authenticated':
				return false;
			default:
				return true;
		}
	}

	/**
	 * Import the given data into the destination Model.
	 *
	 * @param array   $data            Indexed array of records to import/merge from the external source.
	 * @param array   $options         Any options required for the import, such as merge, key, etc.
	 * @param boolean $output_realtime Set to true to output the log in real time as the import happens.
	 *
	 * @throws Exception
	 *
	 * @return \Core\ModelImportLogger
	 */
	public static function ImportData($data, $options, $output_realtime = false) {
		$log = new \Core\ModelImportLogger('User Group Importer', $output_realtime);

		$merge = isset($options['merge']) ? $options['merge'] : true;
		$pk    = isset($options['key']) ? $options['key'] : null;

		if(!$pk) {
			throw new Exception(
				'Import requires a "key" field on options containing the primary key to compare against locally.'
			);
		}

		foreach($data as $dat) {

			if($pk == 'name' || $pk == 'id' || $pk == 'ldap_dn') {
				// These are the only two fields on the User object itself.
				$group = UserGroupModel::Find([$pk . ' = ' . $dat[ $pk ]], 1);
			}
			else {
				$group = UserGroupModel::Find(['name = ' . $dat['name']], 1);
			}

			$status_type = $group ? 'Updated' : 'Created';

			if($group && !$merge) {
				$log->duplicate('Skipped group ' . $group->getLabel() . ', already exists and merge not requested');
				// Skip to the next record.
				continue;
			}

			if(!$group) {
				// All incoming groups must have a name!
				if(!isset($dat['name'])) {
					$log->error('Unable to import groups without a name!');
					// Skip to the next record.
					continue;
				}

				// New user!
				$group = new UserGroupModel();
			}
			// No else needed, else is there IS a valid $user object and it's setup ready to go.


			// Handle all the properties for this user!
			foreach($dat as $key => $val){
				// Default Behaviour,
				// save the key into whatever field it was set to go to.
				$group->set($key, $val);
			}

			try {
				$status = $group->save();
			}
			catch(Exception $e) {
				$log->error($e->getMessage());
				// Skip to the next.
				continue;
			}

			if($status) {
				$log->success($status_type . ' group ' . $group->getLabel() . ' successfully!');
			}
			else {
				$log->skip('Skipped group ' . $group->getLabel() . ', no changes detected.');
			}
		}

		$log->finalize();

		return $log;
	}
	
} // END class UserGroupModel extends Model
