#!/bin/bash
#
# Base Script
#
# @author Charlie Powell <charlie@evalagency.com>
# @version 2020.07.31
#
# @license All Rights Reserved - Closed source, it's not licensed yet!
#          Why?  Because I haven't gone back through recently and cleaned it up :p
#
#@############################################################################@#
# Variables defined herein
#@----------------------------------------------------------------------------@#
# DEBUG            - Variable used to determine if the debug arguments should be used
#                    By default, this is just some feedback.
#                    Set to "0" by default if not set otherwise.
#
# QUIET            - Variable used to determine if standard output should be silent
#                    Set to "0" by default if not set otherwise.
#@############################################################################@#
# @changelog
# 2020.07.31
# * Add support for detecting Pop! OS
# * Add support for detecting Debian Bullseye
# * Add support for detecting Debian 1.1 - 4.0
# 
# 2020.04.17
# * Fix printlinestart to print a set amount of chars
# * Add printend{error,warn,success} to complement printlinestart
# 
# 2020.02.20
# * Add aliases for libcurl to libcurl 3/4 on Debian
# 
# 2019.06.11
# * Add prompttext function along with promptyesno
# 
# 2018.09.14
# * Initial support for INTERACTIVE mode
# 
# 2018.04.30
# * Initial support for install on FreeBSD
# * Initial support for pre and post installation scripts
# * Add jsvc/bsd to list of aliases
# * Add openjdk{,8}/{bsd,debian,redhat} to list of aliases
# * Add oracle-java{,8}/debian to list of aliases
# * Add python2-dateutil/{bsd,debian} to list of aliases
# 
# 2018.03.29
# * Add .zip support for downloadextract
# 
# 2018.03.28
# * Minor fix in install to check for curl
# 
# 2018.03.24
# * New install scripts for repositories
# * Added Ubuntu 18.* install aliases for php to php-7.2
# 
# 2018.02.02
# * Added automatic mapping support for packages across different distributions
# 
# 2017.10.01
# * Minor fix or $ARCH detection on arm hardware
# * Added the $INSTALLCMD global variable
# * Added the option for running the script with the '-d' flag to dump variables to stdout
# 
# 2017.08.19
# * Minor fix to support OpenJDK in the Java version detection.
# * Add support for detecting SuSE 13 version.
# 
# 2017.04.20
# * Include support for detecting JAVA versions installed,
# (only the binary present in $PATH)
# * Fix for Debian 9.0 detection.
# 
# 2017.02.13
# * Update author email address.
# * Switch Ubuntu check to a non-zero-length-string check.
# * Add support for Debian 7.x numeric strings in /etc/debian_version
# 
# 2016.06.10
# * Add support for versions in OpenSuSE and SLES.
# * Updated inline-doc for download().
# * Added --smart-override to download() for use with HTTP/HTTPS.
# * Improved systemctl/service logic in the svc() method.
# 
# 2016.02.26
# * Add check for debian version files containing only the major/minor version.
# 
# 2015.11.09
# * Add version checking for Debian lenny - stretch.
# 
# 2015.09.19
# * Add method "printlinestart" for allowing for shorter [action/success] messages
# 
# 2015.08.24
# * Fix support for CPU count AGAIN, this time it works on both BSD AND Linux!
# * Add a new md5file function for use with calculating the MD5 sum of files
# across Linux and BSD.
# 
# 2015.07.18
# * Fix support with BSD when determining the CPU count
# 
# 2014.09.02
# * Fix error message when ran on Debian
# 
# 2013.11.02
# * Finish implementation of the git:// download method.
# 
# 2013.09.30
# * Add a new function, printsubheader.
# * Compress the output of the logged version, while still maintaining readability.
# 
# 2013.07.22
# * Set the email subject to include the command itself.
# 
# 2013.05.21
# * Added KERNEL, KERNELVERSION, KERNELVERSIONMAJ, and KERNELVERSIONMIN variables.
# 
# 2013.04.06
# * Provided empty color support when running in "log" mode.
# This is because log files don't support color :/
# * Added some extra information to the header and footer for log mode, including
# architecture type, cpu type, and execution duration.
# 
# 2013.03.18
# * Renamed "debugprint" to "printdebug" to match the other print functions.
# * Completely rewrote how download behaves.
# * NOW IN COLOR!
# 
# 2012.09.13
# * Added log and email functions, useful for systems that do not email crontabs.
# 
# 2012.08.11
# * Added documentation and a debugprint function.
# 
# 2012.06.20
# * Fixed a minor bug in the md5 validation with filenames that contain an asterisk.
#
################################################################################
#                       BEGIN BUILDSCRIPT BASE CODE
################################################################################


##
# @var string
# Architecture of the machine, either i386, or x86_64, (or amd64 on BSD)
# @global
ARCH=$(uname -m)
if [ "$ARCH" == "arm" ]; then
	# Sometimes arm processors have more detailed information in sysctl than the kernel.
	A2=$(sysctl -n 'hw.machine_arch' 2>/dev/null)
	if [ -n "$A2" ]; then
		ARCH="$A2"
	fi
fi

##
# @var string
# Kernel currently running, "Linux" or "FreeBSD" usually.
# @global
KERNEL="$(uname -s)"

##
# @var string
# Full kernel version, obviously depends on the kernel
# "3.8.0-19-blah", "2.6", "8.2-RELEASE", etc.
# @global
KERNELVERSION="$(uname -r)"

##
# @var int
# Major number of the kernel version
# "3", "2", etc.
# @global
KERNELVERSIONMAJ=$(echo $KERNELVERSION | sed 's:^\([0-9]*\).*:\1:')

##
# @var int
# Minor number of the kernel version
# "8", "6", "2", etc.
# @global
KERNELVERSIONMIN=$(echo $KERNELVERSION | sed 's:^[0-9]\.\([0-9]*\).*:\1:')

##
# @var string
# @global
# Operating system
# freebsd, centos, redhat, fedora, ubuntu, debian, arch, suse, amazon
OS=""

##
# @var string
# Family of operating system, ie: base
# bsd, redhat, debian, arch, suse
# @global
OSFAMILY=""

##
# @var string
# Operating system full version string
# 6.3, 12.04, etc.
# @global
OSVERSION=""

##
# @var int
# Major version of Operating System
# 6, 12, etc.  (Not supported on all operating systems yet)
# @global
OSVERSIONMAJ=0

##
# @var int
# Processor count, (virtual and/or physical)
# @global
PROCCOUNT=0

##
# @var int
# Thread count, (for compiling)
# Always $PROCCOUNT + 1
# @global
THREADCOUNT=1

##
# @var string
# @global
# Full java version, if java is present
JAVAVERSION=""

##
# @var int
# @global
# Major version of Java available, if java is present
#
# Sun's versioning includes a 1. prefix for all versions <= 8,
# so this digit is dropped.  This means that Java 1.8 is major version 8
JAVAVERSIONMAJ=0

##
# @var string
# @global
# Install command line to use for installing packages
INSTALLCMD=""

##
# @var string
# @global
# Username of the user running this process
USER=""

##
# @var string
# @global
# For administrative tasks that require root/sudo access, this is the necessary
# prefix to those commands.
#
# For example, if the user is already root or sudo is not available, then this contains an empty string.
#
# Otherwise, "sudo " will be the contents.
SUDOPREFIX=""

##
# @var int
# @global
# Set to 1 if an X Display is available
XAVAILABLE=0

if xhost >& /dev/null; then
	XAVAILABLE=1
fi


if [ -z "$USER" ]; then
	# Ensure that USER has a name, if this script didn't run through a login shell, then it may not have been set.
	USER="$(whoami)"
fi

if [ "$EUID" == "0" -o -z "$(which sudo 2>/dev/null)" ]; then
	SUDOPREFIX=""
else
	SUDOPREFIX="sudo "
fi


if [ -n "$(which java 2>/dev/null)" ]; then
	JAVAVERSION="$(java -version 2>&1 | egrep 'java version|openjdk version' | sed 's:.* version "::' | sed 's:"::')"
	if [ -n "$(echo $JAVAVERSION | egrep '^1.[6-9].*')" ]; then
		JAVAVERSIONMAJ=$(echo $JAVAVERSION | sed 's:^1.\([6-9]\).*:\1:')
	elif [ -n "$(echo $JAVAVERSION | egrep '^9.*')" ]; then
		JAVAVERSIONMAJ=9
	fi
fi


# Default options that can be set from the calling script
if [ "$DEBUG" == "" ]; then DEBUG="0"; fi
if [ "$QUIET" == "" ]; then QUIET="0"; fi
if [ "$INTERACTIVE" == "" ]; then INTERACTIVE="0"; fi


# Check for FreeBSD in the uname output
# If it's not FreeBSD, then we move on!
if [ "$(uname -s)" == 'FreeBSD' ]; then
	OS='freebsd'
	OSFAMILY='bsd'

# Check for a redhat-release file and see if we can
# tell which Red Hat variant it is
elif [ -f "/etc/redhat-release" ]; then
	RHV=$(egrep -o 'Fedora|CentOS|Red.Hat' /etc/redhat-release)
	case $RHV in
		Fedora)  OS='fedora';;
		CentOS)  OS='centos';;
		Red.Hat)  OS='redhat';;
	esac
	OSVERSION="$(cat /etc/redhat-release | sed 's:^.* release \([0-9.]*\).*:\1:')"
	OSVERSIONMAJ="$(echo $OSVERSION | sed 's:^\([0-9]*\).*:\1:')"
	OSFAMILY="redhat"

# Check for Ubuntu
elif [ -f "/etc/debian_version" -a -f "/etc/lsb-release" -a -n "$(grep 'DISTRIB_ID=Ubuntu' /etc/lsb-release 2>/dev/null)" ]; then
	OS='ubuntu'
	OSFAMILY='debian'
	OSVERSION="$(grep 'DISTRIB_RELEASE' /etc/lsb-release | sed 's:DISTRIB_RELEASE=::')"
	OSVERSIONMAJ="$(echo $OSVERSION | sed 's:^\([0-9]*\).*:\1:')"

# Check for PopOS
elif [ -f "/etc/debian_version" -a -f "/etc/lsb-release" -a -n "$(grep 'DISTRIB_ID=Pop' /etc/lsb-release 2>/dev/null)" ]; then
	OS='popos'
	OSFAMILY='debian'
	OSVERSION="$(grep 'DISTRIB_RELEASE' /etc/lsb-release | sed 's:DISTRIB_RELEASE=::')"
	OSVERSIONMAJ="$(echo $OSVERSION | sed 's:^\([0-9]*\).*:\1:')"

# Check for debian_version
elif [ -f "/etc/debian_version" ]; then
	OS='debian'
	OSFAMILY="debian"

  if [ -n "$(grep bullseye /etc/debian_version)" ]; then
		OSVERSION="11.0";
		OSVERSIONMAJ=11;
	elif [ -n "$(grep buster /etc/debian_version)" ]; then
		OSVERSION="10.0";
		OSVERSIONMAJ=10;
	elif [ -n "$(grep stretch /etc/debian_version)" ]; then
		OSVERSION="9.0";
		OSVERSIONMAJ=9;
	elif [ -n "$(grep jessie /etc/debian_version)" ]; then
		OSVERSION="8.0";
		OSVERSIONMAJ=8;
	elif [ -n "$(grep wheezy /etc/debian_version)" ]; then
		OSVERSION="7.0";
		OSVERSIONMAJ=7;
	elif [ -n "$(grep squeeze /etc/debian_version)" ]; then
		OSVERSION="6.0";
		OSVERSIONMAJ=6;
	elif [ -n "$(grep lenny /etc/debian_version)" ]; then
		OSVERSION="5.0";
		OSVERSIONMAJ=5;
	elif [ -n "$(grep etch /etc/debian_version)" ]; then
		OSVERSION="4.0";
		OSVERSIONMAJ=4;
	elif [ -n "$(grep sarge /etc/debian_version)" ]; then
		OSVERSION="3.1";
		OSVERSIONMAJ=3;
	elif [ -n "$(grep woody /etc/debian_version)" ]; then
		OSVERSION="3.0";
		OSVERSIONMAJ=3;
	elif [ -n "$(grep potato /etc/debian_version)" ]; then
		OSVERSION="2.2";
		OSVERSIONMAJ=2;
	elif [ -n "$(grep slink /etc/debian_version)" ]; then
		OSVERSION="2.1";
		OSVERSIONMAJ=2;
	elif [ -n "$(grep hamm /etc/debian_version)" ]; then
		OSVERSION="2.0";
		OSVERSIONMAJ=2;
	elif [ -n "$(grep bo /etc/debian_version)" ]; then
		OSVERSION="1.3";
		OSVERSIONMAJ=1;
	elif [ -n "$(grep rex /etc/debian_version)" ]; then
		OSVERSION="1.2";
		OSVERSIONMAJ=1;
	elif [ -n "$(grep buzz /etc/debian_version)" ]; then
		OSVERSION="1.1";
		OSVERSIONMAJ=1;
	elif [ -n "$(egrep '^[5-9]\.[0-9]' /etc/debian_version)" ]; then
		OSVERSION="$(cat /etc/debian_version)"
		OSVERSIONMAJ=$(echo $OSVERSION | sed 's:^\([5-9]\).*:\1:')
	fi

# Check for arch-release
elif [ -f "/etc/arch-release" ]; then
	OS='arch'
	OSFAMILY='arch'

# Check for SuSE-release
elif [ -f "/etc/SuSE-release" ]; then
	OS='suse'
	OSFAMILY='suse'
	OSVERSION="$(cat /etc/SuSE-release | grep 'VERSION' | sed 's:^.*= \([0-9\.]*\).*:\1:')"
	OSVERSIONMAJ="$(cat /etc/SuSE-release | grep 'VERSION' | sed 's:^.*= \([0-9]*\)\..*:\1:')"

# SuSE 13 changed the location of this file.
elif [ -f "/etc/SUSE-brand" ]; then
	OS='suse'
	OSFAMILY='suse'
	OSVERSION="$(cat /etc/SUSE-brand | grep 'VERSION' | sed 's:^.*= \([0-9\.]*\).*:\1:')"
	OSVERSIONMAJ="$(cat /etc/SUSE-brand | grep 'VERSION' | sed 's:^.*= \([0-9]*\)\..*:\1:')"

# Check for Amazon instance, (this is a CentOS-type machine)
elif [ -f "/etc/system-release" ]; then
	OS="amazon"
	OSFAMILY="redhat"
	OSVERSION="$(cat /etc/system-release | sed 's:^.* release \([0-9.]*\).*:\1:')"

else
	OS="unknown"
	OSFAMILY="unknown"
	OSVERSION="0.0"
	OSVERSIONMAJ="0"
fi

# Detect the install command line to use, primarily useful for the -d option
case $OSFAMILY in
	redhat )
		INSTALLCMD="${SUDOPREFIX}yum -y install"
		;;
	debian )
		INSTALLCMD="${SUDOPREFIX}apt-get -y install"
		;;
	suse )
		INSTALLCMD="${SUDOPREFIX}zypper -n install"
		;;
esac


# Number of processors available.
# Try sysctl first.
PROCCOUNT=$(sysctl hw.ncpu 2>/dev/null | sed 's#hw\.ncpu: ##')

# If no results, see if /proc exists.
if [ -z "$PROCCOUNT" -a -e /proc/cpuinfo ]; then
	PROCCOUNT=$(cat /proc/cpuinfo | egrep ^processor | wc -l)
else
	# ??? What else is available?
	PROCCOUNT=1
fi

# Threads to use for compiling
let "THREADCOUNT=$PROCCOUNT+1"

##
# @var string
# @global
# Command for wget, with quiet flags based on $QUIET
# If QUIET is set to "1", then the quiet flag is set automatically
CMDWGET="$(which wget)"

# Reset wget to be quiet if in quiet mode.
if [ "$QUIET" == "1" ]; then
	CMDWGET="$CMDWGET -q"
fi


# Gimme some colors!
# These are used to prettify the terminal.
# Color 1 is always standard and
# Color 2 is always the bold version.

##
# @var string
# @global
# Standard black colour
C_BLK1="\033[0;30m"

##
# @var string
# @global
# Bold black colour
C_BLK2="\033[1;30m"

##
# @var string
# @global
# Standard red colour
C_RED1="\033[0;31m"

##
# @var string
# @global
# Bold red colour
C_RED2="\033[1;31m"

##
# @var string
# @global
# Standard green colour
C_GRN1="\033[0;32m"

##
# @var string
# @global
# Bold green colour
C_GRN2="\033[1;32m"

##
# @var string
# @global
# Standard yellow colour
C_YLW1="\033[0;33m"

##
# @var string
# @global
# Bold yellow colour
C_YLW2="\033[1;33m"

##
# @var string
# @global
# Standard blue colour
C_BLU1="\033[0;34m"

##
# @var string
# @global
# Bold blue colour
C_BLU2="\033[1;34m"

##
# @var string
# @global
# Standard purple colour
C_PRP1="\033[0;35m"

##
# @var string
# @global
# Bold purple colour
C_PRP2="\033[1;35m"

##
# @var string
# @global
# Standard cyan colour
C_CYN1="\033[0;36m"

##
# @var string
# @global
# Bold cyan colour
C_CYN2="\033[1;36m"

##
# @var string
# @global
# Standard white colour
C_WHT1="\033[0;37m"

##
# @var string
# @global
# Bold white colour
C_WHT2="\033[1;37m"

##
# @var string
# @global
# Standard "reset" colour
C_RESET="\033[0m"

##
# @var string
# Empty colour string
C_NONE=""

##
# @var string
# @global
# Line color, the separating characters
C_LN="$C_BLK1"

##
# @var string
# @global
# Heading color
C_HD="$C_CYN2"

##
# @var string
# @global
# Success color
C_SC="$C_GRN2"

##
# @var string
# @global
# Warning color
C_WN="$C_YLW2"

##
# @var string
# @global
# Error color
C_ER="$C_RED2"

##
# @var string
# @global
# Prompt / Question color
C_PR="$C_PRP2"

##
# @var string
# @global
# Debug color
C_DB="$C_BLU1"

##
# Print text to STDOUT if and only if $DEBUG is set to 1.
#
# @param $1 string
#        The text to print
#
function printdebug (){
	if [ $DEBUG == "1" ]; then
		echo -en "$C_LN"
		echo "+~ ## DEBUG ## ~"
		echo "| "
		echo -en "$C_DB"
		echo "$1";
		echo -en "$C_RESET"
	fi
}

##
# Print text to STDOUT unless $QUIET is set to 1.
#
# @param $1 string
#        The text to print
#
function printline (){
	if [ "$QUIET" == "0" ]; then
		if [ "$_LOGSTARTED" == "1" ]; then
			echo "[$(date)] $1"
		else
			echo -en "$C_LN| $C_RESET"
			echo "$1"
		fi
	fi
}

##
# Print text to STDOUT unless $QUIET is set to 1.
#
# This DOES NOT include a newline character!
#
# @param $1 string
#        The text to print
#
function printlinestart (){
	if [ "$QUIET" == "0" ]; then
		if [ "$_LOGSTARTED" == "1" ]; then
			echo -n "[$(date)] $1"
		else
			echo -en "$C_LN| $C_RESET"
			__line="$1"
			__columns=55
			printf "%s%*s" "$__line" $(($__columns-${#__line}))
		fi
	fi
}

##
# Print text as a header to STDOUT unless $QUIET is set to 1.
#
# @param $1 string
#        The text to print
#
function printheader (){
	if [ "$QUIET" == "0" ]; then
		echo -en "$C_LN"
		echo "| "
		echo "+============================================================================="
		echo -en "| $C_HD"
		__header="$1"
		__columns=75
		printf "%*s\n" $(((${#__header}+$__columns)/2)) "$__header"
		echo -en "$C_LN"
		if [ "$_LOGSTARTED" == "1" ]; then
			echo "| $(date)"
		fi
		echo -en "$C_LN"
		echo "+============================================================================="
		echo -en "$C_RESET"
	fi
}

##
# Print text as a sub-header to STDOUT unless $QUIET is set to 1.
#
# @param $1 string
#        The text to print
#
function printsubheader (){
	if [ "$QUIET" == "0" ]; then
		if [ "$_LOGSTARTED" == "1" ]; then
			echo "+-----------------------------------------------------------------------------"
			echo "[$(date)] * $1"
		else
			echo -en "$C_LN"
			echo "| "
			echo "+-----------------------------------------------------------------------------"
			echo -en "| $C_HD"
			echo "$1"
			echo -en "$C_RESET"
                fi
	fi
}


##
# Print a warning to STDOUT after printlinestart has ran
#
# @param $1 string
#        The text to print
#
function printendwarn (){
	if [ "$QUIET" == "0" ]; then
		if [ "$_LOGSTARTED" == "1" ]; then
			echo "[$(date)] WARN - $1"
		else
			echo -en "$C_WN"
			echo "[ $1 ]"
			echo -en "$C_RESET"
                fi
	else
		echo "[WARN] - $1"
	fi
}


##
# Print an error to STDOUT after printlinestart has ran
#
# @param $1 string
#        The text to print
#
function printenderror (){

	if [ "$_LOGSTARTED" == "1" ]; then
		# Log is started, send the message to the log and to stderr
		echo "[$(date)] ERROR - $1"
		echo "$1" 1>&2
	elif [ "$QUIET" == "0" ]; then
		# No log started, but quiet is disabled, so print to stdout
		echo -en "$C_ER"
		echo "[ $1 ]"
		echo -en "$C_RESET"
	else
		# Logging is disabled and quiet is requested, so just print to stderr
		echo "$1" 1>&2
	fi
}

##
# Print a success message to STDOUT after printlinestart has ran
#
# @param $1 string
#        The text to print
#
function printendsuccess (){
	 if [ "$QUIET" == "0" ]; then
		if [ "$_LOGSTARTED" == "1" ]; then
			echo "[$(date)] SUCCESS - $1"
		else
			echo -en "$C_SC"
			echo "[ $1 ]"
			echo -en "$C_RESET"
		fi
	fi
}

##
# Print a warning to STDOUT
#
# @param $1 string
#        The text to print
#
function printwarn (){
	if [ "$QUIET" == "0" ]; then
		if [ "$_LOGSTARTED" == "1" ]; then
			echo "[$(date)] WARN - $1"
		else
			echo -en "$C_LN| [WARN] $C_WN"
			echo "$1"
			echo -en "$C_LN"
			echo -en "$C_RESET"
                fi
	else
		echo "[WARN] - $1"
	fi
}


##
# Print an error to STDOUT
#
# @param $1 string
#        The text to print
#
function printerror (){

	if [ "$_LOGSTARTED" == "1" ]; then
		# Log is started, send the message to the log and to stderr
		echo "[$(date)] ERROR - $1"
		echo "$1" 1>&2
	elif [ "$QUIET" == "0" ]; then
		# No log started, but quiet is disabled, so print to stdout
		echo -en "$C_LN| [ERROR] $C_ER"
		echo "$1"
		echo -en "$C_RESET"
	else
		# Logging is disabled and quiet is requested, so just print to stderr
		echo "$1" 1>&2
	fi
}

##
# Print a success message to STDOUT
#
# @param $1 string
#        The text to print
#
function printsuccess (){
	 if [ "$QUIET" == "0" ]; then
		if [ "$_LOGSTARTED" == "1" ]; then
			echo "[$(date)] SUCCESS - $1"
		else
			echo -en "$C_LN| $C_SC"
			echo "$1"
			echo -en "$C_RESET"
		fi
	fi
}

##
# Display all colors available in this script!
#
# Useful for debugging.
function printcolortest (){
	printheader "Color Test!"

	echo -en "$C_LN| ${C_RESET}Black:   "
	echo -e "${C_BLK1}regular${C_BLK2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}Red:     "
	echo -e "${C_RED1}regular${C_RED2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}Green:   "
	echo -e "${C_GRN1}regular${C_GRN2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}Yellow:  "
	echo -e "${C_YLW1}regular${C_YLW2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}Blue:    "
	echo -e "${C_BLU1}regular${C_BLU2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}Purple:  "
	echo -e "${C_PRP1}regular${C_PRP2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}Cyan:    "
	echo -e "${C_CYN1}regular${C_CYN2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}White:   "
	echo -e "${C_WHT1}regular${C_WHT2} bold${C_RESET}"

	echo -en "$C_LN| ${C_RESET}Lines:   "
	echo -e "${C_LN}default${C_RESET} (\$C_LN)"

	echo -en "$C_LN| ${C_RESET}Headers: "
	echo -e "${C_HD}default${C_RESET} (\$C_HD)"

	echo -en "$C_LN| ${C_RESET}Prompt:  "
	echo -e "${C_PR}default${C_RESET} (\$C_PR)"

	echo -en "$C_LN| ${C_RESET}Success: "
	echo -e "${C_SC}default${C_RESET} (\$C_SC)"

	echo -en "$C_LN| ${C_RESET}Warning: "
	echo -e "${C_WN}default${C_RESET} (\$C_WN)"

	echo -en "$C_LN| ${C_RESET}Error:   "
	echo -e "${C_ER}default${C_RESET} (\$C_ER)"

	echo -en "$C_LN| ${C_RESET}Debug:   "
	echo -e "${C_DB}default${C_RESET} (\$C_DB)"
}

##
# Prompt the user with a question and return 0 (yes) or 1 (no) based on the response
#
# if the user does not respond within 15 seconds, the default response is returned.
#
# Required Positional Arguments:
# $1 "Message to display to the user"
# $2 "yes" or "no" for the default response
#
# Optional Named Arguments:
# --title "Title to display"
# --timeout N number of seconds to wait for a response
#
function promptyesno (){
	# Required positional arguments, message and default
	MSG="$1"
	DEF="$2"
	shift 2

	# Optional arguments
	TITLE=""
	TIMEOUT=15

	# Process any additional arguments passed in
	while [ ! -z "$1" ]; do
		case "$1" in
			"--title" )
				shift # Bump to the next argument to get the value
				TITLE="$1"
				;;
			"--timeout" )
				shift # Bump to the next argument to get the value
				TIMEOUT="$1"
		esac

		shift;
	done


	if [ "$(_evaluate_yes_no $DEF)" == "y" ]; then
		DEF=0
	else
		DEF=1
	fi

	if [ $XAVAILABLE -eq 1 -a -n "$(which zenity 2>/dev/null)" ]; then
		# Use Zenity to display the prompt
		zenity --title "$TITLE" --text "$MSG" --timeout $TIMEOUT --question 2>/dev/null

		# Zenity already returns 1 for no and 0 for yes, no need for conversions.
		RET=$?
		if [ $RET -eq 5 ]; then
			# User timeout, returns default
			return $DEF
		else
			return $RET
		fi
	else
		# Zenity or X isn't available, display the prompt in the shell
		echo -en "$C_LN| ${C_RESET}${C_PR}"
		echo "$MSG"
		if [ "$DEF" == "y" ]; then
			echo -en "Y/n: ${C_RESET}"
		else
			echo -en "y/N: ${C_RESET}"
		fi

		read -t$TIMEOUT ANS
		ANS="$(_evaluate_yes_no $ANS)"

		if [ -z "$ANS" ]; then
			# User timeout, returns default
			return $DEF
		elif [ "$ANS" == "y" ]; then
			return 0
		else
			return 1
		fi
	fi
}


##
# Prompt the user with a question and return the text response
#
# Required Positional Arguments:
# $1 "Message to display to the user"
#
# Optional Named Arguments:
# --title "Title to display"
#
function prompttext (){
	# Required positional arguments, message and default
	MSG="$1"
	shift 1

	# Optional arguments
	TITLE=""

	# Process any additional arguments passed in
	while [ ! -z "$1" ]; do
		case "$1" in
			"--title" )
				shift # Bump to the next argument to get the value
				TITLE="$1"
				;;
		esac

		shift;
	done

	if [ $XAVAILABLE -eq 1 -a -n "$(which zenity 2>/dev/null)" ]; then
		# Use Zenity to display the prompt
		zenity --title "$TITLE" --text "$MSG" --entry 2>/dev/null
	else
		# Zenity or X isn't available, display the prompt in the shell
		echo -en "$C_LN| ${C_RESET}${C_PR}" >&2
		echo "$MSG" >&2
		echo -en ": ${C_RESET}" >&2

		read ANS

		# Print it to stdout so the calling script can get the response.
		echo $ANS
	fi
}



##
# Check the exit status and behave appropriately
#
# This expects the argument passed in to be "0" on success
# and anything else on error.
#
# If the argument passed in is 0, printsuccess is called with "OK!"
# Otherwise, an error is sent to stderr and the script exits.
#
# @param $1 int
#        Return status to check
#
function checkexitstatus (){
	# The result from the called script, usually 0 indicates success.
	RES="$1"

	if [ "$RES" == "0" ]; then
		printsuccess "OK!"
	else
		printerror "Command failed with a status of ${RES}"
		# Set the email subject to something... noticeable
		EMAILLOGSUBJECT="!!!FAILED!!! $EMAILLOGSUBJECT"

		if [ "$INTERACTIVE" == "1" ]; then
			# Allow the user to continue when in interactive mode.
			promptyesno 'Do you want to ignore and continue?'
			if [ "$?" == 0 ]; then
				printdebug "Continuing"
			else
				exit $RES
			fi
		else
			exit $RES
		fi
	fi
}

##
# Start logging all output to a file
#
# Useful if you need to email the result of this script to someone
# or just recording a log of this operation locally.
#
# If the first parameter is set, that destination is used to write the log to,
# as opposed to the default behaviour of logging to a /tmp file.
#
# This is REQUIRED if you wish to email the output to someone!
#
# @param $1 string [optional]
#        Filename to log everything to.
#
# Uses $LOGFILE Filename to log everything to (overwrote by the parameter $1)
# Uses $EMAILLOG Email address to send the log contents to upon exit
# Uses $EMAILLOGSUBJCT Subject to use for email log
#
function startlog (){
	# Redirect stdout so I can redirect it after this runs.
	exec 3>&1

	# In this case, I also need to revert all the colors back to NONE, since logs don't support colors.
	C_LN="$C_NONE"
	C_HD="$C_NONE"
	C_SC="$C_NONE"
	C_WN="$C_NONE"
	C_ER="$C_NONE"
	C_DB="$C_NONE"
	C_RESET="$C_NONE"

	# Is a log location requested?  If so, keep that file.
	# otherwise it's a temporary file that will probably get emailed to someone.
	if [ "$1" != "" ]; then
		LOGFILE="$1"
	fi

	if [ "$LOGFILE" == "" ]; then
		LOGFILE="/tmp/$$.log"
		_LOGDELONEXIT="1"
		trap "_logexit" EXIT
	fi

	exec >>"$LOGFILE" 2>&1
	echo "
+=======================================================
|                  STARTING LOG
|
|     Date: $(date)
|     User: $(whoami)
|   Server: $(hostname -f)
|  Command: $0
|  OS Info: $OS $OSVERSION
| CPU Info: $PROCCOUNT cores of $ARCH architecture
+-------------------------------------------------------"
	_LOGSTARTED=1
	_LOGDATESTART="$(date +%s)"
}

##
# If startlog is called, this function will be called to close up everything
# and send the emails if necessary.
#
function _logexit (){
	# How long did this script take to run?
	# This should be human-friendly.
	_LOGDATEEND="$(date +%s)"
	let "_LOGDIFF=$_LOGDATEEND-$_LOGDATESTART"
	if [ $_LOGDIFF -lt 90 ]; then
		_LOGEXECTIME="$_LOGDIFF seconds"
	elif [ $_LOGDIFF -lt 7200 ]; then
		let "_LOGDIFF=$_LOGDIFF/60"
		_LOGEXECTIME="$_LOGDIFF minutes"
	else
		let "_LOGDIFF=$_LOGDIFF/3600"
		_LOGEXECTIME="$_LOGDIFF hours"
	fi

	echo "
+-------------------------------------------------------
|                   ENDING LOG
|
|     Date: $(date)
| Duration: about $_LOGEXECTIME
+======================================================="
	exec 1>&3;
	# Should I email someone about what happened?
	if [ "$EMAILLOG" != "" ]; then
		if [ "$EMAILLOGSUBJECT" == "" ]; then
			EMAILLOGSUBJECT="Command Log [$0]"
		fi
		cat $LOGFILE | mail -s "$EMAILLOGSUBJECT" $EMAILLOG
	fi

	# Is this a temporary log?
	if [ "$LOGDELONEXIT" == "1" ]; then
		rm -fr $LOGFILE
	fi
}


##
# Safely create a directory if it does not exist
#
# Will not try to create the directory if it already exists.
# Will create any parent directory too.
#
# @param $1 string
#        The full path to create.
#
function safemkdir (){
	if [ ! -e "$1" ]; then
		printline "Creating dir $1"
        	mkdir -p "$1"
		checkexitstatus "$?"
	fi
}


##
# Safely `ln` to a target
#
# Will create a symbolic link from source file ($1) at location ($2)
# Will also create the directory if it doesn't exist.
#
# @param $1 string
#        Destination symlink location
#
# @param $2 string
#        Source directory to point link to
#
function safeln (){
	safemkdir "$(dirname $2)"

	if [ ! -e "$2" ]; then
		printline "Symlinking $2 -> $1"
		ln -s "$1" "$2"
		checkexitstatus "$?"
	fi
}

##
# Download a given file into a given destination
#
# Source file can be http://, https://, git://, or ssh://
#
# Destination must be a local destination, and will be auto-created if it does not exist.
#
# @param $1 string source file (REQUIRED)
#        Fully resolved source file to download
#
# @param $2 string destination file (REQUIRED)
#        Fully resolved local file destination
#
# @param --user string
#        Short argument is -u
#        Username to use for the download
#        example, download ... --user bob
#
# @param --password string
#        Short argument is -p or --pass
#        Password to use for the download
#        example, download ... --pass password123
#        Supported with http, https, and git.  NOT supported for SSH!
#
# @param --port int
#        Short argument is -P
#        Port to connect to remotely for the download.
#
# @param --overwrite
#        Short argument is -o
#        Include this argument to overwrite an existing download, defaults to false
#
# @param --smart-overwrite
#        Only overwrite the file if it's changed.
#        Supported with http and https
#
function download (){

	_SRC=""
	_DEST=""
	_USER=""
	_PASS=""
	_PORT=""
	_OVERWRITE=0

	C=0
	while [ ! -z "$1" ]; do
		let "C=$C+1"

		if [ "$C" == "1" ]; then
			_SRC="$1"
		elif [ "$C" == "2" ]; then
			_DEST="$1"
		else
			case "$1" in
				"--overwrite" | "-o" )
					_OVERWRITE=1
					;;
				"--smart-overwrite" )
					_OVERWRITE="S"
					;;
				"--user" | "-u" )
					# Pull the next field!
					shift
					_USER="$1"
					;;
				"--pass" | "--password" | "-p" )
					# Pull the next field!
					shift
					_PASS="$1"
					;;
				"--port" | "-P" )
					# Pull the next field!
					shift
					_PORT="$1"
					;;
			esac
		fi

		shift;
	done


	printheader "Downloading $_SRC..."
	printdebug "Downloading to $_DEST"

	# Get the protocol of this file
	# (make sure there's a valid protocol)
	if [ "$(echo $_SRC | egrep '^[a-z]+://')" == "" ]; then
		printerror "No valid protocol given, unable to determine file source!"
		exit 1
	fi
	_PROT="$(echo $_SRC | sed 's#^\([a-z]*\)://\([^/:]*\)[/:]\(.*\)#\1#')"
	_HOST="$(echo $_SRC | sed 's#^\([a-z]*\)://\([^/:]*\)[/:]\(.*\)#\2#')"
	_FILE="$(echo $_SRC | sed 's#^\([a-z]*\)://\([^/:]*\)[/:]\(.*\)#\3#')"

	if [ ! -e "$(dirname "$_DEST")" ]; then
		mkdir -p "$(dirname "$_DEST")"
	fi

	if [ -e "$_DEST" -a "$_OVERWRITE" == "0" ]; then
		# Overwite is set to NO and the file exists.
		printline "$(basename $_DEST) already exists, not overwriting"
		printdebug "(use --overwrite to force a download regardless)"
		return
	fi

	case "$_PROT" in
		'http' | 'https' )
#			_SRC="$DOWNLOADPROTOCOL://$DOWNLOADHOST:$DOWNLOADPORT/$1"
			_OPTS=""
			if [ "$_USER" != "" ]; then _OPTS="$_OPTS --user=$_USER"; fi
			if [ "$_PASS" != "" ]; then _OPTS="$_OPTS --password='$_PASS'"; fi

			if [ "$_DEST" == "-" ]; then
				$CMDWGET $_OPTS $_SRC -O -
				checkexitstatus $?
			else
				# Standard download, check and see if it exists if the overwrite is set to smart.
				if [ -e "${_DEST}" -a "${_OVERWRITE}" == "S" ]; then
					_REMSIZE="$($CMDWGET -S --spider ${_SRC} 2>&1 | grep 'Content-Length' | sed 's#^[ ]*Content-Length: ##')"
					_LOCSIZE="$(stat -c "%s" ${_DEST})"
					if [ $_REMSIZE -eq $_LOCSIZE ]; then
						printline "Skipping download, files are the same size."
						return
					fi
				fi
				$CMDWGET $_OPTS $_SRC -O "$_DEST"
				checkexitstatus $?
			fi
			;;
		'ssh' )
			# SSH does not support passwords!
			if [ -z "$_PORT" ]; then _PORT=22; fi
			if [ -z "$_USER" ]; then
				printerror "Please set a user with the '-u username' arguments when calling download with ssh://"
				checkexitstatus 1
			fi

			if [ "$_DEST" == "-" ]; then
				ssh -o "StrictHostKeyChecking=no" -l$_USER -p$_PORT $_HOST "cat $_FILE"
				checkexitstatus $?
			else
				printdebug "executing scp command with arguments -P $_PORT $_USER@$_HOST:$_FILE \"$_DEST\""
				scp -o "StrictHostKeyChecking=no" -P $_PORT $_USER@$_HOST:$_FILE "$_DEST"
				checkexitstatus $?
			fi
			;;
		'git' )
			if [ -n "$(echo $_HOST | grep '://')" ]; then
				# If there is another :// in the HOST, then it's a standard HTTP/HTTPS request and treat it as such.
				_GITSRC="$_HOST/$_FILE"
			else
				# Otherwise no :// in the url, then read it as an SSH connection, which use the ":" symbol.
				_GITSRC="$_HOST:$_FILE"
			fi

			git clone "$_GITSRC" "$_DEST"
			checkexitstatus "$?"
			;;
		* )
			printerror "Unsupported DOWNLOADTYPE [ $DOWNLOADTYPE ]"
			checkexitstatus 1
	esac

}

##
# Download and extract a given file into a given destination
#
# Source file can be http://, https://, git://, or ssh://
#
# Destination must be a local destination, and will be auto-created if it does not exist.
#
# This is a helper function that extends download that also extracts a tarball or zip automatically.
# It does *not* support the advanced options that download provides, so if you need those, you may still
# call download directly and handle the decompression manually.
#
# @param $1 string source file (REQUIRED)
#        Fully resolved source file to download
#
# @param $2 string destination directory (REQUIRED)
#        Fully resolved local parent directory
#
function downloadextract (){
	_P="$(dirname $1)"
	_F="$(basename $1)"
	_T="/tmp/$(dirname $_P | sed 's:[^a-z]:_:g')"

	download $1 "$_T/$_F"
	safemkdir "$2"

	printheader "Extracting $_F..."
	if [ -n "$(echo $_F | egrep '(tar\.gz|tgz|tar\.Z)$')" ]; then
		tar xzf "$_T/$_F" -C "$2"

	elif [ -n "$(echo $1 | egrep 'gz$')" ]; then
		gunzip -f "$_T/$_F"
		# Since gzip modifies the original source.
		mv "$_T/*" "$2"

	elif [ -n "$(echo $1 | egrep 'tar\.bz2$')" ]; then
		tar xjf "$_T/$_F" -C "$2"

	elif [ -n "$(echo $1 | egrep '\.zip$')" ]; then
		if [ -z "$(which unzip 2>/dev/null)" ]; then
			printerror "Unable to extract $1 as unzip is not installed!"
			return 1
		fi
		unzip -u "$_T/$_F" -d "$2"

	else
		echo "Unsupported filetype for extraction, $1"
		exit 1
	fi

	checkexitstatus "$?"
}

# @todo Retool this function
function downloadexecute (){
	download $1
	printheader "Executing $1..."
	source $TEMPDIR/$1
}

##
# Install a package with the system's package manager.
#
# Uses Redhat's yum, Debian's apt-get, and SuSE's zypper.
#
# Usage:
#
# ```{.syntax-shell}
# install apache2 php7.0 mariadb-server
# ```
#
# @param $1..$N string
#        Package, (or packages), to install.  Accepts multiple packages at once.
#
function install (){
	if [ "$#" == "1" ]; then printheader "Installing $1..."
	else printheader "Installing $# packages..."
	fi

	PACKAGES=""
	for pkg in $@; do
		PACKAGES+=" $(resolve_install_package $pkg)"
	done

	for pkg in $PACKAGES; do
		# Preinstalls
		_do_install_script "PRE" $pkg
	done

	case $OSFAMILY in
		bsd )
			${SUDOPREFIX}pkg install -y $PACKAGES
			checkexitstatus "$?"
			;;
		debian )
			${SUDOPREFIX}apt-get -y update
			checkexitstatus "$?"
			${SUDOPREFIX}apt-get -y install $PACKAGES
			checkexitstatus "$?"
			;;
		redhat )
			${SUDOPREFIX}yum -y update
			checkexitstatus "$?"
			${SUDOPREFIX}yum -y install $PACKAGES
			checkexitstatus "$?"
			;;
		suse )
			${SUDOPREFIX}zypper -n install $PACKAGES
			checkexitstatus "$?"
			;;
		* )
			echo "Unsupported OS, $OS.  I do not know which package manager $OS uses."
			exit 1
	esac

	for pkg in $PACKAGES; do
		# Post-installs
		_do_install_script "POST" $pkg
	done
}

function resolve_install_package (){
	if [ -z "$INSTALLALIASES" ]; then
		INSTALLALIASES=$(mktemp)
		if [ -n "$(which curl 2>/dev/null)" ]; then
			curl -s https://git.eval.bz/eval/bash-basescript/raw/master/src/install.aliases \
			| egrep -v '^#|^$' \
			> $INSTALLALIASES
		elif [ -n "$(which wget 2>/dev/null)" ]; then
			wget -q https://git.eval.bz/eval/bash-basescript/raw/master/src/install.aliases -O - \
			| egrep -v '^#|^$' \
			> $INSTALLALIASES
		fi
	fi

	pkg="$1"

	if [ -n "$(egrep ^$pkg/$OSFAMILY/$OS/$OSVERSIONMAJ: $INSTALLALIASES)" ]; then
		# Exact match for this specific OS version X.
		egrep ^$pkg/$OSFAMILY/$OS/$OSVERSIONMAJ: $INSTALLALIASES | sed 's#.*: ##'

	elif [ -n "$(egrep ^$pkg/$OSFAMILY/*/$OSVERSIONMAJ: $INSTALLALIASES)" ]; then
		# Family match for generic version X.
		egrep ^$pkg/$OSFAMILY/*/$OSVERSIONMAJ: $INSTALLALIASES | sed 's#.*: ##'

	elif [ -n "$(egrep ^$pkg/$OSFAMILY: $INSTALLALIASES)" ]; then
		# Family match for any version
		egrep ^$pkg/$OSFAMILY: $INSTALLALIASES | sed 's#.*: ##'

	else
		# Default state, just return the package name itself.
		echo $pkg
	fi
}

##
# Install a given repository based on alias scripts defined in repo_installs.
#
function install_repo (){
	repo="$1"
	DEF=$(mktemp)
	URL="https://git.eval.bz/eval/bash-basescript/raw/master/src/repo_installs/$repo/$OSFAMILY/$OS/$OSVERSIONMAJ/install.sh"

	curl -s $URL > $DEF

	if [ $? -ne 0 ]; then
		printwarning "Unable to locate repository install script for $repo/$OSFAMILY/$OS/$OSVERSIONMAJ"
		return
	fi

	# Check if this file was a symlink to another file in the repo, tends to happen.
	if [ -z "$(head -n1 $DEF | egrep '^#!')" ]; then
		curl -s "$(readlink -m $(dirname $URL)/$(head -n1 $DEF) | sed 's#.*https:/#https://#')" > $DEF
	fi

	printline "Running repo installer for $repo/$OSFAMILY/$OS/$OSVERSIONMAJ"
	source $DEF
}

##
# Compile, make, and install source code in the requested directory.
#
# This function CD's into the requested directory and executes the following:
#
# 1. ./autogen.sh if present
# 2. ./configure
# 3. make
# 4. make install
#
# @param $1 string
#        Directory to compile sourcode in
#
# Usage:
#
# ```{.syntax-shell}
# compile "/path/to/sourcecode"
# ```
#
function compile (){
	cd "$1"
	shift
	printheader "Compiling $(basename `pwd`)..."
	# Is there an autogen file here?
	if [ -e autogen.sh ]; then
		./autogen.sh
		checkexitstatus "$?"
	fi

	./configure $*
	checkexitstatus "$?"
	make -j $THREADCOUNT
	checkexitstatus "$?"
	make install
	checkexitstatus "$?"
	cd -
}

# @todo Retool this function
function mysqlrootquery (){
	printheader "Execute mysql command $1..."
	mysql -uroot -p"$MYSQLPASS" -e "$1"
	checkexitstatus "$?"
}

# @todo Retool this function
function mysqlrootquery_nocheck (){
	printheader "Execute mysql command $1..."
	mysql -uroot -p"$MYSQLPASS" -e "$1"
}

# @todo Retool this function
function mysqlfile (){
	printheader "Execute mysql file $2 on database $1..."
	mysql -u$APPUSER -p"$APPPASS" "$1" < "$2"
	checkexitstatus "$?"
}


##
# Service Manager!
#
# Allows the start/stop and enable/disable of services across Debian and RH based systems.
#
# Usage:
# ```{.syntax-shell}
# # Start Apache2
# svc apache2 start
#
# # Restart ServiceFoo
# svc servicefoo restart
#
# # Enable mysql
# svc mysql enable
# ```
#
# @param $1 string
#        Service name to manipulate
#
# @param $2 string
#        Action to take, either start, stop, restart, enable, or disable.
#
function svc (){
	if [ -n "$(which systemctl 2>/dev/null)" ]; then
		# Systemctl already uses these naming conventions!
		systemctl $2 $1
		checkexitstatus "$?"
	else
		case $OSFAMILY in
			redhat )
				if [ "$2" == "start" ]; then /etc/init.d/$1 start;
				elif [ "$2" == "stop" ]; then /etc/init.d/$1 stop;
				elif [ "$2" == "restart" ]; then /etc/init.d/$1 restart;
				elif [ "$2" == "enable" ]; then chkconfig $1 on;
				elif [ "$2" == "disable" ]; then chkconfig $1 off;
				else echo "Unsupported svc request [$2], aborting"; exit 1;
				fi
				;;
			debian )
				if [ "$2" == "start" ]; then service $1 start;
				elif [ "$2" == "stop" ]; then service $1 stop;
				elif [ "$2" == "restart" ]; then service $1 restart;
				elif [ "$2" == "enable" ]; then update-rc.d -f $1 enable;
				elif [ "$2" == "disable" ]; then update-rc.d -f $1 disable;
				else echo "Unsupported svc request [$2], aborting"; exit 1;
				fi
				;;
		esac
		checkexitstatus "$?"
	fi
}



##
# Internal function to perform the PRE or POST installation logic for a given package.
#
# This takes the OSFAMILY, OS, and OSVERSIONMAJ into account when checking for the script.
#
# @param $1 PRE or POST
# @param $2 Package name
function _do_install_script (){
	case "$1" in
		"PRE" )
			T="Pre-Installation Script for $2"
			F="preinst.sh"
			;;
		"POST" )
			T="Post-Installation Script for $2"
			F="postinst.sh"
			;;
		* )
			print_error "Unsupported operation $1, please set either 'PRE' or 'POST' as the first argument for _do_install_script."
			return 1
	esac

	if [ -n "$(which wget 2>/dev/null)" ]; then
		# Use wget if it's available
		C1="wget -q --spider -O -"
		C2="wget -q -O -"
	else
		# Fallback to curl if not
		C1="curl -sf --head"
		C2="curl -s"
	fi


	for TRY in $2/$OSFAMILY/$OS/$OSVERSIONMAJ $2/$OSFAMILY/$OS $2/$OSFAMILY; do
		S="$($C1 https://git.eval.bz/eval/bash-basescript/raw/master/src/install.scripts/$TRY/$F 2>/dev/null)"
		if [ $? -eq 0 ]; then
			# Download succeeded!
			print_line "Running $T"
			SC=$(mktemp)
			$C2 https://git.eval.bz/eval/bash-basescript/raw/master/src/install.scripts/$TRY/$F > $SC
			source $SC
			if [ -n "$SC" ]; then
				rm -fr $SC
			fi
			return 0
		fi
	done
}

##
# Private function to evaluate "yes" or "no" from human input
#
# @echo "y", "n", or "" if invalid.
function _evaluate_yes_no (){
	case "$(echo $1 | tr '[:upper:]' '[:lower:]')" in
		"y" | "yes" | "ye" )
			echo "y"
			;;
		"n" | "no" )
			echo "n"
			;;
		* )
			echo ""
	esac
}

##
# Calculate and echo the MD5 hash on a given file.
#
# This is extracted out to a function because BSD has a different syntax than Linux.
#
# @param $1 string
#        The filename to md5
#
# @echo string
#
function md5file (){
	if [ "$OSFAMILY" == "bsd" ]; then
		# BSD format: MD5 (install.sh) = 18db3d3f3d5651af5e9a5465ebc889f7
		echo $(md5 "$1" | sed 's:.* \([a-z0-9]*\)$:\1:')
	else
		# Linux format: 18db3d3f3d5651af5e9a5465ebc889f7  install.sh
		echo $(md5sum "$1" | sed 's:^\([a-z0-9]*\) .*:\1:')
	fi
}

# Allow this script to be ran as a query mechanism for external scripts!
if [ "$(basename $0)" == "basescript.sh" -a $# -eq 1 -a "$1" == "-d" ]; then
	cat <<EOD
{
"ARCH": "$ARCH",
"INSTALLCMD": "$INSTALLCMD",
"JAVAVERSION": "$JAVAVERSION",
"JAVAVERSIONMAJ": $JAVAVERSIONMAJ,
"KERNEL": "$KERNEL",
"KERNELVERSION": "$KERNELVERSION",
"KERNELVERSIONMAJ": $KERNELVERSIONMAJ,
"KERNELVERSIONMIN": $KERNELVERSIONMIN,
"OS": "$OS",
"OSFAMILY": "$OSFAMILY",
"OSVERSION": "$OSVERSION",
"OSVERSIONMAJ": $OSVERSIONMAJ,
"XAVAILABLE": $XAVAILABLE
}
EOD
fi
